/* Generated by Cython 0.29.14 */

#ifndef __PYX_HAVE__breezy___simple_set_pyx
#define __PYX_HAVE__breezy___simple_set_pyx

#include "Python.h"
struct SimpleSetObject;

/* "breezy/_simple_set_pyx.pxd":29
 * 
 * 
 * cdef public api class SimpleSet [object SimpleSetObject, type SimpleSet_Type]:             # <<<<<<<<<<<<<<
 *     """A class similar to PySet, but with simpler implementation.
 * 
 */
struct SimpleSetObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_15_simple_set_pyx_SimpleSet *__pyx_vtab;
  Py_ssize_t _used;
  Py_ssize_t _fill;
  Py_ssize_t _mask;
  PyObject **_table;
};

#ifndef __PYX_HAVE_API__breezy___simple_set_pyx

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C DL_IMPORT(PyTypeObject) SimpleSet_Type;

#endif /* !__PYX_HAVE_API__breezy___simple_set_pyx */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_simple_set_pyx(void);
#else
PyMODINIT_FUNC PyInit__simple_set_pyx(void);
#endif

#endif /* !__PYX_HAVE__breezy___simple_set_pyx */
