/* Generated by Cython 0.29.14 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_14"
#define CYTHON_HEX_VERSION 0x001D0EF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__breezy__bzr___btree_serializer_pyx
#define __PYX_HAVE_API__breezy__bzr___btree_serializer_pyx
/* Early includes */
#include "python-compat.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "_static_tuple_c.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "breezy/bzr/_btree_serializer_pyx.pyx",
  "stringsource",
  "breezy/bzr/_str_helpers.pxd",
  "type.pxd",
  "breezy/_static_tuple_c.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record;
typedef struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record;

/* "breezy/bzr/_btree_serializer_pyx.pyx":305
 * #       One slightly ugly option would be to cache block offsets in a global.
 * #       However, that leads to thread-safety issues, etc.
 * ctypedef struct gc_chk_sha1_record:             # <<<<<<<<<<<<<<
 *     unsigned long long block_offset
 *     unsigned int block_length
 */
struct __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record {
  unsigned PY_LONG_LONG block_offset;
  unsigned int block_length;
  unsigned int record_start;
  unsigned int record_end;
  char sha1[20];
};

/* "breezy/bzr/_btree_serializer_pyx.pyx":91
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
 */
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtab;
  PyObject *data;
  int key_length;
  int ref_list_length;
  PyObject *keys;
  char *_cur_str;
  char *_end_str;
  char *_start;
  int _header_found;
};


/* "breezy/bzr/_btree_serializer_pyx.pyx":485
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
 */
struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtab;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *records;
  PyObject *last_key;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *last_record;
  int num_records;
  unsigned char common_shift;
  unsigned char offsets[0x101];
};



/* "breezy/bzr/_btree_serializer_pyx.pyx":91
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
 */

struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject *(*extract_key)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, char *);
  int (*process_line)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;


/* "breezy/bzr/_btree_serializer_pyx.pyx":485
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
 */

struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  StaticTuple *(*_record_to_value_and_refs)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  StaticTuple *(*_record_to_item)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *(*_lookup_record)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  int (*_count_records)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *);
  PyObject *(*_parse_bytes)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *);
  char *(*_parse_one_entry)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *);
  int (*_offset_for_sha1)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  PyObject *(*_compute_common)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto*/
static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_data); /* proto*/
static char *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record); /* proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'breezy.bzr._str_helpers' */
static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *, int, size_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *, Py_ssize_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *, Py_ssize_t); /*proto*/

/* Module declarations from 'breezy._static_tuple_c' */

/* Module declarations from 'breezy._static_tuple_c' */
static PyTypeObject *__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple = 0;

/* Module declarations from 'breezy.bzr._btree_serializer_pyx' */
static PyTypeObject *__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = 0;
static PyTypeObject *__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = 0;
static int __pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[0x100];
static char *__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf;
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(void); /*proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(char *, char *); /*proto*/
static void __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(char *, char *); /*proto*/
static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(PyObject *, char *); /*proto*/
static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(char *); /*proto*/
static unsigned int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(char *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "breezy.bzr._btree_serializer_pyx"
extern int __pyx_module_is_main_breezy__bzr___btree_serializer_pyx;
int __pyx_module_is_main_breezy__bzr___btree_serializer_pyx = 0;

/* Implementation of 'breezy.bzr._btree_serializer_pyx' */
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_d[] = "%d";
static const char __pyx_k_i[] = "i";
static const char __pyx_k__20[] = "\000";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_parse[] = "parse";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_as_bin[] = "as_bin";
static const char __pyx_k_as_hex[] = "as_hex";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_parser[] = "parser";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_key_len[] = "key_len";
static const char __pyx_k_ref_bit[] = "ref_bit";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_flat_len[] = "flat_len";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_next_len[] = "next_len";
static const char __pyx_k_node_len[] = "node_len";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_ref_list[] = "ref_list";
static const char __pyx_k_refs_len[] = "refs_len";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_sha1_bin[] = "sha1_bin";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ref_lists[] = "ref_lists";
static const char __pyx_k_reference[] = "reference";
static const char __pyx_k_value_len[] = "value_len";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_as_bin_sha[] = "as_bin_sha";
static const char __pyx_k_key_length[] = "key_length";
static const char __pyx_k_py_hexlify[] = "_py_hexlify";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_string_key[] = "string_key";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_ref_bit_len[] = "ref_bit_len";
static const char __pyx_k_flatten_node[] = "_flatten_node";
static const char __pyx_k_py_unhexlify[] = "_py_unhexlify";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_first_ref_list[] = "first_ref_list";
static const char __pyx_k_parse_into_chk[] = "_parse_into_chk";
static const char __pyx_k_py_key_to_sha1[] = "_py_key_to_sha1";
static const char __pyx_k_py_sha1_to_key[] = "_py_sha1_to_key";
static const char __pyx_k_BTreeLeafParser[] = "BTreeLeafParser";
static const char __pyx_k_first_reference[] = "first_reference";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_ref_list_length[] = "ref_list_length";
static const char __pyx_k_reference_lists[] = "reference_lists";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_last_self__start[] = "last < self._start";
static const char __pyx_k_parse_leaf_lines[] = "_parse_leaf_lines";
static const char __pyx_k_GCCHKSHA1LeafNode[] = "GCCHKSHA1LeafNode";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_have_reference_lists[] = "have_reference_lists";
static const char __pyx_k_key_r_is_not_present[] = "key %r is not present";
static const char __pyx_k_We_failed_to_unhexlify[] = "We failed to unhexlify";
static const char __pyx_k_not_a_40_byte_hex_digest[] = "not a 40-byte hex digest";
static const char __pyx_k_Failed_to_parse_record_end[] = "Failed to parse record end";
static const char __pyx_k_not_a_20_byte_binary_digest[] = "not a 20-byte binary digest";
static const char __pyx_k_Failed_to_parse_block_length[] = "Failed to parse block length";
static const char __pyx_k_Failed_to_parse_block_offset[] = "Failed to parse block offset";
static const char __pyx_k_pyx_unpickle_BTreeLeafParser[] = "__pyx_unpickle_BTreeLeafParser";
static const char __pyx_k_Failed_to_find_the_value_area[] = "Failed to find the value area";
static const char __pyx_k_only_1_null_not_2_as_expected[] = "only 1 null, not 2 as expected";
static const char __pyx_k_Expected_bytes_for_value_not_r[] = "Expected bytes for value not: %r";
static const char __pyx_k_line_did_not_start_with_sha1_r[] = "line did not start with sha1: %r";
static const char __pyx_k_self_data_is_not_a_byte_string[] = "self.data is not a byte string.";
static const char __pyx_k_invalid_key_wanted_segment_from[] = "invalid key, wanted segment from ";
static const char __pyx_k_Incompatible_checksums_s_vs_0x9b[] = "Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))";
static const char __pyx_k_Line_did_not_contain_40_hex_byte[] = "Line did not contain 40 hex bytes";
static const char __pyx_k_Node_did_not_start_with_type_lea[] = "Node did not start with \"type=leaf\": %r";
static const char __pyx_k_Pyrex_extensions_to_btree_node_p[] = "Pyrex extensions to btree node parsing.";
static const char __pyx_k_Something_went_wrong_while_parsi[] = "Something went wrong while parsing.";
static const char __pyx_k_We_expect_reference_bits_to_be_b[] = "We expect reference bits to be bytes not: %r";
static const char __pyx_k_We_expect_references_to_be_tuple[] = "We expect references to be tuples not: %r";
static const char __pyx_k_We_expected_a_tuple_or_StaticTup[] = "We expected a tuple() or StaticTuple() for node not: %s";
static const char __pyx_k_We_only_support_parsing_byte_str[] = "We only support parsing byte strings.";
static const char __pyx_k_With_ref_lists_we_expected_4_ent[] = "With ref_lists, we expected 4 entries not: %s";
static const char __pyx_k_Without_ref_lists_we_need_at_lea[] = "Without ref_lists, we need at least 3 entries not: %s";
static const char __pyx_k_breezy_bzr__btree_serializer_pyx[] = "breezy/bzr/_btree_serializer_pyx.pyx";
static const char __pyx_k_bytes_did_not_start_with_type_le[] = "bytes did not start with 'type=leaf\\n': %r";
static const char __pyx_k_invalid_key_loop_counter_self_re[] = "invalid key, loop_counter != self.ref_list_length";
static const char __pyx_k_self_last_record_self_records_ca[] = "self.last_record,self.records cannot be converted to a Python object for pickling";
static const char __pyx_k_sha1_bin_must_be_a_str_of_exactl[] = "sha1_bin must be a str of exactly 20 bytes";
static const char __pyx_k_tried_to_create_a_string_with_an[] = "tried to create a string with an invalid size: %d";
static const char __pyx_k_unexpected_reference_data_presen[] = "unexpected reference data present";
static const char __pyx_k_breezy_bzr__btree_serializer_pyx_2[] = "breezy.bzr._btree_serializer_pyx";
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_BTreeLeafParser;
static PyObject *__pyx_kp_s_Expected_bytes_for_value_not_r;
static PyObject *__pyx_kp_s_Failed_to_find_the_value_area;
static PyObject *__pyx_kp_s_Failed_to_parse_block_length;
static PyObject *__pyx_kp_s_Failed_to_parse_block_offset;
static PyObject *__pyx_kp_s_Failed_to_parse_record_end;
static PyObject *__pyx_n_s_GCCHKSHA1LeafNode;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x9b;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_Line_did_not_contain_40_hex_byte;
static PyObject *__pyx_kp_s_Node_did_not_start_with_type_lea;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_kp_s_Something_went_wrong_while_parsi;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s_We_expect_reference_bits_to_be_b;
static PyObject *__pyx_kp_s_We_expect_references_to_be_tuple;
static PyObject *__pyx_kp_s_We_expected_a_tuple_or_StaticTup;
static PyObject *__pyx_kp_s_We_failed_to_unhexlify;
static PyObject *__pyx_kp_s_We_only_support_parsing_byte_str;
static PyObject *__pyx_kp_s_With_ref_lists_we_expected_4_ent;
static PyObject *__pyx_kp_s_Without_ref_lists_we_need_at_lea;
static PyObject *__pyx_kp_b__20;
static PyObject *__pyx_n_s_as_bin;
static PyObject *__pyx_n_s_as_bin_sha;
static PyObject *__pyx_n_s_as_hex;
static PyObject *__pyx_kp_s_breezy_bzr__btree_serializer_pyx;
static PyObject *__pyx_n_s_breezy_bzr__btree_serializer_pyx_2;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_kp_s_bytes_did_not_start_with_type_le;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_b_d;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_first_ref_list;
static PyObject *__pyx_n_s_first_reference;
static PyObject *__pyx_n_s_flat_len;
static PyObject *__pyx_n_s_flatten_node;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_have_reference_lists;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_s_invalid_key_loop_counter_self_re;
static PyObject *__pyx_kp_s_invalid_key_wanted_segment_from;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_len;
static PyObject *__pyx_n_s_key_length;
static PyObject *__pyx_kp_s_key_r_is_not_present;
static PyObject *__pyx_kp_s_last_self__start;
static PyObject *__pyx_n_s_line;
static PyObject *__pyx_kp_s_line_did_not_start_with_sha1_r;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_next_len;
static PyObject *__pyx_n_s_node;
static PyObject *__pyx_n_s_node_len;
static PyObject *__pyx_kp_s_not_a_20_byte_binary_digest;
static PyObject *__pyx_kp_s_not_a_40_byte_hex_digest;
static PyObject *__pyx_kp_s_only_1_null_not_2_as_expected;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_parse;
static PyObject *__pyx_n_s_parse_into_chk;
static PyObject *__pyx_n_s_parse_leaf_lines;
static PyObject *__pyx_n_s_parser;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_py_hexlify;
static PyObject *__pyx_n_s_py_key_to_sha1;
static PyObject *__pyx_n_s_py_sha1_to_key;
static PyObject *__pyx_n_s_py_unhexlify;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_BTreeLeafParser;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_ref_bit;
static PyObject *__pyx_n_s_ref_bit_len;
static PyObject *__pyx_n_s_ref_list;
static PyObject *__pyx_n_s_ref_list_length;
static PyObject *__pyx_n_s_ref_lists;
static PyObject *__pyx_n_s_reference;
static PyObject *__pyx_n_s_reference_lists;
static PyObject *__pyx_n_s_refs_len;
static PyObject *__pyx_kp_s_self_data_is_not_a_byte_string;
static PyObject *__pyx_kp_s_self_last_record_self_records_ca;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sha1_bin;
static PyObject *__pyx_kp_s_sha1_bin_must_be_a_str_of_exactl;
static PyObject *__pyx_n_s_string_key;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_tried_to_create_a_string_with_an;
static PyObject *__pyx_kp_s_unexpected_reference_data_presen;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_value_len;
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static void __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_163054242;
static PyObject *__pyx_int_2147483648;
static PyObject *__pyx_int_4294967295;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
/* Late includes */

/* "breezy/bzr/_btree_serializer_pyx.pyx":121
 *     cdef int _header_found
 * 
 *     def __init__(self, data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.data = data
 *         self.key_length = key_length
 */

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 121, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 121, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 121, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_data = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 121, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self), __pyx_v_data, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":122
 * 
 *     def __init__(self, data, key_length, ref_list_length):
 *         self.data = data             # <<<<<<<<<<<<<<
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
 */
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  __Pyx_GOTREF(__pyx_v_self->data);
  __Pyx_DECREF(__pyx_v_self->data);
  __pyx_v_self->data = __pyx_v_data;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":123
 *     def __init__(self, data, key_length, ref_list_length):
 *         self.data = data
 *         self.key_length = key_length             # <<<<<<<<<<<<<<
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_key_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_v_self->key_length = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":124
 *         self.data = data
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length             # <<<<<<<<<<<<<<
 *         self.keys = []
 *         self._cur_str = NULL
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_ref_list_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_v_self->ref_list_length = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":125
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
 *         self.keys = []             # <<<<<<<<<<<<<<
 *         self._cur_str = NULL
 *         self._end_str = NULL
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":126
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
 *         self._cur_str = NULL             # <<<<<<<<<<<<<<
 *         self._end_str = NULL
 *         self._header_found = 0
 */
  __pyx_v_self->_cur_str = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":127
 *         self.keys = []
 *         self._cur_str = NULL
 *         self._end_str = NULL             # <<<<<<<<<<<<<<
 *         self._header_found = 0
 *         # keys are tuples
 */
  __pyx_v_self->_end_str = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":128
 *         self._cur_str = NULL
 *         self._end_str = NULL
 *         self._header_found = 0             # <<<<<<<<<<<<<<
 *         # keys are tuples
 * 
 */
  __pyx_v_self->_header_found = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":121
 *     cdef int _header_found
 * 
 *     def __init__(self, data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.data = data
 *         self.key_length = key_length
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":131
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last) {
  char *__pyx_v_temp_ptr;
  int __pyx_v_loop_counter;
  StaticTuple *__pyx_v_key = 0;
  PyObject *__pyx_v_failure_string = NULL;
  PyObject *__pyx_v_key_element = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("extract_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":141
 *         cdef StaticTuple key
 * 
 *         key = StaticTuple_New(self.key_length)             # <<<<<<<<<<<<<<
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(__pyx_v_self->key_length)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":142
 * 
 *         key = StaticTuple_New(self.key_length)
 *         for loop_counter from 0 <= loop_counter < self.key_length:             # <<<<<<<<<<<<<<
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 */
  __pyx_t_2 = __pyx_v_self->key_length;
  for (__pyx_v_loop_counter = 0; __pyx_v_loop_counter < __pyx_t_2; __pyx_v_loop_counter++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":144
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:
 */
    __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":145
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 */
    __pyx_t_3 = ((__pyx_v_temp_ptr == NULL) != 0);
    if (__pyx_t_3) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":146
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
 */
      __pyx_t_3 = (((__pyx_v_loop_counter + 1) == __pyx_v_self->key_length) != 0);
      if (likely(__pyx_t_3)) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":148
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 *                     temp_ptr = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # Invalid line
 */
        __pyx_v_temp_ptr = __pyx_v_last;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":146
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
 */
        goto __pyx_L6;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":151
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 */
      /*else*/ {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":152
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +
 *                         repr(safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)
 */
        __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":151
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 */
        __pyx_t_1 = PyNumber_Add(__pyx_kp_s_invalid_key_wanted_segment_from, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_failure_string = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":154
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)             # <<<<<<<<<<<<<<
 *             # capture the key string
 *             if (self.key_length == 1
 */
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_v_failure_string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 154, __pyx_L1_error)
      }
      __pyx_L6:;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":145
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":156
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 */
    __pyx_t_5 = ((__pyx_v_self->key_length == 1) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":157
 *             # capture the key string
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45             # <<<<<<<<<<<<<<
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,
 */
    __pyx_t_5 = (((__pyx_v_temp_ptr - __pyx_v_self->_start) == 45) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":158
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):             # <<<<<<<<<<<<<<
 *                 key_element = safe_string_from_size(self._start,
 *                                                     temp_ptr - self._start)
 */
    __pyx_t_5 = ((strncmp(__pyx_v_self->_start, ((char const *)"sha1:"), 5) == 0) != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":156
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 */
    if (__pyx_t_3) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":159
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                     temp_ptr - self._start)
 *             else:
 */
      __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":156
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, b'sha1:', 5) == 0):
 */
      goto __pyx_L7;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":162
 *                                                     temp_ptr - self._start)
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
 */
    /*else*/ {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":163
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,
 *                                                          temp_ptr - self._start)             # <<<<<<<<<<<<<<
 *             # advance our pointer
 *             self._start = temp_ptr + 1
 */
      __pyx_t_1 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L7:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":165
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
 *             self._start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 */
    __pyx_v_self->_start = (__pyx_v_temp_ptr + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":166
 *             # advance our pointer
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)             # <<<<<<<<<<<<<<
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)
 */
    Py_INCREF(__pyx_v_key_element);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":167
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)             # <<<<<<<<<<<<<<
 *         key = StaticTuple_Intern(key)
 *         return key
 */
    StaticTuple_SET_ITEM(__pyx_v_key, __pyx_v_loop_counter, __pyx_v_key_element);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":168
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)             # <<<<<<<<<<<<<<
 *         return key
 * 
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_Intern(__pyx_v_key)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_key, ((StaticTuple *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":169
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)
 *         return key             # <<<<<<<<<<<<<<
 * 
 *     cdef int process_line(self) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_key));
  __pyx_r = ((PyObject *)__pyx_v_key);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":131
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.extract_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_failure_string);
  __Pyx_XDECREF(__pyx_v_key_element);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":171
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
 */

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  char *__pyx_v_last;
  char *__pyx_v_temp_ptr;
  char *__pyx_v_ref_ptr;
  char *__pyx_v_next_start;
  int __pyx_v_loop_counter;
  Py_ssize_t __pyx_v_str_len;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  StaticTuple *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  StaticTuple *__pyx_v_node_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("process_line", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":180
 *         cdef Py_ssize_t str_len
 * 
 *         self._start = self._cur_str             # <<<<<<<<<<<<<<
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 */
  __pyx_t_1 = __pyx_v_self->_cur_str;
  __pyx_v_self->_start = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":182
 *         self._start = self._cur_str
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)             # <<<<<<<<<<<<<<
 *         if last == NULL:
 *             # Process until the end of the file
 */
  __pyx_v_last = ((char *)memchr(__pyx_v_self->_start, '\n', (__pyx_v_self->_end_str - __pyx_v_self->_start)));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":183
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
 */
  __pyx_t_2 = ((__pyx_v_last == NULL) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":185
 *         if last == NULL:
 *             # Process until the end of the file
 *             last = self._end_str             # <<<<<<<<<<<<<<
 *             self._cur_str = self._end_str
 *         else:
 */
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_last = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":186
 *             # Process until the end of the file
 *             last = self._end_str
 *             self._cur_str = self._end_str             # <<<<<<<<<<<<<<
 *         else:
 *             # And the next string is right after it
 */
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_self->_cur_str = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":183
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":189
 *         else:
 *             # And the next string is right after it
 *             self._cur_str = last + 1             # <<<<<<<<<<<<<<
 *             # The last character is right before the '\n'
 * 
 */
  /*else*/ {
    __pyx_v_self->_cur_str = (__pyx_v_last + 1);
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":192
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
 */
  __pyx_t_2 = ((__pyx_v_last == __pyx_v_self->_start) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":194
 *         if last == self._start:
 *             # parsed it all.
 *             return 0             # <<<<<<<<<<<<<<
 *         if last < self._start:
 *             # Unexpected error condition - fail
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":192
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":195
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 */
  __pyx_t_2 = ((__pyx_v_last < __pyx_v_self->_start) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":197
 *         if last < self._start:
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 197, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":195
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":198
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 */
  __pyx_t_2 = ((0 == __pyx_v_self->_header_found) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":200
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
 */
    __pyx_t_2 = ((strncmp(((char const *)"type=leaf"), __pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)) == 0) != 0);
    if (likely(__pyx_t_2)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":201
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1             # <<<<<<<<<<<<<<
 *                 return 0
 *             else:
 */
      __pyx_v_self->_header_found = 1;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":202
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1
 *                 return 0             # <<<<<<<<<<<<<<
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":200
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":204
 *                 return 0
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'             # <<<<<<<<<<<<<<
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 */
    /*else*/ {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":205
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
 *                     % (safe_string_from_size(self._start, last - self._start)))             # <<<<<<<<<<<<<<
 * 
 *         key = self.extract_key(last)
 */
      __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Node_did_not_start_with_type_lea, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":204
 *                 return 0
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'             # <<<<<<<<<<<<<<
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 204, __pyx_L1_error)
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":198
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp(b"type=leaf", self._start, last - self._start) == 0:
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":207
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 *         key = self.extract_key(last)             # <<<<<<<<<<<<<<
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_last); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_key = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":209
 *         key = self.extract_key(last)
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *         if temp_ptr == NULL:
 *             # Invalid line
 */
  __pyx_v_temp_ptr = ((char *)__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":210
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
 */
  __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":212
 *         if temp_ptr == NULL:
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")             # <<<<<<<<<<<<<<
 *         else:
 *             # Because of how conversions were done, we ended up with *lots* of
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 212, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":210
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":220
 *             # something like '12607215 328306 0 0', which ends up consuming 1MB
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1             # <<<<<<<<<<<<<<
 *             if (str_len > 4
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 */
  /*else*/ {
    __pyx_v_str_len = ((__pyx_v_last - __pyx_v_temp_ptr) - 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":221
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
    __pyx_t_5 = ((__pyx_v_str_len > 4) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L10_bool_binop_done;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":222
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):             # <<<<<<<<<<<<<<
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
 */
    __pyx_t_5 = ((strncmp(((char const *)" 0 0"), (__pyx_v_last - 4), 4) == 0) != 0);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L10_bool_binop_done:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":221
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":225
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
 */
      __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_value = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":221
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(b" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
      goto __pyx_L9;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":227
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             # shrink the references end point
 *             last = temp_ptr
 */
    /*else*/ {
      __pyx_t_3 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_value = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L9:;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":229
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
 *             # shrink the references end point
 *             last = temp_ptr             # <<<<<<<<<<<<<<
 * 
 *         if self.ref_list_length:
 */
    __pyx_v_last = __pyx_v_temp_ptr;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":231
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 */
  __pyx_t_2 = (__pyx_v_self->ref_list_length != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":232
 * 
 *         if self.ref_list_length:
 *             ref_lists = StaticTuple_New(self.ref_list_length)             # <<<<<<<<<<<<<<
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
 */
    __pyx_t_3 = ((PyObject *)StaticTuple_New(__pyx_v_self->ref_list_length)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ref_lists = ((StaticTuple *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":233
 *         if self.ref_list_length:
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0             # <<<<<<<<<<<<<<
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []
 */
    __pyx_v_loop_counter = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":234
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:             # <<<<<<<<<<<<<<
 *                 ref_list = []
 *                 # extract a reference list
 */
    while (1) {
      __pyx_t_2 = ((__pyx_v_loop_counter < __pyx_v_self->ref_list_length) != 0);
      if (!__pyx_t_2) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":235
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []             # <<<<<<<<<<<<<<
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":237
 *                 ref_list = []
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1             # <<<<<<<<<<<<<<
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")
 */
      __pyx_v_loop_counter = (__pyx_v_loop_counter + 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":238
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 */
      __pyx_t_2 = ((__pyx_v_last < __pyx_v_self->_start) != 0);
      if (unlikely(__pyx_t_2)) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":239
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 */
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(0, 239, __pyx_L1_error)

        /* "breezy/bzr/_btree_serializer_pyx.pyx":238
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 */
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":241
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)             # <<<<<<<<<<<<<<
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 */
      __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\t', (__pyx_v_last - __pyx_v_self->_start)));

      /* "breezy/bzr/_btree_serializer_pyx.pyx":242
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
 */
      __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
      if (__pyx_t_2) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":244
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
 */
        __pyx_t_2 = ((__pyx_v_loop_counter != __pyx_v_self->ref_list_length) != 0);
        if (unlikely(__pyx_t_2)) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":246
 *                     if loop_counter != self.ref_list_length:
 *                         # Invalid line
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "invalid key, loop_counter != self.ref_list_length")
 *                     else:
 */
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 246, __pyx_L1_error)

          /* "breezy/bzr/_btree_serializer_pyx.pyx":244
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
 */
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":250
 *                     else:
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last             # <<<<<<<<<<<<<<
 *                         next_start = last
 *                 else:
 */
        /*else*/ {
          __pyx_v_ref_ptr = __pyx_v_last;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":251
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last
 *                         next_start = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # scan to the end of this ref list
 */
          __pyx_v_next_start = __pyx_v_last;
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":242
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
 */
        goto __pyx_L16;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":254
 *                 else:
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr             # <<<<<<<<<<<<<<
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
 */
      /*else*/ {
        __pyx_v_ref_ptr = __pyx_v_temp_ptr;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":255
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr
 *                     next_start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:
 */
        __pyx_v_next_start = (__pyx_v_temp_ptr + 1);
      }
      __pyx_L16:;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":257
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:             # <<<<<<<<<<<<<<
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 */
      while (1) {
        __pyx_t_2 = ((__pyx_v_self->_start < __pyx_v_ref_ptr) != 0);
        if (!__pyx_t_2) break;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":259
 *                 while self._start < ref_ptr:
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',             # <<<<<<<<<<<<<<
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:
 */
        __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\r', (__pyx_v_ref_ptr - __pyx_v_self->_start)));

        /* "breezy/bzr/_btree_serializer_pyx.pyx":261
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
 */
        __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":263
 *                     if temp_ptr == NULL:
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr             # <<<<<<<<<<<<<<
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 */
          __pyx_v_temp_ptr = __pyx_v_ref_ptr;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":261
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
 */
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":265
 *                         temp_ptr = ref_ptr
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))             # <<<<<<<<<<<<<<
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)
 */
        __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_temp_ptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = PyList_Append(__pyx_v_ref_list, __pyx_t_3); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 265, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":266
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))             # <<<<<<<<<<<<<<
 *                 Py_INCREF(ref_list)
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 */
      __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_v_ref_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = ((PyObject *)StaticTuple_Intern(((StaticTuple *)__pyx_t_4))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":267
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)             # <<<<<<<<<<<<<<
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
 */
      Py_INCREF(__pyx_v_ref_list);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":268
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)             # <<<<<<<<<<<<<<
 *                 # prepare for the next reference list
 *                 self._start = next_start
 */
      StaticTuple_SET_ITEM(__pyx_v_ref_lists, (__pyx_v_loop_counter - 1), __pyx_v_ref_list);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":270
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
 *                 self._start = next_start             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 */
      __pyx_v_self->_start = __pyx_v_next_start;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":271
 *                 # prepare for the next reference list
 *                 self._start = next_start
 *             node_value = StaticTuple(value, ref_lists)             # <<<<<<<<<<<<<<
 *         else:
 *             if last != self._start:
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_value);
    __Pyx_INCREF(((PyObject *)__pyx_v_ref_lists));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_ref_lists));
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_ref_lists));
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_node_value = ((StaticTuple *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":231
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 */
    goto __pyx_L12;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":273
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 */
  /*else*/ {
    __pyx_t_2 = ((__pyx_v_last != __pyx_v_self->_start) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":275
 *             if last != self._start:
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 275, __pyx_L1_error)

      /* "breezy/bzr/_btree_serializer_pyx.pyx":273
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":276
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = StaticTuple(value, StaticTuple())             # <<<<<<<<<<<<<<
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 *         return 0
 */
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_value);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_node_value = ((StaticTuple *)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L12:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":277
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_4 = __pyx_v_self->keys;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_key);
  __Pyx_INCREF(((PyObject *)__pyx_v_node_value));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_node_value));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_node_value));
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = PyList_Append(__pyx_t_4, __pyx_t_7); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":278
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def parse(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":171
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.process_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF((PyObject *)__pyx_v_node_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":280
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  Py_ssize_t __pyx_v_byte_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("parse", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":282
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
 */
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((!(PyBytes_CheckExact(__pyx_t_1) != 0)) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":283
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
 *             raise AssertionError('self.data is not a byte string.')             # <<<<<<<<<<<<<<
 *         byte_count = PyBytes_GET_SIZE(self.data)
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 283, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":282
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":284
 *         if not PyBytes_CheckExact(self.data):
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)             # <<<<<<<<<<<<<<
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 *         # This points to the last character in the string
 */
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_byte_count = PyBytes_GET_SIZE(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":285
 *             raise AssertionError('self.data is not a byte string.')
 *         byte_count = PyBytes_GET_SIZE(self.data)
 *         self._cur_str = PyBytes_AS_STRING(self.data)             # <<<<<<<<<<<<<<
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
 */
  __pyx_t_1 = __pyx_v_self->data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_self->_cur_str = PyBytes_AS_STRING(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":287
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count             # <<<<<<<<<<<<<<
 *         while self._cur_str < self._end_str:
 *             self.process_line()
 */
  __pyx_v_self->_end_str = (__pyx_v_self->_cur_str + __pyx_v_byte_count);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":288
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:             # <<<<<<<<<<<<<<
 *             self.process_line()
 *         return self.keys
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_self->_cur_str < __pyx_v_self->_end_str) != 0);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":289
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:
 *             self.process_line()             # <<<<<<<<<<<<<<
 *         return self.keys
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->process_line(__pyx_v_self); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 289, __pyx_L1_error)
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":290
 *         while self._cur_str < self._end_str:
 *             self.process_line()
 *         return self.keys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->keys);
  __pyx_r = __pyx_v_self->keys;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":280
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->_cur_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->_end_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_header_found); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->key_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_self->ref_list_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyTuple_New(8); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_v_self->data);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_self->data);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->keys);
  __Pyx_GIVEREF(__pyx_v_self->keys);
  PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_v_self->keys);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 7, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_7 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v__dict = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "(tree fragment)":7
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_8 = (__pyx_v__dict != Py_None);
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v__dict);
    __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._cur_str, self._end_str, self._header_found, self._start, self.data, self.key_length, self.keys, self.ref_list_length)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, None), state
 */
  /*else*/ {
    __pyx_t_8 = (__pyx_v_self->data != Py_None);
    __pyx_t_10 = (__pyx_t_8 != 0);
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_10 = (__pyx_v_self->keys != Py_None);
    __pyx_t_8 = (__pyx_t_10 != 0);
    __pyx_t_9 = __pyx_t_8;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_9;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, None), state
 *     else:
 */
  __pyx_t_9 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_9) {

    /* "(tree fragment)":13
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_BTreeLeafParser); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_163054242);
    __Pyx_GIVEREF(__pyx_int_163054242);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_163054242);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_7, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.data is not None or self.keys is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, None), state
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_BTreeLeafParser); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_163054242);
    __Pyx_GIVEREF(__pyx_int_163054242);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_int_163054242);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_state);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
    __pyx_t_5 = 0;
    __pyx_t_7 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_BTreeLeafParser, (type(self), 0x9b802a2, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BTreeLeafParser__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.BTreeLeafParser.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":293
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines = {"_parse_leaf_lines", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_leaf_lines (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, 1); __PYX_ERR(0, 293, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, 2); __PYX_ERR(0, 293, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parse_leaf_lines") < 0)) __PYX_ERR(0, 293, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_data = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 293, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(__pyx_self, __pyx_v_data, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_data, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_parser = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_parse_leaf_lines", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":294
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)             # <<<<<<<<<<<<<<
 *     return parser.parse()
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_data);
  __Pyx_INCREF(__pyx_v_key_length);
  __Pyx_GIVEREF(__pyx_v_key_length);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_key_length);
  __Pyx_INCREF(__pyx_v_ref_list_length);
  __Pyx_GIVEREF(__pyx_v_ref_list_length);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_ref_list_length);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_parser = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":295
 * def _parse_leaf_lines(data, key_length, ref_list_length):
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_parser), __pyx_n_s_parse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":293
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parser);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":317
 * _hexbuf = b'0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(void) {
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_populate_unhexbuf", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":319
 * cdef _populate_unhexbuf():
 *     cdef int i
 *     for i from 0 <= i < 256:             # <<<<<<<<<<<<<<
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 */
  for (__pyx_v_i = 0; __pyx_v_i < 0x100; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":320
 *     cdef int i
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
 */
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[__pyx_v_i]) = -1;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":321
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 */
  for (__pyx_v_i = 0; __pyx_v_i < 10; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":322
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
 */
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[(__pyx_v_i + '0')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":323
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 */
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":324
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
 */
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'a')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":325
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()
 */
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":326
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i             # <<<<<<<<<<<<<<
 * _populate_unhexbuf()
 * 
 */
    (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'A')]) = __pyx_v_i;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":317
 * _hexbuf = b'0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":330
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
 */

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(char *__pyx_v_as_hex, char *__pyx_v_as_bin) {
  int __pyx_v_top;
  int __pyx_v_bot;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_unhexlify_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":342
 *     # binascii does this using isupper() and tolower() and ?: syntax. I'm
 *     # guessing a simple lookup array should be faster.
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 */
  __pyx_v_j = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":343
 *     # guessing a simple lookup array should be faster.
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 */
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":344
 *     j = 0
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 */
    __pyx_v_top = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":345
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":346
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         if top == -1 or bot == -1:
 */
    __pyx_v_bot = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":347
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         if top == -1 or bot == -1:
 *             return 0
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":348
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 */
    __pyx_t_2 = ((__pyx_v_top == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_bot == -1L) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":349
 *         j = j + 1
 *         if top == -1 or bot == -1:
 *             return 0             # <<<<<<<<<<<<<<
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":348
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":350
 *         if top == -1 or bot == -1:
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    (__pyx_v_as_bin[__pyx_v_i]) = ((unsigned char)((__pyx_v_top << 4) + __pyx_v_bot));
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":351
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":330
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":354
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_hex); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify[] = "For the test infrastructure, just thunks to _unhexlify_sha1";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify = {"_py_unhexlify", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify, METH_O, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_hex) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_unhexlify (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(__pyx_self, ((PyObject *)__pyx_v_as_hex));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex) {
  PyObject *__pyx_v_as_bin = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_py_unhexlify", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":356
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 */
  __pyx_t_2 = ((!(PyBytes_CheckExact(__pyx_v_as_hex) != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((PyBytes_GET_SIZE(__pyx_v_as_hex) != 40) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":357
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 *         raise ValueError('not a 40-byte hex digest')             # <<<<<<<<<<<<<<
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 357, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":356
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":358
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin
 */
  __pyx_t_3 = PyBytes_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_as_bin = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":359
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
 */
  __pyx_t_1 = (__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(PyBytes_AS_STRING(__pyx_v_as_hex), PyBytes_AS_STRING(__pyx_v_as_bin)) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":360
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin);
    __pyx_r = __pyx_v_as_bin;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":359
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":361
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 *         return as_bin
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":354
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_unhexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":364
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
 */

static void __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(char *__pyx_v_as_bin, char *__pyx_v_as_hex) {
  int __pyx_v_i;
  int __pyx_v_j;
  char __pyx_v_c;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hexlify_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":368
 *     cdef char c
 * 
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
 */
  __pyx_v_j = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":369
 * 
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 */
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":370
 *     j = 0
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
 */
    __pyx_v_c = (__pyx_v_as_bin[__pyx_v_i]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":371
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
 */
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf[((__pyx_v_c >> 4) & 0xf)]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":372
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":373
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 * 
 */
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf[(__pyx_v_c & 0xf)]);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":374
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_j = (__pyx_v_j + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":364
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":377
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_bin); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify[] = "For test infrastructure, thunk to _hexlify_sha1";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify = {"_py_hexlify", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify, METH_O, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_bin) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_hexlify (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(__pyx_self, ((PyObject *)__pyx_v_as_bin));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin) {
  PyObject *__pyx_v_as_hex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_py_hexlify", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":379
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_as_bin); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 != 20) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((!(PyBytes_CheckExact(__pyx_v_as_bin) != 0)) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":380
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')             # <<<<<<<<<<<<<<
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 380, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":379
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":381
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 *     return as_hex
 */
  __pyx_t_4 = PyBytes_FromStringAndSize(NULL, 40); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_as_hex = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":382
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))             # <<<<<<<<<<<<<<
 *     return as_hex
 * 
 */
  __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(PyBytes_AS_STRING(__pyx_v_as_bin), PyBytes_AS_STRING(__pyx_v_as_hex));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":383
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 *     return as_hex             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_as_hex);
  __pyx_r = __pyx_v_as_hex;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":377
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_hexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_hex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":386
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
 */

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(PyObject *__pyx_v_key, char *__pyx_v_sha1) {
  char *__pyx_v_c_val;
  PyObject *__pyx_v_p_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_key_to_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":396
 *     cdef PyObject *p_val
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:
 */
  __pyx_t_2 = (StaticTuple_CheckExact(__pyx_v_key) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  if (!(likely(((__pyx_v_key) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_key, __pyx_ptype_6breezy_15_static_tuple_c_StaticTuple))))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_t_2 = ((StaticTuple_GET_SIZE(((StaticTuple *)__pyx_v_key)) == 1) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":397
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)             # <<<<<<<<<<<<<<
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:
 *         p_val = PyTuple_GET_ITEM(key, 0)
 */
    if (!(likely(((__pyx_v_key) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_key, __pyx_ptype_6breezy_15_static_tuple_c_StaticTuple))))) __PYX_ERR(0, 397, __pyx_L1_error)
    __pyx_v_p_val = ((PyObject *)StaticTuple_GET_ITEM(((StaticTuple *)__pyx_v_key), 0));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":396
 *     cdef PyObject *p_val
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":398
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = PyTuple_GET_ITEM(key, 0)
 *     else:
 */
  __pyx_t_2 = (PyTuple_CheckExact(__pyx_v_key) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = ((PyTuple_GET_SIZE(__pyx_v_key) == 1) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":399
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:
 *         p_val = PyTuple_GET_ITEM(key, 0)             # <<<<<<<<<<<<<<
 *     else:
 *         # Not a tuple or a StaticTuple
 */
    __pyx_v_p_val = PyTuple_GET_ITEM(__pyx_v_key, 0);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":398
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = PyTuple_GET_ITEM(key, 0)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":402
 *     else:
 *         # Not a tuple or a StaticTuple
 *         return 0             # <<<<<<<<<<<<<<
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":403
 *         # Not a tuple or a StaticTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 */
  __pyx_t_2 = (PyBytes_CheckExact(((PyObject *)__pyx_v_p_val)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":404
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):             # <<<<<<<<<<<<<<
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 *     else:
 */
  __pyx_t_2 = ((PyBytes_GET_SIZE(((PyObject *)__pyx_v_p_val)) == 45) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":403
 *         # Not a tuple or a StaticTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 */
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":405
 *     if (PyBytes_CheckExact(<object>p_val)
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)             # <<<<<<<<<<<<<<
 *     else:
 *         return 0
 */
    __pyx_v_c_val = PyBytes_AS_STRING(((PyObject *)__pyx_v_p_val));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":403
 *         # Not a tuple or a StaticTuple
 *         return 0
 *     if (PyBytes_CheckExact(<object>p_val)             # <<<<<<<<<<<<<<
 *             and PyBytes_GET_SIZE(<object>p_val) == 45):
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 */
    goto __pyx_L8;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":407
 *         c_val = PyBytes_AS_STRING(<object>p_val)
 *     else:
 *         return 0             # <<<<<<<<<<<<<<
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L8:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":408
 *     else:
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 */
  __pyx_t_1 = ((strncmp(__pyx_v_c_val, ((char const *)"sha1:"), 5) != 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":409
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0             # <<<<<<<<<<<<<<
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":408
 *     else:
 *         return 0
 *     if strncmp(c_val, b'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":410
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
  __pyx_t_1 = ((!(__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1((__pyx_v_c_val + 5), __pyx_v_sha1) != 0)) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":411
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":410
 *     if strncmp(c_val, b'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":412
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":386
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("breezy.bzr._btree_serializer_pyx._key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":415
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1[] = "Map a key to a simple sha1 string.\n\n    This is a testing thunk to the C function.\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1 = {"_py_key_to_sha1", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1, METH_O, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_key_to_sha1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(__pyx_self, ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_as_bin_sha = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_py_key_to_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":420
 *     This is a testing thunk to the C function.
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
 */
  __pyx_t_1 = PyBytes_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_as_bin_sha = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":421
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
 */
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, PyBytes_AS_STRING(__pyx_v_as_bin_sha)) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":422
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin_sha);
    __pyx_r = __pyx_v_as_bin_sha;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":421
 *     """
 *     as_bin_sha = PyBytes_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":423
 *     if _key_to_sha1(key, PyBytes_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":415
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin_sha);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":426
 * 
 * 
 * cdef StaticTuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef StaticTuple key
 */

static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(char *__pyx_v_sha1) {
  StaticTuple *__pyx_v_key = 0;
  PyObject *__pyx_v_hexxed = 0;
  char *__pyx_v_c_buf;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_sha1_to_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":431
 *     cdef object hexxed
 *     cdef char *c_buf
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)             # <<<<<<<<<<<<<<
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)
 */
  __pyx_t_1 = PyBytes_FromStringAndSize(NULL, 45); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hexxed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":432
 *     cdef char *c_buf
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)
 *     c_buf = PyBytes_AS_STRING(hexxed)             # <<<<<<<<<<<<<<
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
 */
  __pyx_v_c_buf = PyBytes_AS_STRING(__pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":433
 *     hexxed = PyBytes_FromStringAndSize(NULL, 45)
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)
 */
  (void)(memcpy(__pyx_v_c_buf, ((char *)"sha1:"), 5));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":434
 *     c_buf = PyBytes_AS_STRING(hexxed)
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)             # <<<<<<<<<<<<<<
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)
 */
  __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__hexlify_sha1(__pyx_v_sha1, (__pyx_v_c_buf + 5));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":435
 *     memcpy(c_buf, b'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)             # <<<<<<<<<<<<<<
 *     Py_INCREF(hexxed)
 *     StaticTuple_SET_ITEM(key, 0, hexxed)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":436
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)             # <<<<<<<<<<<<<<
 *     StaticTuple_SET_ITEM(key, 0, hexxed)
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
 */
  Py_INCREF(__pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":437
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)
 *     StaticTuple_SET_ITEM(key, 0, hexxed)             # <<<<<<<<<<<<<<
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
 *     # can be done in 66.6us (so 18.6us to build them all).
 */
  StaticTuple_SET_ITEM(__pyx_v_key, 0, __pyx_v_hexxed);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":447
 *     # We *could* hang the PyObject form off of the gc_chk_sha1_record for ones
 *     # that we have deserialized. Something to think about, at least.
 *     key = StaticTuple_Intern(key)             # <<<<<<<<<<<<<<
 *     return key
 * 
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_Intern(__pyx_v_key)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_key, ((StaticTuple *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":448
 *     # that we have deserialized. Something to think about, at least.
 *     key = StaticTuple_Intern(key)
 *     return key             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_key));
  __pyx_r = __pyx_v_key;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":426
 * 
 * 
 * cdef StaticTuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef StaticTuple key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_hexxed);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":451
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key[] = "Test thunk to check the sha1 mapping.";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key = {"_py_sha1_to_key", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key, METH_O, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_sha1_to_key (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(__pyx_self, ((PyObject *)__pyx_v_sha1_bin));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_py_sha1_to_key", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":453
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
 */
  __pyx_t_2 = ((!(PyBytes_CheckExact(__pyx_v_sha1_bin) != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((PyBytes_GET_SIZE(__pyx_v_sha1_bin) != 20) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":454
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')             # <<<<<<<<<<<<<<
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 454, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":453
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":455
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = ((PyObject *)__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(PyBytes_AS_STRING(__pyx_v_sha1_bin))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":451
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._py_sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":458
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
 */

static unsigned int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(char *__pyx_v_sha1) {
  unsigned int __pyx_v_val;
  unsigned int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sha1_to_uint", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":464
 *            | ((<unsigned int>(sha1[1]) & 0xff) << 16)
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))             # <<<<<<<<<<<<<<
 *     return val
 * 
 */
  __pyx_v_val = (((((((unsigned int)(__pyx_v_sha1[0])) & 0xff) << 24) | ((((unsigned int)(__pyx_v_sha1[1])) & 0xff) << 16)) | ((((unsigned int)(__pyx_v_sha1[2])) & 0xff) << 8)) | ((((unsigned int)(__pyx_v_sha1[3])) & 0xff) << 0));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":465
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))
 *     return val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":458
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":468
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  PyObject *__pyx_v_block_offset_str = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("_format_record", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":471
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 */
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":474
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 *         block_offset_str = b'%d' % record.block_offset             # <<<<<<<<<<<<<<
 *         value = PyBytes_FromFormat(
 *             '%s %u %u %u', PyBytes_AS_STRING(block_offset_str),
 */
    __pyx_t_2 = __Pyx_PyInt_From_unsigned_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Remainder(__pyx_kp_b_d, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_block_offset_str = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":475
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 *         block_offset_str = b'%d' % record.block_offset
 *         value = PyBytes_FromFormat(             # <<<<<<<<<<<<<<
 *             '%s %u %u %u', PyBytes_AS_STRING(block_offset_str),
 *             record.block_length, record.record_start, record.record_end)
 */
    __pyx_t_1 = PyBytes_FromFormat(((char *)"%s %u %u %u"), PyBytes_AS_STRING(__pyx_v_block_offset_str), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":471
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # Could use %llu which was added to Python 2.7 but it oddly is missing
 *         # from the Python 3 equivalent functions, so hack still needed. :(
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":479
 *             record.block_length, record.record_start, record.record_end)
 *     else:
 *         value = PyBytes_FromFormat(             # <<<<<<<<<<<<<<
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)
 */
  /*else*/ {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":481
 *         value = PyBytes_FromFormat(
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
    __pyx_t_1 = PyBytes_FromFormat(((char *)"%lu %u %u %u"), ((unsigned long)__pyx_v_record->block_offset), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_value = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":482
 *             '%lu %u %u %u', <unsigned long>record.block_offset,
 *             record.block_length, record.record_start, record.record_end)
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":468
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a bytes object,
 *     # but it makes things work a bit better internally for now.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._format_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_block_offset_str);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":503
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sizeof__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__sizeof__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":504
 * 
 *     def __sizeof__(self):
 *         return (             # <<<<<<<<<<<<<<
 *             sizeof(GCCHKSHA1LeafNode) +
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":505
 *     def __sizeof__(self):
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +             # <<<<<<<<<<<<<<
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode)) + ((sizeof(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record)) * __pyx_v_self->num_records))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":503
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         return (
 *             sizeof(GCCHKSHA1LeafNode) +
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__sizeof__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":508
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 */

/* Python wrapper */
static void __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":509
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->records != NULL) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":510
 *     def __dealloc__(self):
 *         if self.records != NULL:
 *             PyMem_Free(self.records)             # <<<<<<<<<<<<<<
 *             self.records = NULL
 * 
 */
    PyMem_Free(__pyx_v_self->records);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":511
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 *             self.records = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, bytes):
 */
    __pyx_v_self->records = NULL;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":509
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":508
 *             sizeof(gc_chk_sha1_record) * self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":513
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 */

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 513, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_bytes = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 513, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), __pyx_v_bytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":514
 * 
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)             # <<<<<<<<<<<<<<
 *         self.last_key = None
 *         self.last_record = NULL
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_bytes(__pyx_v_self, __pyx_v_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":515
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)
 *         self.last_key = None             # <<<<<<<<<<<<<<
 *         self.last_record = NULL
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":516
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 *         self.last_record = NULL             # <<<<<<<<<<<<<<
 * 
 *     property min_key:
 */
  __pyx_v_self->last_record = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":513
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":519
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":520
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
 */
  __pyx_t_1 = ((__pyx_v_self->num_records > 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":521
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[0]).sha1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":520
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":522
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     property max_key:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":519
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.min_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":525
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":526
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
 */
  __pyx_t_1 = ((__pyx_v_self->num_records > 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":527
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[(__pyx_v_self->num_records - 1)]).sha1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":526
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":528
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":525
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.max_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":530
 *             return None
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
 */

static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  StaticTuple *__pyx_v_value_and_refs = 0;
  StaticTuple *__pyx_v_empty = 0;
  PyObject *__pyx_v_value = NULL;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_record_to_value_and_refs", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":535
 *         cdef StaticTuple value_and_refs
 *         cdef StaticTuple empty
 *         value_and_refs = StaticTuple_New(2)             # <<<<<<<<<<<<<<
 *         value = _format_record(record)
 *         Py_INCREF(value)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":536
 *         cdef StaticTuple empty
 *         value_and_refs = StaticTuple_New(2)
 *         value = _format_record(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 */
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__format_record(__pyx_v_record); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":537
 *         value_and_refs = StaticTuple_New(2)
 *         value = _format_record(record)
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
 */
  Py_INCREF(__pyx_v_value);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":538
 *         value = _format_record(record)
 *         Py_INCREF(value)
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)             # <<<<<<<<<<<<<<
 *         # Always empty refs
 *         empty = StaticTuple_New(0)
 */
  StaticTuple_SET_ITEM(__pyx_v_value_and_refs, 0, __pyx_v_value);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":540
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
 *         empty = StaticTuple_New(0)             # <<<<<<<<<<<<<<
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_empty = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":541
 *         # Always empty refs
 *         empty = StaticTuple_New(0)
 *         Py_INCREF(empty)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs
 */
  Py_INCREF(((PyObject *)__pyx_v_empty));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":542
 *         empty = StaticTuple_New(0)
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)             # <<<<<<<<<<<<<<
 *         return value_and_refs
 * 
 */
  StaticTuple_SET_ITEM(__pyx_v_value_and_refs, 1, ((PyObject *)__pyx_v_empty));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":543
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs             # <<<<<<<<<<<<<<
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_value_and_refs));
  __pyx_r = __pyx_v_value_and_refs;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":530
 *             return None
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_value_and_refs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_value_and_refs);
  __Pyx_XDECREF((PyObject *)__pyx_v_empty);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":545
 *         return value_and_refs
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
 */

static StaticTuple *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  StaticTuple *__pyx_v_item = 0;
  StaticTuple *__pyx_v_key = 0;
  StaticTuple *__pyx_v_value_and_refs = 0;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_record_to_item", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":552
 *         cdef StaticTuple value_and_refs
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)             # <<<<<<<<<<<<<<
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key(__pyx_v_record->sha1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":553
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)
 *         item = StaticTuple_New(2)             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":554
 *         key = _sha1_to_key(record.sha1)
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
 */
  Py_INCREF(((PyObject *)__pyx_v_key));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":555
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)             # <<<<<<<<<<<<<<
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
 */
  StaticTuple_SET_ITEM(__pyx_v_item, 0, ((PyObject *)__pyx_v_key));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":556
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":557
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item
 */
  Py_INCREF(((PyObject *)__pyx_v_value_and_refs));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":558
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)             # <<<<<<<<<<<<<<
 *         return item
 * 
 */
  StaticTuple_SET_ITEM(__pyx_v_item, 1, ((PyObject *)__pyx_v_value_and_refs));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":559
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item             # <<<<<<<<<<<<<<
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_item));
  __pyx_r = __pyx_v_item;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":545
 *         return value_and_refs
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_item", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_item);
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_value_and_refs);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":561
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
 */

static __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_lo;
  int __pyx_v_hi;
  int __pyx_v_mid;
  int __pyx_v_the_cmp;
  int __pyx_v_offset;
  CYTHON_UNUSED long __pyx_v_local_n_cmp;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_lookup_record", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":575
 *         # the offset array dropped us from 23us to 20us and 156 comparisions
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)             # <<<<<<<<<<<<<<
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 575, __pyx_L1_error)
  __pyx_v_offset = __pyx_t_1;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":576
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]             # <<<<<<<<<<<<<<
 *         hi = self.offsets[offset+1]
 *         if hi == 255:
 */
  __pyx_v_lo = (__pyx_v_self->offsets[__pyx_v_offset]);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":577
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]             # <<<<<<<<<<<<<<
 *         if hi == 255:
 *             # if hi == 255 that means we potentially ran off the end of the
 */
  __pyx_v_hi = (__pyx_v_self->offsets[(__pyx_v_offset + 1)]);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":578
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
 */
  __pyx_t_2 = ((__pyx_v_hi == 0xFF) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":583
 *             # note that if 'lo' == 255, that is ok, because we can start
 *             # searching from that part of the list.
 *             hi = self.num_records             # <<<<<<<<<<<<<<
 *         local_n_cmp = 0
 *         while lo < hi:
 */
    __pyx_t_1 = __pyx_v_self->num_records;
    __pyx_v_hi = __pyx_t_1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":578
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":584
 *             # searching from that part of the list.
 *             hi = self.num_records
 *         local_n_cmp = 0             # <<<<<<<<<<<<<<
 *         while lo < hi:
 *             mid = (lo + hi) // 2
 */
  __pyx_v_local_n_cmp = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":585
 *             hi = self.num_records
 *         local_n_cmp = 0
 *         while lo < hi:             # <<<<<<<<<<<<<<
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_lo < __pyx_v_hi) != 0);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":586
 *         local_n_cmp = 0
 *         while lo < hi:
 *             mid = (lo + hi) // 2             # <<<<<<<<<<<<<<
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
 */
    __pyx_v_mid = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":587
 *         while lo < hi:
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)             # <<<<<<<<<<<<<<
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 */
    __pyx_v_the_cmp = memcmp((__pyx_v_self->records[__pyx_v_mid]).sha1, __pyx_v_sha1, 20);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":588
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 */
    __pyx_t_2 = ((__pyx_v_the_cmp == 0) != 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":589
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
 *                 return &self.records[mid]             # <<<<<<<<<<<<<<
 *             elif the_cmp < 0:
 *                 lo = mid + 1
 */
      __pyx_r = (&(__pyx_v_self->records[__pyx_v_mid]));
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":588
 *             mid = (lo + hi) // 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":590
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
 */
    __pyx_t_2 = ((__pyx_v_the_cmp < 0) != 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":591
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 *                 lo = mid + 1             # <<<<<<<<<<<<<<
 *             else:
 *                 hi = mid
 */
      __pyx_v_lo = (__pyx_v_mid + 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":590
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
 */
      goto __pyx_L6;
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":593
 *                 lo = mid + 1
 *             else:
 *                 hi = mid             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    /*else*/ {
      __pyx_v_hi = __pyx_v_mid;
    }
    __pyx_L6:;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":594
 *             else:
 *                 hi = mid
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":561
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._lookup_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":596
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_2;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":599
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 */
  __pyx_t_1 = (__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":601
 *         if _key_to_sha1(key, sha1):
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *             if record != NULL:
 *                 self.last_key = key
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_2 == ((__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 601, __pyx_L1_error)
    __pyx_v_record = __pyx_t_2;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":602
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
 */
    __pyx_t_1 = ((__pyx_v_record != NULL) != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":603
 *             record = self._lookup_record(sha1)
 *             if record != NULL:
 *                 self.last_key = key             # <<<<<<<<<<<<<<
 *                 self.last_record = record
 *                 return True
 */
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      __Pyx_GOTREF(__pyx_v_self->last_key);
      __Pyx_DECREF(__pyx_v_self->last_key);
      __pyx_v_self->last_key = __pyx_v_key;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":604
 *             if record != NULL:
 *                 self.last_key = key
 *                 self.last_record = record             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
      __pyx_v_self->last_record = __pyx_v_record;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":605
 *                 self.last_key = key
 *                 self.last_record = record
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":602
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":599
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":606
 *                 self.last_record = record
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":596
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":608
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":611
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         record = NULL             # <<<<<<<<<<<<<<
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 */
  __pyx_v_record = NULL;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":612
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 */
  __pyx_t_2 = ((__pyx_v_self->last_record != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_key == __pyx_v_self->last_key);
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":613
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record             # <<<<<<<<<<<<<<
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 */
    __pyx_t_4 = __pyx_v_self->last_record;
    __pyx_v_record = __pyx_t_4;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":612
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":614
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 */
  __pyx_t_1 = (__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":615
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_4 == ((__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 615, __pyx_L1_error)
    __pyx_v_record = __pyx_t_4;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":614
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 */
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":616
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
 */
  __pyx_t_1 = ((__pyx_v_record == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":617
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))             # <<<<<<<<<<<<<<
 *         return self._record_to_value_and_refs(record)
 * 
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_key);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_key_r_is_not_present, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 617, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":616
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":618
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":608
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":620
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":621
 * 
 *     def __len__(self):
 *         return self.num_records             # <<<<<<<<<<<<<<
 * 
 *     def all_keys(self):
 */
  __pyx_r = __pyx_v_self->num_records;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":620
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":623
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_keys (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("all_keys", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":625
 *     def all_keys(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":626
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":627
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[__pyx_v_i]).sha1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 627, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":628
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def all_items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":623
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.all_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":630
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_items (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  StaticTuple *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("all_items", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":632
 *     def all_items(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":633
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":634
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])             # <<<<<<<<<<<<<<
 *             PyList_Append(result, item)
 *         return result
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_item(__pyx_v_self, (&(__pyx_v_self->records[__pyx_v_i])))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_item, ((StaticTuple *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":635
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_3 = PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_item)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 635, __pyx_L1_error)
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":636
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":630
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.all_items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":638
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
 */

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end) {
  char *__pyx_v_c_cur;
  int __pyx_v_num_records;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_count_records", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":643
 *         cdef int num_records
 * 
 *         c_cur = c_content             # <<<<<<<<<<<<<<
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
 */
  __pyx_v_c_cur = __pyx_v_c_content;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":644
 * 
 *         c_cur = c_content
 *         num_records = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 */
  __pyx_v_num_records = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":645
 *         c_cur = c_content
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:             # <<<<<<<<<<<<<<
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_c_cur != NULL) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_c_cur < __pyx_v_c_end) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":646
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);             # <<<<<<<<<<<<<<
 *             if c_cur == NULL:
 *                 break
 */
    __pyx_v_c_cur = ((char *)memchr(__pyx_v_c_cur, '\n', (__pyx_v_c_end - __pyx_v_c_cur)));

    /* "breezy/bzr/_btree_serializer_pyx.pyx":647
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
 */
    __pyx_t_1 = ((__pyx_v_c_cur == NULL) != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":648
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
 *                 break             # <<<<<<<<<<<<<<
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
 */
      goto __pyx_L4_break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":647
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":649
 *             if c_cur == NULL:
 *                 break
 *             c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *             num_records = num_records + 1
 *         return num_records
 */
    __pyx_v_c_cur = (__pyx_v_c_cur + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":650
 *                 break
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1             # <<<<<<<<<<<<<<
 *         return num_records
 * 
 */
    __pyx_v_num_records = (__pyx_v_num_records + 1);
  }
  __pyx_L4_break:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":651
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
 *         return num_records             # <<<<<<<<<<<<<<
 * 
 *     cdef _parse_bytes(self, data):
 */
  __pyx_r = __pyx_v_num_records;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":638
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":653
 *         return num_records
 * 
 *     cdef _parse_bytes(self, data):             # <<<<<<<<<<<<<<
 *         """Parse the bytes 'data' into content."""
 *         cdef char *c_bytes
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_data) {
  char *__pyx_v_c_bytes;
  char *__pyx_v_c_cur;
  char *__pyx_v_c_end;
  Py_ssize_t __pyx_v_n_bytes;
  int __pyx_v_num_records;
  int __pyx_v_entry;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_4;
  int __pyx_t_5;
  char *__pyx_t_6;
  __Pyx_RefNannySetupContext("_parse_bytes", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":663
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_data) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":664
 * 
 *         if not PyBytes_CheckExact(data):
 *             raise TypeError('We only support parsing byte strings.')             # <<<<<<<<<<<<<<
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 664, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":663
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyBytes_CheckExact(data):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":666
 *             raise TypeError('We only support parsing byte strings.')
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)             # <<<<<<<<<<<<<<
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
 */
  __pyx_v_n_bytes = PyBytes_GET_SIZE(__pyx_v_data);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":667
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)
 *         c_bytes = PyBytes_AS_STRING(data)             # <<<<<<<<<<<<<<
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 */
  __pyx_v_c_bytes = PyBytes_AS_STRING(__pyx_v_data);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":668
 *         n_bytes = PyBytes_GET_SIZE(data)
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes             # <<<<<<<<<<<<<<
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 */
  __pyx_v_c_end = (__pyx_v_c_bytes + __pyx_v_n_bytes);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":669
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
 */
  __pyx_t_1 = (strncmp(__pyx_v_c_bytes, ((char const *)"type=leaf\n"), 10) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":671
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
 */
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_data, 0, 10, NULL, NULL, &__pyx_slice__10, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_bytes_did_not_start_with_type_le, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":670
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"             # <<<<<<<<<<<<<<
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 670, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":669
 *         c_bytes = PyBytes_AS_STRING(data)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, b'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":672
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10             # <<<<<<<<<<<<<<
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
 */
  __pyx_v_c_cur = (__pyx_v_c_bytes + 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":673
 *                              % (data[:10],))
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)             # <<<<<<<<<<<<<<
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
 */
  __pyx_v_num_records = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_count_records(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":675
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *             # <<<<<<<<<<<<<<
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
 */
  __pyx_v_self->records = ((__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *)PyMem_Malloc((__pyx_v_num_records * ((sizeof(unsigned short)) + (sizeof(__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record))))));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":677
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records             # <<<<<<<<<<<<<<
 *         cur_record = self.records
 *         entry = 0
 */
  __pyx_v_self->num_records = __pyx_v_num_records;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":678
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
 *         cur_record = self.records             # <<<<<<<<<<<<<<
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 */
  __pyx_t_4 = __pyx_v_self->records;
  __pyx_v_cur_record = __pyx_t_4;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":679
 *         self.num_records = num_records
 *         cur_record = self.records
 *         entry = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 */
  __pyx_v_entry = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":680
 *         cur_record = self.records
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:             # <<<<<<<<<<<<<<
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
 */
  while (1) {
    __pyx_t_5 = ((__pyx_v_c_cur != NULL) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = ((__pyx_v_c_cur < __pyx_v_c_end) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = ((__pyx_v_entry < __pyx_v_num_records) != 0);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":681
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)             # <<<<<<<<<<<<<<
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_one_entry(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end, __pyx_v_cur_record); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(0, 681, __pyx_L1_error)
    __pyx_v_c_cur = __pyx_t_6;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":682
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1             # <<<<<<<<<<<<<<
 *             entry = entry + 1
 *         if (entry != self.num_records
 */
    __pyx_v_cur_record = (__pyx_v_cur_record + 1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":683
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
 *             entry = entry + 1             # <<<<<<<<<<<<<<
 *         if (entry != self.num_records
 *             or c_cur != c_end
 */
    __pyx_v_entry = (__pyx_v_entry + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":684
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  __pyx_t_5 = ((__pyx_v_entry != __pyx_v_self->num_records) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":685
 *             entry = entry + 1
 *         if (entry != self.num_records
 *             or c_cur != c_end             # <<<<<<<<<<<<<<
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')
 */
  __pyx_t_5 = ((__pyx_v_c_cur != __pyx_v_c_end) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":686
 *         if (entry != self.num_records
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):             # <<<<<<<<<<<<<<
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
 */
  __pyx_t_5 = ((__pyx_v_cur_record != (__pyx_v_self->records + __pyx_v_self->num_records)) != 0);
  __pyx_t_1 = __pyx_t_5;
  __pyx_L11_bool_binop_done:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":684
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":687
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')             # <<<<<<<<<<<<<<
 *         # Pass 3: build the offset map
 *         self._compute_common()
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 687, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":684
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":689
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
 *         self._compute_common()             # <<<<<<<<<<<<<<
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_compute_common(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":653
 *         return num_records
 * 
 *     cdef _parse_bytes(self, data):             # <<<<<<<<<<<<<<
 *         """Parse the bytes 'data' into content."""
 *         cdef char *c_bytes
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":691
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
 */

static char *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record) {
  char *__pyx_v_c_next;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_parse_one_entry", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":699
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 */
  __pyx_t_1 = (strncmp(__pyx_v_c_cur, ((char const *)"sha1:"), 5) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":701
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 */
    __pyx_t_2 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_c_cur, 10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_line_did_not_start_with_sha1_r, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":700
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'             # <<<<<<<<<<<<<<
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 700, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":699
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":702
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5             # <<<<<<<<<<<<<<
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 */
  __pyx_v_c_cur = (__pyx_v_c_cur + 5);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":703
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)             # <<<<<<<<<<<<<<
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 */
  __pyx_v_c_next = ((char *)memchr(__pyx_v_c_cur, '\x00', (__pyx_v_c_end - __pyx_v_c_cur)));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":704
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 */
  __pyx_t_4 = ((__pyx_v_c_next == NULL) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next - __pyx_v_c_cur) != 40) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":705
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')             # <<<<<<<<<<<<<<
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 705, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":704
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":706
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 */
  __pyx_t_1 = ((!(__pyx_f_6breezy_3bzr_21_btree_serializer_pyx__unhexlify_sha1(__pyx_v_c_cur, __pyx_v_cur_record->sha1) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":707
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 707, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":706
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":708
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":709
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 */
  __pyx_t_1 = (((__pyx_v_c_cur[0]) != '\x00') != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":710
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 710, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":709
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":711
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_cur + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":712
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
 */
  __pyx_v_cur_record->block_offset = strtoull(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":713
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":714
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 714, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":713
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":715
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":716
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 */
  __pyx_v_cur_record->block_length = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":717
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L13_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":718
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 718, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":717
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":719
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":720
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 */
  __pyx_v_cur_record->record_start = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":721
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":722
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 722, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":721
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":723
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":724
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
 */
  __pyx_v_cur_record->record_end = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":725
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != '\n') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L19_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":726
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         return c_cur
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 726, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":725
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":727
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         return c_cur
 * 
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":728
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 *         return c_cur             # <<<<<<<<<<<<<<
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:
 */
  __pyx_r = __pyx_v_c_cur;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":691
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_one_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":730
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
 */

static int __pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_this_offset;
  unsigned int __pyx_v_as_uint;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_offset_for_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":734
 *         cdef int this_offset
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)             # <<<<<<<<<<<<<<
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset
 */
  __pyx_v_as_uint = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint(__pyx_v_sha1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":735
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF             # <<<<<<<<<<<<<<
 *         return this_offset
 * 
 */
  __pyx_v_this_offset = ((__pyx_v_as_uint >> __pyx_v_self->common_shift) & 0xFF);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":736
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset             # <<<<<<<<<<<<<<
 * 
 *     def _get_offset_for_sha1(self, sha1):
 */
  __pyx_r = __pyx_v_this_offset;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":730
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":738
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, PyObject *__pyx_v_sha1); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, PyObject *__pyx_v_sha1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offset_for_sha1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_sha1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_get_offset_for_sha1", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":739
 * 
 *     def _get_offset_for_sha1(self, sha1):
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))             # <<<<<<<<<<<<<<
 * 
 *     cdef _compute_common(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, PyBytes_AS_STRING(__pyx_v_sha1)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":738
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offset_for_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":741
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  unsigned int __pyx_v_first;
  unsigned int __pyx_v_this;
  unsigned int __pyx_v_common_mask;
  unsigned char __pyx_v_common_shift;
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_v_this_offset;
  int __pyx_v_max_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_compute_common", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":755
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 */
  __pyx_t_1 = ((__pyx_v_self->num_records < 2) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":758
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 *             self.common_shift = 24             # <<<<<<<<<<<<<<
 *         else:
 *             common_mask = 0xFFFFFFFF
 */
    __pyx_v_self->common_shift = 24;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":755
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":760
 *             self.common_shift = 24
 *         else:
 *             common_mask = 0xFFFFFFFF             # <<<<<<<<<<<<<<
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
 */
  /*else*/ {
    __pyx_v_common_mask = 0xFFFFFFFF;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":761
 *         else:
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
 */
    __pyx_v_first = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[0]).sha1);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":762
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:             # <<<<<<<<<<<<<<
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
 */
    __pyx_t_2 = __pyx_v_self->num_records;
    for (__pyx_v_i = 0+1; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":763
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
 */
      __pyx_v_this = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[__pyx_v_i]).sha1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":764
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask             # <<<<<<<<<<<<<<
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
 */
      __pyx_v_common_mask = ((~(__pyx_v_first ^ __pyx_v_this)) & __pyx_v_common_mask);
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":765
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24             # <<<<<<<<<<<<<<
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
 */
    __pyx_v_common_shift = 24;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":766
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:             # <<<<<<<<<<<<<<
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_common_mask); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyNumber_And(__pyx_t_3, __pyx_int_2147483648); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {
      } else {
        __pyx_t_1 = __pyx_t_5;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_5 = ((__pyx_v_common_shift > 0) != 0);
      __pyx_t_1 = __pyx_t_5;
      __pyx_L8_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":767
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1             # <<<<<<<<<<<<<<
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
 */
      __pyx_v_common_mask = (__pyx_v_common_mask << 1);

      /* "breezy/bzr/_btree_serializer_pyx.pyx":768
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1             # <<<<<<<<<<<<<<
 *             self.common_shift = common_shift
 *         offset = 0
 */
      __pyx_v_common_shift = (__pyx_v_common_shift - 1);
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":769
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift             # <<<<<<<<<<<<<<
 *         offset = 0
 *         max_offset = self.num_records
 */
    __pyx_v_self->common_shift = __pyx_v_common_shift;
  }
  __pyx_L3:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":770
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
 *         offset = 0             # <<<<<<<<<<<<<<
 *         max_offset = self.num_records
 *         # We cap this loop at 254 records. All the other offsets just get
 */
  __pyx_v_offset = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":771
 *             self.common_shift = common_shift
 *         offset = 0
 *         max_offset = self.num_records             # <<<<<<<<<<<<<<
 *         # We cap this loop at 254 records. All the other offsets just get
 *         # filled with 0xff as the singleton saying 'too many'.
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  __pyx_v_max_offset = __pyx_t_2;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":776
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 */
  __pyx_t_1 = ((__pyx_v_max_offset > 0xFF) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":777
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:
 *             max_offset = 255             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 */
    __pyx_v_max_offset = 0xFF;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":776
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":778
 *         if max_offset > 255:
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:             # <<<<<<<<<<<<<<
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
 */
  __pyx_t_2 = __pyx_v_max_offset;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":779
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, (__pyx_v_self->records[__pyx_v_i]).sha1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 779, __pyx_L1_error)
    __pyx_v_this_offset = __pyx_t_6;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":780
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:             # <<<<<<<<<<<<<<
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
 */
    while (1) {
      __pyx_t_1 = ((__pyx_v_offset <= __pyx_v_this_offset) != 0);
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":781
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i             # <<<<<<<<<<<<<<
 *                 offset = offset + 1
 *         while offset < 257:
 */
      (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_i;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":782
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
 *                 offset = offset + 1             # <<<<<<<<<<<<<<
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
 */
      __pyx_v_offset = (__pyx_v_offset + 1);
    }
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":783
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
 *         while offset < 257:             # <<<<<<<<<<<<<<
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_offset < 0x101) != 0);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":784
 *                 offset = offset + 1
 *         while offset < 257:
 *             self.offsets[offset] = max_offset             # <<<<<<<<<<<<<<
 *             offset = offset + 1
 * 
 */
    (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_max_offset;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":785
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1             # <<<<<<<<<<<<<<
 * 
 *     def _get_offsets(self):
 */
    __pyx_v_offset = (__pyx_v_offset + 1);
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":741
 *         return self._offset_for_sha1(PyBytes_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._compute_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":787
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offsets (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_get_offsets", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":789
 *     def _get_offsets(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":790
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < 257:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, self.offsets[i])
 *         return result
 */
  for (__pyx_v_i = 0; __pyx_v_i < 0x101; __pyx_v_i++) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":791
 *         result = []
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_unsigned_char((__pyx_v_self->offsets[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":792
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":787
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":489
 * 
 *     cdef gc_chk_sha1_record *records
 *     cdef public object last_key             # <<<<<<<<<<<<<<
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->last_key);
  __pyx_r = __pyx_v_self->last_key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":491
 *     cdef public object last_key
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records             # <<<<<<<<<<<<<<
 *     # This is the number of bits to shift to get to the interesting byte. A
 *     # value of 24 means that the very first byte changes across all keys.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->num_records); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L1_error)
  __pyx_v_self->num_records = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":497
 *     # ignore. 0 means that at least the first 3 bytes are identical, though
 *     # that is going to be very rare
 *     cdef public unsigned char common_shift             # <<<<<<<<<<<<<<
 *     # This maps an interesting byte to the first record that matches.
 *     # Equivalent to bisect.bisect_left(self.records, sha1), though only taking
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->common_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_v_self->common_shift = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_20__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_22__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":795
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk[] = "Parse into a format optimized for chk records.";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk = {"_parse_into_chk", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_into_chk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, 1); __PYX_ERR(0, 795, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, 2); __PYX_ERR(0, 795, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parse_into_chk") < 0)) __PYX_ERR(0, 795, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_bytes = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 795, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(__pyx_self, __pyx_v_bytes, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_parse_into_chk", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":797
 * def _parse_into_chk(bytes, key_length, ref_list_length):
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1             # <<<<<<<<<<<<<<
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_key_length, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 797, __pyx_L1_error)
    }
  }
  #endif

  /* "breezy/bzr/_btree_serializer_pyx.pyx":798
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 *     assert ref_list_length == 0             # <<<<<<<<<<<<<<
 *     return GCCHKSHA1LeafNode(bytes)
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ref_list_length, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 798, __pyx_L1_error)
    }
  }
  #endif

  /* "breezy/bzr/_btree_serializer_pyx.pyx":799
 *     assert key_length == 1
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode), __pyx_v_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":795
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_btree_serializer_pyx.pyx":802
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node[] = "Convert a node into the serialized form.\n\n    :param node: A tuple representing a node:\n        (index, key_tuple, value, references)\n    :param reference_lists: Does this index have reference lists?\n    :return: (string_key, flattened)\n        string_key  The serialized key for referencing this node\n        flattened   A string with the serialized form for the contents\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node = {"_flatten_node", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_reference_lists = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_flatten_node (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_reference_lists,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reference_lists)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, 1); __PYX_ERR(0, 802, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_flatten_node") < 0)) __PYX_ERR(0, 802, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_node = values[0];
    __pyx_v_reference_lists = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 802, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(__pyx_self, __pyx_v_node, __pyx_v_reference_lists);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists) {
  int __pyx_v_have_reference_lists;
  Py_ssize_t __pyx_v_flat_len;
  Py_ssize_t __pyx_v_key_len;
  Py_ssize_t __pyx_v_node_len;
  char *__pyx_v_value;
  Py_ssize_t __pyx_v_value_len;
  char *__pyx_v_out;
  Py_ssize_t __pyx_v_refs_len;
  Py_ssize_t __pyx_v_next_len;
  int __pyx_v_first_ref_list;
  int __pyx_v_first_reference;
  int __pyx_v_i;
  Py_ssize_t __pyx_v_ref_bit_len;
  PyObject *__pyx_v_string_key = NULL;
  PyObject *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  PyObject *__pyx_v_reference = NULL;
  PyObject *__pyx_v_ref_bit = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  char *__pyx_t_15;
  __Pyx_RefNannySetupContext("_flatten_node", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":826
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 */
  __pyx_t_2 = ((!(PyTuple_CheckExact(__pyx_v_node) != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((!(StaticTuple_CheckExact(__pyx_v_node) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":828
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))             # <<<<<<<<<<<<<<
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 */
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_We_expected_a_tuple_or_StaticTup, ((PyObject *)Py_TYPE(__pyx_v_node))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "breezy/bzr/_btree_serializer_pyx.pyx":827
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'             # <<<<<<<<<<<<<<
 *             % type(node))
 *     node_len = len(node)
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 827, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":826
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":829
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 *     node_len = len(node)             # <<<<<<<<<<<<<<
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 829, __pyx_L1_error)
  __pyx_v_node_len = __pyx_t_5;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":830
 *             % type(node))
 *     node_len = len(node)
 *     have_reference_lists = reference_lists             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         if node_len != 4:
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_reference_lists); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 830, __pyx_L1_error)
  __pyx_v_have_reference_lists = __pyx_t_6;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":831
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 */
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":832
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 */
    __pyx_t_1 = ((__pyx_v_node_len != 4) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":834
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))             # <<<<<<<<<<<<<<
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 */
      __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 834, __pyx_L1_error)
      __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_With_ref_lists_we_expected_4_ent, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":833
 *     if have_reference_lists:
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'             # <<<<<<<<<<<<<<
 *                 % len(node))
 *     elif node_len < 3:
 */
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 833, __pyx_L1_error)

      /* "breezy/bzr/_btree_serializer_pyx.pyx":832
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":831
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 */
    goto __pyx_L6;
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":835
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
 */
  __pyx_t_1 = ((__pyx_v_node_len < 3) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":837
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))             # <<<<<<<<<<<<<<
 *     # TODO: We can probably do better than string.join(), namely
 *     #       when key has only 1 item, we can just grab that string
 */
    __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 837, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Without_ref_lists_we_need_at_lea, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":836
 *                 % len(node))
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'             # <<<<<<<<<<<<<<
 *             % len(node))
 *     # TODO: We can probably do better than string.join(), namely
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 836, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 836, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":835
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
 */
  }
  __pyx_L6:;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":848
 *     #       We *could* do more work on our own, and grab the actual items
 *     #       lists. For now, just ask people to use a better compiler. :)
 *     string_key = b'\0'.join(node[1])             # <<<<<<<<<<<<<<
 * 
 *     # TODO: instead of using string joins, precompute the final string length,
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_node, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyBytes_Join(__pyx_kp_b__20, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_string_key = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":862
 *     # ref := BYTES (NULL BYTES)*
 *     # value := BYTES
 *     refs_len = 0             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
 */
  __pyx_v_refs_len = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":863
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 */
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":865
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]             # <<<<<<<<<<<<<<
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_node, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ref_lists = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":866
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function             # <<<<<<<<<<<<<<
 *         if next_len > 0:
 *             # If there are no nodes, we don't need to do any work
 */
    __pyx_t_5 = PyObject_Length(__pyx_v_ref_lists); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 866, __pyx_L1_error)
    __pyx_v_next_len = __pyx_t_5;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":867
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 */
    __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_btree_serializer_pyx.pyx":871
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 */
      __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

      /* "breezy/bzr/_btree_serializer_pyx.pyx":872
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
 */
      if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
        __pyx_t_3 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 872, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 872, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 872, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 872, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_7(__pyx_t_3);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 872, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":873
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)             # <<<<<<<<<<<<<<
 *                 if next_len > 0:
 *                     # We will need (len - 1) '\r' characters to separate the
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_ref_list); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 873, __pyx_L1_error)
        __pyx_v_next_len = __pyx_t_8;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":874
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 */
        __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":877
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)
 */
          __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

          /* "breezy/bzr/_btree_serializer_pyx.pyx":878
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 */
          if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
            __pyx_t_4 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_4); __pyx_t_8 = 0;
            __pyx_t_9 = NULL;
          } else {
            __pyx_t_8 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 878, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 878, __pyx_L1_error)
          }
          for (;;) {
            if (likely(!__pyx_t_9)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_10 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 878, __pyx_L1_error)
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 878, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              } else {
                if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 878, __pyx_L1_error)
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 878, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              }
            } else {
              __pyx_t_10 = __pyx_t_9(__pyx_t_4);
              if (unlikely(!__pyx_t_10)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 878, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_10);
            }
            __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "breezy/bzr/_btree_serializer_pyx.pyx":879
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            __pyx_t_2 = ((!(PyTuple_CheckExact(__pyx_v_reference) != 0)) != 0);
            if (__pyx_t_2) {
            } else {
              __pyx_t_1 = __pyx_t_2;
              goto __pyx_L16_bool_binop_done;
            }

            /* "breezy/bzr/_btree_serializer_pyx.pyx":880
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):             # <<<<<<<<<<<<<<
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %r'
 */
            __pyx_t_2 = ((!(StaticTuple_CheckExact(__pyx_v_reference) != 0)) != 0);
            __pyx_t_1 = __pyx_t_2;
            __pyx_L16_bool_binop_done:;

            /* "breezy/bzr/_btree_serializer_pyx.pyx":879
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            if (unlikely(__pyx_t_1)) {

              /* "breezy/bzr/_btree_serializer_pyx.pyx":883
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))             # <<<<<<<<<<<<<<
 *                         next_len = len(reference)
 *                         if next_len > 0:
 */
              __pyx_t_10 = __Pyx_PyString_FormatSafe(__pyx_kp_s_We_expect_references_to_be_tuple, ((PyObject *)Py_TYPE(__pyx_v_reference))); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 883, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);

              /* "breezy/bzr/_btree_serializer_pyx.pyx":881
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(             # <<<<<<<<<<<<<<
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))
 */
              __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 881, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_Raise(__pyx_t_11, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __PYX_ERR(0, 881, __pyx_L1_error)

              /* "breezy/bzr/_btree_serializer_pyx.pyx":879
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            }

            /* "breezy/bzr/_btree_serializer_pyx.pyx":884
 *                                 'We expect references to be tuples not: %r'
 *                                 % type(reference))
 *                         next_len = len(reference)             # <<<<<<<<<<<<<<
 *                         if next_len > 0:
 *                             # We will need (len - 1) '\x00' characters to
 */
            __pyx_t_12 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 884, __pyx_L1_error)
            __pyx_v_next_len = __pyx_t_12;

            /* "breezy/bzr/_btree_serializer_pyx.pyx":885
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 */
            __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
            if (__pyx_t_1) {

              /* "breezy/bzr/_btree_serializer_pyx.pyx":888
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):
 */
              __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

              /* "breezy/bzr/_btree_serializer_pyx.pyx":889
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(
 */
              if (likely(PyList_CheckExact(__pyx_v_reference)) || PyTuple_CheckExact(__pyx_v_reference)) {
                __pyx_t_11 = __pyx_v_reference; __Pyx_INCREF(__pyx_t_11); __pyx_t_12 = 0;
                __pyx_t_13 = NULL;
              } else {
                __pyx_t_12 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_v_reference); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 889, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_11);
                __pyx_t_13 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 889, __pyx_L1_error)
              }
              for (;;) {
                if (likely(!__pyx_t_13)) {
                  if (likely(PyList_CheckExact(__pyx_t_11))) {
                    if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_11)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_10 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_12); __Pyx_INCREF(__pyx_t_10); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 889, __pyx_L1_error)
                    #else
                    __pyx_t_10 = PySequence_ITEM(__pyx_t_11, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 889, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    #endif
                  } else {
                    if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_12); __Pyx_INCREF(__pyx_t_10); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 889, __pyx_L1_error)
                    #else
                    __pyx_t_10 = PySequence_ITEM(__pyx_t_11, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 889, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_10);
                    #endif
                  }
                } else {
                  __pyx_t_10 = __pyx_t_13(__pyx_t_11);
                  if (unlikely(!__pyx_t_10)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 889, __pyx_L1_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_10);
                }
                __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_10);
                __pyx_t_10 = 0;

                /* "breezy/bzr/_btree_serializer_pyx.pyx":890
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
 */
                __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_ref_bit) != 0)) != 0);
                if (unlikely(__pyx_t_1)) {

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":893
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))             # <<<<<<<<<<<<<<
 *                                 refs_len = refs_len + PyBytes_GET_SIZE(ref_bit)
 * 
 */
                  __pyx_t_10 = __Pyx_PyString_FormatSafe(__pyx_kp_s_We_expect_reference_bits_to_be_b, ((PyObject *)Py_TYPE(__pyx_v_ref_bit))); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 893, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_10);

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":891
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(             # <<<<<<<<<<<<<<
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))
 */
                  __pyx_t_14 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 891, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
                  __Pyx_Raise(__pyx_t_14, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __PYX_ERR(0, 891, __pyx_L1_error)

                  /* "breezy/bzr/_btree_serializer_pyx.pyx":890
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyBytes_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError(
 *                                         'We expect reference bits to be bytes'
 */
                }

                /* "breezy/bzr/_btree_serializer_pyx.pyx":894
 *                                         'We expect reference bits to be bytes'
 *                                         ' not: %r' % type(ref_bit))
 *                                 refs_len = refs_len + PyBytes_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 * 
 *     # So we have the (key NULL refs NULL value LF)
 */
                __pyx_v_refs_len = (__pyx_v_refs_len + PyBytes_GET_SIZE(__pyx_v_ref_bit));

                /* "breezy/bzr/_btree_serializer_pyx.pyx":889
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyBytes_CheckExact(ref_bit):
 *                                     raise TypeError(
 */
              }
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "breezy/bzr/_btree_serializer_pyx.pyx":885
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 */
            }

            /* "breezy/bzr/_btree_serializer_pyx.pyx":878
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":874
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 */
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":872
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":867
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 */
    }

    /* "breezy/bzr/_btree_serializer_pyx.pyx":863
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":897
 * 
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyBytes_Size(string_key)             # <<<<<<<<<<<<<<
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):
 */
  __pyx_t_5 = PyBytes_Size(__pyx_v_string_key); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 897, __pyx_L1_error)
  __pyx_v_key_len = __pyx_t_5;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":898
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_node, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_val = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":899
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_val) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":900
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))             # <<<<<<<<<<<<<<
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)
 */
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Expected_bytes_for_value_not_r, ((PyObject *)Py_TYPE(__pyx_v_val))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 900, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 900, __pyx_L1_error)

    /* "breezy/bzr/_btree_serializer_pyx.pyx":899
 *     key_len = PyBytes_Size(string_key)
 *     val = node[2]
 *     if not PyBytes_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":901
 *     if not PyBytes_CheckExact(val):
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)             # <<<<<<<<<<<<<<
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 */
  __pyx_v_value = PyBytes_AS_STRING(__pyx_v_val);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":902
 *         raise TypeError('Expected bytes for value not: %r' % type(val))
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)             # <<<<<<<<<<<<<<
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
 */
  __pyx_v_value_len = PyBytes_GET_SIZE(__pyx_v_val);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":903
 *     value = PyBytes_AS_STRING(val)
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)             # <<<<<<<<<<<<<<
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
 */
  __pyx_v_flat_len = (((((__pyx_v_key_len + 1) + __pyx_v_refs_len) + 1) + __pyx_v_value_len) + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":904
 *     value_len = PyBytes_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)             # <<<<<<<<<<<<<<
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)
 */
  __pyx_t_4 = PyBytes_FromStringAndSize(NULL, __pyx_v_flat_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_line = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":906
 *     line = PyBytes_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)             # <<<<<<<<<<<<<<
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len
 */
  __pyx_t_15 = PyBytes_AsString(__pyx_v_line); if (unlikely(__pyx_t_15 == ((char *)NULL))) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_v_out = __pyx_t_15;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":907
 *     # Get a pointer to the new buffer
 *     out = PyBytes_AsString(line)
 *     memcpy(out, PyBytes_AsString(string_key), key_len)             # <<<<<<<<<<<<<<
 *     out = out + key_len
 *     out[0] = c'\0'
 */
  __pyx_t_15 = PyBytes_AsString(__pyx_v_string_key); if (unlikely(__pyx_t_15 == ((char *)NULL))) __PYX_ERR(0, 907, __pyx_L1_error)
  (void)(memcpy(__pyx_v_out, __pyx_t_15, __pyx_v_key_len));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":908
 *     out = PyBytes_AsString(line)
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out + 1
 */
  __pyx_v_out = (__pyx_v_out + __pyx_v_key_len);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":909
 *     memcpy(out, PyBytes_AsString(string_key), key_len)
 *     out = out + key_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out + 1
 *     if refs_len > 0:
 */
  (__pyx_v_out[0]) = '\x00';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":910
 *     out = out + key_len
 *     out[0] = c'\0'
 *     out = out + 1             # <<<<<<<<<<<<<<
 *     if refs_len > 0:
 *         first_ref_list = 1
 */
  __pyx_v_out = (__pyx_v_out + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":911
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 */
  __pyx_t_1 = ((__pyx_v_refs_len > 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_btree_serializer_pyx.pyx":912
 *     out = out + 1
 *     if refs_len > 0:
 *         first_ref_list = 1             # <<<<<<<<<<<<<<
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
 */
    __pyx_v_first_ref_list = 1;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":913
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 */
    if (unlikely(!__pyx_v_ref_lists)) { __Pyx_RaiseUnboundLocalError("ref_lists"); __PYX_ERR(0, 913, __pyx_L1_error) }
    if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
      __pyx_t_4 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 913, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 913, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 913, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 913, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 913, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 913, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 913, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":914
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
 */
      __pyx_t_1 = ((__pyx_v_first_ref_list == 0) != 0);
      if (__pyx_t_1) {

        /* "breezy/bzr/_btree_serializer_pyx.pyx":915
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'             # <<<<<<<<<<<<<<
 *                 out = out + 1
 *             first_ref_list = 0
 */
        (__pyx_v_out[0]) = '\t';

        /* "breezy/bzr/_btree_serializer_pyx.pyx":916
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 *                 out = out + 1             # <<<<<<<<<<<<<<
 *             first_ref_list = 0
 *             first_reference = 1
 */
        __pyx_v_out = (__pyx_v_out + 1);

        /* "breezy/bzr/_btree_serializer_pyx.pyx":914
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
 */
      }

      /* "breezy/bzr/_btree_serializer_pyx.pyx":917
 *                 out[0] = c'\t'
 *                 out = out + 1
 *             first_ref_list = 0             # <<<<<<<<<<<<<<
 *             first_reference = 1
 *             for reference in ref_list:
 */
      __pyx_v_first_ref_list = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":918
 *                 out = out + 1
 *             first_ref_list = 0
 *             first_reference = 1             # <<<<<<<<<<<<<<
 *             for reference in ref_list:
 *                 if first_reference == 0:
 */
      __pyx_v_first_reference = 1;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":919
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 */
      if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
        __pyx_t_3 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 919, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 919, __pyx_L1_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 919, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_11); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 919, __pyx_L1_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 919, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          }
        } else {
          __pyx_t_11 = __pyx_t_9(__pyx_t_3);
          if (unlikely(!__pyx_t_11)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 919, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":920
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
 */
        __pyx_t_1 = ((__pyx_v_first_reference == 0) != 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":921
 *             for reference in ref_list:
 *                 if first_reference == 0:
 *                     out[0] = c'\r'             # <<<<<<<<<<<<<<
 *                     out = out + 1
 *                 first_reference = 0
 */
          (__pyx_v_out[0]) = '\r';

          /* "breezy/bzr/_btree_serializer_pyx.pyx":922
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 *                     out = out + 1             # <<<<<<<<<<<<<<
 *                 first_reference = 0
 *                 next_len = len(reference)
 */
          __pyx_v_out = (__pyx_v_out + 1);

          /* "breezy/bzr/_btree_serializer_pyx.pyx":920
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
 */
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":923
 *                     out[0] = c'\r'
 *                     out = out + 1
 *                 first_reference = 0             # <<<<<<<<<<<<<<
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 */
        __pyx_v_first_reference = 0;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":924
 *                     out = out + 1
 *                 first_reference = 0
 *                 next_len = len(reference)             # <<<<<<<<<<<<<<
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
 */
        __pyx_t_12 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 924, __pyx_L1_error)
        __pyx_v_next_len = __pyx_t_12;

        /* "breezy/bzr/_btree_serializer_pyx.pyx":925
 *                 first_reference = 0
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:             # <<<<<<<<<<<<<<
 *                     if i != 0:
 *                         out[0] = c'\x00'
 */
        __pyx_t_12 = __pyx_v_next_len;
        for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_12; __pyx_v_i++) {

          /* "breezy/bzr/_btree_serializer_pyx.pyx":926
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
 */
          __pyx_t_1 = ((__pyx_v_i != 0) != 0);
          if (__pyx_t_1) {

            /* "breezy/bzr/_btree_serializer_pyx.pyx":927
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
 *                         out[0] = c'\x00'             # <<<<<<<<<<<<<<
 *                         out = out + 1
 *                     ref_bit = reference[i]
 */
            (__pyx_v_out[0]) = '\x00';

            /* "breezy/bzr/_btree_serializer_pyx.pyx":928
 *                     if i != 0:
 *                         out[0] = c'\x00'
 *                         out = out + 1             # <<<<<<<<<<<<<<
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 */
            __pyx_v_out = (__pyx_v_out + 1);

            /* "breezy/bzr/_btree_serializer_pyx.pyx":926
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
 */
          }

          /* "breezy/bzr/_btree_serializer_pyx.pyx":929
 *                         out[0] = c'\x00'
 *                         out = out + 1
 *                     ref_bit = reference[i]             # <<<<<<<<<<<<<<
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 */
          __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_reference, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 929, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_11);
          __pyx_t_11 = 0;

          /* "breezy/bzr/_btree_serializer_pyx.pyx":930
 *                         out = out + 1
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
 */
          __pyx_v_ref_bit_len = PyBytes_GET_SIZE(__pyx_v_ref_bit);

          /* "breezy/bzr/_btree_serializer_pyx.pyx":931
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)             # <<<<<<<<<<<<<<
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
 */
          (void)(memcpy(__pyx_v_out, PyBytes_AS_STRING(__pyx_v_ref_bit), __pyx_v_ref_bit_len));

          /* "breezy/bzr/_btree_serializer_pyx.pyx":932
 *                     ref_bit_len = PyBytes_GET_SIZE(ref_bit)
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out  + 1
 */
          __pyx_v_out = (__pyx_v_out + __pyx_v_ref_bit_len);
        }

        /* "breezy/bzr/_btree_serializer_pyx.pyx":919
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/bzr/_btree_serializer_pyx.pyx":913
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/bzr/_btree_serializer_pyx.pyx":911
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 */
  }

  /* "breezy/bzr/_btree_serializer_pyx.pyx":933
 *                     memcpy(out, PyBytes_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out  + 1
 *     memcpy(out, value, value_len)
 */
  (__pyx_v_out[0]) = '\x00';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":934
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
 *     out = out  + 1             # <<<<<<<<<<<<<<
 *     memcpy(out, value, value_len)
 *     out = out + value_len
 */
  __pyx_v_out = (__pyx_v_out + 1);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":935
 *     out[0] = c'\0'
 *     out = out  + 1
 *     memcpy(out, value, value_len)             # <<<<<<<<<<<<<<
 *     out = out + value_len
 *     out[0] = c'\n'
 */
  (void)(memcpy(__pyx_v_out, __pyx_v_value, __pyx_v_value_len));

  /* "breezy/bzr/_btree_serializer_pyx.pyx":936
 *     out = out  + 1
 *     memcpy(out, value, value_len)
 *     out = out + value_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\n'
 *     return string_key, line
 */
  __pyx_v_out = (__pyx_v_out + __pyx_v_value_len);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":937
 *     memcpy(out, value, value_len)
 *     out = out + value_len
 *     out[0] = c'\n'             # <<<<<<<<<<<<<<
 *     return string_key, line
 */
  (__pyx_v_out[0]) = '\n';

  /* "breezy/bzr/_btree_serializer_pyx.pyx":938
 *     out = out + value_len
 *     out[0] = c'\n'
 *     return string_key, line             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_string_key);
  __Pyx_GIVEREF(__pyx_v_string_key);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_string_key);
  __Pyx_INCREF(__pyx_v_line);
  __Pyx_GIVEREF(__pyx_v_line);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_line);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":802
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_string_key);
  __Pyx_XDECREF(__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF(__pyx_v_reference);
  __Pyx_XDECREF(__pyx_v_ref_bit);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser = {"__pyx_unpickle_BTreeLeafParser", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BTreeLeafParser", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BTreeLeafParser", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_BTreeLeafParser") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BTreeLeafParser", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_btree_serializer_pyx_14__pyx_unpickle_BTreeLeafParser(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9b802a2:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x9b802a2) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9b802a2:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x9b802a2:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x9b, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x9b802a2:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x9b802a2 = (_cur_str, _end_str, _header_found, _start, data, key_length, keys, ref_list_length))" % __pyx_checksum)
 *     __pyx_result = BTreeLeafParser.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6breezy_3bzr_21_btree_serializer_pyx___pyx_unpickle_BTreeLeafParser__set_state(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BTreeLeafParser__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_cur_str = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_end_str = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_header_found = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_start = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->data);
  __Pyx_DECREF(__pyx_v___pyx_result->data);
  __pyx_v___pyx_result->data = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->key_length = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->keys);
  __Pyx_DECREF(__pyx_v___pyx_result->keys);
  __pyx_v___pyx_result->keys = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->ref_list_length = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 8) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[8])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[8])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_BTreeLeafParser__set_state(<BTreeLeafParser> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BTreeLeafParser__set_state(BTreeLeafParser __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._cur_str = __pyx_state[0]; __pyx_result._end_str = __pyx_state[1]; __pyx_result._header_found = __pyx_state[2]; __pyx_result._start = __pyx_state[3]; __pyx_result.data = __pyx_state[4]; __pyx_result.key_length = __pyx_state[5]; __pyx_result.keys = __pyx_state[6]; __pyx_result.ref_list_length = __pyx_state[7]
 *     if len(__pyx_state) > 8 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy.bzr._btree_serializer_pyx.__pyx_unpickle_BTreeLeafParser__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
 */

static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *__pyx_v_s, int __pyx_v_c, size_t __pyx_v_n) {
  char *__pyx_v_pos;
  char *__pyx_v_start;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_my_memrchr", 0);

  /* "_str_helpers.pxd":32
 *     cdef char *start
 * 
 *     start = <char*>s             # <<<<<<<<<<<<<<
 *     pos = start + n - 1
 *     while pos >= start:
 */
  __pyx_v_start = ((char *)__pyx_v_s);

  /* "_str_helpers.pxd":33
 * 
 *     start = <char*>s
 *     pos = start + n - 1             # <<<<<<<<<<<<<<
 *     while pos >= start:
 *         if pos[0] == c:
 */
  __pyx_v_pos = ((__pyx_v_start + __pyx_v_n) - 1);

  /* "_str_helpers.pxd":34
 *     start = <char*>s
 *     pos = start + n - 1
 *     while pos >= start:             # <<<<<<<<<<<<<<
 *         if pos[0] == c:
 *             return <void*>pos
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_pos >= __pyx_v_start) != 0);
    if (!__pyx_t_1) break;

    /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    __pyx_t_1 = (((__pyx_v_pos[0]) == __pyx_v_c) != 0);
    if (__pyx_t_1) {

      /* "_str_helpers.pxd":36
 *     while pos >= start:
 *         if pos[0] == c:
 *             return <void*>pos             # <<<<<<<<<<<<<<
 *         pos = pos - 1
 *     return NULL
 */
      __pyx_r = ((void *)__pyx_v_pos);
      goto __pyx_L0;

      /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    }

    /* "_str_helpers.pxd":37
 *         if pos[0] == c:
 *             return <void*>pos
 *         pos = pos - 1             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
    __pyx_v_pos = (__pyx_v_pos - 1);
  }

  /* "_str_helpers.pxd":38
 *             return <void*>pos
 *         pos = pos - 1
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("safe_string_from_size", 0);

  /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":44
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(s, size)
 * 
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_str_helpers.pxd":43
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 43, __pyx_L1_error)

    /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  }

  /* "_str_helpers.pxd":45
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("safe_interned_string_from_size", 0);

  /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":51
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_InternFromStringAndSize(s, size)
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_str_helpers.pxd":50
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 50, __pyx_L1_error)

    /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  }

  /* "_str_helpers.pxd":52
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_InternFromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_interned_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  p->data = Py_None; Py_INCREF(Py_None);
  p->keys = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->data);
  Py_CLEAR(p->keys);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  if (p->data) {
    e = (*v)(p->data, a); if (e) return e;
  }
  if (p->keys) {
    e = (*v)(p->keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *)o;
  tmp = ((PyObject*)p->data);
  p->data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->keys);
  p->keys = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser[] = {
  {"parse", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_3parse, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._btree_serializer_pyx.BTreeLeafParser", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Parse the leaf nodes of a BTree index.\n\n    :ivar data: The PyBytes object containing the uncompressed text for the\n        node.\n    :ivar key_length: An integer describing how many pieces the keys have for\n        this index.\n    :ivar ref_list_length: An integer describing how many references this index\n        contains.\n    :ivar keys: A PyList of keys found in this node.\n\n    :ivar _cur_str: A pointer to the start of the next line to parse\n    :ivar _end_str: A pointer to the end of bytes\n    :ivar _start: Pointer to the location within the current line while\n        parsing.\n    :ivar _header_found: True when we have parsed the header for this node\n    ", /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  p->last_key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->last_key);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  if (p->last_key) {
    e = (*v)(p->last_key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  tmp = ((PyObject*)p->last_key);
  p->last_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(o, v);
  }
  else {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(o);
}

static int __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {"__sizeof__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__, METH_NOARGS, 0},
  {"all_keys", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys, METH_NOARGS, 0},
  {"all_items", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items, METH_NOARGS, 0},
  {"_get_offset_for_sha1", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1, METH_O, 0},
  {"_get_offsets", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_21__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_23__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {(char *)"min_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key, 0, (char *)0, 0},
  {(char *)"max_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key, 0, (char *)0, 0},
  {(char *)"last_key", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, (char *)0, 0},
  {(char *)"num_records", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, (char *)0, 0},
  {(char *)"common_shift", __pyx_getprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, __pyx_setprop_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_GCCHKSHA1LeafNode = {
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GCCHKSHA1LeafNode = {
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*mp_length*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._btree_serializer_pyx.GCCHKSHA1LeafNode", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_GCCHKSHA1LeafNode, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GCCHKSHA1LeafNode, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Track all the entries for a given leaf node.", /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__btree_serializer_pyx(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__btree_serializer_pyx},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_btree_serializer_pyx",
    __pyx_k_Pyrex_extensions_to_btree_node_p, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_BTreeLeafParser, __pyx_k_BTreeLeafParser, sizeof(__pyx_k_BTreeLeafParser), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_bytes_for_value_not_r, __pyx_k_Expected_bytes_for_value_not_r, sizeof(__pyx_k_Expected_bytes_for_value_not_r), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_find_the_value_area, __pyx_k_Failed_to_find_the_value_area, sizeof(__pyx_k_Failed_to_find_the_value_area), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_block_length, __pyx_k_Failed_to_parse_block_length, sizeof(__pyx_k_Failed_to_parse_block_length), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_block_offset, __pyx_k_Failed_to_parse_block_offset, sizeof(__pyx_k_Failed_to_parse_block_offset), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_record_end, __pyx_k_Failed_to_parse_record_end, sizeof(__pyx_k_Failed_to_parse_record_end), 0, 0, 1, 0},
  {&__pyx_n_s_GCCHKSHA1LeafNode, __pyx_k_GCCHKSHA1LeafNode, sizeof(__pyx_k_GCCHKSHA1LeafNode), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x9b, __pyx_k_Incompatible_checksums_s_vs_0x9b, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x9b), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_Line_did_not_contain_40_hex_byte, __pyx_k_Line_did_not_contain_40_hex_byte, sizeof(__pyx_k_Line_did_not_contain_40_hex_byte), 0, 0, 1, 0},
  {&__pyx_kp_s_Node_did_not_start_with_type_lea, __pyx_k_Node_did_not_start_with_type_lea, sizeof(__pyx_k_Node_did_not_start_with_type_lea), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_kp_s_Something_went_wrong_while_parsi, __pyx_k_Something_went_wrong_while_parsi, sizeof(__pyx_k_Something_went_wrong_while_parsi), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s_We_expect_reference_bits_to_be_b, __pyx_k_We_expect_reference_bits_to_be_b, sizeof(__pyx_k_We_expect_reference_bits_to_be_b), 0, 0, 1, 0},
  {&__pyx_kp_s_We_expect_references_to_be_tuple, __pyx_k_We_expect_references_to_be_tuple, sizeof(__pyx_k_We_expect_references_to_be_tuple), 0, 0, 1, 0},
  {&__pyx_kp_s_We_expected_a_tuple_or_StaticTup, __pyx_k_We_expected_a_tuple_or_StaticTup, sizeof(__pyx_k_We_expected_a_tuple_or_StaticTup), 0, 0, 1, 0},
  {&__pyx_kp_s_We_failed_to_unhexlify, __pyx_k_We_failed_to_unhexlify, sizeof(__pyx_k_We_failed_to_unhexlify), 0, 0, 1, 0},
  {&__pyx_kp_s_We_only_support_parsing_byte_str, __pyx_k_We_only_support_parsing_byte_str, sizeof(__pyx_k_We_only_support_parsing_byte_str), 0, 0, 1, 0},
  {&__pyx_kp_s_With_ref_lists_we_expected_4_ent, __pyx_k_With_ref_lists_we_expected_4_ent, sizeof(__pyx_k_With_ref_lists_we_expected_4_ent), 0, 0, 1, 0},
  {&__pyx_kp_s_Without_ref_lists_we_need_at_lea, __pyx_k_Without_ref_lists_we_need_at_lea, sizeof(__pyx_k_Without_ref_lists_we_need_at_lea), 0, 0, 1, 0},
  {&__pyx_kp_b__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 0, 0},
  {&__pyx_n_s_as_bin, __pyx_k_as_bin, sizeof(__pyx_k_as_bin), 0, 0, 1, 1},
  {&__pyx_n_s_as_bin_sha, __pyx_k_as_bin_sha, sizeof(__pyx_k_as_bin_sha), 0, 0, 1, 1},
  {&__pyx_n_s_as_hex, __pyx_k_as_hex, sizeof(__pyx_k_as_hex), 0, 0, 1, 1},
  {&__pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_k_breezy_bzr__btree_serializer_pyx, sizeof(__pyx_k_breezy_bzr__btree_serializer_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_breezy_bzr__btree_serializer_pyx_2, __pyx_k_breezy_bzr__btree_serializer_pyx_2, sizeof(__pyx_k_breezy_bzr__btree_serializer_pyx_2), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_kp_s_bytes_did_not_start_with_type_le, __pyx_k_bytes_did_not_start_with_type_le, sizeof(__pyx_k_bytes_did_not_start_with_type_le), 0, 0, 1, 0},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_b_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 0, 0},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_first_ref_list, __pyx_k_first_ref_list, sizeof(__pyx_k_first_ref_list), 0, 0, 1, 1},
  {&__pyx_n_s_first_reference, __pyx_k_first_reference, sizeof(__pyx_k_first_reference), 0, 0, 1, 1},
  {&__pyx_n_s_flat_len, __pyx_k_flat_len, sizeof(__pyx_k_flat_len), 0, 0, 1, 1},
  {&__pyx_n_s_flatten_node, __pyx_k_flatten_node, sizeof(__pyx_k_flatten_node), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_have_reference_lists, __pyx_k_have_reference_lists, sizeof(__pyx_k_have_reference_lists), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_s_invalid_key_loop_counter_self_re, __pyx_k_invalid_key_loop_counter_self_re, sizeof(__pyx_k_invalid_key_loop_counter_self_re), 0, 0, 1, 0},
  {&__pyx_kp_s_invalid_key_wanted_segment_from, __pyx_k_invalid_key_wanted_segment_from, sizeof(__pyx_k_invalid_key_wanted_segment_from), 0, 0, 1, 0},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_len, __pyx_k_key_len, sizeof(__pyx_k_key_len), 0, 0, 1, 1},
  {&__pyx_n_s_key_length, __pyx_k_key_length, sizeof(__pyx_k_key_length), 0, 0, 1, 1},
  {&__pyx_kp_s_key_r_is_not_present, __pyx_k_key_r_is_not_present, sizeof(__pyx_k_key_r_is_not_present), 0, 0, 1, 0},
  {&__pyx_kp_s_last_self__start, __pyx_k_last_self__start, sizeof(__pyx_k_last_self__start), 0, 0, 1, 0},
  {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
  {&__pyx_kp_s_line_did_not_start_with_sha1_r, __pyx_k_line_did_not_start_with_sha1_r, sizeof(__pyx_k_line_did_not_start_with_sha1_r), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_next_len, __pyx_k_next_len, sizeof(__pyx_k_next_len), 0, 0, 1, 1},
  {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
  {&__pyx_n_s_node_len, __pyx_k_node_len, sizeof(__pyx_k_node_len), 0, 0, 1, 1},
  {&__pyx_kp_s_not_a_20_byte_binary_digest, __pyx_k_not_a_20_byte_binary_digest, sizeof(__pyx_k_not_a_20_byte_binary_digest), 0, 0, 1, 0},
  {&__pyx_kp_s_not_a_40_byte_hex_digest, __pyx_k_not_a_40_byte_hex_digest, sizeof(__pyx_k_not_a_40_byte_hex_digest), 0, 0, 1, 0},
  {&__pyx_kp_s_only_1_null_not_2_as_expected, __pyx_k_only_1_null_not_2_as_expected, sizeof(__pyx_k_only_1_null_not_2_as_expected), 0, 0, 1, 0},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_parse, __pyx_k_parse, sizeof(__pyx_k_parse), 0, 0, 1, 1},
  {&__pyx_n_s_parse_into_chk, __pyx_k_parse_into_chk, sizeof(__pyx_k_parse_into_chk), 0, 0, 1, 1},
  {&__pyx_n_s_parse_leaf_lines, __pyx_k_parse_leaf_lines, sizeof(__pyx_k_parse_leaf_lines), 0, 0, 1, 1},
  {&__pyx_n_s_parser, __pyx_k_parser, sizeof(__pyx_k_parser), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_py_hexlify, __pyx_k_py_hexlify, sizeof(__pyx_k_py_hexlify), 0, 0, 1, 1},
  {&__pyx_n_s_py_key_to_sha1, __pyx_k_py_key_to_sha1, sizeof(__pyx_k_py_key_to_sha1), 0, 0, 1, 1},
  {&__pyx_n_s_py_sha1_to_key, __pyx_k_py_sha1_to_key, sizeof(__pyx_k_py_sha1_to_key), 0, 0, 1, 1},
  {&__pyx_n_s_py_unhexlify, __pyx_k_py_unhexlify, sizeof(__pyx_k_py_unhexlify), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_BTreeLeafParser, __pyx_k_pyx_unpickle_BTreeLeafParser, sizeof(__pyx_k_pyx_unpickle_BTreeLeafParser), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ref_bit, __pyx_k_ref_bit, sizeof(__pyx_k_ref_bit), 0, 0, 1, 1},
  {&__pyx_n_s_ref_bit_len, __pyx_k_ref_bit_len, sizeof(__pyx_k_ref_bit_len), 0, 0, 1, 1},
  {&__pyx_n_s_ref_list, __pyx_k_ref_list, sizeof(__pyx_k_ref_list), 0, 0, 1, 1},
  {&__pyx_n_s_ref_list_length, __pyx_k_ref_list_length, sizeof(__pyx_k_ref_list_length), 0, 0, 1, 1},
  {&__pyx_n_s_ref_lists, __pyx_k_ref_lists, sizeof(__pyx_k_ref_lists), 0, 0, 1, 1},
  {&__pyx_n_s_reference, __pyx_k_reference, sizeof(__pyx_k_reference), 0, 0, 1, 1},
  {&__pyx_n_s_reference_lists, __pyx_k_reference_lists, sizeof(__pyx_k_reference_lists), 0, 0, 1, 1},
  {&__pyx_n_s_refs_len, __pyx_k_refs_len, sizeof(__pyx_k_refs_len), 0, 0, 1, 1},
  {&__pyx_kp_s_self_data_is_not_a_byte_string, __pyx_k_self_data_is_not_a_byte_string, sizeof(__pyx_k_self_data_is_not_a_byte_string), 0, 0, 1, 0},
  {&__pyx_kp_s_self_last_record_self_records_ca, __pyx_k_self_last_record_self_records_ca, sizeof(__pyx_k_self_last_record_self_records_ca), 0, 0, 1, 0},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sha1_bin, __pyx_k_sha1_bin, sizeof(__pyx_k_sha1_bin), 0, 0, 1, 1},
  {&__pyx_kp_s_sha1_bin_must_be_a_str_of_exactl, __pyx_k_sha1_bin_must_be_a_str_of_exactl, sizeof(__pyx_k_sha1_bin_must_be_a_str_of_exactl), 0, 0, 1, 0},
  {&__pyx_n_s_string_key, __pyx_k_string_key, sizeof(__pyx_k_string_key), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_k_tried_to_create_a_string_with_an, sizeof(__pyx_k_tried_to_create_a_string_with_an), 0, 0, 1, 0},
  {&__pyx_kp_s_unexpected_reference_data_presen, __pyx_k_unexpected_reference_data_presen, sizeof(__pyx_k_unexpected_reference_data_presen), 0, 0, 1, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_value_len, __pyx_k_value_len, sizeof(__pyx_k_value_len), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 154, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 617, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 664, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":197
 *         if last < self._start:
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_last_self__start); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":212
 *         if temp_ptr == NULL:
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")             # <<<<<<<<<<<<<<
 *         else:
 *             # Because of how conversions were done, we ended up with *lots* of
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_find_the_value_area); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":246
 *                     if loop_counter != self.ref_list_length:
 *                         # Invalid line
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "invalid key, loop_counter != self.ref_list_length")
 *                     else:
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_invalid_key_loop_counter_self_re); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":275
 *             if last != self._start:
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_unexpected_reference_data_presen); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":283
 *         cdef Py_ssize_t byte_count
 *         if not PyBytes_CheckExact(self.data):
 *             raise AssertionError('self.data is not a byte string.')             # <<<<<<<<<<<<<<
 *         byte_count = PyBytes_GET_SIZE(self.data)
 *         self._cur_str = PyBytes_AS_STRING(self.data)
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self_data_is_not_a_byte_string); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":357
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 *         raise ValueError('not a 40-byte hex digest')             # <<<<<<<<<<<<<<
 *     as_bin = PyBytes_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyBytes_AS_STRING(as_hex), PyBytes_AS_STRING(as_bin)):
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_not_a_40_byte_hex_digest); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":380
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')             # <<<<<<<<<<<<<<
 *     as_hex = PyBytes_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyBytes_AS_STRING(as_bin), PyBytes_AS_STRING(as_hex))
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_not_a_20_byte_binary_digest); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":454
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')             # <<<<<<<<<<<<<<
 *     return _sha1_to_key(PyBytes_AS_STRING(sha1_bin))
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_sha1_bin_must_be_a_str_of_exactl); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":664
 * 
 *         if not PyBytes_CheckExact(data):
 *             raise TypeError('We only support parsing byte strings.')             # <<<<<<<<<<<<<<
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyBytes_GET_SIZE(data)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_We_only_support_parsing_byte_str); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":671
 *         if strncmp(c_bytes, b'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (data[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
 */
  __pyx_slice__10 = PySlice_New(Py_None, __pyx_int_10, Py_None); if (unlikely(!__pyx_slice__10)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__10);
  __Pyx_GIVEREF(__pyx_slice__10);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":687
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')             # <<<<<<<<<<<<<<
 *         # Pass 3: build the offset map
 *         self._compute_common()
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_Something_went_wrong_while_parsi); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":705
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')             # <<<<<<<<<<<<<<
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Line_did_not_contain_40_hex_byte); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":707
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_We_failed_to_unhexlify); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":710
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_only_1_null_not_2_as_expected); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":714
 *         cur_record.block_offset = strtoull(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_block_offset); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":718
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_block_length); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":726
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         return c_cur
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_record_end); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_self_last_record_self_records_ca); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.last_record,self.records cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_self_last_record_self_records_ca); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "breezy/bzr/_btree_serializer_pyx.pyx":293
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
 */
  __pyx_tuple__21 = PyTuple_Pack(4, __pyx_n_s_data, __pyx_n_s_key_length, __pyx_n_s_ref_list_length, __pyx_n_s_parser); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_parse_leaf_lines, 293, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 293, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":354
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 */
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_n_s_as_hex, __pyx_n_s_as_bin); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_py_unhexlify, 354, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 354, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":377
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_n_s_as_bin, __pyx_n_s_as_hex); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_py_hexlify, 377, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 377, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":415
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_key, __pyx_n_s_as_bin_sha); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_py_key_to_sha1, 415, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 415, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":451
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_sha1_bin); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_py_sha1_to_key, 451, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 451, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":795
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_n_s_bytes, __pyx_n_s_key_length, __pyx_n_s_ref_list_length); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_parse_into_chk, 795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":802
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */
  __pyx_tuple__33 = PyTuple_Pack(22, __pyx_n_s_node, __pyx_n_s_reference_lists, __pyx_n_s_have_reference_lists, __pyx_n_s_flat_len, __pyx_n_s_key_len, __pyx_n_s_node_len, __pyx_n_s_value, __pyx_n_s_value_len, __pyx_n_s_out, __pyx_n_s_refs_len, __pyx_n_s_next_len, __pyx_n_s_first_ref_list, __pyx_n_s_first_reference, __pyx_n_s_i, __pyx_n_s_ref_bit_len, __pyx_n_s_string_key, __pyx_n_s_ref_lists, __pyx_n_s_ref_list, __pyx_n_s_reference, __pyx_n_s_ref_bit, __pyx_n_s_val, __pyx_n_s_line); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 22, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__btree_serializer_pyx, __pyx_n_s_flatten_node, 802, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 802, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BTreeLeafParser, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_163054242 = PyInt_FromLong(163054242L); if (unlikely(!__pyx_int_163054242)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483648 = PyInt_FromString((char *)"2147483648", 0, 0); if (unlikely(!__pyx_int_2147483648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.extract_key = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_extract_key;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.process_line = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_15BTreeLeafParser_process_line;
  if (PyType_Ready(&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.tp_dictoffset && __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser.tp_dict, __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BTreeLeafParser, (PyObject *)&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_BTreeLeafParser;
  __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_value_and_refs = (StaticTuple *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_item = (StaticTuple *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._lookup_record = (__pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._count_records = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_bytes = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_one_entry = (char *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6breezy_3bzr_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._offset_for_sha1 = (int (*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1;
  __pyx_vtable_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode._compute_common = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode *))__pyx_f_6breezy_3bzr_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common;
  if (PyType_Ready(&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) __PYX_ERR(0, 485, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_dictoffset && __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_dict, __pyx_vtabptr_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) __PYX_ERR(0, 485, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GCCHKSHA1LeafNode, (PyObject *)&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) __PYX_ERR(0, 485, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) __PYX_ERR(0, 485, __pyx_L1_error)
  __pyx_ptype_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_type_6breezy_3bzr_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("breezy._static_tuple_c"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6breezy_15_static_tuple_c_StaticTuple = __Pyx_ImportType(__pyx_t_1, "breezy._static_tuple_c", "StaticTuple", sizeof(StaticTuple), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple) __PYX_ERR(4, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_btree_serializer_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_btree_serializer_pyx(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__btree_serializer_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__btree_serializer_pyx(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__btree_serializer_pyx(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_btree_serializer_pyx' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__btree_serializer_pyx(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_btree_serializer_pyx", __pyx_methods, __pyx_k_Pyrex_extensions_to_btree_node_p, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_breezy__bzr___btree_serializer_pyx) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "breezy.bzr._btree_serializer_pyx")) {
      if (unlikely(PyDict_SetItemString(modules, "breezy.bzr._btree_serializer_pyx", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "breezy/bzr/_btree_serializer_pyx.pyx":84
 *     )
 * 
 * import sys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":88
 * 
 * # This sets up the StaticTuple C_API functionality
 * import_static_tuple_c()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = import_static_tuple_c(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 88, __pyx_L1_error)

  /* "breezy/bzr/_btree_serializer_pyx.pyx":293
 * 
 * 
 * def _parse_leaf_lines(data, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(data, key_length, ref_list_length)
 *     return parser.parse()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_1_parse_leaf_lines, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_leaf_lines, __pyx_t_1) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":315
 * cdef int _unhexbuf[256]
 * cdef char *_hexbuf
 * _hexbuf = b'0123456789abcdef'             # <<<<<<<<<<<<<<
 * 
 * cdef _populate_unhexbuf():
 */
  __pyx_v_6breezy_3bzr_21_btree_serializer_pyx__hexbuf = ((char *)"0123456789abcdef");

  /* "breezy/bzr/_btree_serializer_pyx.pyx":327
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_btree_serializer_pyx__populate_unhexbuf(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":354
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if not PyBytes_CheckExact(as_hex) or PyBytes_GET_SIZE(as_hex) != 40:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_3_py_unhexlify, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_unhexlify, __pyx_t_1) < 0) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":377
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyBytes_CheckExact(as_bin):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_5_py_hexlify, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_hexlify, __pyx_t_1) < 0) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":415
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_7_py_key_to_sha1, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_key_to_sha1, __pyx_t_1) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":451
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyBytes_CheckExact(sha1_bin) or PyBytes_GET_SIZE(sha1_bin) != 20:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_9_py_sha1_to_key, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_sha1_to_key, __pyx_t_1) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":795
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_11_parse_into_chk, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_into_chk, __pyx_t_1) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":802
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_13_flatten_node, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flatten_node, __pyx_t_1) < 0) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BTreeLeafParser(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_btree_serializer_pyx_15__pyx_unpickle_BTreeLeafParser, NULL, __pyx_n_s_breezy_bzr__btree_serializer_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BTreeLeafParser, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_btree_serializer_pyx.pyx":1
 * # Copyright (C) 2008, 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init breezy.bzr._btree_serializer_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init breezy.bzr._btree_serializer_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) ((unsigned PY_LONG_LONG) 0 - (unsigned PY_LONG_LONG) 1), const_zero = (unsigned PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value) {
    const unsigned char neg_one = (unsigned char) ((unsigned char) 0 - (unsigned char) 1), const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *x) {
    const unsigned char neg_one = (unsigned char) ((unsigned char) 0 - (unsigned char) 1), const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned char) -1;
        }
    } else {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyInt_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
