/* Generated by Cython 0.29.14 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_14"
#define CYTHON_HEX_VERSION 0x001D0EF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__breezy__bzr___dirstate_helpers_pyx
#define __PYX_HAVE_API__breezy__bzr___dirstate_helpers_pyx
/* Early includes */
#include "python-compat.h"
#include "_dirstate_helpers_pyx.h"
#include "stdlib.h"
#include "sys/stat.h"
#include "string.h"
#include "_static_tuple_c.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "breezy/bzr/_dirstate_helpers_pyx.pyx",
  "stringsource",
  "breezy/bzr/_str_helpers.pxd",
  "breezy/_static_tuple_c.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":517
 * 
 * 
 * cdef class Reader:             # <<<<<<<<<<<<<<
 *     """Maintain the current location, and return fields as you parse them."""
 * 
 */
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_vtab;
  PyObject *state;
  PyObject *text;
  char *text_cstr;
  int text_size;
  char *end_cstr;
  char *cur_cstr;
  char *next;
};


/* "breezy/bzr/_dirstate_helpers_pyx.pyx":994
 * 
 * 
 * cdef class ProcessEntryC:             # <<<<<<<<<<<<<<
 * 
 *     cdef int doing_consistency_expansion
 */
struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC {
  PyObject_HEAD
  struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_vtab;
  int doing_consistency_expansion;
  PyObject *old_dirname_to_file_id;
  PyObject *new_dirname_to_file_id;
  PyObject *last_source_parent;
  PyObject *last_target_parent;
  int include_unchanged;
  int partial;
  PyObject *use_filesystem_for_exec;
  PyObject *utf8_decode;
  PyObject *searched_specific_files;
  PyObject *searched_exact_paths;
  PyObject *search_specific_files;
  PyObject *search_specific_file_parents;
  PyObject *state;
  PyObject *current_root;
  PyObject *current_root_unicode;
  PyObject *root_entries;
  int root_entries_pos;
  int root_entries_len;
  PyObject *root_abspath;
  int source_index;
  int target_index;
  int want_unversioned;
  PyObject *tree;
  PyObject *dir_iterator;
  int block_index;
  PyObject *current_block;
  int current_block_pos;
  PyObject *current_block_list;
  PyObject *current_dir_info;
  PyObject *current_dir_list;
  PyObject *_pending_consistent_entries;
  int path_index;
  PyObject *root_dir_info;
  PyObject *bisect_left;
  PyObject *pathjoin;
  PyObject *fstat;
  PyObject *seen_ids;
  PyObject *sha_file;
};



/* "breezy/bzr/_dirstate_helpers_pyx.pyx":517
 * 
 * 
 * cdef class Reader:             # <<<<<<<<<<<<<<
 *     """Maintain the current location, and return fields as you parse them."""
 * 
 */

struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader {
  char *(*get_next)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int *);
  PyObject *(*get_next_str)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *);
  int (*_init)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *);
  PyObject *(*_get_entry)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int, void **, int *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;


/* "breezy/bzr/_dirstate_helpers_pyx.pyx":994
 * 
 * 
 * cdef class ProcessEntryC:             # <<<<<<<<<<<<<<
 * 
 *     cdef int doing_consistency_expansion
 */

struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC {
  PyObject *(*_process_entry)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *);
  int (*_gather_result_for_consistency)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *);
  int (*_update_current_block)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_iter_next)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_maybe_tree_ref)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *);
  PyObject *(*_loop_one_block)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_next_consistent_entries)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *);
  PyObject *(*_path_info)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* bytes_tailmatch.proto */
static int __Pyx_PyBytes_SingleTailmatch(PyObject* self, PyObject* arg,
                                         Py_ssize_t start, Py_ssize_t end, int direction);
static int __Pyx_PyBytes_Tailmatch(PyObject* self, PyObject* substr,
                                   Py_ssize_t start, Py_ssize_t end, int direction);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DelItemInt.proto */
#define __Pyx_DelItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_DelItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_DelItem_Generic(o, to_py_func(i))))
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j);
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               int is_list, int wraparound);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_ptrdiff_t(ptrdiff_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static char *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int *__pyx_v_size); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int __pyx_v_num_trees, void **__pyx_v_p_current_dirname, int *__pyx_v_new_block); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_path_info); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_result); /* proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_current_path_info); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_utf8_path, PyObject *__pyx_v_unicode_path); /* proto*/

/* Module declarations from 'breezy.bzr._str_helpers' */
static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *, int, size_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *, Py_ssize_t); /*proto*/

/* Module declarations from 'breezy._static_tuple_c' */

/* Module declarations from 'breezy._static_tuple_c' */
static PyTypeObject *__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple = 0;

/* Module declarations from 'breezy.bzr._dirstate_helpers_pyx' */
static PyTypeObject *__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = 0;
static PyTypeObject *__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = 0;
static int __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND;
static int __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated = 0;
static PyObject *__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference = 0;
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(void *); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(char *, int, char *, int); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(char *, int, char *, int); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(int); /*proto*/
static unsigned long __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(PyObject *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static char __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(char); /*proto*/
static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(char); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, PyObject *); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "breezy.bzr._dirstate_helpers_pyx"
extern int __pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx;
int __pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx = 0;

/* Implementation of 'breezy.bzr._dirstate_helpers_pyx' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_UnicodeDecodeError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_d[] = "d";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k__3[] = "\n";
static const char __pyx_k__4[] = "";
static const char __pyx_k_hi[] = "hi";
static const char __pyx_k_lo[] = "lo";
static const char __pyx_k_ne[] = "__ne__";
static const char __pyx_k_os[] = "os";
static const char __pyx_k__16[] = "/";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_bzr[] = ".bzr";
static const char __pyx_k_c_2[] = "_c";
static const char __pyx_k_cur[] = "cur";
static const char __pyx_k_int[] = "__int__";
static const char __pyx_k_mid[] = "_mid";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_s_2[] = "_s";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_hi_2[] = "_hi";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_lo_2[] = "_lo";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_seek[] = "seek";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_stat[] = "stat";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_tree[] = "tree";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_cache[] = "cache";
static const char __pyx_k_entry[] = "entry";
static const char __pyx_k_errno[] = "errno";
static const char __pyx_k_found[] = "found";
static const char __pyx_k_fstat[] = "fstat";
static const char __pyx_k_lstat[] = "lstat";
static const char __pyx_k_path1[] = "path1";
static const char __pyx_k_path2[] = "path2";
static const char __pyx_k_paths[] = "paths";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_win32[] = "win32";
static const char __pyx_k_EINVAL[] = "EINVAL";
static const char __pyx_k_ENOENT[] = "ENOENT";
static const char __pyx_k_Reader[] = "Reader";
static const char __pyx_k_absent[] = "absent";
static const char __pyx_k_bisect[] = "bisect";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "_encode";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_fs_enc[] = "_fs_enc";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_prefix[] = "prefix";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_rsplit[] = "rsplit";
static const char __pyx_k_st_dev[] = "st_dev";
static const char __pyx_k_st_ino[] = "st_ino";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_ENOTDIR[] = "ENOTDIR";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_S_IEXEC[] = "S_IEXEC";
static const char __pyx_k_S_ISREG[] = "S_ISREG";
static const char __pyx_k_abspath[] = "abspath";
static const char __pyx_k_dirname[] = "dirname";
static const char __pyx_k_file_id[] = "file_id";
static const char __pyx_k_osutils[] = "osutils";
static const char __pyx_k_st_mode[] = "st_mode";
static const char __pyx_k_st_size[] = "st_size";
static const char __pyx_k_symlink[] = "symlink";
static const char __pyx_k_BzrError[] = "BzrError";
static const char __pyx_k_DirState[] = "DirState";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_NULLSTAT[] = "NULLSTAT";
static const char __pyx_k_binascii[] = "binascii";
static const char __pyx_k_cur_cstr[] = "cur_cstr";
static const char __pyx_k_cur_size[] = "cur_size";
static const char __pyx_k_dirstate[] = "dirstate";
static const char __pyx_k_encode_2[] = "encode";
static const char __pyx_k_filename[] = "_filename";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pathjoin[] = "pathjoin";
static const char __pyx_k_platform[] = "platform";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_sha_file[] = "sha_file";
static const char __pyx_k_st_ctime[] = "st_ctime";
static const char __pyx_k_st_mtime[] = "st_mtime";
static const char __pyx_k_winerror[] = "winerror";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_dirblocks[] = "dirblocks";
static const char __pyx_k_directory[] = "directory";
static const char __pyx_k_get_entry[] = "_get_entry";
static const char __pyx_k_is_inside[] = "is_inside";
static const char __pyx_k_pack_stat[] = "pack_stat";
static const char __pyx_k_path_cstr[] = "path_cstr";
static const char __pyx_k_path_size[] = "path_size";
static const char __pyx_k_path_utf8[] = "path_utf8";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_read_link[] = "_read_link";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_relocated[] = "relocated";
static const char __pyx_k_sha1_file[] = "_sha1_file";
static const char __pyx_k_splitpath[] = "splitpath";
static const char __pyx_k_TreeChange[] = "TreeChange";
static const char __pyx_k_b2a_base64[] = "b2a_base64";
static const char __pyx_k_cache_utf8[] = "cache_utf8";
static const char __pyx_k_lt_by_dirs[] = "lt_by_dirs";
static const char __pyx_k_py_memrchr[] = "_py_memrchr";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_stat_value[] = "stat_value";
static const char __pyx_k_state_file[] = "_state_file";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_bisect_left[] = "bisect_left";
static const char __pyx_k_cutoff_time[] = "_cutoff_time";
static const char __pyx_k_dirblocks_2[] = "_dirblocks";
static const char __pyx_k_num_entries[] = "_num_entries";
static const char __pyx_k_utf8_decode[] = "_utf8_decode";
static const char __pyx_k_dirname_cstr[] = "dirname_cstr";
static const char __pyx_k_dirname_size[] = "dirname_size";
static const char __pyx_k_ensure_block[] = "_ensure_block";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_source_index[] = "source_index";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_target_index[] = "target_index";
static const char __pyx_k_ProcessEntryC[] = "ProcessEntryC";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_end_of_header[] = "_end_of_header";
static const char __pyx_k_is_executable[] = "_is_executable";
static const char __pyx_k_is_inside_any[] = "is_inside_any";
static const char __pyx_k_mark_modified[] = "_mark_modified";
static const char __pyx_k_observed_sha1[] = "_observed_sha1";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_sha1_provider[] = "_sha1_provider";
static const char __pyx_k_stat_and_sha1[] = "stat_and_sha1";
static const char __pyx_k_walkdirs_utf8[] = "_walkdirs_utf8";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_dirblock_state[] = "_dirblock_state";
static const char __pyx_k_read_dirblocks[] = "_read_dirblocks";
static const char __pyx_k_tree_reference[] = "tree-reference";
static const char __pyx_k_DirstateCorrupt[] = "DirstateCorrupt";
static const char __pyx_k_bisect_dirblock[] = "bisect_dirblock";
static const char __pyx_k_parse_dirblocks[] = "_parse_dirblocks";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_sha_cutoff_time[] = "_sha_cutoff_time";
static const char __pyx_k_BadFileKindError[] = "BadFileKindError";
static const char __pyx_k_bisect_path_left[] = "_bisect_path_left";
static const char __pyx_k_entries_for_path[] = "_entries_for_path";
static const char __pyx_k_want_unversioned[] = "want_unversioned";
static const char __pyx_k_bisect_path_right[] = "_bisect_path_right";
static const char __pyx_k_include_unchanged[] = "include_unchanged";
static const char __pyx_k_UnicodeDecodeError[] = "UnicodeDecodeError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_parent_directories[] = "parent_directories";
static const char __pyx_k_BadFilenameEncoding[] = "BadFilenameEncoding";
static const char __pyx_k_NULL_PARENT_DETAILS[] = "NULL_PARENT_DETAILS";
static const char __pyx_k_lt_path_by_dirblock[] = "_lt_path_by_dirblock";
static const char __pyx_k_num_present_parents[] = "_num_present_parents";
static const char __pyx_k_pyx_unpickle_Reader[] = "__pyx_unpickle_Reader";
static const char __pyx_k_IN_MEMORY_UNMODIFIED[] = "IN_MEMORY_UNMODIFIED";
static const char __pyx_k_result_is_not_None_r[] = "result is not None: %r";
static const char __pyx_k_get_block_entry_index[] = "_get_block_entry_index";
static const char __pyx_k_search_specific_files[] = "search_specific_files";
static const char __pyx_k_use_filesystem_for_exec[] = "use_filesystem_for_exec";
static const char __pyx_k_file_kind_from_stat_mode[] = "file_kind_from_stat_mode";
static const char __pyx_k_unsupported_target_index[] = "unsupported target index";
static const char __pyx_k_find_block_index_from_key[] = "_find_block_index_from_key";
static const char __pyx_k_Unsupported_target_index_d[] = "Unsupported target index %d";
static const char __pyx_k_pyx_unpickle_ProcessEntryC[] = "__pyx_unpickle_ProcessEntryC";
static const char __pyx_k_directory_is_tree_reference[] = "_directory_is_tree_reference";
static const char __pyx_k_Bad_parse_we_expected_to_end_on[] = "Bad parse, we expected to end on \\n, not: %d %s: %s";
static const char __pyx_k_Could_not_find_target_parent_in[] = "Could not find target parent in wt: %s\nparent of: %s";
static const char __pyx_k_Missing_entry_for_specific_path[] = "Missing entry for specific path parent %r, %r";
static const char __pyx_k_entry_s_s_is_considered_renamed[] = "entry '%s/%s' is considered renamed from %r but source does not exist\nentry: %s";
static const char __pyx_k_get_next_called_when_cur_str_is[] = "get_next() called when cur_str is NULL";
static const char __pyx_k_path1_must_be_a_bytes_string_no[] = "'path1' must be a bytes string, not %s: %r";
static const char __pyx_k_path1_must_be_a_plain_string_no[] = "'path1' must be a plain string, not %s: %r";
static const char __pyx_k_path2_must_be_a_bytes_string_no[] = "'path2' must be a bytes string, not %s: %r";
static const char __pyx_k_path2_must_be_a_plain_string_no[] = "'path2' must be a plain string, not %s: %r";
static const char __pyx_k_split_root_dirblock_into_conten[] = "_split_root_dirblock_into_contents";
static const char __pyx_k_you_must_pass_a_python_list_for[] = "you must pass a python list for 'paths' not: %s %r";
static const char __pyx_k_you_must_pass_a_string_for_path[] = "you must pass a string for 'path' not: %s %r";
static const char __pyx_k_Expected_single_character_string[] = "Expected single character string not %r";
static const char __pyx_k_First_character_should_be_null_n[] = "First character should be null not: %s";
static const char __pyx_k_Got_entry_path_mismatch_for_spec[] = "Got entry<->path mismatch for specific path %r entry %r path_info %r ";
static const char __pyx_k_Helper_functions_for_DirState_Th[] = "Helper functions for DirState.\n\nThis is the python implementation for DirState functions.\n";
static const char __pyx_k_Incompatible_checksums_s_vs_0x21[] = "Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x65[] = "Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))";
static const char __pyx_k_We_could_not_find_the_parent_ent[] = "We could not find the parent entry in index %d for the entry: %s";
static const char __pyx_k_We_read_the_wrong_number_of_entr[] = "We read the wrong number of entries. We expected to read %s, but read %s";
static const char __pyx_k_breezy_bzr__dirstate_helpers_pyx[] = "breezy/bzr/_dirstate_helpers_pyx.pyx";
static const char __pyx_k_don_t_know_how_to_compare_source[] = "don't know how to compare source_minikind=%r, target_minikind=%r";
static const char __pyx_k_failed_to_find_trailing_NULL_0_T[] = "failed to find trailing NULL (\\0). Trailing garbage: %r";
static const char __pyx_k_get_next_called_when_there_are_n[] = "get_next() called when there are no chars left";
static const char __pyx_k_tried_to_create_a_string_with_an[] = "tried to create a string with an invalid size: %d";
static const char __pyx_k_you_must_pass_a_string_for_dirna[] = "you must pass a string for dirname not: %s %r";
static const char __pyx_k_you_must_pass_a_python_list_for_2[] = "you must pass a python list for 'dirblocks' not: %s %r";
static const char __pyx_k_breezy_bzr__dirstate_helpers_pyx_2[] = "breezy.bzr._dirstate_helpers_pyx";
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_BadFileKindError;
static PyObject *__pyx_n_s_BadFilenameEncoding;
static PyObject *__pyx_kp_s_Bad_parse_we_expected_to_end_on;
static PyObject *__pyx_n_s_BzrError;
static PyObject *__pyx_kp_s_Could_not_find_target_parent_in;
static PyObject *__pyx_n_s_DirState;
static PyObject *__pyx_n_s_DirstateCorrupt;
static PyObject *__pyx_n_s_EINVAL;
static PyObject *__pyx_n_s_ENOENT;
static PyObject *__pyx_n_s_ENOTDIR;
static PyObject *__pyx_kp_s_Expected_single_character_string;
static PyObject *__pyx_kp_s_First_character_should_be_null_n;
static PyObject *__pyx_kp_s_Got_entry_path_mismatch_for_spec;
static PyObject *__pyx_n_s_IN_MEMORY_UNMODIFIED;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x21;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x65;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_Missing_entry_for_specific_path;
static PyObject *__pyx_n_s_NULLSTAT;
static PyObject *__pyx_n_s_NULL_PARENT_DETAILS;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_ProcessEntryC;
static PyObject *__pyx_n_s_Reader;
static PyObject *__pyx_n_s_S_IEXEC;
static PyObject *__pyx_n_s_S_ISREG;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TreeChange;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UnicodeDecodeError;
static PyObject *__pyx_kp_s_Unsupported_target_index_d;
static PyObject *__pyx_kp_s_We_could_not_find_the_parent_ent;
static PyObject *__pyx_kp_s_We_read_the_wrong_number_of_entr;
static PyObject *__pyx_kp_b__16;
static PyObject *__pyx_kp_b__3;
static PyObject *__pyx_kp_b__4;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_n_b_a;
static PyObject *__pyx_n_s_absent;
static PyObject *__pyx_n_s_abspath;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_b2a_base64;
static PyObject *__pyx_n_s_binascii;
static PyObject *__pyx_n_s_bisect;
static PyObject *__pyx_n_s_bisect_dirblock;
static PyObject *__pyx_n_s_bisect_left;
static PyObject *__pyx_n_s_bisect_path_left;
static PyObject *__pyx_n_s_bisect_path_right;
static PyObject *__pyx_kp_s_breezy_bzr__dirstate_helpers_pyx;
static PyObject *__pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2;
static PyObject *__pyx_kp_b_bzr;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_c_2;
static PyObject *__pyx_n_s_cache;
static PyObject *__pyx_n_s_cache_utf8;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cur;
static PyObject *__pyx_n_s_cur_cstr;
static PyObject *__pyx_n_s_cur_size;
static PyObject *__pyx_n_s_cutoff_time;
static PyObject *__pyx_n_b_d;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dirblock_state;
static PyObject *__pyx_n_s_dirblocks;
static PyObject *__pyx_n_s_dirblocks_2;
static PyObject *__pyx_n_s_directory;
static PyObject *__pyx_n_s_directory_is_tree_reference;
static PyObject *__pyx_n_s_dirname;
static PyObject *__pyx_n_s_dirname_cstr;
static PyObject *__pyx_n_s_dirname_size;
static PyObject *__pyx_n_s_dirstate;
static PyObject *__pyx_kp_s_don_t_know_how_to_compare_source;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encode_2;
static PyObject *__pyx_n_s_end_of_header;
static PyObject *__pyx_n_s_ensure_block;
static PyObject *__pyx_n_s_entries_for_path;
static PyObject *__pyx_n_s_entry;
static PyObject *__pyx_kp_s_entry_s_s_is_considered_renamed;
static PyObject *__pyx_n_s_errno;
static PyObject *__pyx_n_s_errors;
static PyObject *__pyx_n_b_f;
static PyObject *__pyx_kp_s_failed_to_find_trailing_NULL_0_T;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_file_id;
static PyObject *__pyx_n_s_file_kind_from_stat_mode;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_find_block_index_from_key;
static PyObject *__pyx_n_s_found;
static PyObject *__pyx_n_s_fs_enc;
static PyObject *__pyx_n_s_fstat;
static PyObject *__pyx_n_s_get_block_entry_index;
static PyObject *__pyx_n_s_get_entry;
static PyObject *__pyx_kp_s_get_next_called_when_cur_str_is;
static PyObject *__pyx_kp_s_get_next_called_when_there_are_n;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_hi;
static PyObject *__pyx_n_s_hi_2;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_include_unchanged;
static PyObject *__pyx_n_s_int;
static PyObject *__pyx_n_s_is_executable;
static PyObject *__pyx_n_s_is_inside;
static PyObject *__pyx_n_s_is_inside_any;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_b_l;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_lo;
static PyObject *__pyx_n_s_lo_2;
static PyObject *__pyx_n_s_lstat;
static PyObject *__pyx_n_s_lt_by_dirs;
static PyObject *__pyx_n_s_lt_path_by_dirblock;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mark_modified;
static PyObject *__pyx_n_s_mid;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_ne;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_num_entries;
static PyObject *__pyx_n_s_num_present_parents;
static PyObject *__pyx_n_s_observed_sha1;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_osutils;
static PyObject *__pyx_n_s_pack_stat;
static PyObject *__pyx_n_s_parent_directories;
static PyObject *__pyx_n_s_parse_dirblocks;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_path1;
static PyObject *__pyx_kp_s_path1_must_be_a_bytes_string_no;
static PyObject *__pyx_kp_s_path1_must_be_a_plain_string_no;
static PyObject *__pyx_n_s_path2;
static PyObject *__pyx_kp_s_path2_must_be_a_bytes_string_no;
static PyObject *__pyx_kp_s_path2_must_be_a_plain_string_no;
static PyObject *__pyx_n_s_path_cstr;
static PyObject *__pyx_n_s_path_size;
static PyObject *__pyx_n_s_path_utf8;
static PyObject *__pyx_n_s_pathjoin;
static PyObject *__pyx_n_s_paths;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_platform;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_prefix;
static PyObject *__pyx_n_s_py_memrchr;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_ProcessEntryC;
static PyObject *__pyx_n_s_pyx_unpickle_Reader;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_b_r;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_read_dirblocks;
static PyObject *__pyx_n_s_read_link;
static PyObject *__pyx_n_s_reader;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_relocated;
static PyObject *__pyx_kp_s_result_is_not_None_r;
static PyObject *__pyx_n_s_rsplit;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_s_2;
static PyObject *__pyx_n_s_search_specific_files;
static PyObject *__pyx_n_s_seek;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sha1_file;
static PyObject *__pyx_n_s_sha1_provider;
static PyObject *__pyx_n_s_sha_cutoff_time;
static PyObject *__pyx_n_s_sha_file;
static PyObject *__pyx_n_s_source_index;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_split_root_dirblock_into_conten;
static PyObject *__pyx_n_s_splitpath;
static PyObject *__pyx_n_s_st_ctime;
static PyObject *__pyx_n_s_st_dev;
static PyObject *__pyx_n_s_st_ino;
static PyObject *__pyx_n_s_st_mode;
static PyObject *__pyx_n_s_st_mtime;
static PyObject *__pyx_n_s_st_size;
static PyObject *__pyx_n_s_stat;
static PyObject *__pyx_n_s_stat_and_sha1;
static PyObject *__pyx_n_s_stat_value;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_state_file;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_symlink;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_target_index;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_tree;
static PyObject *__pyx_kp_s_tree_reference;
static PyObject *__pyx_kp_s_tried_to_create_a_string_with_an;
static PyObject *__pyx_kp_s_unsupported_target_index;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_use_filesystem_for_exec;
static PyObject *__pyx_n_s_utf8;
static PyObject *__pyx_n_s_utf8_decode;
static PyObject *__pyx_n_s_walkdirs_utf8;
static PyObject *__pyx_n_s_want_unversioned;
static PyObject *__pyx_n_s_win32;
static PyObject *__pyx_n_s_winerror;
static PyObject *__pyx_kp_s_you_must_pass_a_python_list_for;
static PyObject *__pyx_kp_s_you_must_pass_a_python_list_for_2;
static PyObject *__pyx_kp_s_you_must_pass_a_string_for_dirna;
static PyObject *__pyx_kp_s_you_must_pass_a_string_for_path;
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dirblocks, PyObject *__pyx_v_dirname, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, CYTHON_UNUSED PyObject *__pyx_v_cache); /* proto */
static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_stat_value); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value); /* proto */
static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_include_unchanged, PyObject *__pyx_v_use_filesystem_for_exec, PyObject *__pyx_v_search_specific_files, PyObject *__pyx_v_state, PyObject *__pyx_v_source_index, PyObject *__pyx_v_target_index, PyObject *__pyx_v_want_unversioned, PyObject *__pyx_v_tree); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_35494349;
static PyObject *__pyx_int_106938589;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__5;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__13;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
/* Late includes */

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":139
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr[] = "Just to expose _my_memrchr for testing.\n\n    :param s: The Python string to search\n    :param c: The character to search for\n    :return: The offset to the last instance of 'c' in s\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr = {"_py_memrchr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_c = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_memrchr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_s,&__pyx_n_s_c,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_c)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_py_memrchr", 1, 2, 2, 1); __PYX_ERR(0, 139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_py_memrchr") < 0)) __PYX_ERR(0, 139, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_s = values[0];
    __pyx_v_c = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_py_memrchr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 139, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._py_memrchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(__pyx_self, __pyx_v_s, __pyx_v_c);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx__py_memrchr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s, PyObject *__pyx_v_c) {
  void *__pyx_v__s;
  void *__pyx_v_found;
  int __pyx_v_length;
  char *__pyx_v__c;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_py_memrchr", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":151
 *     cdef char *_c
 * 
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c             # <<<<<<<<<<<<<<
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((PyBytes_Size(__pyx_v_c) == 1) != 0))) {
      __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Expected_single_character_string, __pyx_v_c); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      PyErr_SetObject(PyExc_AssertionError, __pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 151, __pyx_L1_error)
    }
  }
  #endif

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":152
 * 
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c
 *     _c = PyBytes_AsString(c)             # <<<<<<<<<<<<<<
 *     _s = PyBytes_AsString(s)
 *     length = PyBytes_Size(s)
 */
  __pyx_v__c = PyBytes_AsString(__pyx_v_c);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":153
 *     assert PyBytes_Size(c) == 1, 'Expected single character string not %r' % c
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)             # <<<<<<<<<<<<<<
 *     length = PyBytes_Size(s)
 * 
 */
  __pyx_v__s = PyBytes_AsString(__pyx_v_s);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":154
 *     _c = PyBytes_AsString(c)
 *     _s = PyBytes_AsString(s)
 *     length = PyBytes_Size(s)             # <<<<<<<<<<<<<<
 * 
 *     found = _my_memrchr(_s, _c[0], length)
 */
  __pyx_v_length = PyBytes_Size(__pyx_v_s);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":156
 *     length = PyBytes_Size(s)
 * 
 *     found = _my_memrchr(_s, _c[0], length)             # <<<<<<<<<<<<<<
 *     if found == NULL:
 *         return None
 */
  __pyx_v_found = __pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v__s, (__pyx_v__c[0]), __pyx_v_length);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":157
 * 
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <char*>found - <char*>_s
 */
  __pyx_t_2 = ((__pyx_v_found == NULL) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":158
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <char*>found - <char*>_s
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":157
 * 
 *     found = _my_memrchr(_s, _c[0], length)
 *     if found == NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <char*>found - <char*>_s
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":159
 *     if found == NULL:
 *         return None
 *     return <char*>found - <char*>_s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_ptrdiff_t((((char *)__pyx_v_found) - ((char *)__pyx_v__s))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":139
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._py_memrchr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":162
 * 
 * 
 * cdef int _is_aligned(void *ptr): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Is this pointer aligned to an integer size offset?
 * 
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(void *__pyx_v_ptr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_aligned", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":167
 *     :return: 1 if this pointer is aligned, 0 otherwise.
 *     """
 *     return ((<intptr_t>ptr) & ((sizeof(int))-1)) == 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((((intptr_t)__pyx_v_ptr) & ((sizeof(int)) - 1)) == 0);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":162
 * 
 * 
 * cdef int _is_aligned(void *ptr): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Is this pointer aligned to an integer size offset?
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":170
 * 
 * 
 * cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned char *cur1
 *     cdef unsigned char *cur2
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(char *__pyx_v_path1, int __pyx_v_size1, char *__pyx_v_path2, int __pyx_v_size2) {
  unsigned char *__pyx_v_cur1;
  unsigned char *__pyx_v_cur2;
  unsigned char *__pyx_v_end1;
  unsigned char *__pyx_v_end2;
  int *__pyx_v_cur_int1;
  int *__pyx_v_cur_int2;
  int *__pyx_v_end_int1;
  int *__pyx_v_end_int2;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  size_t __pyx_t_3;
  __Pyx_RefNannySetupContext("_cmp_by_dirs", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":180
 *     cdef int *end_int2
 * 
 *     if path1 == path2 and size1 == size2:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_path1 == __pyx_v_path2) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_size1 == __pyx_v_size2) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":181
 * 
 *     if path1 == path2 and size1 == size2:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     end1 = <unsigned char*>path1+size1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":180
 *     cdef int *end_int2
 * 
 *     if path1 == path2 and size1 == size2:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":183
 *         return 0
 * 
 *     end1 = <unsigned char*>path1+size1             # <<<<<<<<<<<<<<
 *     end2 = <unsigned char*>path2+size2
 * 
 */
  __pyx_v_end1 = (((unsigned char *)__pyx_v_path1) + __pyx_v_size1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":184
 * 
 *     end1 = <unsigned char*>path1+size1
 *     end2 = <unsigned char*>path2+size2             # <<<<<<<<<<<<<<
 * 
 *     # Use 32-bit comparisons for the matching portion of the string.
 */
  __pyx_v_end2 = (((unsigned char *)__pyx_v_path2) + __pyx_v_size2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":191
 *     # 99% of the time, these will be aligned, but in case they aren't just skip
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):             # <<<<<<<<<<<<<<
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
 */
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(__pyx_v_path1) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__is_aligned(__pyx_v_path2) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":192
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):
 *         cur_int1 = <int*>path1             # <<<<<<<<<<<<<<
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
 */
    __pyx_v_cur_int1 = ((int *)__pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":193
 *     if _is_aligned(path1) and _is_aligned(path2):
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2             # <<<<<<<<<<<<<<
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
 */
    __pyx_v_cur_int2 = ((int *)__pyx_v_path2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":194
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))             # <<<<<<<<<<<<<<
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
 * 
 */
    __pyx_t_3 = (sizeof(int));
    if (unlikely(__pyx_t_3 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(0, 194, __pyx_L1_error)
    }
    __pyx_v_end_int1 = ((int *)((__pyx_v_path1 + __pyx_v_size1) - (__pyx_v_size1 % __pyx_t_3)));

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":195
 *         cur_int2 = <int*>path2
 *         end_int1 = <int*>(path1 + size1 - (size1 % sizeof(int)))
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))             # <<<<<<<<<<<<<<
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 */
    __pyx_t_3 = (sizeof(int));
    if (unlikely(__pyx_t_3 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(0, 195, __pyx_L1_error)
    }
    __pyx_v_end_int2 = ((int *)((__pyx_v_path2 + __pyx_v_size2) - (__pyx_v_size2 % __pyx_t_3)));

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":197
 *         end_int2 = <int*>(path2 + size2 - (size2 % sizeof(int)))
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:             # <<<<<<<<<<<<<<
 *             if cur_int1[0] != cur_int2[0]:
 *                 break
 */
    while (1) {
      __pyx_t_2 = ((__pyx_v_cur_int1 < __pyx_v_end_int1) != 0);
      if (__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_2 = ((__pyx_v_cur_int2 < __pyx_v_end_int2) != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":198
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:             # <<<<<<<<<<<<<<
 *                 break
 *             cur_int1 = cur_int1 + 1
 */
      __pyx_t_1 = (((__pyx_v_cur_int1[0]) != (__pyx_v_cur_int2[0])) != 0);
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":199
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:
 *                 break             # <<<<<<<<<<<<<<
 *             cur_int1 = cur_int1 + 1
 *             cur_int2 = cur_int2 + 1
 */
        goto __pyx_L10_break;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":198
 * 
 *         while cur_int1 < end_int1 and cur_int2 < end_int2:
 *             if cur_int1[0] != cur_int2[0]:             # <<<<<<<<<<<<<<
 *                 break
 *             cur_int1 = cur_int1 + 1
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":200
 *             if cur_int1[0] != cur_int2[0]:
 *                 break
 *             cur_int1 = cur_int1 + 1             # <<<<<<<<<<<<<<
 *             cur_int2 = cur_int2 + 1
 * 
 */
      __pyx_v_cur_int1 = (__pyx_v_cur_int1 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":201
 *                 break
 *             cur_int1 = cur_int1 + 1
 *             cur_int2 = cur_int2 + 1             # <<<<<<<<<<<<<<
 * 
 *         cur1 = <unsigned char*>cur_int1
 */
      __pyx_v_cur_int2 = (__pyx_v_cur_int2 + 1);
    }
    __pyx_L10_break:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":203
 *             cur_int2 = cur_int2 + 1
 * 
 *         cur1 = <unsigned char*>cur_int1             # <<<<<<<<<<<<<<
 *         cur2 = <unsigned char*>cur_int2
 *     else:
 */
    __pyx_v_cur1 = ((unsigned char *)__pyx_v_cur_int1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":204
 * 
 *         cur1 = <unsigned char*>cur_int1
 *         cur2 = <unsigned char*>cur_int2             # <<<<<<<<<<<<<<
 *     else:
 *         cur1 = <unsigned char*>path1
 */
    __pyx_v_cur2 = ((unsigned char *)__pyx_v_cur_int2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":191
 *     # 99% of the time, these will be aligned, but in case they aren't just skip
 *     # this loop
 *     if _is_aligned(path1) and _is_aligned(path2):             # <<<<<<<<<<<<<<
 *         cur_int1 = <int*>path1
 *         cur_int2 = <int*>path2
 */
    goto __pyx_L6;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":206
 *         cur2 = <unsigned char*>cur_int2
 *     else:
 *         cur1 = <unsigned char*>path1             # <<<<<<<<<<<<<<
 *         cur2 = <unsigned char*>path2
 * 
 */
  /*else*/ {
    __pyx_v_cur1 = ((unsigned char *)__pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":207
 *     else:
 *         cur1 = <unsigned char*>path1
 *         cur2 = <unsigned char*>path2             # <<<<<<<<<<<<<<
 * 
 *     while cur1 < end1 and cur2 < end2:
 */
    __pyx_v_cur2 = ((unsigned char *)__pyx_v_path2);
  }
  __pyx_L6:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":209
 *         cur2 = <unsigned char*>path2
 * 
 *     while cur1 < end1 and cur2 < end2:             # <<<<<<<<<<<<<<
 *         if cur1[0] == cur2[0]:
 *             # This character matches, just go to the next one
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_cur1 < __pyx_v_end1) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L16_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_cur2 < __pyx_v_end2) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L16_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":210
 * 
 *     while cur1 < end1 and cur2 < end2:
 *         if cur1[0] == cur2[0]:             # <<<<<<<<<<<<<<
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
 */
    __pyx_t_1 = (((__pyx_v_cur1[0]) == (__pyx_v_cur2[0])) != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":212
 *         if cur1[0] == cur2[0]:
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1             # <<<<<<<<<<<<<<
 *             cur2 = cur2 + 1
 *             continue
 */
      __pyx_v_cur1 = (__pyx_v_cur1 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":213
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
 *             cur2 = cur2 + 1             # <<<<<<<<<<<<<<
 *             continue
 *         # The current characters do not match
 */
      __pyx_v_cur2 = (__pyx_v_cur2 + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":214
 *             cur1 = cur1 + 1
 *             cur2 = cur2 + 1
 *             continue             # <<<<<<<<<<<<<<
 *         # The current characters do not match
 *         if cur1[0] == b'/':
 */
      goto __pyx_L14_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":210
 * 
 *     while cur1 < end1 and cur2 < end2:
 *         if cur1[0] == cur2[0]:             # <<<<<<<<<<<<<<
 *             # This character matches, just go to the next one
 *             cur1 = cur1 + 1
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":216
 *             continue
 *         # The current characters do not match
 *         if cur1[0] == b'/':             # <<<<<<<<<<<<<<
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
 */
    __pyx_t_1 = (((__pyx_v_cur1[0]) == '/') != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":217
 *         # The current characters do not match
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first             # <<<<<<<<<<<<<<
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":216
 *             continue
 *         # The current characters do not match
 *         if cur1[0] == b'/':             # <<<<<<<<<<<<<<
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":218
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':             # <<<<<<<<<<<<<<
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
 */
    __pyx_t_1 = (((__pyx_v_cur2[0]) == '/') != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":219
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first             # <<<<<<<<<<<<<<
 *         elif cur1[0] < cur2[0]:
 *             return -1
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":218
 *         if cur1[0] == b'/':
 *             return -1 # Reached the end of path1 segment first
 *         elif cur2[0] == b'/':             # <<<<<<<<<<<<<<
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":220
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
 */
    __pyx_t_1 = (((__pyx_v_cur1[0]) < (__pyx_v_cur2[0])) != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":221
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:
 *             return -1             # <<<<<<<<<<<<<<
 *         else:
 *             return 1
 */
      __pyx_r = -1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":220
 *         elif cur2[0] == b'/':
 *             return 1 # Reached the end of path2 segment first
 *         elif cur1[0] < cur2[0]:             # <<<<<<<<<<<<<<
 *             return -1
 *         else:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":223
 *             return -1
 *         else:
 *             return 1             # <<<<<<<<<<<<<<
 * 
 *     # We reached the end of at least one of the strings
 */
    /*else*/ {
      __pyx_r = 1;
      goto __pyx_L0;
    }
    __pyx_L14_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":226
 * 
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:             # <<<<<<<<<<<<<<
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
 */
  __pyx_t_1 = ((__pyx_v_cur1 < __pyx_v_end1) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":227
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2             # <<<<<<<<<<<<<<
 *     if cur2 < end2:
 *         return -1 # At the end of cur1, but not at cur2
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":226
 * 
 *     # We reached the end of at least one of the strings
 *     if cur1 < end1:             # <<<<<<<<<<<<<<
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":228
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:             # <<<<<<<<<<<<<<
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
 */
  __pyx_t_1 = ((__pyx_v_cur2 < __pyx_v_end2) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":229
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:
 *         return -1 # At the end of cur1, but not at cur2             # <<<<<<<<<<<<<<
 *     # We reached the end of both strings
 *     return 0
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":228
 *     if cur1 < end1:
 *         return 1 # Not at the end of cur1, must be at the end of cur2
 *     if cur2 < end2:             # <<<<<<<<<<<<<<
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":231
 *         return -1 # At the end of cur1, but not at cur2
 *     # We reached the end of both strings
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":170
 * 
 * 
 * cdef int _cmp_by_dirs(char *path1, int size1, char *path2, int size2): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned char *cur1
 *     cdef unsigned char *cur2
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("breezy.bzr._dirstate_helpers_pyx._cmp_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":234
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs[] = "Compare two paths directory by directory.\n\n    This is equivalent to doing::\n\n       operator.lt(path1.split('/'), path2.split('/'))\n\n    The idea is that you should compare path components separately. This\n    differs from plain ``path1 < path2`` for paths like ``'a-b'`` and ``a/b``.\n    \"a-b\" comes after \"a\" but would come before \"a/b\" lexically.\n\n    :param path1: first path\n    :param path2: second path\n    :return: True if path1 comes first, otherwise False\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs = {"lt_by_dirs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path1 = 0;
  PyObject *__pyx_v_path2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lt_by_dirs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path1,&__pyx_n_s_path2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lt_by_dirs", 1, 2, 2, 1); __PYX_ERR(0, 234, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lt_by_dirs") < 0)) __PYX_ERR(0, 234, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_path1 = values[0];
    __pyx_v_path2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lt_by_dirs", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 234, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.lt_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(__pyx_self, __pyx_v_path1, __pyx_v_path2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_2lt_by_dirs(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("lt_by_dirs", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":249
 *     :return: True if path1 comes first, otherwise False
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path1) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":251
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path1)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path1)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path1)));
    __Pyx_INCREF(__pyx_v_path1);
    __Pyx_GIVEREF(__pyx_v_path1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path1);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_path1_must_be_a_bytes_string_no, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":250
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 250, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":249
 *     :return: True if path1 comes first, otherwise False
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":252
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path2) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":254
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))             # <<<<<<<<<<<<<<
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),
 *                               PyBytes_Size(path1),
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path2)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path2)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path2)));
    __Pyx_INCREF(__pyx_v_path2);
    __Pyx_GIVEREF(__pyx_v_path2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_path2_must_be_a_bytes_string_no, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":253
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 253, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":252
 *         raise TypeError("'path1' must be a bytes string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":255
 *         raise TypeError("'path2' must be a bytes string, not %s: %r"
 *                         % (type(path2), path2))
 *     return -1 == _cmp_by_dirs(PyBytes_AsString(path1),             # <<<<<<<<<<<<<<
 *                               PyBytes_Size(path1),
 *                               PyBytes_AsString(path2),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":258
 *                               PyBytes_Size(path1),
 *                               PyBytes_AsString(path2),
 *                               PyBytes_Size(path2))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((-1L == __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(PyBytes_AsString(__pyx_v_path1), PyBytes_Size(__pyx_v_path1), PyBytes_AsString(__pyx_v_path2), PyBytes_Size(__pyx_v_path2)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":234
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.lt_by_dirs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":261
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock[] = "Compare two paths based on what directory they are in.\n\n    This generates a sort order, such that all children of a directory are\n    sorted together, and grandchildren are in the same order as the\n    children appear. But all grandchildren come after all children.\n\n    In other words, all entries in a directory are sorted together, and\n    directories are sorted in cmp_by_dirs order.\n\n    :param path1: first path\n    :param path2: the second path\n    :return: True if path1 comes first, otherwise False.\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock = {"_lt_path_by_dirblock", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path1 = 0;
  PyObject *__pyx_v_path2 = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lt_path_by_dirblock (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path1,&__pyx_n_s_path2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_lt_path_by_dirblock", 1, 2, 2, 1); __PYX_ERR(0, 261, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_lt_path_by_dirblock") < 0)) __PYX_ERR(0, 261, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_path1 = values[0];
    __pyx_v_path2 = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_lt_path_by_dirblock", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 261, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._lt_path_by_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(__pyx_self, __pyx_v_path1, __pyx_v_path2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_4_lt_path_by_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path1, PyObject *__pyx_v_path2) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_lt_path_by_dirblock", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":275
 *     :return: True if path1 comes first, otherwise False.
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path1) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":277
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path1)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path1)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path1)));
    __Pyx_INCREF(__pyx_v_path1);
    __Pyx_GIVEREF(__pyx_v_path1);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path1);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_path1_must_be_a_plain_string_no, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":276
 *     """
 *     if not PyBytes_CheckExact(path1):
 *         raise TypeError("'path1' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 276, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":275
 *     :return: True if path1 comes first, otherwise False.
 *     """
 *     if not PyBytes_CheckExact(path1):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":278
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path2) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":280
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))             # <<<<<<<<<<<<<<
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
 *     return (_cmp_path_by_dirblock_intern(PyBytes_AsString(path1),
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path2)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path2)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path2)));
    __Pyx_INCREF(__pyx_v_path2);
    __Pyx_GIVEREF(__pyx_v_path2);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_path2_must_be_a_plain_string_no, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":279
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):
 *         raise TypeError("'path2' must be a plain string, not %s: %r"             # <<<<<<<<<<<<<<
 *                         % (type(path2), path2))
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 279, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":278
 *         raise TypeError("'path1' must be a plain string, not %s: %r"
 *                         % (type(path1), path1))
 *     if not PyBytes_CheckExact(path2):             # <<<<<<<<<<<<<<
 *         raise TypeError("'path2' must be a plain string, not %s: %r"
 *                         % (type(path2), path2))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":282
 *                         % (type(path2), path2))
 *     # GZ 2017-06-09: This internal function really only needs lt as well.
 *     return (_cmp_path_by_dirblock_intern(PyBytes_AsString(path1),             # <<<<<<<<<<<<<<
 *                                          PyBytes_Size(path1),
 *                                          PyBytes_AsString(path2),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":285
 *                                          PyBytes_Size(path1),
 *                                          PyBytes_AsString(path2),
 *                                          PyBytes_Size(path2)) < 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(PyBytes_AsString(__pyx_v_path1), PyBytes_Size(__pyx_v_path1), PyBytes_AsString(__pyx_v_path2), PyBytes_Size(__pyx_v_path2)) < 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":261
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._lt_path_by_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":288
 * 
 * 
 * cdef int _cmp_path_by_dirblock_intern(char *path1, int path1_len,             # <<<<<<<<<<<<<<
 *                                       char *path2, int path2_len): # cannot_raise
 *     """Compare two paths by what directory they are in.
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(char *__pyx_v_path1, int __pyx_v_path1_len, char *__pyx_v_path2, int __pyx_v_path2_len) {
  char *__pyx_v_dirname1;
  int __pyx_v_dirname1_len;
  char *__pyx_v_dirname2;
  int __pyx_v_dirname2_len;
  char *__pyx_v_basename1;
  int __pyx_v_basename1_len;
  char *__pyx_v_basename2;
  int __pyx_v_basename2_len;
  int __pyx_v_cur_len;
  int __pyx_v_cmp_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_cmp_path_by_dirblock_intern", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":305
 *     cdef int cmp_val
 * 
 *     if path1_len == 0 and path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_path1_len == 0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_path2_len == 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":306
 * 
 *     if path1_len == 0 and path2_len == 0:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if path1 == path2 and path1_len == path2_len:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":305
 *     cdef int cmp_val
 * 
 *     if path1_len == 0 and path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":308
 *         return 0
 * 
 *     if path1 == path2 and path1_len == path2_len:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_2 = ((__pyx_v_path1 == __pyx_v_path2) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_path1_len == __pyx_v_path2_len) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":309
 * 
 *     if path1 == path2 and path1_len == path2_len:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     if path1_len == 0:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":308
 *         return 0
 * 
 *     if path1 == path2 and path1_len == path2_len:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":311
 *         return 0
 * 
 *     if path1_len == 0:             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  __pyx_t_1 = ((__pyx_v_path1_len == 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":312
 * 
 *     if path1_len == 0:
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     if path2_len == 0:
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":311
 *         return 0
 * 
 *     if path1_len == 0:             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":314
 *         return -1
 * 
 *     if path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_t_1 = ((__pyx_v_path2_len == 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":315
 * 
 *     if path2_len == 0:
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":314
 *         return -1
 * 
 *     if path2_len == 0:             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":317
 *         return 1
 * 
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)             # <<<<<<<<<<<<<<
 * 
 *     if basename1 == NULL:
 */
  __pyx_v_basename1 = ((char *)__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_path1, '/', __pyx_v_path1_len));

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":319
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
 * 
 *     if basename1 == NULL:             # <<<<<<<<<<<<<<
 *         basename1 = path1
 *         basename1_len = path1_len
 */
  __pyx_t_1 = ((__pyx_v_basename1 == NULL) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":320
 * 
 *     if basename1 == NULL:
 *         basename1 = path1             # <<<<<<<<<<<<<<
 *         basename1_len = path1_len
 *         dirname1 = b''
 */
    __pyx_v_basename1 = __pyx_v_path1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":321
 *     if basename1 == NULL:
 *         basename1 = path1
 *         basename1_len = path1_len             # <<<<<<<<<<<<<<
 *         dirname1 = b''
 *         dirname1_len = 0
 */
    __pyx_v_basename1_len = __pyx_v_path1_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":322
 *         basename1 = path1
 *         basename1_len = path1_len
 *         dirname1 = b''             # <<<<<<<<<<<<<<
 *         dirname1_len = 0
 *     else:
 */
    __pyx_v_dirname1 = ((char *)"");

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":323
 *         basename1_len = path1_len
 *         dirname1 = b''
 *         dirname1_len = 0             # <<<<<<<<<<<<<<
 *     else:
 *         dirname1 = path1
 */
    __pyx_v_dirname1_len = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":319
 *     basename1 = <char*>_my_memrchr(path1, b'/', path1_len)
 * 
 *     if basename1 == NULL:             # <<<<<<<<<<<<<<
 *         basename1 = path1
 *         basename1_len = path1_len
 */
    goto __pyx_L11;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":325
 *         dirname1_len = 0
 *     else:
 *         dirname1 = path1             # <<<<<<<<<<<<<<
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1
 */
  /*else*/ {
    __pyx_v_dirname1 = __pyx_v_path1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":326
 *     else:
 *         dirname1 = path1
 *         dirname1_len = basename1 - path1             # <<<<<<<<<<<<<<
 *         basename1 = basename1 + 1
 *         basename1_len = path1_len - dirname1_len - 1
 */
    __pyx_v_dirname1_len = (__pyx_v_basename1 - __pyx_v_path1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":327
 *         dirname1 = path1
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1             # <<<<<<<<<<<<<<
 *         basename1_len = path1_len - dirname1_len - 1
 * 
 */
    __pyx_v_basename1 = (__pyx_v_basename1 + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":328
 *         dirname1_len = basename1 - path1
 *         basename1 = basename1 + 1
 *         basename1_len = path1_len - dirname1_len - 1             # <<<<<<<<<<<<<<
 * 
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
 */
    __pyx_v_basename1_len = ((__pyx_v_path1_len - __pyx_v_dirname1_len) - 1);
  }
  __pyx_L11:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":330
 *         basename1_len = path1_len - dirname1_len - 1
 * 
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)             # <<<<<<<<<<<<<<
 * 
 *     if basename2 == NULL:
 */
  __pyx_v_basename2 = ((char *)__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(__pyx_v_path2, '/', __pyx_v_path2_len));

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":332
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
 * 
 *     if basename2 == NULL:             # <<<<<<<<<<<<<<
 *         basename2 = path2
 *         basename2_len = path2_len
 */
  __pyx_t_1 = ((__pyx_v_basename2 == NULL) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":333
 * 
 *     if basename2 == NULL:
 *         basename2 = path2             # <<<<<<<<<<<<<<
 *         basename2_len = path2_len
 *         dirname2 = b''
 */
    __pyx_v_basename2 = __pyx_v_path2;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":334
 *     if basename2 == NULL:
 *         basename2 = path2
 *         basename2_len = path2_len             # <<<<<<<<<<<<<<
 *         dirname2 = b''
 *         dirname2_len = 0
 */
    __pyx_v_basename2_len = __pyx_v_path2_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":335
 *         basename2 = path2
 *         basename2_len = path2_len
 *         dirname2 = b''             # <<<<<<<<<<<<<<
 *         dirname2_len = 0
 *     else:
 */
    __pyx_v_dirname2 = ((char *)"");

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":336
 *         basename2_len = path2_len
 *         dirname2 = b''
 *         dirname2_len = 0             # <<<<<<<<<<<<<<
 *     else:
 *         dirname2 = path2
 */
    __pyx_v_dirname2_len = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":332
 *     basename2 = <char*>_my_memrchr(path2, b'/', path2_len)
 * 
 *     if basename2 == NULL:             # <<<<<<<<<<<<<<
 *         basename2 = path2
 *         basename2_len = path2_len
 */
    goto __pyx_L12;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":338
 *         dirname2_len = 0
 *     else:
 *         dirname2 = path2             # <<<<<<<<<<<<<<
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1
 */
  /*else*/ {
    __pyx_v_dirname2 = __pyx_v_path2;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":339
 *     else:
 *         dirname2 = path2
 *         dirname2_len = basename2 - path2             # <<<<<<<<<<<<<<
 *         basename2 = basename2 + 1
 *         basename2_len = path2_len - dirname2_len - 1
 */
    __pyx_v_dirname2_len = (__pyx_v_basename2 - __pyx_v_path2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":340
 *         dirname2 = path2
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1             # <<<<<<<<<<<<<<
 *         basename2_len = path2_len - dirname2_len - 1
 * 
 */
    __pyx_v_basename2 = (__pyx_v_basename2 + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":341
 *         dirname2_len = basename2 - path2
 *         basename2 = basename2 + 1
 *         basename2_len = path2_len - dirname2_len - 1             # <<<<<<<<<<<<<<
 * 
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
 */
    __pyx_v_basename2_len = ((__pyx_v_path2_len - __pyx_v_dirname2_len) - 1);
  }
  __pyx_L12:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":343
 *         basename2_len = path2_len - dirname2_len - 1
 * 
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,             # <<<<<<<<<<<<<<
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:
 */
  __pyx_v_cmp_val = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_dirname1, __pyx_v_dirname1_len, __pyx_v_dirname2, __pyx_v_dirname2_len);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":345
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 * 
 */
  __pyx_t_1 = ((__pyx_v_cmp_val != 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":346
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:
 *         return cmp_val             # <<<<<<<<<<<<<<
 * 
 *     cur_len = basename1_len
 */
    __pyx_r = __pyx_v_cmp_val;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":345
 *     cmp_val = _cmp_by_dirs(dirname1, dirname1_len,
 *                            dirname2, dirname2_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":348
 *         return cmp_val
 * 
 *     cur_len = basename1_len             # <<<<<<<<<<<<<<
 *     if basename2_len < basename1_len:
 *         cur_len = basename2_len
 */
  __pyx_v_cur_len = __pyx_v_basename1_len;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":349
 * 
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:             # <<<<<<<<<<<<<<
 *         cur_len = basename2_len
 * 
 */
  __pyx_t_1 = ((__pyx_v_basename2_len < __pyx_v_basename1_len) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":350
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:
 *         cur_len = basename2_len             # <<<<<<<<<<<<<<
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 */
    __pyx_v_cur_len = __pyx_v_basename2_len;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":349
 * 
 *     cur_len = basename1_len
 *     if basename2_len < basename1_len:             # <<<<<<<<<<<<<<
 *         cur_len = basename2_len
 * 
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":352
 *         cur_len = basename2_len
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)             # <<<<<<<<<<<<<<
 *     if cmp_val != 0:
 *         return cmp_val
 */
  __pyx_v_cmp_val = memcmp(__pyx_v_basename1, __pyx_v_basename2, __pyx_v_cur_len);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":353
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 *     if basename1_len == basename2_len:
 */
  __pyx_t_1 = ((__pyx_v_cmp_val != 0) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":354
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:
 *         return cmp_val             # <<<<<<<<<<<<<<
 *     if basename1_len == basename2_len:
 *         return 0
 */
    __pyx_r = __pyx_v_cmp_val;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":353
 * 
 *     cmp_val = memcmp(basename1, basename2, cur_len)
 *     if cmp_val != 0:             # <<<<<<<<<<<<<<
 *         return cmp_val
 *     if basename1_len == basename2_len:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":355
 *     if cmp_val != 0:
 *         return cmp_val
 *     if basename1_len == basename2_len:             # <<<<<<<<<<<<<<
 *         return 0
 *     if basename1_len < basename2_len:
 */
  __pyx_t_1 = ((__pyx_v_basename1_len == __pyx_v_basename2_len) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":356
 *         return cmp_val
 *     if basename1_len == basename2_len:
 *         return 0             # <<<<<<<<<<<<<<
 *     if basename1_len < basename2_len:
 *         return -1
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":355
 *     if cmp_val != 0:
 *         return cmp_val
 *     if basename1_len == basename2_len:             # <<<<<<<<<<<<<<
 *         return 0
 *     if basename1_len < basename2_len:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":357
 *     if basename1_len == basename2_len:
 *         return 0
 *     if basename1_len < basename2_len:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 1
 */
  __pyx_t_1 = ((__pyx_v_basename1_len < __pyx_v_basename2_len) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":358
 *         return 0
 *     if basename1_len < basename2_len:
 *         return -1             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":357
 *     if basename1_len == basename2_len:
 *         return 0
 *     if basename1_len < basename2_len:             # <<<<<<<<<<<<<<
 *         return -1
 *     return 1
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":359
 *     if basename1_len < basename2_len:
 *         return -1
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":288
 * 
 * 
 * cdef int _cmp_path_by_dirblock_intern(char *path1, int path1_len,             # <<<<<<<<<<<<<<
 *                                       char *path2, int path2_len): # cannot_raise
 *     """Compare two paths by what directory they are in.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":362
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left[] = "Return the index where to insert path into paths.\n\n    This uses a path-wise comparison so we get::\n        a\n        a-b\n        a=b\n        a/b\n    Rather than::\n        a\n        a-b\n        a/b\n        a=b\n    :param paths: A list of paths to search through\n    :param path: A single path to insert\n    :return: An offset where 'path' can be inserted.\n    :seealso: bisect.bisect_left\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left = {"_bisect_path_left", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_paths = 0;
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bisect_path_left (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_paths,&__pyx_n_s_path,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_paths)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bisect_path_left", 1, 2, 2, 1); __PYX_ERR(0, 362, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_bisect_path_left") < 0)) __PYX_ERR(0, 362, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_paths = values[0];
    __pyx_v_path = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bisect_path_left", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 362, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(__pyx_self, __pyx_v_paths, __pyx_v_path);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6_bisect_path_left(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_path_cstr;
  int __pyx_v_path_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_bisect_path_left", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":389
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 */
  __pyx_t_1 = ((!(PyList_CheckExact(__pyx_v_paths) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":391
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_paths)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_paths)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_paths)));
    __Pyx_INCREF(__pyx_v_paths);
    __Pyx_GIVEREF(__pyx_v_paths);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_paths);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_python_list_for, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":390
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 390, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":389
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":392
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":394
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))             # <<<<<<<<<<<<<<
 * 
 *     _hi = len(paths)
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path)));
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_string_for_path, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":393
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 393, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":392
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":396
 *                         % (type(path), path))
 * 
 *     _hi = len(paths)             # <<<<<<<<<<<<<<
 *     _lo = 0
 * 
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_paths); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_v__hi = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":397
 * 
 *     _hi = len(paths)
 *     _lo = 0             # <<<<<<<<<<<<<<
 * 
 *     path_cstr = PyBytes_AsString(path)
 */
  __pyx_v__lo = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":399
 *     _lo = 0
 * 
 *     path_cstr = PyBytes_AsString(path)             # <<<<<<<<<<<<<<
 *     path_size = PyBytes_Size(path)
 * 
 */
  __pyx_v_path_cstr = PyBytes_AsString(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":400
 * 
 *     path_cstr = PyBytes_AsString(path)
 *     path_size = PyBytes_Size(path)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
 */
  __pyx_v_path_size = PyBytes_Size(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":402
 *     path_size = PyBytes_Size(path)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v__lo < __pyx_v__hi) != 0);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":403
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 */
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":404
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)             # <<<<<<<<<<<<<<
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 */
    __pyx_v_cur = PyList_GET_ITEM(__pyx_v_paths, __pyx_v__mid);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":405
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 */
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":406
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:
 */
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":408
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
 */
    __pyx_t_1 = ((__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(__pyx_v_cur_cstr, __pyx_v_cur_size, __pyx_v_path_cstr, __pyx_v_path_size) < 0) != 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":407
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,             # <<<<<<<<<<<<<<
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1
 */
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":409
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *         else:
 *             _hi = _mid
 */
      __pyx_v__lo = (__pyx_v__mid + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":407
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(cur_cstr, cur_size,             # <<<<<<<<<<<<<<
 *                                         path_cstr, path_size) < 0:
 *             _lo = _mid + 1
 */
      goto __pyx_L7;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":411
 *             _lo = _mid + 1
 *         else:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *     return _lo
 * 
 */
    /*else*/ {
      __pyx_v__hi = __pyx_v__mid;
    }
    __pyx_L7:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":412
 *         else:
 *             _hi = _mid
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":362
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_left", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":415
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right[] = "Return the index where to insert path into paths.\n\n    This uses a path-wise comparison so we get::\n        a\n        a-b\n        a=b\n        a/b\n    Rather than::\n        a\n        a-b\n        a/b\n        a=b\n    :param paths: A list of paths to search through\n    :param path: A single path to insert\n    :return: An offset where 'path' can be inserted.\n    :seealso: bisect.bisect_right\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right = {"_bisect_path_right", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_paths = 0;
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bisect_path_right (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_paths,&__pyx_n_s_path,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_paths)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bisect_path_right", 1, 2, 2, 1); __PYX_ERR(0, 415, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_bisect_path_right") < 0)) __PYX_ERR(0, 415, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_paths = values[0];
    __pyx_v_path = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bisect_path_right", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 415, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(__pyx_self, __pyx_v_paths, __pyx_v_path);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_8_bisect_path_right(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_paths, PyObject *__pyx_v_path) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_path_cstr;
  int __pyx_v_path_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_bisect_path_right", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":442
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 */
  __pyx_t_1 = ((!(PyList_CheckExact(__pyx_v_paths) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":444
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_paths)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_paths)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_paths)));
    __Pyx_INCREF(__pyx_v_paths);
    __Pyx_GIVEREF(__pyx_v_paths);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_paths);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_python_list_for, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":443
 * 
 *     if not PyList_CheckExact(paths):
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 443, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":442
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(paths):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":445
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_path) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":447
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))             # <<<<<<<<<<<<<<
 * 
 *     _hi = len(paths)
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_path)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_path)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_path)));
    __Pyx_INCREF(__pyx_v_path);
    __Pyx_GIVEREF(__pyx_v_path);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_path);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_string_for_path, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":446
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):
 *         raise TypeError("you must pass a string for 'path' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(path), path))
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 446, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":445
 *         raise TypeError("you must pass a python list for 'paths' not: %s %r"
 *                         % (type(paths), paths))
 *     if not PyBytes_CheckExact(path):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for 'path' not: %s %r"
 *                         % (type(path), path))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":449
 *                         % (type(path), path))
 * 
 *     _hi = len(paths)             # <<<<<<<<<<<<<<
 *     _lo = 0
 * 
 */
  __pyx_t_4 = PyObject_Length(__pyx_v_paths); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 449, __pyx_L1_error)
  __pyx_v__hi = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":450
 * 
 *     _hi = len(paths)
 *     _lo = 0             # <<<<<<<<<<<<<<
 * 
 *     path_cstr = PyBytes_AsString(path)
 */
  __pyx_v__lo = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":452
 *     _lo = 0
 * 
 *     path_cstr = PyBytes_AsString(path)             # <<<<<<<<<<<<<<
 *     path_size = PyBytes_Size(path)
 * 
 */
  __pyx_v_path_cstr = PyBytes_AsString(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":453
 * 
 *     path_cstr = PyBytes_AsString(path)
 *     path_size = PyBytes_Size(path)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
 */
  __pyx_v_path_size = PyBytes_Size(__pyx_v_path);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":455
 *     path_size = PyBytes_Size(path)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v__lo < __pyx_v__hi) != 0);
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":456
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 */
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":457
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)             # <<<<<<<<<<<<<<
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 */
    __pyx_v_cur = PyList_GET_ITEM(__pyx_v_paths, __pyx_v__mid);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":458
 *         _mid = (_lo + _hi) // 2
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 */
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":459
 *         cur = PyList_GetItem_object_void(paths, _mid)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:
 */
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":461
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:             # <<<<<<<<<<<<<<
 *             _hi = _mid
 *         else:
 */
    __pyx_t_1 = ((__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_path_by_dirblock_intern(__pyx_v_path_cstr, __pyx_v_path_size, __pyx_v_cur_cstr, __pyx_v_cur_size) < 0) != 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,             # <<<<<<<<<<<<<<
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid
 */
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":462
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *         else:
 *             _lo = _mid + 1
 */
      __pyx_v__hi = __pyx_v__mid;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":460
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_path_by_dirblock_intern(path_cstr, path_size,             # <<<<<<<<<<<<<<
 *                                         cur_cstr, cur_size) < 0:
 *             _hi = _mid
 */
      goto __pyx_L7;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":464
 *             _hi = _mid
 *         else:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *     return _lo
 * 
 */
    /*else*/ {
      __pyx_v__lo = (__pyx_v__mid + 1);
    }
    __pyx_L7:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":465
 *         else:
 *             _lo = _mid + 1
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":415
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._bisect_path_right", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":468
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock[] = "Return the index where to insert dirname into the dirblocks.\n\n    The return value idx is such that all directories blocks in dirblock[:idx]\n    have names < dirname, and all blocks in dirblock[idx:] have names >=\n    dirname.\n\n    Optional args lo (default 0) and hi (default len(dirblocks)) bound the\n    slice of a to be searched.\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock = {"bisect_dirblock", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dirblocks = 0;
  PyObject *__pyx_v_dirname = 0;
  PyObject *__pyx_v_lo = 0;
  PyObject *__pyx_v_hi = 0;
  CYTHON_UNUSED PyObject *__pyx_v_cache = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bisect_dirblock (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dirblocks,&__pyx_n_s_dirname,&__pyx_n_s_lo,&__pyx_n_s_hi,&__pyx_n_s_cache,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[2] = ((PyObject *)__pyx_int_0);
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dirblocks)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dirname)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("bisect_dirblock", 0, 2, 5, 1); __PYX_ERR(0, 468, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lo);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hi);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cache);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bisect_dirblock") < 0)) __PYX_ERR(0, 468, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_dirblocks = values[0];
    __pyx_v_dirname = values[1];
    __pyx_v_lo = values[2];
    __pyx_v_hi = values[3];
    __pyx_v_cache = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bisect_dirblock", 0, 2, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 468, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.bisect_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(__pyx_self, __pyx_v_dirblocks, __pyx_v_dirname, __pyx_v_lo, __pyx_v_hi, __pyx_v_cache);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_10bisect_dirblock(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dirblocks, PyObject *__pyx_v_dirname, PyObject *__pyx_v_lo, PyObject *__pyx_v_hi, CYTHON_UNUSED PyObject *__pyx_v_cache) {
  int __pyx_v__lo;
  int __pyx_v__hi;
  int __pyx_v__mid;
  char *__pyx_v_dirname_cstr;
  int __pyx_v_dirname_size;
  char *__pyx_v_cur_cstr;
  int __pyx_v_cur_size;
  void *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("bisect_dirblock", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":487
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(dirblocks):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 */
  __pyx_t_1 = ((!(PyList_CheckExact(__pyx_v_dirblocks) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":489
 *     if not PyList_CheckExact(dirblocks):
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))             # <<<<<<<<<<<<<<
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_dirblocks)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_dirblocks)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_dirblocks)));
    __Pyx_INCREF(__pyx_v_dirblocks);
    __Pyx_GIVEREF(__pyx_v_dirblocks);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_dirblocks);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_python_list_for_2, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":488
 * 
 *     if not PyList_CheckExact(dirblocks):
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 488, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":487
 *     cdef void *cur
 * 
 *     if not PyList_CheckExact(dirblocks):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":490
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 */
  __pyx_t_1 = ((!(PyBytes_CheckExact(__pyx_v_dirname) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":492
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))             # <<<<<<<<<<<<<<
 *     if hi is None:
 *         _hi = len(dirblocks)
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_dirname)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_dirname)));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(__pyx_v_dirname)));
    __Pyx_INCREF(__pyx_v_dirname);
    __Pyx_GIVEREF(__pyx_v_dirname);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_dirname);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_you_must_pass_a_string_for_dirna, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":491
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):
 *         raise TypeError("you must pass a string for dirname not: %s %r"             # <<<<<<<<<<<<<<
 *                         % (type(dirname), dirname))
 *     if hi is None:
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 491, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":490
 *         raise TypeError("you must pass a python list for 'dirblocks' not: %s %r"
 *                         % (type(dirblocks), dirblocks))
 *     if not PyBytes_CheckExact(dirname):             # <<<<<<<<<<<<<<
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":493
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 *     if hi is None:             # <<<<<<<<<<<<<<
 *         _hi = len(dirblocks)
 *     else:
 */
  __pyx_t_1 = (__pyx_v_hi == Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":494
 *                         % (type(dirname), dirname))
 *     if hi is None:
 *         _hi = len(dirblocks)             # <<<<<<<<<<<<<<
 *     else:
 *         _hi = hi
 */
    __pyx_t_5 = PyObject_Length(__pyx_v_dirblocks); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 494, __pyx_L1_error)
    __pyx_v__hi = __pyx_t_5;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":493
 *         raise TypeError("you must pass a string for dirname not: %s %r"
 *                         % (type(dirname), dirname))
 *     if hi is None:             # <<<<<<<<<<<<<<
 *         _hi = len(dirblocks)
 *     else:
 */
    goto __pyx_L5;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":496
 *         _hi = len(dirblocks)
 *     else:
 *         _hi = hi             # <<<<<<<<<<<<<<
 * 
 *     _lo = lo
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_hi); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 496, __pyx_L1_error)
    __pyx_v__hi = __pyx_t_6;
  }
  __pyx_L5:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":498
 *         _hi = hi
 * 
 *     _lo = lo             # <<<<<<<<<<<<<<
 *     dirname_cstr = PyBytes_AsString(dirname)
 *     dirname_size = PyBytes_Size(dirname)
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_lo); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_v__lo = __pyx_t_6;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":499
 * 
 *     _lo = lo
 *     dirname_cstr = PyBytes_AsString(dirname)             # <<<<<<<<<<<<<<
 *     dirname_size = PyBytes_Size(dirname)
 * 
 */
  __pyx_v_dirname_cstr = PyBytes_AsString(__pyx_v_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":500
 *     _lo = lo
 *     dirname_cstr = PyBytes_AsString(dirname)
 *     dirname_size = PyBytes_Size(dirname)             # <<<<<<<<<<<<<<
 * 
 *     while _lo < _hi:
 */
  __pyx_v_dirname_size = PyBytes_Size(__pyx_v_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":502
 *     dirname_size = PyBytes_Size(dirname)
 * 
 *     while _lo < _hi:             # <<<<<<<<<<<<<<
 *         _mid = (_lo + _hi) // 2
 *         # Grab the dirname for the current dirblock
 */
  while (1) {
    __pyx_t_4 = ((__pyx_v__lo < __pyx_v__hi) != 0);
    if (!__pyx_t_4) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":503
 * 
 *     while _lo < _hi:
 *         _mid = (_lo + _hi) // 2             # <<<<<<<<<<<<<<
 *         # Grab the dirname for the current dirblock
 *         # cur = dirblocks[_mid][0]
 */
    __pyx_v__mid = __Pyx_div_long((__pyx_v__lo + __pyx_v__hi), 2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":506
 *         # Grab the dirname for the current dirblock
 *         # cur = dirblocks[_mid][0]
 *         cur = PyTuple_GetItem_void_void(             # <<<<<<<<<<<<<<
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 */
    __pyx_v_cur = PyTuple_GET_ITEM(PyList_GET_ITEM(__pyx_v_dirblocks, __pyx_v__mid), 0);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":508
 *         cur = PyTuple_GetItem_void_void(
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)             # <<<<<<<<<<<<<<
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
 */
    __pyx_v_cur_cstr = PyBytes_AS_STRING(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":509
 *                 PyList_GetItem_object_void(dirblocks, _mid), 0)
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)             # <<<<<<<<<<<<<<
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
 *             _lo = _mid + 1
 */
    __pyx_v_cur_size = PyBytes_GET_SIZE(__pyx_v_cur);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":510
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
 */
    __pyx_t_4 = ((__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_cur_cstr, __pyx_v_cur_size, __pyx_v_dirname_cstr, __pyx_v_dirname_size) < 0) != 0);
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":511
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:
 *             _lo = _mid + 1             # <<<<<<<<<<<<<<
 *         else:
 *             _hi = _mid
 */
      __pyx_v__lo = (__pyx_v__mid + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":510
 *         cur_cstr = PyBytes_AS_STRING_void(cur)
 *         cur_size = PyBytes_GET_SIZE_void(cur)
 *         if _cmp_by_dirs(cur_cstr, cur_size, dirname_cstr, dirname_size) < 0:             # <<<<<<<<<<<<<<
 *             _lo = _mid + 1
 *         else:
 */
      goto __pyx_L8;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":513
 *             _lo = _mid + 1
 *         else:
 *             _hi = _mid             # <<<<<<<<<<<<<<
 *     return _lo
 * 
 */
    /*else*/ {
      __pyx_v__hi = __pyx_v__mid;
    }
    __pyx_L8:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":514
 *         else:
 *             _hi = _mid
 *     return _lo             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v__lo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":468
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.bisect_dirblock", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":529
 *     cdef char *next # Pointer to the end of this record
 * 
 *     def __init__(self, text, state):             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.text = text
 */

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_state = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_text,&__pyx_n_s_state,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_text)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 529, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 529, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_text = values[0];
    __pyx_v_state = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 529, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self), __pyx_v_text, __pyx_v_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_state) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":530
 * 
 *     def __init__(self, text, state):
 *         self.state = state             # <<<<<<<<<<<<<<
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)
 */
  __Pyx_INCREF(__pyx_v_state);
  __Pyx_GIVEREF(__pyx_v_state);
  __Pyx_GOTREF(__pyx_v_self->state);
  __Pyx_DECREF(__pyx_v_self->state);
  __pyx_v_self->state = __pyx_v_state;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":531
 *     def __init__(self, text, state):
 *         self.state = state
 *         self.text = text             # <<<<<<<<<<<<<<
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)
 */
  __Pyx_INCREF(__pyx_v_text);
  __Pyx_GIVEREF(__pyx_v_text);
  __Pyx_GOTREF(__pyx_v_self->text);
  __Pyx_DECREF(__pyx_v_self->text);
  __pyx_v_self->text = __pyx_v_text;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":532
 *         self.state = state
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)             # <<<<<<<<<<<<<<
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size
 */
  __pyx_v_self->text_cstr = PyBytes_AsString(__pyx_v_text);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":533
 *         self.text = text
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)             # <<<<<<<<<<<<<<
 *         self.end_cstr = self.text_cstr + self.text_size
 *         self.cur_cstr = self.text_cstr
 */
  __pyx_v_self->text_size = PyBytes_Size(__pyx_v_text);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":534
 *         self.text_cstr = PyBytes_AsString(text)
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size             # <<<<<<<<<<<<<<
 *         self.cur_cstr = self.text_cstr
 * 
 */
  __pyx_v_self->end_cstr = (__pyx_v_self->text_cstr + __pyx_v_self->text_size);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":535
 *         self.text_size = PyBytes_Size(text)
 *         self.end_cstr = self.text_cstr + self.text_size
 *         self.cur_cstr = self.text_cstr             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_next(self, int *size) except NULL:
 */
  __pyx_t_1 = __pyx_v_self->text_cstr;
  __pyx_v_self->cur_cstr = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":529
 *     cdef char *next # Pointer to the end of this record
 * 
 *     def __init__(self, text, state):             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.text = text
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":537
 *         self.cur_cstr = self.text_cstr
 * 
 *     cdef char *get_next(self, int *size) except NULL:             # <<<<<<<<<<<<<<
 *         """Return a pointer to the start of the next field."""
 *         cdef char *next
 */

static char *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int *__pyx_v_size) {
  char *__pyx_v_next;
  Py_ssize_t __pyx_v_extra_len;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("get_next", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":542
 *         cdef Py_ssize_t extra_len
 * 
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
 */
  __pyx_t_1 = ((__pyx_v_self->cur_cstr == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":543
 * 
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')             # <<<<<<<<<<<<<<
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 543, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":542
 *         cdef Py_ssize_t extra_len
 * 
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":544
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
 */
  __pyx_t_1 = ((__pyx_v_self->cur_cstr >= __pyx_v_self->end_cstr) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":545
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'             # <<<<<<<<<<<<<<
 *                                  ' left')
 *         next = self.cur_cstr
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 545, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":544
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:             # <<<<<<<<<<<<<<
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":547
 *             raise AssertionError('get_next() called when there are no chars'
 *                                  ' left')
 *         next = self.cur_cstr             # <<<<<<<<<<<<<<
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:
 */
  __pyx_t_3 = __pyx_v_self->cur_cstr;
  __pyx_v_next = __pyx_t_3;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":548
 *                                  ' left')
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)             # <<<<<<<<<<<<<<
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next
 */
  __pyx_v_self->cur_cstr = ((char *)memchr(__pyx_v_next, '\x00', (__pyx_v_self->end_cstr - __pyx_v_next)));

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":549
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,
 */
  __pyx_t_1 = ((__pyx_v_self->cur_cstr == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":550
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'failed to find trailing NULL (\\0).'
 */
    __pyx_v_extra_len = (__pyx_v_self->end_cstr - __pyx_v_next);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":551
 *         if self.cur_cstr == NULL:
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                 'failed to find trailing NULL (\\0).'
 *                 ' Trailing garbage: %r'
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":554
 *                 'failed to find trailing NULL (\\0).'
 *                 ' Trailing garbage: %r'
 *                 % safe_string_from_size(next, extra_len))             # <<<<<<<<<<<<<<
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1
 */
    __pyx_t_5 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_next, __pyx_v_extra_len); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyString_FormatSafe(__pyx_kp_s_failed_to_find_trailing_NULL_0_T, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self->state, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self->state, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_v_self->state);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 551, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":549
 *         next = self.cur_cstr
 *         self.cur_cstr = <char*>memchr(next, b'\0', self.end_cstr - next)
 *         if self.cur_cstr == NULL:             # <<<<<<<<<<<<<<
 *             extra_len = self.end_cstr - next
 *             raise DirstateCorrupt(self.state,
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":555
 *                 ' Trailing garbage: %r'
 *                 % safe_string_from_size(next, extra_len))
 *         size[0] = self.cur_cstr - next             # <<<<<<<<<<<<<<
 *         self.cur_cstr = self.cur_cstr + 1
 *         return next
 */
  (__pyx_v_size[0]) = (__pyx_v_self->cur_cstr - __pyx_v_next);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":556
 *                 % safe_string_from_size(next, extra_len))
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1             # <<<<<<<<<<<<<<
 *         return next
 * 
 */
  __pyx_v_self->cur_cstr = (__pyx_v_self->cur_cstr + 1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":557
 *         size[0] = self.cur_cstr - next
 *         self.cur_cstr = self.cur_cstr + 1
 *         return next             # <<<<<<<<<<<<<<
 * 
 *     cdef object get_next_str(self):
 */
  __pyx_r = __pyx_v_next;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":537
 *         self.cur_cstr = self.text_cstr
 * 
 *     cdef char *get_next(self, int *size) except NULL:             # <<<<<<<<<<<<<<
 *         """Return a pointer to the start of the next field."""
 *         cdef char *next
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.get_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":559
 *         return next
 * 
 *     cdef object get_next_str(self):             # <<<<<<<<<<<<<<
 *         """Get the next field as a Python string."""
 *         cdef int size
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  int __pyx_v_size;
  char *__pyx_v_next;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("get_next_str", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":563
 *         cdef int size
 *         cdef char *next
 *         next = self.get_next(&size)             # <<<<<<<<<<<<<<
 *         return safe_string_from_size(next, size)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 563, __pyx_L1_error)
  __pyx_v_next = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":564
 *         cdef char *next
 *         next = self.get_next(&size)
 *         return safe_string_from_size(next, size)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _init(self) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_next, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":559
 *         return next
 * 
 *     cdef object get_next_str(self):             # <<<<<<<<<<<<<<
 *         """Get the next field as a Python string."""
 *         cdef int size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.get_next_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":566
 *         return safe_string_from_size(next, size)
 * 
 *     cdef int _init(self) except -1:             # <<<<<<<<<<<<<<
 *         """Get the pointer ready.
 * 
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  char *__pyx_v_first;
  int __pyx_v_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_init", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":577
 *         cdef int size
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)             # <<<<<<<<<<<<<<
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 577, __pyx_L1_error)
  __pyx_v_first = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":578
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:             # <<<<<<<<<<<<<<
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
 */
  __pyx_t_3 = (((__pyx_v_first[0]) != '\x00') != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_size == 0) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":580
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_first); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_First_character_should_be_null_n, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":579
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:
 *             raise AssertionError('First character should be null not: %s'             # <<<<<<<<<<<<<<
 *                                  % (first,))
 *         return 0
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 579, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":578
 *         # The first field should be an empty string left over from the Header
 *         first = self.get_next(&size)
 *         if first[0] != b'\0' and size == 0:             # <<<<<<<<<<<<<<
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":581
 *             raise AssertionError('First character should be null not: %s'
 *                                  % (first,))
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":566
 *         return safe_string_from_size(next, size)
 * 
 *     cdef int _init(self) except -1:             # <<<<<<<<<<<<<<
 *         """Get the pointer ready.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":583
 *         return 0
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,             # <<<<<<<<<<<<<<
 *                            int *new_block):
 *         """Extract the next entry.
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, int __pyx_v_num_trees, void **__pyx_v_p_current_dirname, int *__pyx_v_new_block) {
  StaticTuple *__pyx_v_path_name_file_id_key = 0;
  StaticTuple *__pyx_v_tmp = 0;
  char *__pyx_v_entry_size_cstr;
  unsigned long __pyx_v_entry_size;
  char *__pyx_v_executable_cstr;
  int __pyx_v_is_executable;
  char *__pyx_v_dirname_cstr;
  char *__pyx_v_trailing;
  int __pyx_v_cur_size;
  CYTHON_UNUSED int __pyx_v_i;
  PyObject *__pyx_v_minikind = 0;
  PyObject *__pyx_v_fingerprint = 0;
  PyObject *__pyx_v_info = 0;
  PyObject *__pyx_v_dirname = NULL;
  PyObject *__pyx_v_cur_dirname = NULL;
  PyObject *__pyx_v_cur_basename = NULL;
  PyObject *__pyx_v_cur_file_id = NULL;
  PyObject *__pyx_v_trees = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("_get_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":619
 * 
 *         # Read the 'key' information (dirname, name, file_id)
 *         dirname_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *         # Check to see if we have started a new directory block.
 *         # If so, then we need to create a new dirname PyBytes, so that it can
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 619, __pyx_L1_error)
  __pyx_v_dirname_cstr = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":627
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
 */
  __pyx_t_3 = ((__pyx_v_cur_size != PyBytes_GET_SIZE((__pyx_v_p_current_dirname[0]))) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":635
 *                        # <object>
 *                        PyBytes_AS_STRING_void(p_current_dirname[0]),
 *                        cur_size+1) != 0):             # <<<<<<<<<<<<<<
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname
 */
  __pyx_t_3 = ((strncmp(__pyx_v_dirname_cstr, PyBytes_AS_STRING((__pyx_v_p_current_dirname[0])), (__pyx_v_cur_size + 1)) != 0) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":627
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
 */
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":636
 *                        PyBytes_AS_STRING_void(p_current_dirname[0]),
 *                        cur_size+1) != 0):
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)             # <<<<<<<<<<<<<<
 *             p_current_dirname[0] = <void*>dirname
 *             new_block[0] = 1
 */
    __pyx_t_4 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_dirname_cstr, __pyx_v_cur_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_dirname = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":637
 *                        cur_size+1) != 0):
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname             # <<<<<<<<<<<<<<
 *             new_block[0] = 1
 *         else:
 */
    (__pyx_v_p_current_dirname[0]) = ((void *)__pyx_v_dirname);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":638
 *             dirname = safe_string_from_size(dirname_cstr, cur_size)
 *             p_current_dirname[0] = <void*>dirname
 *             new_block[0] = 1             # <<<<<<<<<<<<<<
 *         else:
 *             new_block[0] = 0
 */
    (__pyx_v_new_block[0]) = 1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":627
 *         # Do the cheap 'length of string' check first. If the string is a
 *         # different length, then we *have* to be a different directory.
 *         if (cur_size != PyBytes_GET_SIZE_void(p_current_dirname[0])             # <<<<<<<<<<<<<<
 *             or strncmp(dirname_cstr,
 *                        # Extract the char* from our current dirname string.  We
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":640
 *             new_block[0] = 1
 *         else:
 *             new_block[0] = 0             # <<<<<<<<<<<<<<
 * 
 *         # Build up the key that will be used.
 */
  /*else*/ {
    (__pyx_v_new_block[0]) = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":645
 *         # By using <object>(void *) Pyrex will automatically handle the
 *         # Py_INCREF that we need.
 *         cur_dirname = <object>p_current_dirname[0]             # <<<<<<<<<<<<<<
 *         # Use StaticTuple_New to pre-allocate, rather than creating a regular
 *         # tuple and passing it to the StaticTuple constructor.
 */
  __pyx_t_4 = ((PyObject *)(__pyx_v_p_current_dirname[0]));
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_cur_dirname = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":652
 *         #                          self.get_next_str(),
 *         #                         )
 *         tmp = StaticTuple_New(3)             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_dirname); StaticTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()
 */
  __pyx_t_4 = ((PyObject *)StaticTuple_New(3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_tmp = ((StaticTuple *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":653
 *         #                         )
 *         tmp = StaticTuple_New(3)
 *         Py_INCREF(cur_dirname); StaticTuple_SET_ITEM(tmp, 0, cur_dirname)             # <<<<<<<<<<<<<<
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()
 */
  Py_INCREF(__pyx_v_cur_dirname);
  StaticTuple_SET_ITEM(__pyx_v_tmp, 0, __pyx_v_cur_dirname);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":654
 *         tmp = StaticTuple_New(3)
 *         Py_INCREF(cur_dirname); StaticTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()             # <<<<<<<<<<<<<<
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); StaticTuple_SET_ITEM(tmp, 1, cur_basename)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_cur_basename = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":655
 *         Py_INCREF(cur_dirname); StaticTuple_SET_ITEM(tmp, 0, cur_dirname)
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_basename); StaticTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); StaticTuple_SET_ITEM(tmp, 2, cur_file_id)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_cur_file_id = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":656
 *         cur_basename = self.get_next_str()
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); StaticTuple_SET_ITEM(tmp, 1, cur_basename)             # <<<<<<<<<<<<<<
 *         Py_INCREF(cur_file_id); StaticTuple_SET_ITEM(tmp, 2, cur_file_id)
 *         path_name_file_id_key = tmp
 */
  Py_INCREF(__pyx_v_cur_basename);
  StaticTuple_SET_ITEM(__pyx_v_tmp, 1, __pyx_v_cur_basename);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":657
 *         cur_file_id = self.get_next_str()
 *         Py_INCREF(cur_basename); StaticTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); StaticTuple_SET_ITEM(tmp, 2, cur_file_id)             # <<<<<<<<<<<<<<
 *         path_name_file_id_key = tmp
 * 
 */
  Py_INCREF(__pyx_v_cur_file_id);
  StaticTuple_SET_ITEM(__pyx_v_tmp, 2, __pyx_v_cur_file_id);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":658
 *         Py_INCREF(cur_basename); StaticTuple_SET_ITEM(tmp, 1, cur_basename)
 *         Py_INCREF(cur_file_id); StaticTuple_SET_ITEM(tmp, 2, cur_file_id)
 *         path_name_file_id_key = tmp             # <<<<<<<<<<<<<<
 * 
 *         # Parse all of the per-tree information. current has the information in
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_tmp));
  __pyx_v_path_name_file_id_key = __pyx_v_tmp;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":673
 *         #       Especially since this code is pretty much fixed at a max of
 *         #       4GB.
 *         trees = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_trees = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":674
 *         #       4GB.
 *         trees = []
 *         for i from 0 <= i < num_trees:             # <<<<<<<<<<<<<<
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()
 */
  __pyx_t_5 = __pyx_v_num_trees;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_5; __pyx_v_i++) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":675
 *         trees = []
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()             # <<<<<<<<<<<<<<
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_minikind, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":676
 *         for i from 0 <= i < num_trees:
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()             # <<<<<<<<<<<<<<
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_fingerprint, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":677
 *             minikind = self.get_next_str()
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 677, __pyx_L1_error)
    __pyx_v_entry_size_cstr = __pyx_t_1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":678
 *             fingerprint = self.get_next_str()
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)             # <<<<<<<<<<<<<<
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')
 */
    __pyx_v_entry_size = strtoul(__pyx_v_entry_size_cstr, NULL, 10);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":679
 *             entry_size_cstr = self.get_next(&cur_size)
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *             is_executable = (executable_cstr[0] == b'y')
 *             info = self.get_next_str()
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 679, __pyx_L1_error)
    __pyx_v_executable_cstr = __pyx_t_1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":680
 *             entry_size = strtoul(entry_size_cstr, NULL, 10)
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')             # <<<<<<<<<<<<<<
 *             info = self.get_next_str()
 *             # TODO: If we want to use StaticTuple_New here we need to be pretty
 */
    __pyx_v_is_executable = ((__pyx_v_executable_cstr[0]) == 'y');

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":681
 *             executable_cstr = self.get_next(&cur_size)
 *             is_executable = (executable_cstr[0] == b'y')
 *             info = self.get_next_str()             # <<<<<<<<<<<<<<
 *             # TODO: If we want to use StaticTuple_New here we need to be pretty
 *             #       careful. We are relying on a bit of Pyrex
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next_str(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_info, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":698
 *                 minikind,     # minikind
 *                 fingerprint,  # fingerprint
 *                 entry_size,   # size             # <<<<<<<<<<<<<<
 *                 is_executable,# executable
 *                 info,         # packed_stat or revision_id
 */
    __pyx_t_4 = __Pyx_PyInt_From_unsigned_long(__pyx_v_entry_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":699
 *                 fingerprint,  # fingerprint
 *                 entry_size,   # size
 *                 is_executable,# executable             # <<<<<<<<<<<<<<
 *                 info,         # packed_stat or revision_id
 *             ))
 */
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_is_executable); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":695
 *             # Py_INCREF(info); StaticTuple_SET_ITEM(tmp, 4, info)
 *             # PyList_Append(trees, tmp)
 *             PyList_Append(trees, StaticTuple(             # <<<<<<<<<<<<<<
 *                 minikind,     # minikind
 *                 fingerprint,  # fingerprint
 */
    __pyx_t_7 = PyTuple_New(5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_minikind);
    __Pyx_GIVEREF(__pyx_v_minikind);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_minikind);
    __Pyx_INCREF(__pyx_v_fingerprint);
    __Pyx_GIVEREF(__pyx_v_fingerprint);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_fingerprint);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_6);
    __Pyx_INCREF(__pyx_v_info);
    __Pyx_GIVEREF(__pyx_v_info);
    PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_info);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple), __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = PyList_Append(__pyx_v_trees, __pyx_t_6); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 695, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":704
 * 
 *         # The returned tuple is (key, [trees])
 *         ret = (path_name_file_id_key, trees)             # <<<<<<<<<<<<<<
 *         # Ignore the trailing newline, but assert that it does exist, this
 *         # ensures that we always finish parsing a line on an end-of-entry
 */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_path_name_file_id_key));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_path_name_file_id_key));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_path_name_file_id_key));
  __Pyx_INCREF(__pyx_v_trees);
  __Pyx_GIVEREF(__pyx_v_trees);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_trees);
  __pyx_v_ret = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":708
 *         # ensures that we always finish parsing a line on an end-of-entry
 *         # marker.
 *         trailing = self.get_next(&cur_size)             # <<<<<<<<<<<<<<
 *         if cur_size != 1 or not trailing.startswith(b'\n'):
 *             raise DirstateCorrupt(self.state,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->get_next(__pyx_v_self, (&__pyx_v_cur_size)); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(0, 708, __pyx_L1_error)
  __pyx_v_trailing = __pyx_t_1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":709
 *         # marker.
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 */
  __pyx_t_3 = ((__pyx_v_cur_size != 1) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_trailing); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyBytes_Tailmatch(__pyx_t_6, __pyx_kp_b__3, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_9 = ((!(__pyx_t_3 != 0)) != 0);
  __pyx_t_2 = __pyx_t_9;
  __pyx_L9_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":710
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 710, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":712
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),             # <<<<<<<<<<<<<<
 *                    ret))
 *         return ret
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_cur_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(__pyx_v_trailing, __pyx_v_cur_size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":713
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 *                    ret))             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_10);
    __Pyx_INCREF(__pyx_v_ret);
    __Pyx_GIVEREF(__pyx_v_ret);
    PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_ret);
    __pyx_t_4 = 0;
    __pyx_t_10 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":712
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),             # <<<<<<<<<<<<<<
 *                    ret))
 *         return ret
 */
    __pyx_t_10 = __Pyx_PyString_Format(__pyx_kp_s_Bad_parse_we_expected_to_end_on, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_self->state, __pyx_t_10};
      __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_self->state, __pyx_t_10};
      __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_v_self->state);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 710, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 710, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":709
 *         # marker.
 *         trailing = self.get_next(&cur_size)
 *         if cur_size != 1 or not trailing.startswith(b'\n'):             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                 'Bad parse, we expected to end on \\n, not: %d %s: %s'
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":714
 *                 % (cur_size, safe_string_from_size(trailing, cur_size),
 *                    ret))
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def _parse_dirblocks(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":583
 *         return 0
 * 
 *     cdef object _get_entry(self, int num_trees, void **p_current_dirname,             # <<<<<<<<<<<<<<
 *                            int *new_block):
 *         """Extract the next entry.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._get_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_path_name_file_id_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmp);
  __Pyx_XDECREF(__pyx_v_minikind);
  __Pyx_XDECREF(__pyx_v_fingerprint);
  __Pyx_XDECREF(__pyx_v_info);
  __Pyx_XDECREF(__pyx_v_dirname);
  __Pyx_XDECREF(__pyx_v_cur_dirname);
  __Pyx_XDECREF(__pyx_v_cur_basename);
  __Pyx_XDECREF(__pyx_v_cur_file_id);
  __Pyx_XDECREF(__pyx_v_trees);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":716
 *         return ret
 * 
 *     def _parse_dirblocks(self):             # <<<<<<<<<<<<<<
 *         """Parse all dirblocks in the state file."""
 *         cdef int num_trees
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks[] = "Parse all dirblocks in the state file.";
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_dirblocks (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  int __pyx_v_num_trees;
  PyObject *__pyx_v_current_block = 0;
  PyObject *__pyx_v_entry = 0;
  void *__pyx_v_current_dirname;
  int __pyx_v_new_block;
  int __pyx_v_expected_entry_count;
  int __pyx_v_entry_count;
  PyObject *__pyx_v_dirblocks = NULL;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_parse_dirblocks", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":726
 *         cdef int entry_count
 * 
 *         num_trees = self.state._num_present_parents() + 1             # <<<<<<<<<<<<<<
 *         expected_entry_count = self.state._num_entries
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_num_present_parents); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_num_trees = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":727
 * 
 *         num_trees = self.state._num_present_parents() + 1
 *         expected_entry_count = self.state._num_entries             # <<<<<<<<<<<<<<
 * 
 *         # Ignore the first record
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_num_entries); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_expected_entry_count = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":730
 * 
 *         # Ignore the first record
 *         self._init()             # <<<<<<<<<<<<<<
 * 
 *         current_block = []
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 730, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":732
 *         self._init()
 * 
 *         current_block = []             # <<<<<<<<<<<<<<
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_current_block = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":733
 * 
 *         current_block = []
 *         dirblocks = [(b'', current_block), (b'', [])]             # <<<<<<<<<<<<<<
 *         self.state._dirblocks = dirblocks
 *         obj = b''
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_b__4);
  __Pyx_GIVEREF(__pyx_kp_b__4);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_b__4);
  __Pyx_INCREF(__pyx_v_current_block);
  __Pyx_GIVEREF(__pyx_v_current_block);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_current_block);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_b__4);
  __Pyx_GIVEREF(__pyx_kp_b__4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_b__4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_v_dirblocks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":734
 *         current_block = []
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks             # <<<<<<<<<<<<<<
 *         obj = b''
 *         current_dirname = <void*>obj
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2, __pyx_v_dirblocks) < 0) __PYX_ERR(0, 734, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":735
 *         dirblocks = [(b'', current_block), (b'', [])]
 *         self.state._dirblocks = dirblocks
 *         obj = b''             # <<<<<<<<<<<<<<
 *         current_dirname = <void*>obj
 *         new_block = 0
 */
  __Pyx_INCREF(__pyx_kp_b__4);
  __pyx_v_obj = __pyx_kp_b__4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":736
 *         self.state._dirblocks = dirblocks
 *         obj = b''
 *         current_dirname = <void*>obj             # <<<<<<<<<<<<<<
 *         new_block = 0
 *         entry_count = 0
 */
  __pyx_v_current_dirname = ((void *)__pyx_v_obj);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":737
 *         obj = b''
 *         current_dirname = <void*>obj
 *         new_block = 0             # <<<<<<<<<<<<<<
 *         entry_count = 0
 * 
 */
  __pyx_v_new_block = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":738
 *         current_dirname = <void*>obj
 *         new_block = 0
 *         entry_count = 0             # <<<<<<<<<<<<<<
 * 
 *         # TODO: jam 2007-05-07 Consider pre-allocating some space for the
 */
  __pyx_v_entry_count = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":747
 *         #       so), and then truncate. That would give us a malloc + realloc,
 *         #       rather than lots of reallocs.
 *         while self.cur_cstr < self.end_cstr:             # <<<<<<<<<<<<<<
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:
 */
  while (1) {
    __pyx_t_5 = ((__pyx_v_self->cur_cstr < __pyx_v_self->end_cstr) != 0);
    if (!__pyx_t_5) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":748
 *         #       rather than lots of reallocs.
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)             # <<<<<<<<<<<<<<
 *             if new_block:
 *                 # new block - different dirname
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self->__pyx_vtab)->_get_entry(__pyx_v_self, __pyx_v_num_trees, (&__pyx_v_current_dirname), (&__pyx_v_new_block)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":749
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:             # <<<<<<<<<<<<<<
 *                 # new block - different dirname
 *                 current_block = []
 */
    __pyx_t_5 = (__pyx_v_new_block != 0);
    if (__pyx_t_5) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":751
 *             if new_block:
 *                 # new block - different dirname
 *                 current_block = []             # <<<<<<<<<<<<<<
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_current_block, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":753
 *                 current_block = []
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))             # <<<<<<<<<<<<<<
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_current_dirname));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_current_dirname));
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_current_dirname));
      __Pyx_INCREF(__pyx_v_current_block);
      __Pyx_GIVEREF(__pyx_v_current_block);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_current_block);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":752
 *                 # new block - different dirname
 *                 current_block = []
 *                 PyList_Append(dirblocks,             # <<<<<<<<<<<<<<
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)
 */
      __pyx_t_4 = PyList_Append(__pyx_v_dirblocks, __pyx_t_1); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 752, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":749
 *         while self.cur_cstr < self.end_cstr:
 *             entry = self._get_entry(num_trees, &current_dirname, &new_block)
 *             if new_block:             # <<<<<<<<<<<<<<
 *                 # new block - different dirname
 *                 current_block = []
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":754
 *                 PyList_Append(dirblocks,
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)             # <<<<<<<<<<<<<<
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:
 */
    __pyx_t_4 = PyList_Append(__pyx_v_current_block, __pyx_v_entry); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 754, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":755
 *                               (<object>current_dirname, current_block))
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1             # <<<<<<<<<<<<<<
 *         if entry_count != expected_entry_count:
 *             raise DirstateCorrupt(self.state,
 */
    __pyx_v_entry_count = (__pyx_v_entry_count + 1);
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":756
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                     'We read the wrong number of entries.'
 */
  __pyx_t_5 = ((__pyx_v_entry_count != __pyx_v_expected_entry_count) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":757
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:
 *             raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                     'We read the wrong number of entries.'
 *                     ' We expected to read %s, but read %s'
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":760
 *                     'We read the wrong number of entries.'
 *                     ' We expected to read %s, but read %s'
 *                     % (expected_entry_count, entry_count))             # <<<<<<<<<<<<<<
 *         self.state._split_root_dirblock_into_contents()
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_expected_entry_count); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_entry_count); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_6);
    __pyx_t_2 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_We_read_the_wrong_number_of_entr, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->state, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->state, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_v_self->state);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_4, __pyx_v_self->state);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_4, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 757, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":756
 *             PyList_Append(current_block, entry)
 *             entry_count = entry_count + 1
 *         if entry_count != expected_entry_count:             # <<<<<<<<<<<<<<
 *             raise DirstateCorrupt(self.state,
 *                     'We read the wrong number of entries.'
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":761
 *                     ' We expected to read %s, but read %s'
 *                     % (expected_entry_count, entry_count))
 *         self.state._split_root_dirblock_into_contents()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_split_root_dirblock_into_conten); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":716
 *         return ret
 * 
 *     def _parse_dirblocks(self):             # <<<<<<<<<<<<<<
 *         """Parse all dirblocks in the state file."""
 *         cdef int num_trees
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader._parse_dirblocks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_block);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_dirblocks);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_4__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->cur_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_self->end_cstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_self->next); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->text_cstr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->text_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->state);
  __Pyx_GIVEREF(__pyx_v_self->state);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_v_self->state);
  __Pyx_INCREF(__pyx_v_self->text);
  __Pyx_GIVEREF(__pyx_v_self->text);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_v_self->text);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_6 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v__dict = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "(tree fragment)":7
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_7 = (__pyx_v__dict != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v__dict);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.cur_cstr, self.end_cstr, self.next, self.state, self.text, self.text_cstr, self.text_size)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.state is not None or self.text is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->state != Py_None);
    __pyx_t_9 = (__pyx_t_7 != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->text != Py_None);
    __pyx_t_7 = (__pyx_t_9 != 0);
    __pyx_t_8 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_8;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, None), state
 *     else:
 */
  __pyx_t_8 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_8) {

    /* "(tree fragment)":13
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Reader); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_106938589);
    __Pyx_GIVEREF(__pyx_int_106938589);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_106938589);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.state is not None or self.text is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, None), state
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Reader); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_106938589);
    __Pyx_GIVEREF(__pyx_int_106938589);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_106938589);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_6__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Reader, (type(self), 0x65fc0dd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Reader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.Reader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":764
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks(PyObject *__pyx_self, PyObject *__pyx_v_state); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks[] = "Read in the dirblocks for the given DirState object.\n\n    This is tightly bound to the DirState internal representation. It should be\n    thought of as a member function, which is only separated out so that we can\n    re-write it in pyrex.\n\n    :param state: A DirState object.\n    :return: None\n    :postcondition: The dirblocks will be loaded into the appropriate fields in\n        the DirState object.\n    ";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks = {"_read_dirblocks", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks, METH_O, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks(PyObject *__pyx_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read_dirblocks (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(__pyx_self, ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_12_read_dirblocks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_v_text = NULL;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v_reader = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_read_dirblocks", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":776
 *         the DirState object.
 *     """
 *     state._state_file.seek(state._end_of_header)             # <<<<<<<<<<<<<<
 *     text = state._state_file.read()
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_n_s_state_file); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_seek); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_n_s_end_of_header); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":777
 *     """
 *     state._state_file.seek(state._end_of_header)
 *     text = state._state_file.read()             # <<<<<<<<<<<<<<
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_state, __pyx_n_s_state_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_read); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_text = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":780
 *     # TODO: check the crc checksums. crc_measured = zlib.crc32(text)
 * 
 *     reader = Reader(text, state)             # <<<<<<<<<<<<<<
 * 
 *     reader._parse_dirblocks()
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_text);
  __Pyx_GIVEREF(__pyx_v_text);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_text);
  __Pyx_INCREF(__pyx_v_state);
  __Pyx_GIVEREF(__pyx_v_state);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_state);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_reader = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":782
 *     reader = Reader(text, state)
 * 
 *     reader._parse_dirblocks()             # <<<<<<<<<<<<<<
 *     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_reader), __pyx_n_s_parse_dirblocks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":783
 * 
 *     reader._parse_dirblocks()
 *     state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DirState); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_IN_MEMORY_UNMODIFIED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_state, __pyx_n_s_dirblock_state, __pyx_t_1) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":764
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._read_dirblocks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_text);
  __Pyx_XDECREF((PyObject *)__pyx_v_reader);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":786
 * 
 * 
 * cdef int minikind_from_mode(int mode): # cannot_raise             # <<<<<<<<<<<<<<
 *     # in order of frequency:
 *     if S_ISREG(mode):
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(int __pyx_v_mode) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("minikind_from_mode", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":788
 * cdef int minikind_from_mode(int mode): # cannot_raise
 *     # in order of frequency:
 *     if S_ISREG(mode):             # <<<<<<<<<<<<<<
 *         return c"f"
 *     if S_ISDIR(mode):
 */
  __pyx_t_1 = (S_ISREG(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":789
 *     # in order of frequency:
 *     if S_ISREG(mode):
 *         return c"f"             # <<<<<<<<<<<<<<
 *     if S_ISDIR(mode):
 *         return c"d"
 */
    __pyx_r = 'f';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":788
 * cdef int minikind_from_mode(int mode): # cannot_raise
 *     # in order of frequency:
 *     if S_ISREG(mode):             # <<<<<<<<<<<<<<
 *         return c"f"
 *     if S_ISDIR(mode):
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":790
 *     if S_ISREG(mode):
 *         return c"f"
 *     if S_ISDIR(mode):             # <<<<<<<<<<<<<<
 *         return c"d"
 *     if S_ISLNK(mode):
 */
  __pyx_t_1 = (S_ISDIR(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":791
 *         return c"f"
 *     if S_ISDIR(mode):
 *         return c"d"             # <<<<<<<<<<<<<<
 *     if S_ISLNK(mode):
 *         return c"l"
 */
    __pyx_r = 'd';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":790
 *     if S_ISREG(mode):
 *         return c"f"
 *     if S_ISDIR(mode):             # <<<<<<<<<<<<<<
 *         return c"d"
 *     if S_ISLNK(mode):
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":792
 *     if S_ISDIR(mode):
 *         return c"d"
 *     if S_ISLNK(mode):             # <<<<<<<<<<<<<<
 *         return c"l"
 *     return 0
 */
  __pyx_t_1 = (S_ISLNK(__pyx_v_mode) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":793
 *         return c"d"
 *     if S_ISLNK(mode):
 *         return c"l"             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 'l';
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":792
 *     if S_ISDIR(mode):
 *         return c"d"
 *     if S_ISLNK(mode):             # <<<<<<<<<<<<<<
 *         return c"l"
 *     return 0
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":794
 *     if S_ISLNK(mode):
 *         return c"l"
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":786
 * 
 * 
 * cdef int minikind_from_mode(int mode): # cannot_raise             # <<<<<<<<<<<<<<
 *     # in order of frequency:
 *     if S_ISREG(mode):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":800
 * 
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise             # <<<<<<<<<<<<<<
 *     if PyFloat_Check(t):
 *         t = t.__int__()
 */

static unsigned long __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(PyObject *__pyx_v_t) {
  unsigned long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned long __pyx_t_5;
  __Pyx_RefNannySetupContext("_time_to_unsigned", 0);
  __Pyx_INCREF(__pyx_v_t);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":801
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):             # <<<<<<<<<<<<<<
 *         t = t.__int__()
 *     return PyInt_AsUnsignedLongMask(t)
 */
  __pyx_t_1 = (PyFloat_Check(__pyx_v_t) != 0);
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":802
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):
 *         t = t.__int__()             # <<<<<<<<<<<<<<
 *     return PyInt_AsUnsignedLongMask(t)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_t, __pyx_n_s_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_t, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":801
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise
 *     if PyFloat_Check(t):             # <<<<<<<<<<<<<<
 *         t = t.__int__()
 *     return PyInt_AsUnsignedLongMask(t)
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":803
 *     if PyFloat_Check(t):
 *         t = t.__int__()
 *     return PyInt_AsUnsignedLongMask(t)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = PyInt_AsUnsignedLongMask(__pyx_v_t); if (unlikely(__pyx_t_5 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 803, __pyx_L1_error)
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":800
 * 
 * 
 * cdef unsigned long _time_to_unsigned(object t):  # cannot_raise             # <<<<<<<<<<<<<<
 *     if PyFloat_Check(t):
 *         t = t.__int__()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("breezy.bzr._dirstate_helpers_pyx._time_to_unsigned", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":806
 * 
 * 
 * cdef _pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """return a string representing the stat value's key fields.
 * 
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(PyObject *__pyx_v_stat_value) {
  char __pyx_v_result[(6 * 4)];
  int *__pyx_v_aliased;
  PyObject *__pyx_v_packed = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_pack_stat", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":814
 *     cdef char result[6*4] # 6 long ints
 *     cdef int *aliased
 *     aliased = <int *>result             # <<<<<<<<<<<<<<
 *     aliased[0] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_size))
 *     # mtime and ctime will often be floats but get converted to PyInt within
 */
  __pyx_v_aliased = ((int *)__pyx_v_result);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":815
 *     cdef int *aliased
 *     aliased = <int *>result
 *     aliased[0] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_size))             # <<<<<<<<<<<<<<
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[0]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":817
 *     aliased[0] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_size))
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))             # <<<<<<<<<<<<<<
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_dev))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mtime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  (__pyx_v_aliased[1]) = htonl(__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":818
 *     # mtime and ctime will often be floats but get converted to PyInt within
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))             # <<<<<<<<<<<<<<
 *     aliased[3] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_ino))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_ctime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  (__pyx_v_aliased[2]) = htonl(__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__time_to_unsigned(__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":819
 *     aliased[1] = htonl(_time_to_unsigned(stat_value.st_mtime))
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_dev))             # <<<<<<<<<<<<<<
 *     aliased[4] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_dev); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[3]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":820
 *     aliased[2] = htonl(_time_to_unsigned(stat_value.st_ctime))
 *     aliased[3] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_ino))             # <<<<<<<<<<<<<<
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_ino); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[4]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":821
 *     aliased[3] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_dev))
 *     aliased[4] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))             # <<<<<<<<<<<<<<
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_AsUnsignedLongMask(__pyx_t_1); if (unlikely(__pyx_t_2 == ((unsigned long)-1L) && PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_aliased[5]) = htonl(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":822
 *     aliased[4] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_ino))
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)             # <<<<<<<<<<<<<<
 *     return _encode(packed)[:-1]
 * 
 */
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v_result, 24); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":823
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_packed) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_packed);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, -1L, NULL, NULL, &__pyx_slice__5, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":806
 * 
 * 
 * cdef _pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """return a string representing the stat value's key fields.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._pack_stat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":826
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat(PyObject *__pyx_self, PyObject *__pyx_v_stat_value); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat[] = "Convert stat value into a packed representation quickly with pyrex";
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat = {"pack_stat", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat, METH_O, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat(PyObject *__pyx_self, PyObject *__pyx_v_stat_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_stat (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(__pyx_self, ((PyObject *)__pyx_v_stat_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_14pack_stat(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_stat_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pack_stat", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":828
 * def pack_stat(stat_value):
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(__pyx_v_stat_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":826
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.pack_stat", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":831
 * 
 * 
 * cpdef update_entry(self, entry, abspath, stat_value):             # <<<<<<<<<<<<<<
 *     """Update the entry based on what is actually on disk.
 * 
 */

static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_minikind;
  int __pyx_v_saved_minikind;
  void *__pyx_v_details;
  int __pyx_v_worth_saving;
  PyObject *__pyx_v_packed_stat = NULL;
  PyObject *__pyx_v_saved_link_or_sha1 = NULL;
  PyObject *__pyx_v_saved_file_size = NULL;
  PyObject *__pyx_v_saved_executable = NULL;
  PyObject *__pyx_v_saved_packed_stat = NULL;
  PyObject *__pyx_v_link_or_sha1 = NULL;
  PyObject *__pyx_v_executable = NULL;
  PyObject *__pyx_v_block_index = NULL;
  PyObject *__pyx_v_entry_index = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_dir_present = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_file_present = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("update_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":850
 *     cdef void * details
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)             # <<<<<<<<<<<<<<
 *     if 0 == minikind:
 *         return None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_minikind = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_minikind_from_mode(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":851
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:             # <<<<<<<<<<<<<<
 *         return None
 *     packed_stat = _pack_stat(stat_value)
 */
  __pyx_t_3 = ((0 == __pyx_v_minikind) != 0);
  if (__pyx_t_3) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":852
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:
 *         return None             # <<<<<<<<<<<<<<
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":851
 *     cdef int worth_saving
 *     minikind = minikind_from_mode(stat_value.st_mode)
 *     if 0 == minikind:             # <<<<<<<<<<<<<<
 *         return None
 *     packed_stat = _pack_stat(stat_value)
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":853
 *     if 0 == minikind:
 *         return None
 *     packed_stat = _pack_stat(stat_value)             # <<<<<<<<<<<<<<
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 */
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__pack_stat(__pyx_v_stat_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed_stat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":854
 *         return None
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)             # <<<<<<<<<<<<<<
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':
 */
  __pyx_v_details = PyList_GET_ITEM(PyTuple_GET_ITEM(((void *)__pyx_v_entry), 1), 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":855
 *     packed_stat = _pack_stat(stat_value)
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]             # <<<<<<<<<<<<<<
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'
 */
  __pyx_v_saved_minikind = (PyBytes_AsString(((PyObject *)PyTuple_GET_ITEM(__pyx_v_details, 0)))[0]);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":856
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':             # <<<<<<<<<<<<<<
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 */
  __pyx_t_4 = ((__pyx_v_minikind == 'd') != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_v_saved_minikind == 't') != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_3) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":857
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'             # <<<<<<<<<<<<<<
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 */
    __pyx_v_minikind = 't';

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":856
 *     details = PyList_GetItem_void_void(PyTuple_GetItem_void_void(<void *>entry, 1), 0)
 *     saved_minikind = PyBytes_AsString_obj(<PyObject *>PyTuple_GetItem_void_void(details, 0))[0]
 *     if minikind == b'd' and saved_minikind == b't':             # <<<<<<<<<<<<<<
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":858
 *     if minikind == b'd' and saved_minikind == b't':
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)             # <<<<<<<<<<<<<<
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_link_or_sha1 = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":859
 *         minikind = b't'
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)             # <<<<<<<<<<<<<<
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_file_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":860
 *     saved_link_or_sha1 = PyTuple_GetItem_void_object(details, 1)
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)             # <<<<<<<<<<<<<<
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
 *     # Deal with pyrex decrefing the objects
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_executable = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":861
 *     saved_file_size = PyTuple_GetItem_void_object(details, 2)
 *     saved_executable = PyTuple_GetItem_void_object(details, 3)
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)             # <<<<<<<<<<<<<<
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)
 */
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v_details, 4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_saved_packed_stat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":863
 *     saved_packed_stat = PyTuple_GetItem_void_object(details, 4)
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)
 */
  Py_INCREF(__pyx_v_saved_link_or_sha1);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":864
 *     # Deal with pyrex decrefing the objects
 *     Py_INCREF(saved_link_or_sha1)
 *     Py_INCREF(saved_file_size)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_executable)
 *     Py_INCREF(saved_packed_stat)
 */
  Py_INCREF(__pyx_v_saved_file_size);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":865
 *     Py_INCREF(saved_link_or_sha1)
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)             # <<<<<<<<<<<<<<
 *     Py_INCREF(saved_packed_stat)
 *     #(saved_minikind, saved_link_or_sha1, saved_file_size,
 */
  Py_INCREF(__pyx_v_saved_executable);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":866
 *     Py_INCREF(saved_file_size)
 *     Py_INCREF(saved_executable)
 *     Py_INCREF(saved_packed_stat)             # <<<<<<<<<<<<<<
 *     #(saved_minikind, saved_link_or_sha1, saved_file_size,
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 */
  Py_INCREF(__pyx_v_saved_packed_stat);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":870
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
 */
  __pyx_t_4 = ((__pyx_v_minikind == __pyx_v_saved_minikind) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L8_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":871
 * 
 *     if (minikind == saved_minikind
 *         and packed_stat == saved_packed_stat):             # <<<<<<<<<<<<<<
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_packed_stat, __pyx_v_saved_packed_stat, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_4;
  __pyx_L8_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":870
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
 */
  if (__pyx_t_3) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":874
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
 *         if minikind == b'd':             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __pyx_t_3 = ((__pyx_v_minikind == 'd') != 0);
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":875
 *         # saved sha hash.
 *         if minikind == b'd':
 *             return None             # <<<<<<<<<<<<<<
 * 
 *         # size should also be in packed_stat
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":874
 *         # The stat hasn't changed since we saved, so we can re-use the
 *         # saved sha hash.
 *         if minikind == b'd':             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":878
 * 
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:             # <<<<<<<<<<<<<<
 *             return saved_link_or_sha1
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_RichCompare(__pyx_v_saved_file_size, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 878, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":879
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:
 *             return saved_link_or_sha1             # <<<<<<<<<<<<<<
 * 
 *     # If we have gotten this far, that means that we need to actually
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_saved_link_or_sha1);
      __pyx_r = __pyx_v_saved_link_or_sha1;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":878
 * 
 *         # size should also be in packed_stat
 *         if saved_file_size == stat_value.st_size:             # <<<<<<<<<<<<<<
 *             return saved_link_or_sha1
 * 
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":870
 *     # saved_executable, saved_packed_stat) = entry[1][0]
 * 
 *     if (minikind == saved_minikind             # <<<<<<<<<<<<<<
 *         and packed_stat == saved_packed_stat):
 *         # The stat hasn't changed since we saved, so we can re-use the
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":883
 *     # If we have gotten this far, that means that we need to actually
 *     # process this entry.
 *     link_or_sha1 = None             # <<<<<<<<<<<<<<
 *     worth_saving = 1
 *     if minikind == b'f':
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_link_or_sha1 = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":884
 *     # process this entry.
 *     link_or_sha1 = None
 *     worth_saving = 1             # <<<<<<<<<<<<<<
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,
 */
  __pyx_v_worth_saving = 1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":885
 *     link_or_sha1 = None
 *     worth_saving = 1
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 */
  switch (__pyx_v_minikind) {
    case 'f':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":886
 *     worth_saving = 1
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,             # <<<<<<<<<<<<<<
 *                                          saved_executable)
 *         if self._cutoff_time is None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_is_executable); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":887
 *     if minikind == b'f':
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)             # <<<<<<<<<<<<<<
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 */
    __pyx_t_7 = NULL;
    __pyx_t_2 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_2 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_saved_executable};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_saved_executable};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_2, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_saved_executable);
      __Pyx_GIVEREF(__pyx_v_saved_executable);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_2, __pyx_v_saved_executable);
      __pyx_t_6 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_executable = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":888
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = (__pyx_t_5 == Py_None);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":889
 *                                          saved_executable)
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()             # <<<<<<<<<<<<<<
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sha_cutoff_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":888
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":890
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mtime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 890, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":891
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and len(entry[1]) > 1
 *             and entry[1][1][0] != b'a'):
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_ctime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_8, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 891, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":892
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1             # <<<<<<<<<<<<<<
 *             and entry[1][1][0] != b'a'):
 *                 # Could check for size changes for further optimised
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 892, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L14_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":893
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
 *             and entry[1][1][0] != b'a'):             # <<<<<<<<<<<<<<
 *                 # Could check for size changes for further optimised
 *                 # avoidance of sha1's. However the most prominent case of
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_t_5, __pyx_n_b_a, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 893, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_4 = __pyx_t_3;
    __pyx_L14_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":890
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
 */
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":897
 *                 # avoidance of sha1's. However the most prominent case of
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)             # <<<<<<<<<<<<<<
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,
 *                            executable, packed_stat)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sha1_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_v_abspath) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_abspath);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":898
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            executable, packed_stat)
 *         else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":899
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,
 *                            executable, packed_stat)             # <<<<<<<<<<<<<<
 *         else:
 *             # This file is not worth caching the sha1. Either it is too new, or
 */
      __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_n_b_f);
      __Pyx_GIVEREF(__pyx_n_b_f);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_b_f);
      __Pyx_INCREF(__pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_v_link_or_sha1);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_executable);
      __Pyx_GIVEREF(__pyx_v_executable);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_executable);
      __Pyx_INCREF(__pyx_v_packed_stat);
      __Pyx_GIVEREF(__pyx_v_packed_stat);
      PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_packed_stat);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":898
 *                 # over-shaing is during initial add, which this catches.
 *             link_or_sha1 = self._sha1_file(abspath)
 *             entry[1][0] = (b'f', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            executable, packed_stat)
 *         else:
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_SetItemInt(__pyx_t_5, 0, __pyx_t_1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":890
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time
 *             and len(entry[1]) > 1
 */
      goto __pyx_L13;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":906
 *             # *not* set the IN_MEMORY_MODIFIED flag. (But we'll save the
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,             # <<<<<<<<<<<<<<
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":907
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,
 *                            DirState.NULLSTAT)             # <<<<<<<<<<<<<<
 *             worth_saving = 0
 *     elif minikind == b'd':
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DirState); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_NULLSTAT); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":906
 *             # *not* set the IN_MEMORY_MODIFIED flag. (But we'll save the
 *             # updated values if there is *other* data worth saving.)
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,             # <<<<<<<<<<<<<<
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
 */
      __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_b_f);
      __Pyx_GIVEREF(__pyx_n_b_f);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_b_f);
      __Pyx_INCREF(__pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_kp_b__4);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_executable);
      __Pyx_GIVEREF(__pyx_v_executable);
      PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_executable);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_8);
      __pyx_t_1 = 0;
      __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (unlikely(__Pyx_SetItemInt(__pyx_t_8, 0, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 906, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":908
 *             entry[1][0] = (b'f', b'', stat_value.st_size, executable,
 *                            DirState.NULLSTAT)
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 */
      __pyx_v_worth_saving = 0;
    }
    __pyx_L13:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":885
 *     link_or_sha1 = None
 *     worth_saving = 1
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         executable = self._is_executable(stat_value.st_mode,
 *                                          saved_executable)
 */
    break;
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":910
 *             worth_saving = 0
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)             # <<<<<<<<<<<<<<
 *         if saved_minikind != b'd':
 *             # This changed from something into a directory. Make sure we
 */
    __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_n_b_d);
    __Pyx_GIVEREF(__pyx_n_b_d);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_b_d);
    __Pyx_INCREF(__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_kp_b__4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_b__4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_int_0);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    PyTuple_SET_ITEM(__pyx_t_5, 3, Py_False);
    __Pyx_INCREF(__pyx_v_packed_stat);
    __Pyx_GIVEREF(__pyx_v_packed_stat);
    PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_packed_stat);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely(__Pyx_SetItemInt(__pyx_t_8, 0, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":911
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':             # <<<<<<<<<<<<<<
 *             # This changed from something into a directory. Make sure we
 *             # have a directory block for it. This doesn't happen very
 */
    __pyx_t_4 = ((__pyx_v_saved_minikind != 'd') != 0);
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":916
 *             # often, so this doesn't have to be super fast.
 *             block_index, entry_index, dir_present, file_present = \
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)             # <<<<<<<<<<<<<<
 *             self._ensure_block(block_index, entry_index,
 *                                pathjoin(entry[0][0], entry[0][1]))
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_get_block_entry_index); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 916, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_2 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_2 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_6, __pyx_t_7, __pyx_int_0};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_2, 3+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_6, __pyx_t_7, __pyx_int_0};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_2, 3+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(3+__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_1) {
          __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_2, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_2, __pyx_t_7);
        __Pyx_INCREF(__pyx_int_0);
        __Pyx_GIVEREF(__pyx_int_0);
        PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_2, __pyx_int_0);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 916, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 4)) {
          if (size > 4) __Pyx_RaiseTooManyValuesError(4);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 915, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 3); 
        } else {
          __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 3); 
        }
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[4] = {&__pyx_t_8,&__pyx_t_10,&__pyx_t_7,&__pyx_t_6};
          for (i=0; i < 4; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 915, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[4] = {&__pyx_t_8,&__pyx_t_10,&__pyx_t_7,&__pyx_t_6};
        __pyx_t_1 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 915, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext;
        for (index=0; index < 4; index++) {
          PyObject* item = __pyx_t_11(__pyx_t_1); if (unlikely(!item)) goto __pyx_L19_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_1), 4) < 0) __PYX_ERR(0, 915, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L20_unpacking_done;
        __pyx_L19_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 915, __pyx_L1_error)
        __pyx_L20_unpacking_done:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":915
 *             # have a directory block for it. This doesn't happen very
 *             # often, so this doesn't have to be super fast.
 *             block_index, entry_index, dir_present, file_present = \             # <<<<<<<<<<<<<<
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,
 */
      __pyx_v_block_index = __pyx_t_8;
      __pyx_t_8 = 0;
      __pyx_v_entry_index = __pyx_t_10;
      __pyx_t_10 = 0;
      __pyx_v_dir_present = __pyx_t_7;
      __pyx_t_7 = 0;
      __pyx_v_file_present = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":917
 *             block_index, entry_index, dir_present, file_present = \
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,             # <<<<<<<<<<<<<<
 *                                pathjoin(entry[0][0], entry[0][1]))
 *         else:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_ensure_block); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":918
 *                 self._get_block_entry_index(entry[0][0], entry[0][1], 0)
 *             self._ensure_block(block_index, entry_index,
 *                                pathjoin(entry[0][0], entry[0][1]))             # <<<<<<<<<<<<<<
 *         else:
 *             # Any changes are derived trivially from the stat object, not worth
 */
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pathjoin); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 918, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_2 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_2 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_12};
        __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_1, __pyx_t_12};
        __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(2+__pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_8) {
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_8); __pyx_t_8 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_2, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_2, __pyx_t_12);
        __pyx_t_1 = 0;
        __pyx_t_12 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 918, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_2 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_2 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_block_index, __pyx_v_entry_index, __pyx_t_7};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_2, 3+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_v_block_index, __pyx_v_entry_index, __pyx_t_7};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_2, 3+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(3+__pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_block_index);
        __Pyx_GIVEREF(__pyx_v_block_index);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_2, __pyx_v_block_index);
        __Pyx_INCREF(__pyx_v_entry_index);
        __Pyx_GIVEREF(__pyx_v_entry_index);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_2, __pyx_v_entry_index);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_2, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 917, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":911
 *     elif minikind == b'd':
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':             # <<<<<<<<<<<<<<
 *             # This changed from something into a directory. Make sure we
 *             # have a directory block for it. This doesn't happen very
 */
      goto __pyx_L18;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":922
 *             # Any changes are derived trivially from the stat object, not worth
 *             # re-writing a dirstate for just this
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':
 */
    /*else*/ {
      __pyx_v_worth_saving = 0;
    }
    __pyx_L18:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":909
 *                            DirState.NULLSTAT)
 *             worth_saving = 0
 *     elif minikind == b'd':             # <<<<<<<<<<<<<<
 *         entry[1][0] = (b'd', b'', 0, False, packed_stat)
 *         if saved_minikind != b'd':
 */
    break;
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":924
 *             worth_saving = 0
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':             # <<<<<<<<<<<<<<
 *             # If the object hasn't changed kind, it isn't worth saving the
 *             # dirstate just for a symlink. The default is 'fast symlinks' which
 */
    __pyx_t_4 = ((__pyx_v_saved_minikind == 'l') != 0);
    if (__pyx_t_4) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":929
 *             # save the target in the inode entry, rather than separately. So to
 *             # stat, we've already read everything off disk.
 *             worth_saving = 0             # <<<<<<<<<<<<<<
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:
 */
      __pyx_v_worth_saving = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":924
 *             worth_saving = 0
 *     elif minikind == b'l':
 *         if saved_minikind == b'l':             # <<<<<<<<<<<<<<
 *             # If the object hasn't changed kind, it isn't worth saving the
 *             # dirstate just for a symlink. The default is 'fast symlinks' which
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":930
 *             # stat, we've already read everything off disk.
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)             # <<<<<<<<<<<<<<
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_read_link); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 930, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = NULL;
    __pyx_t_2 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_2 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_abspath, __pyx_v_saved_link_or_sha1};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_v_abspath, __pyx_v_saved_link_or_sha1};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_2, 2+__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_INCREF(__pyx_v_abspath);
      __Pyx_GIVEREF(__pyx_v_abspath);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_2, __pyx_v_abspath);
      __Pyx_INCREF(__pyx_v_saved_link_or_sha1);
      __Pyx_GIVEREF(__pyx_v_saved_link_or_sha1);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_2, __pyx_v_saved_link_or_sha1);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":931
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__pyx_t_5 == Py_None);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":932
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()             # <<<<<<<<<<<<<<
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_sha_cutoff_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":931
 *             worth_saving = 0
 *         link_or_sha1 = self._read_link(abspath, saved_link_or_sha1)
 *         if self._cutoff_time is None:             # <<<<<<<<<<<<<<
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":933
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_mtime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L24_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":934
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):             # <<<<<<<<<<<<<<
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 *                            False, packed_stat)
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_ctime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_cutoff_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_3 = __pyx_t_4;
    __pyx_L24_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":933
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 */
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":935
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, packed_stat)
 *         else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":936
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 *                            False, packed_stat)             # <<<<<<<<<<<<<<
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 */
      __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_n_b_l);
      __Pyx_GIVEREF(__pyx_n_b_l);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_b_l);
      __Pyx_INCREF(__pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_v_link_or_sha1);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_link_or_sha1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_5);
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      PyTuple_SET_ITEM(__pyx_t_6, 3, Py_False);
      __Pyx_INCREF(__pyx_v_packed_stat);
      __Pyx_GIVEREF(__pyx_v_packed_stat);
      PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_v_packed_stat);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":935
 *         if (stat_value.st_mtime < self._cutoff_time
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, packed_stat)
 *         else:
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_SetItemInt(__pyx_t_5, 0, __pyx_t_6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":933
 *         if self._cutoff_time is None:
 *             self._sha_cutoff_time()
 *         if (stat_value.st_mtime < self._cutoff_time             # <<<<<<<<<<<<<<
 *             and stat_value.st_ctime < self._cutoff_time):
 *             entry[1][0] = (b'l', link_or_sha1, stat_value.st_size,
 */
      goto __pyx_L23;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":938
 *                            False, packed_stat)
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat_value, __pyx_n_s_st_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":939
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)             # <<<<<<<<<<<<<<
 *     if worth_saving:
 *         # Note, even though _mark_modified will only set
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_DirState); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_NULLSTAT); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":938
 *                            False, packed_stat)
 *         else:
 *             entry[1][0] = (b'l', b'', stat_value.st_size,             # <<<<<<<<<<<<<<
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:
 */
      __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_n_b_l);
      __Pyx_GIVEREF(__pyx_n_b_l);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_b_l);
      __Pyx_INCREF(__pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_kp_b__4);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_kp_b__4);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_6);
      __Pyx_INCREF(Py_False);
      __Pyx_GIVEREF(Py_False);
      PyTuple_SET_ITEM(__pyx_t_5, 3, Py_False);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_SetItemInt(__pyx_t_7, 0, __pyx_t_5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_L23:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":923
 *             # re-writing a dirstate for just this
 *             worth_saving = 0
 *     elif minikind == b'l':             # <<<<<<<<<<<<<<
 *         if saved_minikind == b'l':
 *             # If the object hasn't changed kind, it isn't worth saving the
 */
    break;
    default: break;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":940
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:             # <<<<<<<<<<<<<<
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 */
  __pyx_t_3 = (__pyx_v_worth_saving != 0);
  if (__pyx_t_3) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":943
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 *         self._mark_modified([entry])             # <<<<<<<<<<<<<<
 *     return link_or_sha1
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_mark_modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PyList_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_entry);
    __Pyx_GIVEREF(__pyx_v_entry);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_v_entry);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_13, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":940
 *             entry[1][0] = (b'l', b'', stat_value.st_size,
 *                            False, DirState.NULLSTAT)
 *     if worth_saving:             # <<<<<<<<<<<<<<
 *         # Note, even though _mark_modified will only set
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":944
 *         # IN_MEMORY_HASH_MODIFIED, it still isn't worth
 *         self._mark_modified([entry])
 *     return link_or_sha1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_link_or_sha1);
  __pyx_r = __pyx_v_link_or_sha1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":831
 * 
 * 
 * cpdef update_entry(self, entry, abspath, stat_value):             # <<<<<<<<<<<<<<
 *     """Update the entry based on what is actually on disk.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed_stat);
  __Pyx_XDECREF(__pyx_v_saved_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_saved_file_size);
  __Pyx_XDECREF(__pyx_v_saved_executable);
  __Pyx_XDECREF(__pyx_v_saved_packed_stat);
  __Pyx_XDECREF(__pyx_v_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_executable);
  __Pyx_XDECREF(__pyx_v_block_index);
  __Pyx_XDECREF(__pyx_v_entry_index);
  __Pyx_XDECREF(__pyx_v_dir_present);
  __Pyx_XDECREF(__pyx_v_file_present);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry[] = "Update the entry based on what is actually on disk.\n\n    This function only calculates the sha if it needs to - if the entry is\n    uncachable, or clearly different to the first parent's entry, no sha\n    is calculated, and None is returned.\n\n    :param self: The dirstate object this is operating on.\n    :param entry: This is the dirblock entry for the file in question.\n    :param abspath: The path on disk for this file.\n    :param stat_value: The stat value done on the path.\n    :return: None, or The sha1 hexdigest of the file (40 bytes) or link\n        target of a symlink.\n    ";
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_entry = 0;
  PyObject *__pyx_v_abspath = 0;
  PyObject *__pyx_v_stat_value = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_entry (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_entry,&__pyx_n_s_abspath,&__pyx_n_s_stat_value,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_entry)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, 1); __PYX_ERR(0, 831, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_abspath)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, 2); __PYX_ERR(0, 831, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stat_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, 3); __PYX_ERR(0, 831, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "update_entry") < 0)) __PYX_ERR(0, 831, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_self = values[0];
    __pyx_v_entry = values[1];
    __pyx_v_abspath = values[2];
    __pyx_v_stat_value = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update_entry", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 831, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(__pyx_self, __pyx_v_self, __pyx_v_entry, __pyx_v_abspath, __pyx_v_stat_value);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_abspath, PyObject *__pyx_v_stat_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("update_entry", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(__pyx_v_self, __pyx_v_entry, __pyx_v_abspath, __pyx_v_stat_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.update_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":948
 * 
 * # TODO: Do we want to worry about exceptions here?
 * cdef char _minikind_from_string(object string) except? -1:             # <<<<<<<<<<<<<<
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]
 */

static char __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(PyObject *__pyx_v_string) {
  char __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_minikind_from_string", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":950
 * cdef char _minikind_from_string(object string) except? -1:
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (PyBytes_AsString(__pyx_v_string)[0]);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":948
 * 
 * # TODO: Do we want to worry about exceptions here?
 * cdef char _minikind_from_string(object string) except? -1:             # <<<<<<<<<<<<<<
 *     """Convert a python string to a char."""
 *     return PyBytes_AsString(string)[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":967
 * 
 * 
 * cdef object _minikind_to_kind(char minikind):             # <<<<<<<<<<<<<<
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(char __pyx_v_minikind) {
  char __pyx_v__minikind[1];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_minikind_to_kind", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":970
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         return _kind_file
 *     elif minikind == b'd':
 */
  switch (__pyx_v_minikind) {
    case 'f':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":971
 *     cdef char _minikind[1]
 *     if minikind == b'f':
 *         return _kind_file             # <<<<<<<<<<<<<<
 *     elif minikind == b'd':
 *         return _kind_directory
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":970
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 *     if minikind == b'f':             # <<<<<<<<<<<<<<
 *         return _kind_file
 *     elif minikind == b'd':
 */
    break;
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":973
 *         return _kind_file
 *     elif minikind == b'd':
 *         return _kind_directory             # <<<<<<<<<<<<<<
 *     elif minikind == b'a':
 *         return _kind_absent
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":972
 *     if minikind == b'f':
 *         return _kind_file
 *     elif minikind == b'd':             # <<<<<<<<<<<<<<
 *         return _kind_directory
 *     elif minikind == b'a':
 */
    break;
    case 'a':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":975
 *         return _kind_directory
 *     elif minikind == b'a':
 *         return _kind_absent             # <<<<<<<<<<<<<<
 *     elif minikind == b'r':
 *         return _kind_relocated
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":974
 *     elif minikind == b'd':
 *         return _kind_directory
 *     elif minikind == b'a':             # <<<<<<<<<<<<<<
 *         return _kind_absent
 *     elif minikind == b'r':
 */
    break;
    case 'r':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":977
 *         return _kind_absent
 *     elif minikind == b'r':
 *         return _kind_relocated             # <<<<<<<<<<<<<<
 *     elif minikind == b'l':
 *         return _kind_symlink
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":976
 *     elif minikind == b'a':
 *         return _kind_absent
 *     elif minikind == b'r':             # <<<<<<<<<<<<<<
 *         return _kind_relocated
 *     elif minikind == b'l':
 */
    break;
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":979
 *         return _kind_relocated
 *     elif minikind == b'l':
 *         return _kind_symlink             # <<<<<<<<<<<<<<
 *     elif minikind == b't':
 *         return _kind_tree_reference
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":978
 *     elif minikind == b'r':
 *         return _kind_relocated
 *     elif minikind == b'l':             # <<<<<<<<<<<<<<
 *         return _kind_symlink
 *     elif minikind == b't':
 */
    break;
    case 't':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":981
 *         return _kind_symlink
 *     elif minikind == b't':
 *         return _kind_tree_reference             # <<<<<<<<<<<<<<
 *     _minikind[0] = minikind
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference);
    __pyx_r = __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":980
 *     elif minikind == b'l':
 *         return _kind_symlink
 *     elif minikind == b't':             # <<<<<<<<<<<<<<
 *         return _kind_tree_reference
 *     _minikind[0] = minikind
 */
    break;
    default: break;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":982
 *     elif minikind == b't':
 *         return _kind_tree_reference
 *     _minikind[0] = minikind             # <<<<<<<<<<<<<<
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))
 * 
 */
  (__pyx_v__minikind[0]) = __pyx_v_minikind;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":983
 *         return _kind_tree_reference
 *     _minikind[0] = minikind
 *     raise KeyError(PyBytes_FromStringAndSize(_minikind, 1))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyBytes_FromStringAndSize(__pyx_v__minikind, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 983, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 983, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":967
 * 
 * 
 * cdef object _minikind_to_kind(char minikind):             # <<<<<<<<<<<<<<
 *     """Create a string kind for minikind."""
 *     cdef char _minikind[1]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx._minikind_to_kind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":986
 * 
 * 
 * cdef int _versioned_minikind(char minikind): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(char __pyx_v_minikind) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_versioned_minikind", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":988
 * cdef int _versioned_minikind(char minikind): # cannot_raise
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or             # <<<<<<<<<<<<<<
 *             minikind == b'd' or
 *             minikind == b'l' or
 */
  switch (__pyx_v_minikind) {
    case 'f':
    case 'd':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":989
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
 *             minikind == b'd' or             # <<<<<<<<<<<<<<
 *             minikind == b'l' or
 *             minikind == b't')
 */
    case 'l':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":990
 *     return (minikind == b'f' or
 *             minikind == b'd' or
 *             minikind == b'l' or             # <<<<<<<<<<<<<<
 *             minikind == b't')
 * 
 */
    case 't':

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":988
 * cdef int _versioned_minikind(char minikind): # cannot_raise
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or             # <<<<<<<<<<<<<<
 *             minikind == b'd' or
 *             minikind == b'l' or
 */
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":986
 * 
 * 
 * cdef int _versioned_minikind(char minikind): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Return non-zero if minikind is in fltd"""
 *     return (minikind == b'f' or
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1038
 *     cdef object sha_file
 * 
 *     def __init__(self, include_unchanged, use_filesystem_for_exec,             # <<<<<<<<<<<<<<
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
 */

/* Python wrapper */
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_include_unchanged = 0;
  PyObject *__pyx_v_use_filesystem_for_exec = 0;
  PyObject *__pyx_v_search_specific_files = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_source_index = 0;
  PyObject *__pyx_v_target_index = 0;
  PyObject *__pyx_v_want_unversioned = 0;
  PyObject *__pyx_v_tree = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_include_unchanged,&__pyx_n_s_use_filesystem_for_exec,&__pyx_n_s_search_specific_files,&__pyx_n_s_state,&__pyx_n_s_source_index,&__pyx_n_s_target_index,&__pyx_n_s_want_unversioned,&__pyx_n_s_tree,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_include_unchanged)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_use_filesystem_for_exec)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 1); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_search_specific_files)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 2); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 3); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 4); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_target_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 5); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_want_unversioned)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 6); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tree)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, 7); __PYX_ERR(0, 1038, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1038, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_include_unchanged = values[0];
    __pyx_v_use_filesystem_for_exec = values[1];
    __pyx_v_search_specific_files = values[2];
    __pyx_v_state = values[3];
    __pyx_v_source_index = values[4];
    __pyx_v_target_index = values[5];
    __pyx_v_want_unversioned = values[6];
    __pyx_v_tree = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1038, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self), __pyx_v_include_unchanged, __pyx_v_use_filesystem_for_exec, __pyx_v_search_specific_files, __pyx_v_state, __pyx_v_source_index, __pyx_v_target_index, __pyx_v_want_unversioned, __pyx_v_tree);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC___init__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_include_unchanged, PyObject *__pyx_v_use_filesystem_for_exec, PyObject *__pyx_v_search_specific_files, PyObject *__pyx_v_state, PyObject *__pyx_v_source_index, PyObject *__pyx_v_target_index, PyObject *__pyx_v_want_unversioned, PyObject *__pyx_v_tree) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1041
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
 *         self.doing_consistency_expansion = 0             # <<<<<<<<<<<<<<
 *         self.old_dirname_to_file_id = {}
 *         self.new_dirname_to_file_id = {}
 */
  __pyx_v_self->doing_consistency_expansion = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1042
 *         want_unversioned, tree):
 *         self.doing_consistency_expansion = 0
 *         self.old_dirname_to_file_id = {}             # <<<<<<<<<<<<<<
 *         self.new_dirname_to_file_id = {}
 *         # Are we doing a partial iter_changes?
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->old_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v_self->old_dirname_to_file_id);
  __pyx_v_self->old_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1043
 *         self.doing_consistency_expansion = 0
 *         self.old_dirname_to_file_id = {}
 *         self.new_dirname_to_file_id = {}             # <<<<<<<<<<<<<<
 *         # Are we doing a partial iter_changes?
 *         self.partial = set(['']).__ne__(search_specific_files)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->new_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v_self->new_dirname_to_file_id);
  __pyx_v_self->new_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1045
 *         self.new_dirname_to_file_id = {}
 *         # Are we doing a partial iter_changes?
 *         self.partial = set(['']).__ne__(search_specific_files)             # <<<<<<<<<<<<<<
 *         # Using a list so that we can access the values and change them in
 *         # nested scope. Each one is [path, file_id, entry]
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_kp_s__4) < 0) __PYX_ERR(0, 1045, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ne); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_search_specific_files) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_search_specific_files);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->partial = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1048
 *         # Using a list so that we can access the values and change them in
 *         # nested scope. Each one is [path, file_id, entry]
 *         self.last_source_parent = [None, None]             # <<<<<<<<<<<<<<
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 1, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->last_source_parent);
  __Pyx_DECREF(__pyx_v_self->last_source_parent);
  __pyx_v_self->last_source_parent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1049
 *         # nested scope. Each one is [path, file_id, entry]
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]             # <<<<<<<<<<<<<<
 *         if include_unchanged is None:
 *             self.include_unchanged = False
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 1, Py_None);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->last_target_parent);
  __Pyx_DECREF(__pyx_v_self->last_target_parent);
  __pyx_v_self->last_target_parent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1050
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:             # <<<<<<<<<<<<<<
 *             self.include_unchanged = False
 *         else:
 */
  __pyx_t_5 = (__pyx_v_include_unchanged == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1051
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:
 *             self.include_unchanged = False             # <<<<<<<<<<<<<<
 *         else:
 *             self.include_unchanged = int(include_unchanged)
 */
    __pyx_v_self->include_unchanged = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1050
 *         self.last_source_parent = [None, None]
 *         self.last_target_parent = [None, None]
 *         if include_unchanged is None:             # <<<<<<<<<<<<<<
 *             self.include_unchanged = False
 *         else:
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1053
 *             self.include_unchanged = False
 *         else:
 *             self.include_unchanged = int(include_unchanged)             # <<<<<<<<<<<<<<
 *         self.use_filesystem_for_exec = use_filesystem_for_exec
 *         self.utf8_decode = cache_utf8._utf8_decode
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_include_unchanged); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1053, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->include_unchanged = __pyx_t_4;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1054
 *         else:
 *             self.include_unchanged = int(include_unchanged)
 *         self.use_filesystem_for_exec = use_filesystem_for_exec             # <<<<<<<<<<<<<<
 *         self.utf8_decode = cache_utf8._utf8_decode
 *         # for all search_indexs in each path at or under each element of
 */
  __Pyx_INCREF(__pyx_v_use_filesystem_for_exec);
  __Pyx_GIVEREF(__pyx_v_use_filesystem_for_exec);
  __Pyx_GOTREF(__pyx_v_self->use_filesystem_for_exec);
  __Pyx_DECREF(__pyx_v_self->use_filesystem_for_exec);
  __pyx_v_self->use_filesystem_for_exec = __pyx_v_use_filesystem_for_exec;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1055
 *             self.include_unchanged = int(include_unchanged)
 *         self.use_filesystem_for_exec = use_filesystem_for_exec
 *         self.utf8_decode = cache_utf8._utf8_decode             # <<<<<<<<<<<<<<
 *         # for all search_indexs in each path at or under each element of
 *         # search_specific_files, if the detail is relocated: add the id, and
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_cache_utf8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_utf8_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->utf8_decode);
  __Pyx_DECREF(__pyx_v_self->utf8_decode);
  __pyx_v_self->utf8_decode = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1060
 *         # add the relocated path as one to search if its not searched already.
 *         # If the detail is not relocated, add the id.
 *         self.searched_specific_files = set()             # <<<<<<<<<<<<<<
 *         # When we search exact paths without expanding downwards, we record
 *         # that here.
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->searched_specific_files);
  __Pyx_DECREF(__pyx_v_self->searched_specific_files);
  __pyx_v_self->searched_specific_files = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1063
 *         # When we search exact paths without expanding downwards, we record
 *         # that here.
 *         self.searched_exact_paths = set()             # <<<<<<<<<<<<<<
 *         self.search_specific_files = search_specific_files
 *         # The parents up to the root of the paths we are searching.
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->searched_exact_paths);
  __Pyx_DECREF(__pyx_v_self->searched_exact_paths);
  __pyx_v_self->searched_exact_paths = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1064
 *         # that here.
 *         self.searched_exact_paths = set()
 *         self.search_specific_files = search_specific_files             # <<<<<<<<<<<<<<
 *         # The parents up to the root of the paths we are searching.
 *         # After all normal paths are returned, these specific items are returned.
 */
  __Pyx_INCREF(__pyx_v_search_specific_files);
  __Pyx_GIVEREF(__pyx_v_search_specific_files);
  __Pyx_GOTREF(__pyx_v_self->search_specific_files);
  __Pyx_DECREF(__pyx_v_self->search_specific_files);
  __pyx_v_self->search_specific_files = __pyx_v_search_specific_files;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1067
 *         # The parents up to the root of the paths we are searching.
 *         # After all normal paths are returned, these specific items are returned.
 *         self.search_specific_file_parents = set()             # <<<<<<<<<<<<<<
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->search_specific_file_parents);
  __Pyx_DECREF(__pyx_v_self->search_specific_file_parents);
  __pyx_v_self->search_specific_file_parents = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1069
 *         self.search_specific_file_parents = set()
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()             # <<<<<<<<<<<<<<
 *         self.state = state
 *         self.current_root = None
 */
  __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->seen_ids);
  __Pyx_DECREF(__pyx_v_self->seen_ids);
  __pyx_v_self->seen_ids = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1070
 *         # The ids we've sent out in the delta.
 *         self.seen_ids = set()
 *         self.state = state             # <<<<<<<<<<<<<<
 *         self.current_root = None
 *         self.current_root_unicode = None
 */
  __Pyx_INCREF(__pyx_v_state);
  __Pyx_GIVEREF(__pyx_v_state);
  __Pyx_GOTREF(__pyx_v_self->state);
  __Pyx_DECREF(__pyx_v_self->state);
  __pyx_v_self->state = __pyx_v_state;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1071
 *         self.seen_ids = set()
 *         self.state = state
 *         self.current_root = None             # <<<<<<<<<<<<<<
 *         self.current_root_unicode = None
 *         self.root_entries = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_root);
  __Pyx_DECREF(__pyx_v_self->current_root);
  __pyx_v_self->current_root = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1072
 *         self.state = state
 *         self.current_root = None
 *         self.current_root_unicode = None             # <<<<<<<<<<<<<<
 *         self.root_entries = None
 *         self.root_entries_pos = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_root_unicode);
  __Pyx_DECREF(__pyx_v_self->current_root_unicode);
  __pyx_v_self->current_root_unicode = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1073
 *         self.current_root = None
 *         self.current_root_unicode = None
 *         self.root_entries = None             # <<<<<<<<<<<<<<
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_entries);
  __Pyx_DECREF(__pyx_v_self->root_entries);
  __pyx_v_self->root_entries = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1074
 *         self.current_root_unicode = None
 *         self.root_entries = None
 *         self.root_entries_pos = 0             # <<<<<<<<<<<<<<
 *         self.root_entries_len = 0
 *         self.root_abspath = None
 */
  __pyx_v_self->root_entries_pos = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1075
 *         self.root_entries = None
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0             # <<<<<<<<<<<<<<
 *         self.root_abspath = None
 *         if source_index is None:
 */
  __pyx_v_self->root_entries_len = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1076
 *         self.root_entries_pos = 0
 *         self.root_entries_len = 0
 *         self.root_abspath = None             # <<<<<<<<<<<<<<
 *         if source_index is None:
 *             self.source_index = -1
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_abspath);
  __Pyx_DECREF(__pyx_v_self->root_abspath);
  __pyx_v_self->root_abspath = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1077
 *         self.root_entries_len = 0
 *         self.root_abspath = None
 *         if source_index is None:             # <<<<<<<<<<<<<<
 *             self.source_index = -1
 *         else:
 */
  __pyx_t_6 = (__pyx_v_source_index == Py_None);
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1078
 *         self.root_abspath = None
 *         if source_index is None:
 *             self.source_index = -1             # <<<<<<<<<<<<<<
 *         else:
 *             self.source_index = source_index
 */
    __pyx_v_self->source_index = -1;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1077
 *         self.root_entries_len = 0
 *         self.root_abspath = None
 *         if source_index is None:             # <<<<<<<<<<<<<<
 *             self.source_index = -1
 *         else:
 */
    goto __pyx_L4;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1080
 *             self.source_index = -1
 *         else:
 *             self.source_index = source_index             # <<<<<<<<<<<<<<
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_source_index); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1080, __pyx_L1_error)
    __pyx_v_self->source_index = __pyx_t_4;
  }
  __pyx_L4:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1081
 *         else:
 *             self.source_index = source_index
 *         self.target_index = target_index             # <<<<<<<<<<<<<<
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_target_index); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_v_self->target_index = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1082
 *             self.source_index = source_index
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned             # <<<<<<<<<<<<<<
 *         self.tree = tree
 *         self.dir_iterator = None
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_want_unversioned); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1082, __pyx_L1_error)
  __pyx_v_self->want_unversioned = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1083
 *         self.target_index = target_index
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree             # <<<<<<<<<<<<<<
 *         self.dir_iterator = None
 *         self.block_index = -1
 */
  __Pyx_INCREF(__pyx_v_tree);
  __Pyx_GIVEREF(__pyx_v_tree);
  __Pyx_GOTREF(__pyx_v_self->tree);
  __Pyx_DECREF(__pyx_v_self->tree);
  __pyx_v_self->tree = __pyx_v_tree;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1084
 *         self.want_unversioned = want_unversioned
 *         self.tree = tree
 *         self.dir_iterator = None             # <<<<<<<<<<<<<<
 *         self.block_index = -1
 *         self.current_block = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->dir_iterator);
  __Pyx_DECREF(__pyx_v_self->dir_iterator);
  __pyx_v_self->dir_iterator = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1085
 *         self.tree = tree
 *         self.dir_iterator = None
 *         self.block_index = -1             # <<<<<<<<<<<<<<
 *         self.current_block = None
 *         self.current_block_list = None
 */
  __pyx_v_self->block_index = -1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1086
 *         self.dir_iterator = None
 *         self.block_index = -1
 *         self.current_block = None             # <<<<<<<<<<<<<<
 *         self.current_block_list = None
 *         self.current_block_pos = -1
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_block);
  __Pyx_DECREF(__pyx_v_self->current_block);
  __pyx_v_self->current_block = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1087
 *         self.block_index = -1
 *         self.current_block = None
 *         self.current_block_list = None             # <<<<<<<<<<<<<<
 *         self.current_block_pos = -1
 *         self.current_dir_info = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_block_list);
  __Pyx_DECREF(__pyx_v_self->current_block_list);
  __pyx_v_self->current_block_list = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1088
 *         self.current_block = None
 *         self.current_block_list = None
 *         self.current_block_pos = -1             # <<<<<<<<<<<<<<
 *         self.current_dir_info = None
 *         self.current_dir_list = None
 */
  __pyx_v_self->current_block_pos = -1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1089
 *         self.current_block_list = None
 *         self.current_block_pos = -1
 *         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_dir_info);
  __Pyx_DECREF(__pyx_v_self->current_dir_info);
  __pyx_v_self->current_dir_info = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1090
 *         self.current_block_pos = -1
 *         self.current_dir_info = None
 *         self.current_dir_list = None             # <<<<<<<<<<<<<<
 *         self._pending_consistent_entries = []
 *         self.path_index = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->current_dir_list);
  __Pyx_DECREF(__pyx_v_self->current_dir_list);
  __pyx_v_self->current_dir_list = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1091
 *         self.current_dir_info = None
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []             # <<<<<<<<<<<<<<
 *         self.path_index = 0
 *         self.root_dir_info = None
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_pending_consistent_entries);
  __Pyx_DECREF(__pyx_v_self->_pending_consistent_entries);
  __pyx_v_self->_pending_consistent_entries = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1092
 *         self.current_dir_list = None
 *         self._pending_consistent_entries = []
 *         self.path_index = 0             # <<<<<<<<<<<<<<
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left
 */
  __pyx_v_self->path_index = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1093
 *         self._pending_consistent_entries = []
 *         self.path_index = 0
 *         self.root_dir_info = None             # <<<<<<<<<<<<<<
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->root_dir_info);
  __Pyx_DECREF(__pyx_v_self->root_dir_info);
  __pyx_v_self->root_dir_info = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1094
 *         self.path_index = 0
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left             # <<<<<<<<<<<<<<
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_bisect); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_bisect_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->bisect_left);
  __Pyx_DECREF(__pyx_v_self->bisect_left);
  __pyx_v_self->bisect_left = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1095
 *         self.root_dir_info = None
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin             # <<<<<<<<<<<<<<
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_pathjoin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->pathjoin);
  __Pyx_DECREF(__pyx_v_self->pathjoin);
  __pyx_v_self->pathjoin = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1096
 *         self.bisect_left = bisect.bisect_left
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat             # <<<<<<<<<<<<<<
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_fstat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fstat);
  __Pyx_DECREF(__pyx_v_self->fstat);
  __pyx_v_self->fstat = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1097
 *         self.pathjoin = osutils.pathjoin
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file             # <<<<<<<<<<<<<<
 *         if target_index != 0:
 *             # A lot of code in here depends on target_index == 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_sha_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->sha_file);
  __Pyx_DECREF(__pyx_v_self->sha_file);
  __pyx_v_self->sha_file = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1098
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:             # <<<<<<<<<<<<<<
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')
 */
  __pyx_t_3 = __Pyx_PyInt_NeObjC(__pyx_v_target_index, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1100
 *         if target_index != 0:
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')             # <<<<<<<<<<<<<<
 * 
 *     cdef _process_entry(self, entry, path_info):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_errors); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BzrError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_kp_s_unsupported_target_index) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_s_unsupported_target_index);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1100, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1098
 *         self.fstat = os.fstat
 *         self.sha_file = osutils.sha_file
 *         if target_index != 0:             # <<<<<<<<<<<<<<
 *             # A lot of code in here depends on target_index == 0
 *             raise errors.BzrError('unsupported target index')
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1038
 *     cdef object sha_file
 * 
 *     def __init__(self, include_unchanged, use_filesystem_for_exec,             # <<<<<<<<<<<<<<
 *         search_specific_files, state, source_index, target_index,
 *         want_unversioned, tree):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1102
 *             raise errors.BzrError('unsupported target index')
 * 
 *     cdef _process_entry(self, entry, path_info):             # <<<<<<<<<<<<<<
 *         """Compare an entry and real disk to generate delta information.
 * 
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_entry, PyObject *__pyx_v_path_info) {
  char __pyx_v_target_minikind;
  char __pyx_v_source_minikind;
  PyObject *__pyx_v_file_id = 0;
  int __pyx_v_content_change;
  PyObject *__pyx_v_details_list = 0;
  PyObject *__pyx_v_source_details = NULL;
  PyObject *__pyx_v_target_details = NULL;
  PyObject *__pyx_v_link_or_sha1 = NULL;
  PyObject *__pyx_v_old_dirname = NULL;
  PyObject *__pyx_v_old_basename = NULL;
  PyObject *__pyx_v_old_path = NULL;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_old_entry = NULL;
  PyObject *__pyx_v_target_kind = NULL;
  PyObject *__pyx_v_target_exec = NULL;
  PyObject *__pyx_v_statvalue = NULL;
  PyObject *__pyx_v_source_parent_id = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_source_parent_entry = NULL;
  PyObject *__pyx_v_new_dirname = NULL;
  PyObject *__pyx_v_target_parent_id = NULL;
  PyObject *__pyx_v_target_parent_entry = NULL;
  PyObject *__pyx_v_source_exec = NULL;
  PyObject *__pyx_v_changed = NULL;
  PyObject *__pyx_v_old_path_u = NULL;
  PyObject *__pyx_v_path_u = NULL;
  PyObject *__pyx_v_source_kind = NULL;
  PyObject *__pyx_v_parent_entry = NULL;
  PyObject *__pyx_v_parent_id = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  __Pyx_RefNannySetupContext("_process_entry", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1124
 *         cdef int content_change
 *         cdef object details_list
 *         file_id = None             # <<<<<<<<<<<<<<
 *         details_list = entry[1]
 *         if -1 == self.source_index:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_file_id = Py_None;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1125
 *         cdef object details_list
 *         file_id = None
 *         details_list = entry[1]             # <<<<<<<<<<<<<<
 *         if -1 == self.source_index:
 *             source_details = DirState.NULL_PARENT_DETAILS
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_details_list = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1126
 *         file_id = None
 *         details_list = entry[1]
 *         if -1 == self.source_index:             # <<<<<<<<<<<<<<
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
 */
  __pyx_t_2 = ((-1L == __pyx_v_self->source_index) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1127
 *         details_list = entry[1]
 *         if -1 == self.source_index:
 *             source_details = DirState.NULL_PARENT_DETAILS             # <<<<<<<<<<<<<<
 *         else:
 *             source_details = details_list[self.source_index]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_DirState); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_NULL_PARENT_DETAILS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_source_details = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1126
 *         file_id = None
 *         details_list = entry[1]
 *         if -1 == self.source_index:             # <<<<<<<<<<<<<<
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1129
 *             source_details = DirState.NULL_PARENT_DETAILS
 *         else:
 *             source_details = details_list[self.source_index]             # <<<<<<<<<<<<<<
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_source_details = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1130
 *         else:
 *             source_details = details_list[self.source_index]
 *         target_details = details_list[self.target_index]             # <<<<<<<<<<<<<<
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_target_details = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1131
 *             source_details = details_list[self.source_index]
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])             # <<<<<<<<<<<<<<
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_target_details, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_3); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_target_minikind = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1132
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
 */
  __pyx_t_5 = (__pyx_v_path_info != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_6 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind) != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1133
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:             # <<<<<<<<<<<<<<
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
 */
    __pyx_t_2 = ((__pyx_v_self->target_index != 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1135
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)             # <<<<<<<<<<<<<<
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
 *             # The entry may have been modified by update_entry
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1134
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %             # <<<<<<<<<<<<<<
 *                                      self.target_index)
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
 */
      __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Unsupported_target_index_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1134, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1134, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1133
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):
 *             if self.target_index != 0:             # <<<<<<<<<<<<<<
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1136
 *                 raise AssertionError("Unsupported target index %d" %
 *                                      self.target_index)
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])             # <<<<<<<<<<<<<<
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]
 */
    __pyx_t_3 = __pyx_v_self->state;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_update_entry(__pyx_t_3, __pyx_v_entry, __pyx_t_1, __pyx_t_7, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_link_or_sha1 = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1138
 *             link_or_sha1 = update_entry(self.state, entry, path_info[4], path_info[3])
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]             # <<<<<<<<<<<<<<
 *             target_minikind = _minikind_from_string(target_details[0])
 *         else:
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_details_list, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF_SET(__pyx_v_target_details, __pyx_t_8);
    __pyx_t_8 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1139
 *             # The entry may have been modified by update_entry
 *             target_details = details_list[self.target_index]
 *             target_minikind = _minikind_from_string(target_details[0])             # <<<<<<<<<<<<<<
 *         else:
 *             link_or_sha1 = None
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_target_details, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_8); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_target_minikind = __pyx_t_4;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1132
 *         target_details = details_list[self.target_index]
 *         target_minikind = _minikind_from_string(target_details[0])
 *         if path_info is not None and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             if self.target_index != 0:
 *                 raise AssertionError("Unsupported target index %d" %
 */
    goto __pyx_L4;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1141
 *             target_minikind = _minikind_from_string(target_details[0])
 *         else:
 *             link_or_sha1 = None             # <<<<<<<<<<<<<<
 *         # the rest of this function is 0.3 seconds on 50K paths, or
 *         # 0.000006 seconds per call.
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_link_or_sha1 = Py_None;
  }
  __pyx_L4:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1144
 *         # the rest of this function is 0.3 seconds on 50K paths, or
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])             # <<<<<<<<<<<<<<
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')
 *             and _versioned_minikind(target_minikind)):
 */
  __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_source_details, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_8); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_source_minikind = __pyx_t_4;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1145
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
 */
  __pyx_t_6 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind) != 0);
  if (!__pyx_t_6) {
  } else {
    goto __pyx_L10_next_and;
  }
  __pyx_t_6 = ((__pyx_v_source_minikind == 'r') != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_L10_next_and:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1146
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')
 *             and _versioned_minikind(target_minikind)):             # <<<<<<<<<<<<<<
 *             # claimed content in both: diff
 *             #   r    | fdlt   |      | add source to search, add id path move and perform
 */
  __pyx_t_6 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind) != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1145
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
 */
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1152
 *             #   r    | fdlt   |  a   | dangling file that was present in the basis.
 *             #        |        |      | ???
 *             if source_minikind != b'r':             # <<<<<<<<<<<<<<
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
 */
    __pyx_t_2 = ((__pyx_v_source_minikind != 'r') != 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1153
 *             #        |        |      | ???
 *             if source_minikind != b'r':
 *                 old_dirname = entry[0][0]             # <<<<<<<<<<<<<<
 *                 old_basename = entry[0][1]
 *                 old_path = path = None
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_v_old_dirname = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1154
 *             if source_minikind != b'r':
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]             # <<<<<<<<<<<<<<
 *                 old_path = path = None
 *             else:
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_7, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_old_basename = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1155
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
 *                 old_path = path = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # add the source to the search path to find any children it
 */
      __Pyx_INCREF(Py_None);
      __pyx_v_old_path = Py_None;
      __Pyx_INCREF(Py_None);
      __pyx_v_path = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1152
 *             #   r    | fdlt   |  a   | dangling file that was present in the basis.
 *             #        |        |      | ???
 *             if source_minikind != b'r':             # <<<<<<<<<<<<<<
 *                 old_dirname = entry[0][0]
 *                 old_basename = entry[0][1]
 */
      goto __pyx_L12;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1159
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
 */
    /*else*/ {
      __pyx_t_6 = ((!(__pyx_v_self->doing_consistency_expansion != 0)) != 0);
      if (__pyx_t_6) {
      } else {
        __pyx_t_2 = __pyx_t_6;
        goto __pyx_L14_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1160
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_is_inside_any); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1161
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):             # <<<<<<<<<<<<<<
 *                     self.search_specific_files.add(source_details[1])
 *                     # expanding from a user requested path, parent expansion
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self->searched_specific_files, __pyx_t_1};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_self->searched_specific_files, __pyx_t_1};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_v_self->searched_specific_files);
        __Pyx_GIVEREF(__pyx_v_self->searched_specific_files);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_self->searched_specific_files);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1160
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and
 *                     not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1160, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_5 = ((!__pyx_t_6) != 0);
      __pyx_t_2 = __pyx_t_5;
      __pyx_L14_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1159
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
 */
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1162
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
 *                     self.search_specific_files.add(source_details[1])             # <<<<<<<<<<<<<<
 *                     # expanding from a user requested path, parent expansion
 *                     # for delta consistency happens later.
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_files, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1162, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1162, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_1 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_1)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_8 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_1, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_10);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1162, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1159
 *                 # add the source to the search path to find any children it
 *                 # has.  TODO ? : only add if it is a container ?
 *                 if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                     not is_inside_any(self.searched_specific_files,
 *                                              source_details[1])):
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1167
 *                 # generate the old path; this is needed for stating later
 *                 # as well.
 *                 old_path = source_details[1]             # <<<<<<<<<<<<<<
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1167, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_v_old_path = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1168
 *                 # as well.
 *                 old_path = source_details[1]
 *                 old_dirname, old_basename = os.path.split(old_path)             # <<<<<<<<<<<<<<
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_os); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_path); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_8 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_10, __pyx_v_old_path) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_old_path);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
        PyObject* sequence = __pyx_t_8;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1168, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_10);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_11(__pyx_t_1); if (unlikely(!__pyx_t_7)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_10 = __pyx_t_11(__pyx_t_1); if (unlikely(!__pyx_t_10)) goto __pyx_L16_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_10);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_1), 2) < 0) __PYX_ERR(0, 1168, __pyx_L1_error)
        __pyx_t_11 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L17_unpacking_done;
        __pyx_L16_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_11 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1168, __pyx_L1_error)
        __pyx_L17_unpacking_done:;
      }
      __pyx_v_old_dirname = __pyx_t_7;
      __pyx_t_7 = 0;
      __pyx_v_old_basename = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1169
 *                 old_path = source_details[1]
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *                 old_entry = self.state._get_entry(self.source_index,
 *                                              path_utf8=old_path)
 */
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_10, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_v_self->pathjoin);
      __pyx_t_10 = __pyx_v_self->pathjoin; __pyx_t_3 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_1};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1169, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_7, __pyx_t_1};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1169, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1169, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, __pyx_t_1);
        __pyx_t_7 = 0;
        __pyx_t_1 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1169, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_path = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1170
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                              path_utf8=old_path)
 *                 # update the source details variable to be the real
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_get_entry); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1171
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,
 *                                              path_utf8=old_path)             # <<<<<<<<<<<<<<
 *                 # update the source details variable to be the real
 *                 # location.
 */
      __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_path_utf8, __pyx_v_old_path) < 0) __PYX_ERR(0, 1171, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1170
 *                 old_dirname, old_basename = os.path.split(old_path)
 *                 path = self.pathjoin(entry[0][0], entry[0][1])
 *                 old_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                              path_utf8=old_path)
 *                 # update the source details variable to be the real
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_12, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_old_entry = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1174
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_old_entry, __pyx_tuple__6, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1174, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1174, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__pyx_t_2)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1175
 *                 # location.
 *                 if old_entry == (None, None):
 *                     raise DirstateCorrupt(self.state._filename,             # <<<<<<<<<<<<<<
 *                         "entry '%s/%s' is considered renamed from %r"
 *                         " but source does not exist\n"
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_filename); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1178
 *                         "entry '%s/%s' is considered renamed from %r"
 *                         " but source does not exist\n"
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))             # <<<<<<<<<<<<<<
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
 */
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_8, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = PyTuple_New(4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_old_path);
        __Pyx_GIVEREF(__pyx_v_old_path);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_v_old_path);
        __Pyx_INCREF(__pyx_v_entry);
        __Pyx_GIVEREF(__pyx_v_entry);
        PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_v_entry);
        __pyx_t_7 = 0;
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_entry_s_s_is_considered_renamed, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_12, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_12, __pyx_t_3};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1175, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_t_3);
          __pyx_t_12 = 0;
          __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 1175, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1174
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1179
 *                         " but source does not exist\n"
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
 *                 source_details = old_entry[1][self.source_index]             # <<<<<<<<<<<<<<
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_old_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1179, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_source_details, __pyx_t_10);
      __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1180
 *                         "entry: %s" % (entry[0][0], entry[0][1], old_path, entry))
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])             # <<<<<<<<<<<<<<
 *             if path_info is None:
 *                 # the file is missing on disk, show as removed.
 */
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_source_details, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_10); if (unlikely(__pyx_t_4 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_v_source_minikind = __pyx_t_4;
    }
    __pyx_L12:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1181
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:             # <<<<<<<<<<<<<<
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
 */
    __pyx_t_2 = (__pyx_v_path_info == Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (__pyx_t_5) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1183
 *             if path_info is None:
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1             # <<<<<<<<<<<<<<
 *                 target_kind = None
 *                 target_exec = False
 */
      __pyx_v_content_change = 1;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1184
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
 *                 target_kind = None             # <<<<<<<<<<<<<<
 *                 target_exec = False
 *             else:
 */
      __Pyx_INCREF(Py_None);
      __pyx_v_target_kind = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1185
 *                 content_change = 1
 *                 target_kind = None
 *                 target_exec = False             # <<<<<<<<<<<<<<
 *             else:
 *                 # source and target are both versioned and disk file is present.
 */
      __Pyx_INCREF(Py_False);
      __pyx_v_target_exec = Py_False;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1181
 *                 source_details = old_entry[1][self.source_index]
 *                 source_minikind = _minikind_from_string(source_details[0])
 *             if path_info is None:             # <<<<<<<<<<<<<<
 *                 # the file is missing on disk, show as removed.
 *                 content_change = 1
 */
      goto __pyx_L19;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1188
 *             else:
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]             # <<<<<<<<<<<<<<
 *                 if target_kind == 'directory':
 *                     if path is None:
 */
    /*else*/ {
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_v_target_kind = __pyx_t_10;
      __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1189
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':             # <<<<<<<<<<<<<<
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 */
      __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_target_kind, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1189, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1190
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
 */
        __pyx_t_5 = (__pyx_v_path == Py_None);
        __pyx_t_2 = (__pyx_t_5 != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1191
 *                 if target_kind == 'directory':
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
 */
          __Pyx_INCREF(__pyx_v_self->pathjoin);
          __pyx_t_1 = __pyx_v_self->pathjoin; __pyx_t_7 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1191, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1191, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_10);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1191, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_7) {
              __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            }
            __Pyx_INCREF(__pyx_v_old_dirname);
            __Pyx_GIVEREF(__pyx_v_old_dirname);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_9, __pyx_v_old_dirname);
            __Pyx_INCREF(__pyx_v_old_basename);
            __Pyx_GIVEREF(__pyx_v_old_basename);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_9, __pyx_v_old_basename);
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1191, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_t_10);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_10);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1190
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1192
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]             # <<<<<<<<<<<<<<
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':
 */
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_10, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF_SET(__pyx_v_file_id, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1193
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'd':
 *                         content_change = 1
 */
        if (unlikely(PyObject_SetItem(__pyx_v_self->new_dirname_to_file_id, __pyx_v_path, __pyx_v_file_id) < 0)) __PYX_ERR(0, 1193, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1194
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
        __pyx_t_2 = ((__pyx_v_source_minikind != 'd') != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1195
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # directories have no fingerprint
 */
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1194
 *                     file_id = entry[0][2]
 *                     self.new_dirname_to_file_id[path] = file_id
 *                     if source_minikind != b'd':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
          goto __pyx_L22;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1198
 *                     else:
 *                         # directories have no fingerprint
 *                         content_change = 0             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 elif target_kind == 'file':
 */
        /*else*/ {
          __pyx_v_content_change = 0;
        }
        __pyx_L22:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1199
 *                         # directories have no fingerprint
 *                         content_change = 0
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':
 */
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1189
 *                 # source and target are both versioned and disk file is present.
 *                 target_kind = path_info[2]
 *                 if target_kind == 'directory':             # <<<<<<<<<<<<<<
 *                     if path is None:
 *                         old_path = path = self.pathjoin(old_dirname, old_basename)
 */
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1200
 *                         content_change = 0
 *                     target_exec = False
 *                 elif target_kind == 'file':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'f':
 *                         content_change = 1
 */
      __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_target_kind, __pyx_n_s_file, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1200, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1201
 *                     target_exec = False
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
        __pyx_t_2 = ((__pyx_v_source_minikind != 'f') != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1202
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # Check the sha. We can't just rely on the size as
 */
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1201
 *                     target_exec = False
 *                 elif target_kind == 'file':
 *                     if source_minikind != b'f':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
          goto __pyx_L23;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1207
 *                         # content filtering may mean differ sizes actually
 *                         # map to the same content
 *                         if link_or_sha1 is None:             # <<<<<<<<<<<<<<
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
 */
        /*else*/ {
          __pyx_t_2 = (__pyx_v_link_or_sha1 == Py_None);
          __pyx_t_5 = (__pyx_t_2 != 0);
          if (__pyx_t_5) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1210
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
 *                                 self.state._sha1_provider.stat_and_sha1(             # <<<<<<<<<<<<<<
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_sha1_provider); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1210, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_stat_and_sha1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1211
 *                             statvalue, link_or_sha1 = \
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])             # <<<<<<<<<<<<<<
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)
 */
            __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_path_info, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1211, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __pyx_t_7 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_7)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_7);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_7, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_10);
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
              PyObject* sequence = __pyx_t_1;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 1209, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_10);
              #else
              __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1209, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_10 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1209, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              #endif
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1209, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext;
              index = 0; __pyx_t_3 = __pyx_t_11(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L25_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_3);
              index = 1; __pyx_t_10 = __pyx_t_11(__pyx_t_7); if (unlikely(!__pyx_t_10)) goto __pyx_L25_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_10);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_7), 2) < 0) __PYX_ERR(0, 1209, __pyx_L1_error)
              __pyx_t_11 = NULL;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              goto __pyx_L26_unpacking_done;
              __pyx_L25_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_11 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 1209, __pyx_L1_error)
              __pyx_L26_unpacking_done:;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1209
 *                         if link_or_sha1 is None:
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \             # <<<<<<<<<<<<<<
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])
 */
            __pyx_v_statvalue = __pyx_t_3;
            __pyx_t_3 = 0;
            __Pyx_DECREF_SET(__pyx_v_link_or_sha1, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1212
 *                                 self.state._sha1_provider.stat_and_sha1(
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,             # <<<<<<<<<<<<<<
 *                                 statvalue)
 *                         content_change = (link_or_sha1 != source_details[1])
 */
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_observed_sha1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1212, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1213
 *                                 path_info[4])
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)             # <<<<<<<<<<<<<<
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
 */
            __pyx_t_3 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_entry, __pyx_v_link_or_sha1, __pyx_v_statvalue};
              __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_v_entry, __pyx_v_link_or_sha1, __pyx_v_statvalue};
              __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else
            #endif
            {
              __pyx_t_7 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1212, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              if (__pyx_t_3) {
                __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
              }
              __Pyx_INCREF(__pyx_v_entry);
              __Pyx_GIVEREF(__pyx_v_entry);
              PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_9, __pyx_v_entry);
              __Pyx_INCREF(__pyx_v_link_or_sha1);
              __Pyx_GIVEREF(__pyx_v_link_or_sha1);
              PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_9, __pyx_v_link_or_sha1);
              __Pyx_INCREF(__pyx_v_statvalue);
              __Pyx_GIVEREF(__pyx_v_statvalue);
              PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_9, __pyx_v_statvalue);
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1207
 *                         # content filtering may mean differ sizes actually
 *                         # map to the same content
 *                         if link_or_sha1 is None:             # <<<<<<<<<<<<<<
 *                             # Stat cache miss:
 *                             statvalue, link_or_sha1 = \
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1214
 *                             self.state._observed_sha1(entry, link_or_sha1,
 *                                 statvalue)
 *                         content_change = (link_or_sha1 != source_details[1])             # <<<<<<<<<<<<<<
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:
 */
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = PyObject_RichCompare(__pyx_v_link_or_sha1, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1214, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1214, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_v_content_change = __pyx_t_9;
        }
        __pyx_L23:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1216
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
 */
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_self->use_filesystem_for_exec); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1216, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1219
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
 *                         target_exec = bool(S_IXUSR & path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                     else:
 *                         target_exec = target_details[3]
 */
          __pyx_t_10 = __Pyx_PyInt_From_int(S_IXUSR); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = PyNumber_And(__pyx_t_10, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_5))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_v_target_exec = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1216
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     # Target details is updated at update_entry time
 *                     if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                         # We don't need S_ISREG here, because we are sure
 *                         # we are dealing with a file.
 */
          goto __pyx_L27;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1221
 *                         target_exec = bool(S_IXUSR & path_info[3].st_mode)
 *                     else:
 *                         target_exec = target_details[3]             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_target_details, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1221, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_v_target_exec = __pyx_t_1;
          __pyx_t_1 = 0;
        }
        __pyx_L27:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1200
 *                         content_change = 0
 *                     target_exec = False
 *                 elif target_kind == 'file':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'f':
 *                         content_change = 1
 */
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1222
 *                     else:
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'l':
 *                         content_change = 1
 */
      __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_target_kind, __pyx_n_s_symlink, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1222, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1223
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
        __pyx_t_5 = ((__pyx_v_source_minikind != 'l') != 0);
        if (__pyx_t_5) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1224
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])
 */
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1223
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':
 *                     if source_minikind != b'l':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
          goto __pyx_L28;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1226
 *                         content_change = 1
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_source_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = PyObject_RichCompare(__pyx_v_link_or_sha1, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1226, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1226, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v_content_change = __pyx_t_9;
        }
        __pyx_L28:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1227
 *                     else:
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':
 */
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1222
 *                     else:
 *                         target_exec = target_details[3]
 *                 elif target_kind == 'symlink':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b'l':
 *                         content_change = 1
 */
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1228
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b't':
 *                         content_change = 1
 */
      __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_v_target_kind, __pyx_kp_s_tree_reference, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1228, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1229
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
        __pyx_t_5 = ((__pyx_v_source_minikind != 't') != 0);
        if (__pyx_t_5) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1230
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':
 *                         content_change = 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         content_change = 0
 */
          __pyx_v_content_change = 1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1229
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':
 *                     if source_minikind != b't':             # <<<<<<<<<<<<<<
 *                         content_change = 1
 *                     else:
 */
          goto __pyx_L29;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1232
 *                         content_change = 1
 *                     else:
 *                         content_change = 0             # <<<<<<<<<<<<<<
 *                     target_exec = False
 *                 else:
 */
        /*else*/ {
          __pyx_v_content_change = 0;
        }
        __pyx_L29:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1233
 *                     else:
 *                         content_change = 0
 *                     target_exec = False             # <<<<<<<<<<<<<<
 *                 else:
 *                     if path is None:
 */
        __Pyx_INCREF(Py_False);
        __pyx_v_target_exec = Py_False;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1228
 *                         content_change = (link_or_sha1 != source_details[1])
 *                     target_exec = False
 *                 elif target_kind == 'tree-reference':             # <<<<<<<<<<<<<<
 *                     if source_minikind != b't':
 *                         content_change = 1
 */
        goto __pyx_L20;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1235
 *                     target_exec = False
 *                 else:
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 */
      /*else*/ {
        __pyx_t_5 = (__pyx_v_path == Py_None);
        __pyx_t_2 = (__pyx_t_5 != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1236
 *                 else:
 *                     if path is None:
 *                         path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
 */
          __Pyx_INCREF(__pyx_v_self->pathjoin);
          __pyx_t_1 = __pyx_v_self->pathjoin; __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1236, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_old_dirname, __pyx_v_old_basename};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1236, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_7);
          } else
          #endif
          {
            __pyx_t_3 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1236, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_v_old_dirname);
            __Pyx_GIVEREF(__pyx_v_old_dirname);
            PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_9, __pyx_v_old_dirname);
            __Pyx_INCREF(__pyx_v_old_basename);
            __Pyx_GIVEREF(__pyx_v_old_basename);
            PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_9, __pyx_v_old_basename);
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1236, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1235
 *                     target_exec = False
 *                 else:
 *                     if path is None:             # <<<<<<<<<<<<<<
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1237
 *                     if path is None:
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])             # <<<<<<<<<<<<<<
 *             if source_minikind == b'd':
 *                 if path is None:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_errors); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BadFileKindError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_path, __pyx_t_1};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1237, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_path, __pyx_t_1};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1237, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1237, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
          }
          __Pyx_INCREF(__pyx_v_path);
          __Pyx_GIVEREF(__pyx_v_path);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_9, __pyx_v_path);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_9, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1237, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1237, __pyx_L1_error)
      }
      __pyx_L20:;
    }
    __pyx_L19:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1238
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':             # <<<<<<<<<<<<<<
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 */
    __pyx_t_2 = ((__pyx_v_source_minikind == 'd') != 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1239
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
 *                 if path is None:             # <<<<<<<<<<<<<<
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
 */
      __pyx_t_2 = (__pyx_v_path == Py_None);
      __pyx_t_5 = (__pyx_t_2 != 0);
      if (__pyx_t_5) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1240
 *             if source_minikind == b'd':
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                 if file_id is None:
 *                     file_id = entry[0][2]
 */
        __Pyx_INCREF(__pyx_v_self->pathjoin);
        __pyx_t_3 = __pyx_v_self->pathjoin; __pyx_t_12 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_1 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(__pyx_v_old_dirname);
          __Pyx_GIVEREF(__pyx_v_old_dirname);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_9, __pyx_v_old_dirname);
          __Pyx_INCREF(__pyx_v_old_basename);
          __Pyx_GIVEREF(__pyx_v_old_basename);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_9, __pyx_v_old_basename);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_t_7);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1239
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':
 *                 if path is None:             # <<<<<<<<<<<<<<
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1241
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id
 */
      __pyx_t_5 = (__pyx_v_file_id == Py_None);
      __pyx_t_2 = (__pyx_t_5 != 0);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1242
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:
 *                     file_id = entry[0][2]             # <<<<<<<<<<<<<<
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_7, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF_SET(__pyx_v_file_id, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1241
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 *                 if file_id is None:             # <<<<<<<<<<<<<<
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1243
 *                 if file_id is None:
 *                     file_id = entry[0][2]
 *                 self.old_dirname_to_file_id[old_path] = file_id             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:
 */
      if (unlikely(PyObject_SetItem(__pyx_v_self->old_dirname_to_file_id, __pyx_v_old_path, __pyx_v_file_id) < 0)) __PYX_ERR(0, 1243, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1238
 *                         path = self.pathjoin(old_dirname, old_basename)
 *                     raise errors.BadFileKindError(path, path_info[2])
 *             if source_minikind == b'd':             # <<<<<<<<<<<<<<
 *                 if path is None:
 *                     old_path = path = self.pathjoin(old_dirname, old_basename)
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1245
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_old_basename); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1245, __pyx_L1_error)
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L35_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->last_source_parent, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = PyObject_RichCompare(__pyx_v_old_dirname, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1245, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_2 = __pyx_t_5;
    __pyx_L35_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1247
 *             if old_basename and old_dirname == self.last_source_parent[0]:
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]             # <<<<<<<<<<<<<<
 *             else:
 *                 try:
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->last_source_parent, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_v_source_parent_id = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1245
 *                 self.old_dirname_to_file_id[old_path] = file_id
 *             # parent id is the entry for the path in the target tree
 *             if old_basename and old_dirname == self.last_source_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root source entries.
 *                 source_parent_id = self.last_source_parent[1]
 */
      goto __pyx_L34;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1249
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 */
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1250
 *             else:
 *                 try:
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]             # <<<<<<<<<<<<<<
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 */
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_v_self->old_dirname_to_file_id, __pyx_v_old_dirname); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1250, __pyx_L37_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_v_source_parent_id = __pyx_t_7;
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1249
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 */
        }
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        goto __pyx_L42_try_end;
        __pyx_L37_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1251
 *                 try:
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:             # <<<<<<<<<<<<<<
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 1251, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_v__ = __pyx_t_3;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1252
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 */
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_get_entry); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1252, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1252, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1252, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_10);
          PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1253
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)             # <<<<<<<<<<<<<<
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:
 */
          __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1253, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_path_utf8, __pyx_v_old_dirname) < 0) __PYX_ERR(0, 1253, __pyx_L39_except_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1252
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 *                     source_parent_entry = self.state._get_entry(self.source_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 */
          __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_8, __pyx_t_10); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1252, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_v_source_parent_entry = __pyx_t_16;
          __pyx_t_16 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1254
 *                     source_parent_entry = self.state._get_entry(self.source_index,
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]             # <<<<<<<<<<<<<<
 *                 if source_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 */
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_source_parent_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1254, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_16, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1254, __pyx_L39_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF_SET(__pyx_v_source_parent_id, __pyx_t_10);
          __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L38_exception_handled;
        }
        goto __pyx_L39_except_error;
        __pyx_L39_except_error:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1249
 *                 source_parent_id = self.last_source_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     source_parent_id = self.old_dirname_to_file_id[old_dirname]
 *                 except KeyError, _:
 */
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
        goto __pyx_L1_error;
        __pyx_L38_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_14, __pyx_t_15);
        __pyx_L42_try_end:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1255
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_source_parent_id, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1257
 *                 if source_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.last_source_parent[0] = old_dirname
 */
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_source_parent_id, Py_None);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1255
 *                                                            path_utf8=old_dirname)
 *                     source_parent_id = source_parent_entry[0][2]
 *                 if source_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     source_parent_id = None
 */
        goto __pyx_L45;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1259
 *                     source_parent_id = None
 *                 else:
 *                     self.last_source_parent[0] = old_dirname             # <<<<<<<<<<<<<<
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
 */
      /*else*/ {
        if (unlikely(__Pyx_SetItemInt(__pyx_v_self->last_source_parent, 0, __pyx_v_old_dirname, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 1259, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1260
 *                 else:
 *                     self.last_source_parent[0] = old_dirname
 *                     self.last_source_parent[1] = source_parent_id             # <<<<<<<<<<<<<<
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
 */
        if (unlikely(__Pyx_SetItemInt(__pyx_v_self->last_source_parent, 1, __pyx_v_source_parent_id, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 1260, __pyx_L1_error)
      }
      __pyx_L45:;
    }
    __pyx_L34:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1261
 *                     self.last_source_parent[0] = old_dirname
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]             # <<<<<<<<<<<<<<
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
 *                 # use a cached hit for non-root target entries.
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_new_dirname = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1262
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L47_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->last_target_parent, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_new_dirname, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __pyx_t_5;
    __pyx_L47_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1264
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]             # <<<<<<<<<<<<<<
 *             else:
 *                 try:
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->last_target_parent, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_target_parent_id = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1262
 *                     self.last_source_parent[1] = source_parent_id
 *             new_dirname = entry[0][0]
 *             if entry[0][1] and new_dirname == self.last_target_parent[0]:             # <<<<<<<<<<<<<<
 *                 # use a cached hit for non-root target entries.
 *                 target_parent_id = self.last_target_parent[1]
 */
      goto __pyx_L46;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1266
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
 */
    /*else*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_13);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1267
 *             else:
 *                 try:
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]             # <<<<<<<<<<<<<<
 *                 except KeyError, _:
 *                     # TODO: We don't always need to do the lookup, because the
 */
          __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_self->new_dirname_to_file_id, __pyx_v_new_dirname); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L49_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_v_target_parent_id = __pyx_t_3;
          __pyx_t_3 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1266
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
 */
        }
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L54_try_end;
        __pyx_L49_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1268
 *                 try:
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:             # <<<<<<<<<<<<<<
 *                     # TODO: We don't always need to do the lookup, because the
 *                     #       parent entry will be the same as the source entry.
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_7) < 0) __PYX_ERR(0, 1268, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1271
 *                     # TODO: We don't always need to do the lookup, because the
 *                     #       parent entry will be the same as the source entry.
 *                     target_parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_get_entry); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1271, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1271, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1271, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_16);
          PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1272
 *                     #       parent entry will be the same as the source entry.
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)             # <<<<<<<<<<<<<<
 *                     if target_parent_entry == (None, None):
 *                         raise AssertionError(
 */
          __pyx_t_16 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1272, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_path_utf8, __pyx_v_new_dirname) < 0) __PYX_ERR(0, 1272, __pyx_L51_except_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1271
 *                     # TODO: We don't always need to do the lookup, because the
 *                     #       parent entry will be the same as the source entry.
 *                     target_parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):
 */
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_8, __pyx_t_16); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1271, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_v_target_parent_entry = __pyx_t_12;
          __pyx_t_12 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1273
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):             # <<<<<<<<<<<<<<
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 */
          __pyx_t_12 = PyObject_RichCompare(__pyx_v_target_parent_entry, __pyx_tuple__6, Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1273, __pyx_L51_except_error)
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1273, __pyx_L51_except_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(__pyx_t_2)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1276
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))             # <<<<<<<<<<<<<<
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:
 */
            __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1276, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_INCREF(__pyx_v_new_dirname);
            __Pyx_GIVEREF(__pyx_v_new_dirname);
            PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_new_dirname);
            __Pyx_INCREF(__pyx_v_entry);
            __Pyx_GIVEREF(__pyx_v_entry);
            PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_entry);
            __pyx_t_16 = __Pyx_PyString_Format(__pyx_kp_s_Could_not_find_target_parent_in, __pyx_t_12); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1276, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1274
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))
 */
            __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_16); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1274, __pyx_L51_except_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_Raise(__pyx_t_12, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __PYX_ERR(0, 1274, __pyx_L51_except_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1273
 *                     target_parent_entry = self.state._get_entry(self.target_index,
 *                                                            path_utf8=new_dirname)
 *                     if target_parent_entry == (None, None):             # <<<<<<<<<<<<<<
 *                         raise AssertionError(
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1277
 *                             "Could not find target parent in wt: %s\nparent of: %s"
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]             # <<<<<<<<<<<<<<
 *                 if target_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 */
          __pyx_t_12 = __Pyx_GetItemInt(__pyx_v_target_parent_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1277, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_12, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1277, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF_SET(__pyx_v_target_parent_id, __pyx_t_16);
          __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          goto __pyx_L50_exception_handled;
        }
        goto __pyx_L51_except_error;
        __pyx_L51_except_error:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1266
 *                 target_parent_id = self.last_target_parent[1]
 *             else:
 *                 try:             # <<<<<<<<<<<<<<
 *                     target_parent_id = self.new_dirname_to_file_id[new_dirname]
 *                 except KeyError, _:
 */
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
        goto __pyx_L1_error;
        __pyx_L50_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_14, __pyx_t_13);
        __pyx_L54_try_end:;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1278
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_7, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = PyObject_RichCompare(__pyx_v_target_parent_id, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1278, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1278, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1280
 *                 if target_parent_id == entry[0][2]:
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.last_target_parent[0] = new_dirname
 */
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_target_parent_id, Py_None);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1278
 *                             % (new_dirname, entry))
 *                     target_parent_id = target_parent_entry[0][2]
 *                 if target_parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                     # This is the root, so the parent is None
 *                     target_parent_id = None
 */
        goto __pyx_L58;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1282
 *                     target_parent_id = None
 *                 else:
 *                     self.last_target_parent[0] = new_dirname             # <<<<<<<<<<<<<<
 *                     self.last_target_parent[1] = target_parent_id
 * 
 */
      /*else*/ {
        if (unlikely(__Pyx_SetItemInt(__pyx_v_self->last_target_parent, 0, __pyx_v_new_dirname, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 1282, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1283
 *                 else:
 *                     self.last_target_parent[0] = new_dirname
 *                     self.last_target_parent[1] = target_parent_id             # <<<<<<<<<<<<<<
 * 
 *             source_exec = source_details[3]
 */
        if (unlikely(__Pyx_SetItemInt(__pyx_v_self->last_target_parent, 1, __pyx_v_target_parent_id, long, 1, __Pyx_PyInt_From_long, 0, 0, 1) < 0)) __PYX_ERR(0, 1283, __pyx_L1_error)
      }
      __pyx_L58:;
    }
    __pyx_L46:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1285
 *                     self.last_target_parent[1] = target_parent_id
 * 
 *             source_exec = source_details[3]             # <<<<<<<<<<<<<<
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id
 */
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_source_details, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_source_exec = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1286
 * 
 *             source_exec = source_details[3]
 *             changed = (content_change             # <<<<<<<<<<<<<<
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]
 */
    if (!__pyx_v_content_change) {
    } else {
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_content_change); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1287
 *             source_exec = source_details[3]
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id             # <<<<<<<<<<<<<<
 *                 or old_basename != entry[0][1]
 *                 or source_exec != target_exec
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_source_parent_id, __pyx_v_target_parent_id, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1287, __pyx_L1_error)
    if (!__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = __pyx_t_1;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1288
 *             changed = (content_change
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]             # <<<<<<<<<<<<<<
 *                 or source_exec != target_exec
 *                 )
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_old_basename, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1288, __pyx_L1_error)
    if (!__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_7 = __pyx_t_1;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L59_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1289
 *                 or source_parent_id != target_parent_id
 *                 or old_basename != entry[0][1]
 *                 or source_exec != target_exec             # <<<<<<<<<<<<<<
 *                 )
 *             if not changed and not self.include_unchanged:
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_source_exec, __pyx_v_target_exec, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_L59_bool_binop_done:;
    __pyx_v_changed = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1291
 *                 or source_exec != target_exec
 *                 )
 *             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                 return None, False
 *             else:
 */
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1291, __pyx_L1_error)
    __pyx_t_6 = ((!__pyx_t_5) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L64_bool_binop_done;
    }
    __pyx_t_6 = ((!(__pyx_v_self->include_unchanged != 0)) != 0);
    __pyx_t_2 = __pyx_t_6;
    __pyx_L64_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1292
 *                 )
 *             if not changed and not self.include_unchanged:
 *                 return None, False             # <<<<<<<<<<<<<<
 *             else:
 *                 if old_path is None:
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_tuple__7);
      __pyx_r = __pyx_tuple__7;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1291
 *                 or source_exec != target_exec
 *                 )
 *             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                 return None, False
 *             else:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1294
 *                 return None, False
 *             else:
 *                 if old_path is None:             # <<<<<<<<<<<<<<
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_old_path == Py_None);
      __pyx_t_6 = (__pyx_t_2 != 0);
      if (__pyx_t_6) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1295
 *             else:
 *                 if old_path is None:
 *                     path = self.pathjoin(old_dirname, old_basename)             # <<<<<<<<<<<<<<
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path)[0]
 */
        __Pyx_INCREF(__pyx_v_self->pathjoin);
        __pyx_t_1 = __pyx_v_self->pathjoin; __pyx_t_3 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1295, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
          PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_old_dirname, __pyx_v_old_basename};
          __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1295, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_GOTREF(__pyx_t_7);
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1295, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_3) {
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_3); __pyx_t_3 = NULL;
          }
          __Pyx_INCREF(__pyx_v_old_dirname);
          __Pyx_GIVEREF(__pyx_v_old_dirname);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_9, __pyx_v_old_dirname);
          __Pyx_INCREF(__pyx_v_old_basename);
          __Pyx_GIVEREF(__pyx_v_old_basename);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_9, __pyx_v_old_basename);
          __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_16, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1295, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_path, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1296
 *                 if old_path is None:
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path             # <<<<<<<<<<<<<<
 *                     old_path_u = self.utf8_decode(old_path)[0]
 *                     path_u = old_path_u
 */
        __Pyx_INCREF(__pyx_v_path);
        __Pyx_DECREF_SET(__pyx_v_old_path, __pyx_v_path);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1297
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path)[0]             # <<<<<<<<<<<<<<
 *                     path_u = old_path_u
 *                 else:
 */
        __Pyx_INCREF(__pyx_v_self->utf8_decode);
        __pyx_t_1 = __pyx_v_self->utf8_decode; __pyx_t_16 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_7 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_16, __pyx_v_old_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_old_path);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1297, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_old_path_u = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1298
 *                     old_path = path
 *                     old_path_u = self.utf8_decode(old_path)[0]
 *                     path_u = old_path_u             # <<<<<<<<<<<<<<
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path)[0]
 */
        __Pyx_INCREF(__pyx_v_old_path_u);
        __pyx_v_path_u = __pyx_v_old_path_u;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1294
 *                 return None, False
 *             else:
 *                 if old_path is None:             # <<<<<<<<<<<<<<
 *                     path = self.pathjoin(old_dirname, old_basename)
 *                     old_path = path
 */
        goto __pyx_L66;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1300
 *                     path_u = old_path_u
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path)[0]             # <<<<<<<<<<<<<<
 *                     if old_path == path:
 *                         path_u = old_path_u
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_self->utf8_decode);
        __pyx_t_7 = __pyx_v_self->utf8_decode; __pyx_t_16 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_1 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_16, __pyx_v_old_path) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_old_path);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_old_path_u = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1301
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path)[0]
 *                     if old_path == path:             # <<<<<<<<<<<<<<
 *                         path_u = old_path_u
 *                     else:
 */
        __pyx_t_7 = PyObject_RichCompare(__pyx_v_old_path, __pyx_v_path, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1301, __pyx_L1_error)
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1301, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_6) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1302
 *                     old_path_u = self.utf8_decode(old_path)[0]
 *                     if old_path == path:
 *                         path_u = old_path_u             # <<<<<<<<<<<<<<
 *                     else:
 *                         path_u = self.utf8_decode(path)[0]
 */
          __Pyx_INCREF(__pyx_v_old_path_u);
          __pyx_v_path_u = __pyx_v_old_path_u;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1301
 *                 else:
 *                     old_path_u = self.utf8_decode(old_path)[0]
 *                     if old_path == path:             # <<<<<<<<<<<<<<
 *                         path_u = old_path_u
 *                     else:
 */
          goto __pyx_L67;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1304
 *                         path_u = old_path_u
 *                     else:
 *                         path_u = self.utf8_decode(path)[0]             # <<<<<<<<<<<<<<
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return TreeChange(entry[0][2],
 */
        /*else*/ {
          __Pyx_INCREF(__pyx_v_self->utf8_decode);
          __pyx_t_1 = __pyx_v_self->utf8_decode; __pyx_t_16 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_7 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_16, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_v_path_u = __pyx_t_1;
          __pyx_t_1 = 0;
        }
        __pyx_L67:;
      }
      __pyx_L66:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1305
 *                     else:
 *                         path_u = self.utf8_decode(path)[0]
 *                 source_kind = _minikind_to_kind(source_minikind)             # <<<<<<<<<<<<<<
 *                 return TreeChange(entry[0][2],
 *                        (old_path_u, path_u),
 */
      __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(__pyx_v_source_minikind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_source_kind = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1306
 *                         path_u = self.utf8_decode(path)[0]
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (old_path_u, path_u),
 *                        content_change,
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_16, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1307
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return TreeChange(entry[0][2],
 *                        (old_path_u, path_u),             # <<<<<<<<<<<<<<
 *                        content_change,
 *                        (True, True),
 */
      __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_INCREF(__pyx_v_old_path_u);
      __Pyx_GIVEREF(__pyx_v_old_path_u);
      PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_v_old_path_u);
      __Pyx_INCREF(__pyx_v_path_u);
      __Pyx_GIVEREF(__pyx_v_path_u);
      PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_v_path_u);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1308
 *                 return TreeChange(entry[0][2],
 *                        (old_path_u, path_u),
 *                        content_change,             # <<<<<<<<<<<<<<
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),
 */
      __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_content_change); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1310
 *                        content_change,
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),             # <<<<<<<<<<<<<<
 *                        (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
 *                        (source_kind, target_kind),
 */
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_v_source_parent_id);
      __Pyx_GIVEREF(__pyx_v_source_parent_id);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_source_parent_id);
      __Pyx_INCREF(__pyx_v_target_parent_id);
      __Pyx_GIVEREF(__pyx_v_target_parent_id);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_target_parent_id);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1311
 *                        (True, True),
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),             # <<<<<<<<<<<<<<
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
 */
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_17 = __pyx_v_self->utf8_decode; __pyx_t_18 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_17, function);
        }
      }
      __pyx_t_10 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_18, __pyx_v_old_basename) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_v_old_basename);
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_18 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_19 = __Pyx_GetItemInt(__pyx_t_18, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_19);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_18 = __pyx_v_self->utf8_decode; __pyx_t_20 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
        __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_18);
        if (likely(__pyx_t_20)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_20);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_18, function);
        }
      }
      __pyx_t_10 = (__pyx_t_20) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_20, __pyx_t_19) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_19);
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_18);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_18);
      __pyx_t_17 = 0;
      __pyx_t_18 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1312
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
 *                        (source_kind, target_kind),             # <<<<<<<<<<<<<<
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
 */
      __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_INCREF(__pyx_v_source_kind);
      __Pyx_GIVEREF(__pyx_v_source_kind);
      PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_v_source_kind);
      __Pyx_INCREF(__pyx_v_target_kind);
      __Pyx_GIVEREF(__pyx_v_target_kind);
      PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_v_target_kind);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1313
 *                        (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed             # <<<<<<<<<<<<<<
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
 *             # looks like a new file
 */
      __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1313, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_INCREF(__pyx_v_source_exec);
      __Pyx_GIVEREF(__pyx_v_source_exec);
      PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v_source_exec);
      __Pyx_INCREF(__pyx_v_target_exec);
      __Pyx_GIVEREF(__pyx_v_target_exec);
      PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_v_target_exec);
      __pyx_t_19 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_19)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_19);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[9] = {__pyx_t_19, __pyx_t_3, __pyx_t_16, __pyx_t_12, __pyx_tuple__8, __pyx_t_8, __pyx_t_10, __pyx_t_18, __pyx_t_17};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[9] = {__pyx_t_19, __pyx_t_3, __pyx_t_16, __pyx_t_12, __pyx_tuple__8, __pyx_t_8, __pyx_t_10, __pyx_t_18, __pyx_t_17};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      } else
      #endif
      {
        __pyx_t_20 = PyTuple_New(8+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        if (__pyx_t_19) {
          __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_19); __pyx_t_19 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_9, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_16);
        PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_9, __pyx_t_16);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_9, __pyx_t_12);
        __Pyx_INCREF(__pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_tuple__8);
        PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_9, __pyx_tuple__8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_20, 4+__pyx_t_9, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_20, 5+__pyx_t_9, __pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_18);
        PyTuple_SET_ITEM(__pyx_t_20, 6+__pyx_t_9, __pyx_t_18);
        __Pyx_GIVEREF(__pyx_t_17);
        PyTuple_SET_ITEM(__pyx_t_20, 7+__pyx_t_9, __pyx_t_17);
        __pyx_t_3 = 0;
        __pyx_t_16 = 0;
        __pyx_t_12 = 0;
        __pyx_t_8 = 0;
        __pyx_t_10 = 0;
        __pyx_t_18 = 0;
        __pyx_t_17 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1306
 *                         path_u = self.utf8_decode(path)[0]
 *                 source_kind = _minikind_to_kind(source_minikind)
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (old_path_u, path_u),
 *                        content_change,
 */
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_changed);
      __Pyx_GIVEREF(__pyx_v_changed);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_changed);
      __pyx_t_1 = 0;
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      goto __pyx_L0;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1145
 *         # 0.000006 seconds per call.
 *         source_minikind = _minikind_from_string(source_details[0])
 *         if ((_versioned_minikind(source_minikind) or source_minikind == b'r')             # <<<<<<<<<<<<<<
 *             and _versioned_minikind(target_minikind)):
 *             # claimed content in both: diff
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1314
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])
 */
  __pyx_t_2 = ((__pyx_v_source_minikind == 'a') != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_6 = __pyx_t_2;
    goto __pyx_L68_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_target_minikind) != 0);
  __pyx_t_6 = __pyx_t_2;
  __pyx_L68_bool_binop_done:;
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1316
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             # TODO: these are the same for an entire directory: cache em.
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v_self->pathjoin);
    __pyx_t_1 = __pyx_v_self->pathjoin; __pyx_t_18 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_18)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_18);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_t_20, __pyx_t_17};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_t_20, __pyx_t_17};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_18) {
        __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_18); __pyx_t_18 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_20);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_20);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_17);
      __pyx_t_20 = 0;
      __pyx_t_17 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_path = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1319
 *             # parent id is the entry for the path in the target tree
 *             # TODO: these are the same for an entire directory: cache em.
 *             parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_get_entry); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1320
 *             # TODO: these are the same for an entire directory: cache em.
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])             # <<<<<<<<<<<<<<
 *             if parent_entry is None:
 *                 raise DirstateCorrupt(self.state,
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_path_utf8, __pyx_t_20) < 0) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1319
 *             # parent id is the entry for the path in the target tree
 *             # TODO: these are the same for an entire directory: cache em.
 *             parent_entry = self.state._get_entry(self.target_index,             # <<<<<<<<<<<<<<
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:
 */
    __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, __pyx_t_1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_parent_entry = __pyx_t_20;
    __pyx_t_20 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1321
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:             # <<<<<<<<<<<<<<
 *                 raise DirstateCorrupt(self.state,
 *                     "We could not find the parent entry in index %d"
 */
    __pyx_t_6 = (__pyx_v_parent_entry == Py_None);
    __pyx_t_2 = (__pyx_t_6 != 0);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1322
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:
 *                 raise DirstateCorrupt(self.state,             # <<<<<<<<<<<<<<
 *                     "We could not find the parent entry in index %d"
 *                     " for the entry: %s"
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1325
 *                     "We could not find the parent entry in index %d"
 *                     " for the entry: %s"
 *                     % (self.target_index, entry[0]))             # <<<<<<<<<<<<<<
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:
 */
      __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_We_could_not_find_the_parent_ent, __pyx_t_17); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_17)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_v_self->state, __pyx_t_7};
        __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_17, __pyx_v_self->state, __pyx_t_7};
        __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (__pyx_t_17) {
          __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_17); __pyx_t_17 = NULL;
        }
        __Pyx_INCREF(__pyx_v_self->state);
        __Pyx_GIVEREF(__pyx_v_self->state);
        PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_self->state);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1322, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_20, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __PYX_ERR(0, 1322, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1321
 *             parent_entry = self.state._get_entry(self.target_index,
 *                                                  path_utf8=entry[0][0])
 *             if parent_entry is None:             # <<<<<<<<<<<<<<
 *                 raise DirstateCorrupt(self.state,
 *                     "We could not find the parent entry in index %d"
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1326
 *                     " for the entry: %s"
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]             # <<<<<<<<<<<<<<
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 */
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_v_parent_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_20, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_v_parent_id = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1327
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             if path_info is not None:
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_parent_id, __pyx_t_20, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1327, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1328
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:
 *                 parent_id = None             # <<<<<<<<<<<<<<
 *             if path_info is not None:
 *                 # Present on disk:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_parent_id, Py_None);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1327
 *                     % (self.target_index, entry[0]))
 *             parent_id = parent_entry[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             if path_info is not None:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1329
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             if path_info is not None:             # <<<<<<<<<<<<<<
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:
 */
    __pyx_t_2 = (__pyx_v_path_info != Py_None);
    __pyx_t_6 = (__pyx_t_2 != 0);
    if (__pyx_t_6) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1331
 *             if path_info is not None:
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_self->use_filesystem_for_exec); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1331, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1335
 *                     # is a file or not.
 *                     target_exec = bool(
 *                         S_ISREG(path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                         and S_IXUSR & path_info[3].st_mode)
 *                 else:
 */
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_20); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1335, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __pyx_t_2 = (S_ISREG(__pyx_t_9) != 0);
        if (__pyx_t_2) {
        } else {
          __pyx_t_6 = __pyx_t_2;
          goto __pyx_L74_bool_binop_done;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1336
 *                     target_exec = bool(
 *                         S_ISREG(path_info[3].st_mode)
 *                         and S_IXUSR & path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                 else:
 *                     target_exec = target_details[3]
 */
        __pyx_t_20 = __Pyx_PyInt_From_int(S_IXUSR); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = PyNumber_And(__pyx_t_20, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1336, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1336, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_6 = __pyx_t_2;
        __pyx_L74_bool_binop_done:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1334
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
 *                     target_exec = bool(             # <<<<<<<<<<<<<<
 *                         S_ISREG(path_info[3].st_mode)
 *                         and S_IXUSR & path_info[3].st_mode)
 */
        __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_6))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_target_exec = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1331
 *             if path_info is not None:
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:             # <<<<<<<<<<<<<<
 *                     # We need S_ISREG here, because we aren't sure if this
 *                     # is a file or not.
 */
        goto __pyx_L73;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1338
 *                         and S_IXUSR & path_info[3].st_mode)
 *                 else:
 *                     target_exec = target_details[3]             # <<<<<<<<<<<<<<
 *                 return TreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path)[0]),
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_target_details, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_target_exec = __pyx_t_1;
        __pyx_t_1 = 0;
      }
      __pyx_L73:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1339
 *                 else:
 *                     target_exec = target_details[3]
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path)[0]),
 *                        True,
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_20 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_20, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1340
 *                     target_exec = target_details[3]
 *                 return TreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path)[0]),             # <<<<<<<<<<<<<<
 *                        True,
 *                        (False, True),
 */
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_17 = __pyx_v_self->utf8_decode; __pyx_t_18 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_17, function);
        }
      }
      __pyx_t_20 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_18, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_v_path);
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_20, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_20, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_17);
      __pyx_t_17 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1343
 *                        True,
 *                        (False, True),
 *                        (None, parent_id),             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 *                        (None, path_info[2]),
 */
      __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_17, 0, Py_None);
      __Pyx_INCREF(__pyx_v_parent_id);
      __Pyx_GIVEREF(__pyx_v_parent_id);
      PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_v_parent_id);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1344
 *                        (False, True),
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1])[0]),             # <<<<<<<<<<<<<<
 *                        (None, path_info[2]),
 *                        (None, target_exec)), True
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_8 = __pyx_v_self->utf8_decode; __pyx_t_16 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_18 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_16, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_12);
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_18, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_18, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1345
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 *                        (None, path_info[2]),             # <<<<<<<<<<<<<<
 *                        (None, target_exec)), True
 *             else:
 */
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_12, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1346
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 *                        (None, path_info[2]),
 *                        (None, target_exec)), True             # <<<<<<<<<<<<<<
 *             else:
 *                 # Its a missing file, report it as such.
 */
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1346, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_8, 0, Py_None);
      __Pyx_INCREF(__pyx_v_target_exec);
      __Pyx_GIVEREF(__pyx_v_target_exec);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_target_exec);
      __pyx_t_16 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[9] = {__pyx_t_16, __pyx_t_7, __pyx_t_20, Py_True, __pyx_tuple__9, __pyx_t_17, __pyx_t_18, __pyx_t_12, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[9] = {__pyx_t_16, __pyx_t_7, __pyx_t_20, Py_True, __pyx_tuple__9, __pyx_t_17, __pyx_t_18, __pyx_t_12, __pyx_t_8};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(8+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (__pyx_t_16) {
          __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_16); __pyx_t_16 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_9, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_20);
        PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_9, __pyx_t_20);
        __Pyx_INCREF(Py_True);
        __Pyx_GIVEREF(Py_True);
        PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_9, Py_True);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyTuple_SET_ITEM(__pyx_t_3, 3+__pyx_t_9, __pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_t_17);
        PyTuple_SET_ITEM(__pyx_t_3, 4+__pyx_t_9, __pyx_t_17);
        __Pyx_GIVEREF(__pyx_t_18);
        PyTuple_SET_ITEM(__pyx_t_3, 5+__pyx_t_9, __pyx_t_18);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_3, 6+__pyx_t_9, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 7+__pyx_t_9, __pyx_t_8);
        __pyx_t_7 = 0;
        __pyx_t_20 = 0;
        __pyx_t_17 = 0;
        __pyx_t_18 = 0;
        __pyx_t_12 = 0;
        __pyx_t_8 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1339, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1339
 *                 else:
 *                     target_exec = target_details[3]
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path)[0]),
 *                        True,
 */
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1339, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      PyTuple_SET_ITEM(__pyx_t_10, 1, Py_True);
      __pyx_t_1 = 0;
      __pyx_r = __pyx_t_10;
      __pyx_t_10 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1329
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             if path_info is not None:             # <<<<<<<<<<<<<<
 *                 # Present on disk:
 *                 if self.use_filesystem_for_exec:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1349
 *             else:
 *                 # Its a missing file, report it as such.
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path)[0]),
 *                        False,
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_3, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1350
 *                 # Its a missing file, report it as such.
 *                 return TreeChange(entry[0][2],
 *                        (None, self.utf8_decode(path)[0]),             # <<<<<<<<<<<<<<
 *                        False,
 *                        (False, True),
 */
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_12 = __pyx_v_self->utf8_decode; __pyx_t_18 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_18)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_3 = (__pyx_t_18) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_18, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_path);
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1350, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_12);
      __pyx_t_12 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1353
 *                        False,
 *                        (False, True),
 *                        (None, parent_id),             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 *                        (None, None),
 */
      __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_12, 0, Py_None);
      __Pyx_INCREF(__pyx_v_parent_id);
      __Pyx_GIVEREF(__pyx_v_parent_id);
      PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_parent_id);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1354
 *                        (False, True),
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1])[0]),             # <<<<<<<<<<<<<<
 *                        (None, None),
 *                        (None, False)), True
 */
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_17, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_INCREF(__pyx_v_self->utf8_decode);
      __pyx_t_17 = __pyx_v_self->utf8_decode; __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_17);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_17, function);
        }
      }
      __pyx_t_18 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_7, __pyx_t_20) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_t_20);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_18, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_18, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_17);
      __pyx_t_17 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1356
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 *                        (None, None),
 *                        (None, False)), True             # <<<<<<<<<<<<<<
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':
 *             # unversioned, possibly, or possibly not deleted: we dont care.
 */
      __pyx_t_17 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_17)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[9] = {__pyx_t_17, __pyx_t_8, __pyx_t_3, Py_False, __pyx_tuple__9, __pyx_t_12, __pyx_t_18, __pyx_tuple__6, __pyx_tuple__7};
        __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1349, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[9] = {__pyx_t_17, __pyx_t_8, __pyx_t_3, Py_False, __pyx_tuple__9, __pyx_t_12, __pyx_t_18, __pyx_tuple__6, __pyx_tuple__7};
        __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1349, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      } else
      #endif
      {
        __pyx_t_20 = PyTuple_New(8+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        if (__pyx_t_17) {
          __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_17); __pyx_t_17 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_9, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_9, __pyx_t_3);
        __Pyx_INCREF(Py_False);
        __Pyx_GIVEREF(Py_False);
        PyTuple_SET_ITEM(__pyx_t_20, 2+__pyx_t_9, Py_False);
        __Pyx_INCREF(__pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_tuple__9);
        PyTuple_SET_ITEM(__pyx_t_20, 3+__pyx_t_9, __pyx_tuple__9);
        __Pyx_GIVEREF(__pyx_t_12);
        PyTuple_SET_ITEM(__pyx_t_20, 4+__pyx_t_9, __pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_18);
        PyTuple_SET_ITEM(__pyx_t_20, 5+__pyx_t_9, __pyx_t_18);
        __Pyx_INCREF(__pyx_tuple__6);
        __Pyx_GIVEREF(__pyx_tuple__6);
        PyTuple_SET_ITEM(__pyx_t_20, 6+__pyx_t_9, __pyx_tuple__6);
        __Pyx_INCREF(__pyx_tuple__7);
        __Pyx_GIVEREF(__pyx_tuple__7);
        PyTuple_SET_ITEM(__pyx_t_20, 7+__pyx_t_9, __pyx_tuple__7);
        __pyx_t_8 = 0;
        __pyx_t_3 = 0;
        __pyx_t_12 = 0;
        __pyx_t_18 = 0;
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_20, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1349
 *             else:
 *                 # Its a missing file, report it as such.
 *                 return TreeChange(entry[0][2],             # <<<<<<<<<<<<<<
 *                        (None, self.utf8_decode(path)[0]),
 *                        False,
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_10);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      PyTuple_SET_ITEM(__pyx_t_1, 1, Py_True);
      __pyx_t_10 = 0;
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1314
 *                        (source_kind, target_kind),
 *                        (source_exec, target_exec)), changed
 *         elif source_minikind == b'a' and _versioned_minikind(target_minikind):             # <<<<<<<<<<<<<<
 *             # looks like a new file
 *             path = self.pathjoin(entry[0][0], entry[0][1])
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1357
 *                        (None, None),
 *                        (None, False)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':             # <<<<<<<<<<<<<<
 *             # unversioned, possibly, or possibly not deleted: we dont care.
 *             # if its still on disk, *and* theres no other entry at this
 */
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_6 = __pyx_t_2;
    goto __pyx_L76_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_target_minikind == 'a') != 0);
  __pyx_t_6 = __pyx_t_2;
  __pyx_L76_bool_binop_done:;
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1362
 *             # path [we dont know this in this routine at the moment -
 *             # perhaps we should change this - then it would be an unknown.
 *             old_path = self.pathjoin(entry[0][0], entry[0][1])             # <<<<<<<<<<<<<<
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 */
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_10, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1362, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_INCREF(__pyx_v_self->pathjoin);
    __pyx_t_10 = __pyx_v_self->pathjoin; __pyx_t_12 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_20, __pyx_t_18};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1362, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_t_20, __pyx_t_18};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1362, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_12) {
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_12); __pyx_t_12 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_20);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_9, __pyx_t_20);
      __Pyx_GIVEREF(__pyx_t_18);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_9, __pyx_t_18);
      __pyx_t_20 = 0;
      __pyx_t_18 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1362, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_old_path = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1364
 *             old_path = self.pathjoin(entry[0][0], entry[0][1])
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]             # <<<<<<<<<<<<<<
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_get_entry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_18 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_18, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_path_utf8, __pyx_t_20) < 0) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_20, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_t_10, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_parent_id = __pyx_t_20;
    __pyx_t_20 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1365
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             return TreeChange(
 */
    __pyx_t_20 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_20, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __pyx_t_20 = PyObject_RichCompare(__pyx_v_parent_id, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_20); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1365, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_20); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1365, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    if (__pyx_t_6) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1366
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:
 *                 parent_id = None             # <<<<<<<<<<<<<<
 *             return TreeChange(
 *                    entry[0][2],
 */
      __Pyx_INCREF(Py_None);
      __Pyx_DECREF_SET(__pyx_v_parent_id, Py_None);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1365
 *             # parent id is the entry for the path in the target tree
 *             parent_id = self.state._get_entry(self.source_index, path_utf8=entry[0][0])[0][2]
 *             if parent_id == entry[0][2]:             # <<<<<<<<<<<<<<
 *                 parent_id = None
 *             return TreeChange(
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1367
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             return TreeChange(             # <<<<<<<<<<<<<<
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path)[0], None),
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1368
 *                 parent_id = None
 *             return TreeChange(
 *                    entry[0][2],             # <<<<<<<<<<<<<<
 *                    (self.utf8_decode(old_path)[0], None),
 *                    True,
 */
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1369
 *             return TreeChange(
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path)[0], None),             # <<<<<<<<<<<<<<
 *                    True,
 *                    (True, False),
 */
    __Pyx_INCREF(__pyx_v_self->utf8_decode);
    __pyx_t_18 = __pyx_v_self->utf8_decode; __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_18);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_18, function);
      }
    }
    __pyx_t_3 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_12, __pyx_v_old_path) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_v_old_path);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_18);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None);
    __pyx_t_18 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1372
 *                    True,
 *                    (True, False),
 *                    (parent_id, None),             # <<<<<<<<<<<<<<
 *                    (self.utf8_decode(entry[0][1])[0], None),
 *                    (_minikind_to_kind(source_minikind), None),
 */
    __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_v_parent_id);
    __Pyx_GIVEREF(__pyx_v_parent_id);
    PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_v_parent_id);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_18, 1, Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1373
 *                    (True, False),
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1])[0], None),             # <<<<<<<<<<<<<<
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_8, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_INCREF(__pyx_v_self->utf8_decode);
    __pyx_t_8 = __pyx_v_self->utf8_decode; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_12 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_17);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_12, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_8);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_12, 1, Py_None);
    __pyx_t_8 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1374
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1])[0], None),
 *                    (_minikind_to_kind(source_minikind), None),             # <<<<<<<<<<<<<<
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':
 */
    __pyx_t_8 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_to_kind(__pyx_v_source_minikind); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_8);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_17, 1, Py_None);
    __pyx_t_8 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1375
 *                    (self.utf8_decode(entry[0][1])[0], None),
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True             # <<<<<<<<<<<<<<
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':
 *             # a rename; could be a true rename, or a rename inherited from
 */
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_source_details, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_7, 1, Py_None);
    __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[9] = {__pyx_t_8, __pyx_t_1, __pyx_t_3, Py_True, __pyx_tuple__10, __pyx_t_18, __pyx_t_12, __pyx_t_17, __pyx_t_7};
      __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1367, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[9] = {__pyx_t_8, __pyx_t_1, __pyx_t_3, Py_True, __pyx_tuple__10, __pyx_t_18, __pyx_t_12, __pyx_t_17, __pyx_t_7};
      __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_9, 8+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1367, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(8+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_9, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_9, __pyx_t_3);
      __Pyx_INCREF(Py_True);
      __Pyx_GIVEREF(Py_True);
      PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_9, Py_True);
      __Pyx_INCREF(__pyx_tuple__10);
      __Pyx_GIVEREF(__pyx_tuple__10);
      PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_9, __pyx_tuple__10);
      __Pyx_GIVEREF(__pyx_t_18);
      PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_9, __pyx_t_18);
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_9, __pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_9, __pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_9, __pyx_t_7);
      __pyx_t_1 = 0;
      __pyx_t_3 = 0;
      __pyx_t_18 = 0;
      __pyx_t_12 = 0;
      __pyx_t_17 = 0;
      __pyx_t_7 = 0;
      __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_16, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1367, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1367
 *             if parent_id == entry[0][2]:
 *                 parent_id = None
 *             return TreeChange(             # <<<<<<<<<<<<<<
 *                    entry[0][2],
 *                    (self.utf8_decode(old_path)[0], None),
 */
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_20);
    __Pyx_INCREF(Py_True);
    __Pyx_GIVEREF(Py_True);
    PyTuple_SET_ITEM(__pyx_t_10, 1, Py_True);
    __pyx_t_20 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1357
 *                        (None, None),
 *                        (None, False)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'a':             # <<<<<<<<<<<<<<
 *             # unversioned, possibly, or possibly not deleted: we dont care.
 *             # if its still on disk, *and* theres no other entry at this
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1376
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':             # <<<<<<<<<<<<<<
 *             # a rename; could be a true rename, or a rename inherited from
 *             # a renamed parent. TODO: handle this efficiently. Its not
 */
  __pyx_t_2 = (__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__versioned_minikind(__pyx_v_source_minikind) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_6 = __pyx_t_2;
    goto __pyx_L79_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_target_minikind == 'r') != 0);
  __pyx_t_6 = __pyx_t_2;
  __pyx_L79_bool_binop_done:;
  if (__pyx_t_6) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1381
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
 */
    __pyx_t_2 = ((!(__pyx_v_self->doing_consistency_expansion != 0)) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_6 = __pyx_t_2;
      goto __pyx_L82_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1382
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_20, __pyx_n_s_is_inside_any); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1383
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):             # <<<<<<<<<<<<<<
 *                 self.search_specific_files.add(target_details[1])
 *                 # We don't expand the specific files parents list here as
 */
    __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_target_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_20);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_20, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_20)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->searched_specific_files, __pyx_t_16};
      __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_20)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->searched_specific_files, __pyx_t_16};
      __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    } else
    #endif
    {
      __pyx_t_17 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self->searched_specific_files);
      __Pyx_GIVEREF(__pyx_v_self->searched_specific_files);
      PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_9, __pyx_v_self->searched_specific_files);
      __Pyx_GIVEREF(__pyx_t_16);
      PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_9, __pyx_t_16);
      __pyx_t_16 = 0;
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_17, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    }
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1382
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and
 *                 not is_inside_any(self.searched_specific_files,             # <<<<<<<<<<<<<<
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_5 = ((!__pyx_t_2) != 0);
    __pyx_t_6 = __pyx_t_5;
    __pyx_L82_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1381
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
 */
    if (__pyx_t_6) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1384
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
 *                 self.search_specific_files.add(target_details[1])             # <<<<<<<<<<<<<<
 *                 # We don't expand the specific files parents list here as
 *                 # the path is absent in target and won't create a delta with
 */
      __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_files, __pyx_n_s_add); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_target_details, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_16 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_20))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_20);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_20, function);
        }
      }
      __pyx_t_10 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_20, __pyx_t_16, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_t_17);
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1381
 *             # common case to rename dirs though, so a correct but slow
 *             # implementation will do.
 *             if (not self.doing_consistency_expansion and             # <<<<<<<<<<<<<<
 *                 not is_inside_any(self.searched_specific_files,
 *                     target_details[1])):
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1376
 *                    (_minikind_to_kind(source_minikind), None),
 *                    (source_details[3], None)), True
 *         elif _versioned_minikind(source_minikind) and target_minikind == b'r':             # <<<<<<<<<<<<<<
 *             # a rename; could be a true rename, or a rename inherited from
 *             # a renamed parent. TODO: handle this efficiently. Its not
 */
    goto __pyx_L8;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1388
 *                 # the path is absent in target and won't create a delta with
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and             # <<<<<<<<<<<<<<
 *               (target_minikind == b'r' or target_minikind == b'a')):
 *             # neither of the selected trees contain this path,
 */
  switch (__pyx_v_source_minikind) {
    case 'r':
    case 'a':
    __pyx_t_5 = 1;
    break;
    default:
    __pyx_t_5 = 0;
    break;
  }
  if (__pyx_t_5) {
  } else {
    __pyx_t_6 = __pyx_t_5;
    goto __pyx_L84_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1389
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and
 *               (target_minikind == b'r' or target_minikind == b'a')):             # <<<<<<<<<<<<<<
 *             # neither of the selected trees contain this path,
 *             # so skip over it. This is not currently directly tested, but
 */
  switch (__pyx_v_target_minikind) {
    case 'r':
    case 'a':
    __pyx_t_5 = 1;
    break;
    default:
    __pyx_t_5 = 0;
    break;
  }
  __pyx_t_6 = __pyx_t_5;
  __pyx_L84_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1388
 *                 # the path is absent in target and won't create a delta with
 *                 # missing parent.
 *         elif ((source_minikind == b'r' or source_minikind == b'a') and             # <<<<<<<<<<<<<<
 *               (target_minikind == b'r' or target_minikind == b'a')):
 *             # neither of the selected trees contain this path,
 */
  if (likely(__pyx_t_6)) {
    goto __pyx_L8;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1395
 *             pass
 *         else:
 *             raise AssertionError("don't know how to compare "             # <<<<<<<<<<<<<<
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))
 */
  /*else*/ {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1397
 *             raise AssertionError("don't know how to compare "
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))             # <<<<<<<<<<<<<<
 *             ## import pdb;pdb.set_trace()
 *         return None, None
 */
    __pyx_t_10 = __Pyx_PyInt_From_char(__pyx_v_source_minikind); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_20 = __Pyx_PyInt_From_char(__pyx_v_target_minikind); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_20);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_20);
    __pyx_t_10 = 0;
    __pyx_t_20 = 0;
    __pyx_t_20 = __Pyx_PyString_Format(__pyx_kp_s_don_t_know_how_to_compare_source, __pyx_t_17); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1395
 *             pass
 *         else:
 *             raise AssertionError("don't know how to compare "             # <<<<<<<<<<<<<<
 *                 "source_minikind=%r, target_minikind=%r"
 *                 % (source_minikind, target_minikind))
 */
    __pyx_t_17 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_20); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_Raise(__pyx_t_17, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __PYX_ERR(0, 1395, __pyx_L1_error)
  }
  __pyx_L8:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1399
 *                 % (source_minikind, target_minikind))
 *             ## import pdb;pdb.set_trace()
 *         return None, None             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_tuple__6);
  __pyx_r = __pyx_tuple__6;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1102
 *             raise errors.BzrError('unsupported target index')
 * 
 *     cdef _process_entry(self, entry, path_info):             # <<<<<<<<<<<<<<
 *         """Compare an entry and real disk to generate delta information.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._process_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_file_id);
  __Pyx_XDECREF(__pyx_v_details_list);
  __Pyx_XDECREF(__pyx_v_source_details);
  __Pyx_XDECREF(__pyx_v_target_details);
  __Pyx_XDECREF(__pyx_v_link_or_sha1);
  __Pyx_XDECREF(__pyx_v_old_dirname);
  __Pyx_XDECREF(__pyx_v_old_basename);
  __Pyx_XDECREF(__pyx_v_old_path);
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_old_entry);
  __Pyx_XDECREF(__pyx_v_target_kind);
  __Pyx_XDECREF(__pyx_v_target_exec);
  __Pyx_XDECREF(__pyx_v_statvalue);
  __Pyx_XDECREF(__pyx_v_source_parent_id);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_source_parent_entry);
  __Pyx_XDECREF(__pyx_v_new_dirname);
  __Pyx_XDECREF(__pyx_v_target_parent_id);
  __Pyx_XDECREF(__pyx_v_target_parent_entry);
  __Pyx_XDECREF(__pyx_v_source_exec);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_old_path_u);
  __Pyx_XDECREF(__pyx_v_path_u);
  __Pyx_XDECREF(__pyx_v_source_kind);
  __Pyx_XDECREF(__pyx_v_parent_entry);
  __Pyx_XDECREF(__pyx_v_parent_id);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1401
 *         return None, None
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_2__iter__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1402
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def iter_changes(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1401
 *         return None, None
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1404
 *         return self
 * 
 *     def iter_changes(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iter_changes (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_4iter_changes(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iter_changes", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1405
 * 
 *     def iter_changes(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1404
 *         return self
 * 
 *     def iter_changes(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1407
 *         return self
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:             # <<<<<<<<<<<<<<
 *         """Check a result we will yield to make sure we are consistent later.
 * 
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_v_new_path = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("_gather_result_for_consistency", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1414
 *         :param result: A result tuple.
 *         """
 *         if not self.partial or not result.file_id:             # <<<<<<<<<<<<<<
 *             return 0
 *         self.seen_ids.add(result.file_id)
 */
  __pyx_t_2 = ((!(__pyx_v_self->partial != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_file_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = ((!__pyx_t_2) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1415
 *         """
 *         if not self.partial or not result.file_id:
 *             return 0             # <<<<<<<<<<<<<<
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1414
 *         :param result: A result tuple.
 *         """
 *         if not self.partial or not result.file_id:             # <<<<<<<<<<<<<<
 *             return 0
 *         self.seen_ids.add(result.file_id)
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1416
 *         if not self.partial or not result.file_id:
 *             return 0
 *         self.seen_ids.add(result.file_id)             # <<<<<<<<<<<<<<
 *         new_path = result.path[1]
 *         if new_path:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->seen_ids, __pyx_n_s_add); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_file_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1417
 *             return 0
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]             # <<<<<<<<<<<<<<
 *         if new_path:
 *             # Not the root and not a delete: queue up the parents of the path.
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_new_path = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1418
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
 *         if new_path:             # <<<<<<<<<<<<<<
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_new_path); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1418, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1420
 *         if new_path:
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(             # <<<<<<<<<<<<<<
 *                 osutils.parent_directories(new_path.encode('utf8')))
 *             # Add the root directory which parent_directories does not
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_file_parents, __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1421
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
 *                 osutils.parent_directories(new_path.encode('utf8')))             # <<<<<<<<<<<<<<
 *             # Add the root directory which parent_directories does not
 *             # provide.
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_osutils); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_parent_directories); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_new_path, __pyx_n_s_encode_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_7 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_n_s_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_n_s_utf8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1424
 *             # Add the root directory which parent_directories does not
 *             # provide.
 *             self.search_specific_file_parents.add(b'')             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_file_parents, __pyx_n_s_add); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_kp_b__4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_b__4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1418
 *         self.seen_ids.add(result.file_id)
 *         new_path = result.path[1]
 *         if new_path:             # <<<<<<<<<<<<<<
 *             # Not the root and not a delete: queue up the parents of the path.
 *             self.search_specific_file_parents.update(
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1425
 *             # provide.
 *             self.search_specific_file_parents.add(b'')
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _update_current_block(self) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1407
 *         return self
 * 
 *     cdef int _gather_result_for_consistency(self, result) except -1:             # <<<<<<<<<<<<<<
 *         """Check a result we will yield to make sure we are consistent later.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._gather_result_for_consistency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_new_path);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1427
 *         return 0
 * 
 *     cdef int _update_current_block(self) except -1:             # <<<<<<<<<<<<<<
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 */

static int __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("_update_current_block", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1428
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((__pyx_v_self->block_index < __pyx_t_3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1429
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):             # <<<<<<<<<<<<<<
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_is_inside); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_6, __pyx_v_self->block_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->current_root, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_self->current_root, __pyx_t_6};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->current_root);
    __Pyx_GIVEREF(__pyx_v_self->current_root);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_self->current_root);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1428
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
 */
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1430
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]             # <<<<<<<<<<<<<<
 *             self.current_block_list = self.current_block[1]
 *             self.current_block_pos = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_self->block_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->current_block);
    __Pyx_DECREF(__pyx_v_self->current_block);
    __pyx_v_self->current_block = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1431
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]             # <<<<<<<<<<<<<<
 *             self.current_block_pos = 0
 *         else:
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_block, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->current_block_list);
    __Pyx_DECREF(__pyx_v_self->current_block_list);
    __pyx_v_self->current_block_list = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1432
 *             self.current_block = self.state._dirblocks[self.block_index]
 *             self.current_block_list = self.current_block[1]
 *             self.current_block_pos = 0             # <<<<<<<<<<<<<<
 *         else:
 *             self.current_block = None
 */
    __pyx_v_self->current_block_pos = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1428
 * 
 *     cdef int _update_current_block(self) except -1:
 *         if (self.block_index < len(self.state._dirblocks) and             # <<<<<<<<<<<<<<
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 *             self.current_block = self.state._dirblocks[self.block_index]
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1434
 *             self.current_block_pos = 0
 *         else:
 *             self.current_block = None             # <<<<<<<<<<<<<<
 *             self.current_block_list = None
 *         return 0
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_block);
    __Pyx_DECREF(__pyx_v_self->current_block);
    __pyx_v_self->current_block = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1435
 *         else:
 *             self.current_block = None
 *             self.current_block_list = None             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_block_list);
    __Pyx_DECREF(__pyx_v_self->current_block_list);
    __pyx_v_self->current_block_list = Py_None;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1436
 *             self.current_block = None
 *             self.current_block_list = None
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1427
 *         return 0
 * 
 *     cdef int _update_current_block(self) except -1:             # <<<<<<<<<<<<<<
 *         if (self.block_index < len(self.state._dirblocks) and
 *             is_inside(self.current_root, self.state._dirblocks[self.block_index][0])):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._update_current_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1438
 *         return 0
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_6__next__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1440
 *     def __next__(self):
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()             # <<<<<<<<<<<<<<
 * 
 *     cdef _iter_next(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1438
 *         return 0
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         # Simple thunk to allow tail recursion without pyrex confusion
 *         return self._iter_next()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1442
 *         return self._iter_next()
 * 
 *     cdef _iter_next(self):             # <<<<<<<<<<<<<<
 *         """Iterate over the changes."""
 *         # This function single steps through an iterator. As such while loops
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_current_dirname = 0;
  PyObject *__pyx_v_current_blockname = 0;
  char *__pyx_v_current_dirname_c;
  char *__pyx_v_current_blockname_c;
  int __pyx_v_path_handled;
  CYTHON_UNUSED PyObject *__pyx_v_searched_specific_files = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_root_stat = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_new_executable = NULL;
  PyObject *__pyx_v_e_winerror = NULL;
  PyObject *__pyx_v_win_errors = NULL;
  PyObject *__pyx_v_bzr_index = NULL;
  PyObject *__pyx_v_initial_key = NULL;
  PyObject *__pyx_v_current_path_info = NULL;
  PyObject *__pyx_v_current_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  __Pyx_RefNannySetupContext("_iter_next", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1495
 *         cdef int advance_entry, advance_path
 *         cdef int path_handled
 *         searched_specific_files = self.searched_specific_files             # <<<<<<<<<<<<<<
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:
 */
  __pyx_t_1 = __pyx_v_self->searched_specific_files;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_searched_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1497
 *         searched_specific_files = self.searched_specific_files
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:             # <<<<<<<<<<<<<<
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_self->root_entries_pos < __pyx_v_self->root_entries_len) != 0);
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1498
 *         # Are we walking a root?
 *         while self.root_entries_pos < self.root_entries_len:
 *             entry = self.root_entries[self.root_entries_pos]             # <<<<<<<<<<<<<<
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->root_entries, __pyx_v_self->root_entries_pos, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1499
 *         while self.root_entries_pos < self.root_entries_len:
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1             # <<<<<<<<<<<<<<
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
 */
    __pyx_v_self->root_entries_pos = (__pyx_v_self->root_entries_pos + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1500
 *             entry = self.root_entries[self.root_entries_pos]
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)             # <<<<<<<<<<<<<<
 *             if changed is not None:
 *                 if changed:
 */
    __pyx_t_1 = __pyx_v_self->root_dir_info;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1500, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1500, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1500, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1500, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1501
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 */
    __pyx_t_2 = (__pyx_v_changed != Py_None);
    __pyx_t_7 = (__pyx_t_2 != 0);
    if (__pyx_t_7) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1502
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
 */
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1502, __pyx_L1_error)
      if (__pyx_t_7) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1503
 *             if changed is not None:
 *                 if changed:
 *                     self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                 if changed or self.include_unchanged:
 *                     return result
 */
        __pyx_t_8 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1503, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1502
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1504
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1504, __pyx_L1_error)
      if (!__pyx_t_2) {
      } else {
        __pyx_t_7 = __pyx_t_2;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_7 = __pyx_t_2;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_7) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1505
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:
 *                     return result             # <<<<<<<<<<<<<<
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_result);
        __pyx_r = __pyx_v_result;
        goto __pyx_L0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1504
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1501
 *             self.root_entries_pos = self.root_entries_pos + 1
 *             result, changed = self._process_entry(entry, self.root_dir_info)
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 */
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1507
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:             # <<<<<<<<<<<<<<
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 */
  __pyx_t_7 = (__pyx_v_self->current_root == Py_None);
  __pyx_t_2 = (__pyx_t_7 != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1511
 *             # interface doesn't require it.
 *             try:
 *                 self.current_root = self.search_specific_files.pop()             # <<<<<<<<<<<<<<
 *             except KeyError, _:
 *                 raise StopIteration()
 */
        __pyx_t_3 = __Pyx_PyObject_Pop(__pyx_v_self->search_specific_files); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_v_self->current_root);
        __Pyx_DECREF(__pyx_v_self->current_root);
        __pyx_v_self->current_root = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
 */
      }
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L18_try_end;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1512
 *             try:
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:             # <<<<<<<<<<<<<<
 *                 raise StopIteration()
 *             self.searched_specific_files.add(self.current_root)
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 1512, __pyx_L15_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v__ = __pyx_t_4;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1513
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
 *                 raise StopIteration()             # <<<<<<<<<<<<<<
 *             self.searched_specific_files.add(self.current_root)
 *             # process the entries for this containing directory: the rest will be
 */
        __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1513, __pyx_L15_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_Raise(__pyx_t_5, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __PYX_ERR(0, 1513, __pyx_L15_except_error)
      }
      goto __pyx_L15_except_error;
      __pyx_L15_except_error:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1510
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.current_root = self.search_specific_files.pop()
 *             except KeyError, _:
 */
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L18_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1514
 *             except KeyError, _:
 *                 raise StopIteration()
 *             self.searched_specific_files.add(self.current_root)             # <<<<<<<<<<<<<<
 *             # process the entries for this containing directory: the rest will be
 *             # found by their parents recursively.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->searched_specific_files, __pyx_n_s_add); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_self->current_root) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->current_root);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1517
 *             # process the entries for this containing directory: the rest will be
 *             # found by their parents recursively.
 *             self.root_entries = self.state._entries_for_path(self.current_root)             # <<<<<<<<<<<<<<
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_entries_for_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_self->current_root) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->current_root);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->root_entries);
    __Pyx_DECREF(__pyx_v_self->root_entries);
    __pyx_v_self->root_entries = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1518
 *             # found by their parents recursively.
 *             self.root_entries = self.state._entries_for_path(self.current_root)
 *             self.root_entries_len = len(self.root_entries)             # <<<<<<<<<<<<<<
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 */
    __pyx_t_1 = __pyx_v_self->root_entries;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_12 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1518, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->root_entries_len = __pyx_t_12;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1519
 *             self.root_entries = self.state._entries_for_path(self.current_root)
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')             # <<<<<<<<<<<<<<
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->current_root, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_s_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_utf8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1519, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->current_root_unicode);
    __Pyx_DECREF(__pyx_v_self->current_root_unicode);
    __pyx_v_self->current_root_unicode = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1520
 *             self.root_entries_len = len(self.root_entries)
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)             # <<<<<<<<<<<<<<
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_abspath); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_self->current_root_unicode) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_self->current_root_unicode);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->root_abspath);
    __Pyx_DECREF(__pyx_v_self->root_abspath);
    __pyx_v_self->root_abspath = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1521
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1522
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)             # <<<<<<<<<<<<<<
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_os); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1522, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_lstat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1522, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_self->root_abspath) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self->root_abspath);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1522, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_root_stat = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1521
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1531
 *                     raise
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,             # <<<<<<<<<<<<<<
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 */
      /*else:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1532
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,             # <<<<<<<<<<<<<<
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_osutils); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1532, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_file_kind_from_stat_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1532, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_root_stat, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1532, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1532, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1531
 *                     raise
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,             # <<<<<<<<<<<<<<
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 */
        __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1531, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_kp_b__4);
        __Pyx_GIVEREF(__pyx_kp_b__4);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_b__4);
        __Pyx_INCREF(__pyx_v_self->current_root);
        __Pyx_GIVEREF(__pyx_v_self->current_root);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_self->current_root);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_1);
        __Pyx_INCREF(__pyx_v_root_stat);
        __Pyx_GIVEREF(__pyx_v_root_stat);
        PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_root_stat);
        __Pyx_INCREF(__pyx_v_self->root_abspath);
        __Pyx_GIVEREF(__pyx_v_self->root_abspath);
        PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->root_abspath);
        __pyx_t_1 = 0;
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_v_self->root_dir_info);
        __Pyx_DECREF(__pyx_v_self->root_dir_info);
        __pyx_v_self->root_dir_info = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1534
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 */
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1534, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1534, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1535
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_directory_is_tree_reference); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1535, __pyx_L23_except_error)
          __Pyx_GOTREF(__pyx_t_1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1536
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):             # <<<<<<<<<<<<<<
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
 */
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_self->current_root_unicode) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_self->current_root_unicode);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1535, __pyx_L23_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1535
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 */
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1535, __pyx_L23_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1537
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
 */
            __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 0, 2, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1537, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_4);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1538
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *             if not self.root_entries and not self.root_dir_info:
 *                 # this specified path is not present at all, skip it.
 */
            __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_tuple__12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 3, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1538, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1538, __pyx_L23_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1537
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
 */
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_v_self->root_dir_info);
            __Pyx_DECREF(__pyx_v_self->root_dir_info);
            __pyx_v_self->root_dir_info = __pyx_t_3;
            __pyx_t_3 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1535
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':
 *                     if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1534
 *                     osutils.file_kind_from_stat_mode(root_stat.st_mode), root_stat,
 *                     self.root_abspath)
 *                 if self.root_dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 */
        }
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L26_try_end;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1523
 *             try:
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:             # <<<<<<<<<<<<<<
 *                 if e.errno == errno.ENOENT:
 *                     # the path does not exist: let _process_entry know that.
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_1) < 0) __PYX_ERR(0, 1523, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_v_e = __pyx_t_4;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1524
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1524, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_errno); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1524, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_ENOENT); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1524, __pyx_L23_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_5, __pyx_t_14, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1524, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1524, __pyx_L23_except_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (likely(__pyx_t_2)) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1526
 *                 if e.errno == errno.ENOENT:
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None             # <<<<<<<<<<<<<<
 *                 else:
 *                     # some other random error: hand it up.
 */
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          __Pyx_GOTREF(__pyx_v_self->root_dir_info);
          __Pyx_DECREF(__pyx_v_self->root_dir_info);
          __pyx_v_self->root_dir_info = Py_None;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1524
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 *                 if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                     # the path does not exist: let _process_entry know that.
 *                     self.root_dir_info = None
 */
          goto __pyx_L31;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1529
 *                 else:
 *                     # some other random error: hand it up.
 *                     raise             # <<<<<<<<<<<<<<
 *             else:
 *                 self.root_dir_info = (b'', self.current_root,
 */
        /*else*/ {
          __Pyx_GIVEREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          __Pyx_XGIVEREF(__pyx_t_1);
          __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_4, __pyx_t_1);
          __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_1 = 0; 
          __PYX_ERR(0, 1529, __pyx_L23_except_error)
        }
        __pyx_L31:;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L22_exception_handled;
      }
      goto __pyx_L23_except_error;
      __pyx_L23_except_error:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1521
 *             self.current_root_unicode = self.current_root.decode('utf8')
 *             self.root_abspath = self.tree.abspath(self.current_root_unicode)
 *             try:             # <<<<<<<<<<<<<<
 *                 root_stat = os.lstat(self.root_abspath)
 *             except OSError, e:
 */
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L22_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      __pyx_L26_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1539
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 # this specified path is not present at all, skip it.
 *                 # (tail recursion, can do a loop once the full structure is
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_entries); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __pyx_t_15 = ((!__pyx_t_7) != 0);
    if (__pyx_t_15) {
    } else {
      __pyx_t_2 = __pyx_t_15;
      goto __pyx_L33_bool_binop_done;
    }
    __pyx_t_15 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1539, __pyx_L1_error)
    __pyx_t_7 = ((!__pyx_t_15) != 0);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L33_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1543
 *                 # (tail recursion, can do a loop once the full structure is
 *                 # known).
 *                 return self._iter_next()             # <<<<<<<<<<<<<<
 *             path_handled = 0
 *             self.root_entries_pos = 0
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1543, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_r = __pyx_t_1;
      __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1539
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 # this specified path is not present at all, skip it.
 *                 # (tail recursion, can do a loop once the full structure is
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1544
 *                 # known).
 *                 return self._iter_next()
 *             path_handled = 0             # <<<<<<<<<<<<<<
 *             self.root_entries_pos = 0
 *             # XXX Clarity: This loop is duplicated a out the self.current_root
 */
    __pyx_v_path_handled = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1545
 *                 return self._iter_next()
 *             path_handled = 0
 *             self.root_entries_pos = 0             # <<<<<<<<<<<<<<
 *             # XXX Clarity: This loop is duplicated a out the self.current_root
 *             # is None guard above: if we return from it, it completes there
 */
    __pyx_v_self->root_entries_pos = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1550
 *             # (and the following if block cannot trigger because
 *             # path_handled must be true, so the if block is not # duplicated.
 *             while self.root_entries_pos < self.root_entries_len:             # <<<<<<<<<<<<<<
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1
 */
    while (1) {
      __pyx_t_2 = ((__pyx_v_self->root_entries_pos < __pyx_v_self->root_entries_len) != 0);
      if (!__pyx_t_2) break;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1551
 *             # path_handled must be true, so the if block is not # duplicated.
 *             while self.root_entries_pos < self.root_entries_len:
 *                 entry = self.root_entries[self.root_entries_pos]             # <<<<<<<<<<<<<<
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->root_entries, __pyx_v_self->root_entries_pos, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1551, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1552
 *             while self.root_entries_pos < self.root_entries_len:
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1             # <<<<<<<<<<<<<<
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:
 */
      __pyx_v_self->root_entries_pos = (__pyx_v_self->root_entries_pos + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1553
 *                 entry = self.root_entries[self.root_entries_pos]
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)             # <<<<<<<<<<<<<<
 *                 if changed is not None:
 *                     path_handled = -1
 */
      __pyx_t_1 = __pyx_v_self->root_dir_info;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1553, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1553, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_13 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1553, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_6 = Py_TYPE(__pyx_t_13)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_6(__pyx_t_13); if (unlikely(!__pyx_t_1)) goto __pyx_L37_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_13); if (unlikely(!__pyx_t_3)) goto __pyx_L37_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_13), 2) < 0) __PYX_ERR(0, 1553, __pyx_L1_error)
        __pyx_t_6 = NULL;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L38_unpacking_done;
        __pyx_L37_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_6 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1553, __pyx_L1_error)
        __pyx_L38_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1554
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     path_handled = -1
 *                     if changed:
 */
      __pyx_t_2 = (__pyx_v_changed != Py_None);
      __pyx_t_7 = (__pyx_t_2 != 0);
      if (__pyx_t_7) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1555
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:
 *                     path_handled = -1             # <<<<<<<<<<<<<<
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 */
        __pyx_v_path_handled = -1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1556
 *                 if changed is not None:
 *                     path_handled = -1
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 */
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1556, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1557
 *                     path_handled = -1
 *                     if changed:
 *                         self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if changed or self.include_unchanged:
 *                         return result
 */
          __pyx_t_8 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1557, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1556
 *                 if changed is not None:
 *                     path_handled = -1
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1558
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             # handle unversioned specified paths:
 */
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1558, __pyx_L1_error)
        if (!__pyx_t_2) {
        } else {
          __pyx_t_7 = __pyx_t_2;
          goto __pyx_L42_bool_binop_done;
        }
        __pyx_t_2 = (__pyx_v_self->include_unchanged != 0);
        __pyx_t_7 = __pyx_t_2;
        __pyx_L42_bool_binop_done:;
        if (__pyx_t_7) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1559
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 *                         return result             # <<<<<<<<<<<<<<
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_result);
          __pyx_r = __pyx_v_result;
          goto __pyx_L0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1558
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             # handle unversioned specified paths:
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1554
 *                 self.root_entries_pos = self.root_entries_pos + 1
 *                 result, changed = self._process_entry(entry, self.root_dir_info)
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     path_handled = -1
 *                     if changed:
 */
      }
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1561
 *                         return result
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 */
    __pyx_t_2 = (__pyx_v_self->want_unversioned != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_7 = __pyx_t_2;
      goto __pyx_L45_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_path_handled != 0)) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_7 = __pyx_t_2;
      goto __pyx_L45_bool_binop_done;
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1561, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_2;
    __pyx_L45_bool_binop_done:;
    if (__pyx_t_7) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1563
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
 *                 return TreeChange(
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_stat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_S_ISREG); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1563, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_2) {
      } else {
        __pyx_t_7 = __pyx_t_2;
        goto __pyx_L48_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1564
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                 return TreeChange(
 *                        None,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_stat); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_S_IEXEC); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_And(__pyx_t_1, __pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1564, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __pyx_t_2;
      __pyx_L48_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1562
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:
 *                 new_executable = bool(             # <<<<<<<<<<<<<<
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
 */
      __pyx_t_4 = __Pyx_PyBool_FromLong((!(!__pyx_t_7))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1562, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_new_executable = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1565
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 *                     and stat.S_IEXEC & self.root_dir_info[3].st_mode)
 *                 return TreeChange(             # <<<<<<<<<<<<<<
 *                        None,
 *                        (None, self.current_root_unicode),
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1565, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1567
 *                 return TreeChange(
 *                        None,
 *                        (None, self.current_root_unicode),             # <<<<<<<<<<<<<<
 *                        True,
 *                        (False, False),
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_1, 0, Py_None);
      __Pyx_INCREF(__pyx_v_self->current_root_unicode);
      __Pyx_GIVEREF(__pyx_v_self->current_root_unicode);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->current_root_unicode);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1571
 *                        (False, False),
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),             # <<<<<<<<<<<<<<
 *                        (None, self.root_dir_info[2]),
 *                        (None, new_executable)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_splitpath); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_5, __pyx_v_self->current_root_unicode) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_self->current_root_unicode);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_3, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_14);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1572
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),
 *                        (None, self.root_dir_info[2]),             # <<<<<<<<<<<<<<
 *                        (None, new_executable)
 *                       )
 */
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_5, 0, Py_None);
      __Pyx_GIVEREF(__pyx_t_14);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_14);
      __pyx_t_14 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1573
 *                        (None, splitpath(self.current_root_unicode)[-1]),
 *                        (None, self.root_dir_info[2]),
 *                        (None, new_executable)             # <<<<<<<<<<<<<<
 *                       )
 *             # If we reach here, the outer flow continues, which enters into the
 */
      __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyTuple_SET_ITEM(__pyx_t_14, 0, Py_None);
      __Pyx_INCREF(__pyx_v_new_executable);
      __Pyx_GIVEREF(__pyx_v_new_executable);
      PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_v_new_executable);
      __pyx_t_16 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_13)) {
        PyObject *__pyx_temp[9] = {__pyx_t_16, Py_None, __pyx_t_1, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_3, __pyx_t_5, __pyx_t_14};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1565, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
        PyObject *__pyx_temp[9] = {__pyx_t_16, Py_None, __pyx_t_1, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_3, __pyx_t_5, __pyx_t_14};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1565, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      {
        __pyx_t_17 = PyTuple_New(8+__pyx_t_8); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        if (__pyx_t_16) {
          __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_16); __pyx_t_16 = NULL;
        }
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_8, Py_None);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_8, __pyx_t_1);
        __Pyx_INCREF(Py_True);
        __Pyx_GIVEREF(Py_True);
        PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_8, Py_True);
        __Pyx_INCREF(__pyx_tuple__14);
        __Pyx_GIVEREF(__pyx_tuple__14);
        PyTuple_SET_ITEM(__pyx_t_17, 3+__pyx_t_8, __pyx_tuple__14);
        __Pyx_INCREF(__pyx_tuple__6);
        __Pyx_GIVEREF(__pyx_tuple__6);
        PyTuple_SET_ITEM(__pyx_t_17, 4+__pyx_t_8, __pyx_tuple__6);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_17, 5+__pyx_t_8, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_5);
        PyTuple_SET_ITEM(__pyx_t_17, 6+__pyx_t_8, __pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_17, 7+__pyx_t_8, __pyx_t_14);
        __pyx_t_1 = 0;
        __pyx_t_3 = 0;
        __pyx_t_5 = 0;
        __pyx_t_14 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_17, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1565, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      }
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1561
 *                         return result
 *             # handle unversioned specified paths:
 *             if self.want_unversioned and not path_handled and self.root_dir_info:             # <<<<<<<<<<<<<<
 *                 new_executable = bool(
 *                     stat.S_ISREG(self.root_dir_info[3].st_mode)
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1507
 *                     return result
 *         # Have we finished the prior root, or never started one ?
 *         if self.current_root is None:             # <<<<<<<<<<<<<<
 *             # TODO: the pending list should be lexically sorted?  the
 *             # interface doesn't require it.
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1577
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
 */
  __pyx_t_2 = (__pyx_v_self->current_dir_info == Py_None);
  __pyx_t_15 = (__pyx_t_2 != 0);
  if (__pyx_t_15) {
  } else {
    __pyx_t_7 = __pyx_t_15;
    goto __pyx_L51_bool_binop_done;
  }
  __pyx_t_15 = (__pyx_v_self->current_block == Py_None);
  __pyx_t_2 = (__pyx_t_15 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_7 = __pyx_t_2;
    goto __pyx_L51_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1578
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not
 *             self.doing_consistency_expansion):             # <<<<<<<<<<<<<<
 *             # setup iteration of this root:
 *             self.current_dir_list = None
 */
  __pyx_t_2 = ((!(__pyx_v_self->doing_consistency_expansion != 0)) != 0);
  __pyx_t_7 = __pyx_t_2;
  __pyx_L51_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1577
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
 */
  if (__pyx_t_7) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1580
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
 *             self.current_dir_list = None             # <<<<<<<<<<<<<<
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
 *                 self.current_dir_info = None
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_dir_list);
    __Pyx_DECREF(__pyx_v_self->current_dir_list);
    __pyx_v_self->current_dir_list = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1581
 *             # setup iteration of this root:
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                 self.current_dir_info = None
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->root_dir_info); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1581, __pyx_L1_error)
    if (__pyx_t_2) {
    } else {
      __pyx_t_7 = __pyx_t_2;
      goto __pyx_L55_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_self->root_dir_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_4, __pyx_kp_s_tree_reference, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1581, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = __pyx_t_2;
    __pyx_L55_bool_binop_done:;
    if (__pyx_t_7) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1582
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':
 *                 self.current_dir_info = None             # <<<<<<<<<<<<<<
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->current_dir_info);
      __Pyx_DECREF(__pyx_v_self->current_dir_info);
      __pyx_v_self->current_dir_info = Py_None;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1581
 *             # setup iteration of this root:
 *             self.current_dir_list = None
 *             if self.root_dir_info and self.root_dir_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                 self.current_dir_info = None
 *             else:
 */
      goto __pyx_L54;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1584
 *                 self.current_dir_info = None
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,             # <<<<<<<<<<<<<<
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_osutils); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_walkdirs_utf8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_self->root_abspath);
      __Pyx_GIVEREF(__pyx_v_self->root_abspath);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->root_abspath);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1585
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
 *                     prefix=self.current_root)             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 try:
 */
      __pyx_t_17 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1585, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      if (PyDict_SetItem(__pyx_t_17, __pyx_n_s_prefix, __pyx_v_self->current_root) < 0) __PYX_ERR(0, 1585, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1584
 *                 self.current_dir_info = None
 *             else:
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,             # <<<<<<<<<<<<<<
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 */
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_4, __pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_GIVEREF(__pyx_t_14);
      __Pyx_GOTREF(__pyx_v_self->dir_iterator);
      __Pyx_DECREF(__pyx_v_self->dir_iterator);
      __pyx_v_self->dir_iterator = __pyx_t_14;
      __pyx_t_14 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1586
 *                 self.dir_iterator = osutils._walkdirs_utf8(self.root_abspath,
 *                     prefix=self.current_root)
 *                 self.path_index = 0             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 */
      __pyx_v_self->path_index = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1587
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        /*try:*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1588
 *                 self.path_index = 0
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except OSError, e:
 */
          __pyx_t_14 = __pyx_v_self->dir_iterator;
          __Pyx_INCREF(__pyx_t_14);
          __pyx_t_17 = __Pyx_PyIter_Next(__pyx_t_14); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1588, __pyx_L57_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_GIVEREF(__pyx_t_17);
          __Pyx_GOTREF(__pyx_v_self->current_dir_info);
          __Pyx_DECREF(__pyx_v_self->current_dir_info);
          __pyx_v_self->current_dir_info = __pyx_t_17;
          __pyx_t_17 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1589
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                 except OSError, e:
 *                     # there may be directories in the inventory even though
 */
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1589, __pyx_L57_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_GIVEREF(__pyx_t_17);
          __Pyx_GOTREF(__pyx_v_self->current_dir_list);
          __Pyx_DECREF(__pyx_v_self->current_dir_list);
          __pyx_v_self->current_dir_list = __pyx_t_17;
          __pyx_t_17 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1587
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1613
 *                         raise
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':             # <<<<<<<<<<<<<<
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 */
        /*else:*/ {
          __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1613, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1613, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          __pyx_t_7 = (__Pyx_PyBytes_Equals(__pyx_t_14, __pyx_kp_b__4, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1613, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          if (__pyx_t_7) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1615
 *                     if self.current_dir_info[0][0] == b'':
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))             # <<<<<<<<<<<<<<
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
 */
            __Pyx_INCREF(__pyx_v_self->bisect_left);
            __pyx_t_17 = __pyx_v_self->bisect_left; __pyx_t_4 = NULL;
            __pyx_t_8 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_4)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_4);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
                __pyx_t_8 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self->current_dir_list, __pyx_tuple__15};
              __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1615, __pyx_L59_except_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_14);
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self->current_dir_list, __pyx_tuple__15};
              __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1615, __pyx_L59_except_error)
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_14);
            } else
            #endif
            {
              __pyx_t_13 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1615, __pyx_L59_except_error)
              __Pyx_GOTREF(__pyx_t_13);
              if (__pyx_t_4) {
                __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
              }
              __Pyx_INCREF(__pyx_v_self->current_dir_list);
              __Pyx_GIVEREF(__pyx_v_self->current_dir_list);
              PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_8, __pyx_v_self->current_dir_list);
              __Pyx_INCREF(__pyx_tuple__15);
              __Pyx_GIVEREF(__pyx_tuple__15);
              PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_8, __pyx_tuple__15);
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_13, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1615, __pyx_L59_except_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_v_bzr_index = __pyx_t_14;
            __pyx_t_14 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1616
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':             # <<<<<<<<<<<<<<
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
 */
            __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_v_self->current_dir_list, __pyx_v_bzr_index); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1616, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1616, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_7 = (__Pyx_PyBytes_Equals(__pyx_t_17, __pyx_kp_b_bzr, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1616, __pyx_L59_except_error)
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(__pyx_t_7)) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1617
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')
 */
              __pyx_t_17 = __Pyx_PyObject_CallNoArg(__pyx_builtin_AssertionError); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1617, __pyx_L59_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_Raise(__pyx_t_17, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __PYX_ERR(0, 1617, __pyx_L59_except_error)

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1616
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':             # <<<<<<<<<<<<<<
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
 */
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1618
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]             # <<<<<<<<<<<<<<
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 */
            if (unlikely(PyObject_DelItem(__pyx_v_self->current_dir_list, __pyx_v_bzr_index) < 0)) __PYX_ERR(0, 1618, __pyx_L59_except_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1613
 *                         raise
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':             # <<<<<<<<<<<<<<
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))
 */
          }
        }
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L62_try_end;
        __pyx_L57_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1590
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except OSError, e:             # <<<<<<<<<<<<<<
 *                     # there may be directories in the inventory even though
 *                     # this path is not a file on disk: so mark it as end of
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_17, &__pyx_t_14, &__pyx_t_13) < 0) __PYX_ERR(0, 1590, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_17);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_14);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1594
 *                     # this path is not a file on disk: so mark it as end of
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errno); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ENOENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (!__pyx_t_2) {
          } else {
            __pyx_t_7 = __pyx_t_2;
            goto __pyx_L68_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_ENOTDIR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (!__pyx_t_2) {
          } else {
            __pyx_t_7 = __pyx_t_2;
            goto __pyx_L68_bool_binop_done;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_EINVAL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1594, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_7 = __pyx_t_2;
          __pyx_L68_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_2 = (__pyx_t_7 != 0);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1595
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                     elif sys.platform == 'win32':
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
 */
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = Py_None;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1594
 *                     # this path is not a file on disk: so mark it as end of
 *                     # iterator
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':
 */
            goto __pyx_L67;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1596
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':             # <<<<<<<<<<<<<<
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
 *                         # python 2.5 has e.errno == EINVAL,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_sys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1596, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_platform); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1596, __pyx_L59_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_win32, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1596, __pyx_L59_except_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (likely(__pyx_t_2)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1600
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              /*try:*/ {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1601
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:
 *                             e_winerror = e.winerror             # <<<<<<<<<<<<<<
 *                         except AttributeError, _:
 *                             e_winerror = None
 */
                __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_winerror); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1601, __pyx_L71_error)
                __Pyx_GOTREF(__pyx_t_5);
                __pyx_v_e_winerror = __pyx_t_5;
                __pyx_t_5 = 0;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1600
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
 */
              }
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
              __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
              goto __pyx_L78_try_end;
              __pyx_L71_error:;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1602
 *                         try:
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:             # <<<<<<<<<<<<<<
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 */
              __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
              if (__pyx_t_8) {
                __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
                if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 1602, __pyx_L73_except_error)
                __Pyx_GOTREF(__pyx_t_5);
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GOTREF(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_4);
                __pyx_v__ = __pyx_t_4;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1603
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
 *                             e_winerror = None             # <<<<<<<<<<<<<<
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):
 */
                __Pyx_INCREF(Py_None);
                __Pyx_XDECREF_SET(__pyx_v_e_winerror, Py_None);
                __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                goto __pyx_L72_exception_handled;
              }
              goto __pyx_L73_except_error;
              __pyx_L73_except_error:;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1600
 *                         # python 2.5 has e.errno == EINVAL,
 *                         #            and e.winerror == ERROR_DIRECTORY
 *                         try:             # <<<<<<<<<<<<<<
 *                             e_winerror = e.winerror
 *                         except AttributeError, _:
 */
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
              goto __pyx_L59_except_error;
              __pyx_L72_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
              __pyx_L78_try_end:;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1604
 *                         except AttributeError, _:
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)             # <<<<<<<<<<<<<<
 *                         if (e.errno in win_errors or e_winerror in win_errors):
 *                             self.current_dir_info = None
 */
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1604, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1604, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1604, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
            __pyx_t_3 = 0;
            __pyx_t_4 = 0;
            __pyx_v_win_errors = __pyx_t_5;
            __pyx_t_5 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1605
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):             # <<<<<<<<<<<<<<
 *                             self.current_dir_info = None
 *                         else:
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errno); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1605, __pyx_L59_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_t_5, __pyx_v_win_errors, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1605, __pyx_L59_except_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_15 = (__pyx_t_7 != 0);
            if (!__pyx_t_15) {
            } else {
              __pyx_t_2 = __pyx_t_15;
              goto __pyx_L82_bool_binop_done;
            }
            __pyx_t_15 = (__Pyx_PySequence_ContainsTF(__pyx_v_e_winerror, __pyx_v_win_errors, Py_EQ)); if (unlikely(__pyx_t_15 < 0)) __PYX_ERR(0, 1605, __pyx_L59_except_error)
            __pyx_t_7 = (__pyx_t_15 != 0);
            __pyx_t_2 = __pyx_t_7;
            __pyx_L82_bool_binop_done:;
            if (likely(__pyx_t_2)) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1606
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):
 *                             self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                         else:
 *                             # Will this really raise the right exception ?
 */
              __Pyx_INCREF(Py_None);
              __Pyx_GIVEREF(Py_None);
              __Pyx_GOTREF(__pyx_v_self->current_dir_info);
              __Pyx_DECREF(__pyx_v_self->current_dir_info);
              __pyx_v_self->current_dir_info = Py_None;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1605
 *                             e_winerror = None
 *                         win_errors = (ERROR_DIRECTORY, ERROR_PATH_NOT_FOUND)
 *                         if (e.errno in win_errors or e_winerror in win_errors):             # <<<<<<<<<<<<<<
 *                             self.current_dir_info = None
 *                         else:
 */
              goto __pyx_L81;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1609
 *                         else:
 *                             # Will this really raise the right exception ?
 *                             raise             # <<<<<<<<<<<<<<
 *                     else:
 *                         raise
 */
            /*else*/ {
              __Pyx_GIVEREF(__pyx_t_17);
              __Pyx_GIVEREF(__pyx_t_14);
              __Pyx_XGIVEREF(__pyx_t_13);
              __Pyx_ErrRestoreWithState(__pyx_t_17, __pyx_t_14, __pyx_t_13);
              __pyx_t_17 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; 
              __PYX_ERR(0, 1609, __pyx_L59_except_error)
            }
            __pyx_L81:;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1596
 *                     if e.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
 *                         self.current_dir_info = None
 *                     elif sys.platform == 'win32':             # <<<<<<<<<<<<<<
 *                         # on win32, python2.4 has e.errno == ERROR_DIRECTORY, but
 *                         # python 2.5 has e.errno == EINVAL,
 */
            goto __pyx_L67;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1611
 *                             raise
 *                     else:
 *                         raise             # <<<<<<<<<<<<<<
 *                 else:
 *                     if self.current_dir_info[0][0] == b'':
 */
          /*else*/ {
            __Pyx_GIVEREF(__pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_14);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ErrRestoreWithState(__pyx_t_17, __pyx_t_14, __pyx_t_13);
            __pyx_t_17 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; 
            __PYX_ERR(0, 1611, __pyx_L59_except_error)
          }
          __pyx_L67:;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          goto __pyx_L58_exception_handled;
        }
        goto __pyx_L59_except_error;
        __pyx_L59_except_error:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1587
 *                     prefix=self.current_root)
 *                 self.path_index = 0
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        goto __pyx_L1_error;
        __pyx_L58_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
        __pyx_L62_try_end:;
      }
    }
    __pyx_L54:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1619
 *                             raise AssertionError()
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')             # <<<<<<<<<<<<<<
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:
 */
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_v_self->current_root);
    __Pyx_GIVEREF(__pyx_v_self->current_root);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_self->current_root);
    __Pyx_INCREF(__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_kp_b__4);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_kp_b__4);
    __Pyx_INCREF(__pyx_kp_b__4);
    __Pyx_GIVEREF(__pyx_kp_b__4);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_kp_b__4);
    __pyx_v_initial_key = ((PyObject*)__pyx_t_13);
    __pyx_t_13 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1620
 *                         del self.current_dir_list[bzr_index]
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)             # <<<<<<<<<<<<<<
 *             if self.block_index == 0:
 *                 # we have processed the total root already, but because the
 */
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_find_block_index_from_key); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_17 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_17)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_17);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
      }
    }
    __pyx_t_13 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_17, __pyx_v_initial_key) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_v_initial_key);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1620, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_13))) || (PyList_CheckExact(__pyx_t_13))) {
      PyObject* sequence = __pyx_t_13;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1620, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_14 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_17);
      #else
      __pyx_t_14 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      #endif
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1620, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_14 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_14)) goto __pyx_L84_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_14);
      index = 1; __pyx_t_17 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_17)) goto __pyx_L84_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_17);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1620, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L85_unpacking_done;
      __pyx_L84_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1620, __pyx_L1_error)
      __pyx_L85_unpacking_done:;
    }
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_14); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1620, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_v_self->block_index = __pyx_t_8;
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_17);
    __pyx_t_17 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1621
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:             # <<<<<<<<<<<<<<
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
 */
    __pyx_t_2 = ((__pyx_v_self->block_index == 0) != 0);
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1624
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
 *                 self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *             self._update_current_block()
 *         # walk until both the directory listing and the versioned metadata
 */
      __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1621
 *             initial_key = (self.current_root, b'', b'')
 *             self.block_index, _ = self.state._find_block_index_from_key(initial_key)
 *             if self.block_index == 0:             # <<<<<<<<<<<<<<
 *                 # we have processed the total root already, but because the
 *                 # initial key matched it we should skip it here.
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1625
 *                 # initial key matched it we should skip it here.
 *                 self.block_index = self.block_index + 1
 *             self._update_current_block()             # <<<<<<<<<<<<<<
 *         # walk until both the directory listing and the versioned metadata
 *         # are exhausted.
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1625, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1577
 *             # If we reach here, the outer flow continues, which enters into the
 *             # per-root setup logic.
 *         if (self.current_dir_info is None and self.current_block is None and not             # <<<<<<<<<<<<<<
 *             self.doing_consistency_expansion):
 *             # setup iteration of this root:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1628
 *         # walk until both the directory listing and the versioned metadata
 *         # are exhausted.
 *         while (self.current_dir_info is not None             # <<<<<<<<<<<<<<
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 */
  while (1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1629
 *         # are exhausted.
 *         while (self.current_dir_info is not None
 *             or self.current_block is not None):             # <<<<<<<<<<<<<<
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block
 */
    __pyx_t_7 = (__pyx_v_self->current_dir_info != Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1628
 *         # walk until both the directory listing and the versioned metadata
 *         # are exhausted.
 *         while (self.current_dir_info is not None             # <<<<<<<<<<<<<<
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 */
    __pyx_t_15 = (__pyx_t_7 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_2 = __pyx_t_15;
      goto __pyx_L89_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1629
 *         # are exhausted.
 *         while (self.current_dir_info is not None
 *             or self.current_block is not None):             # <<<<<<<<<<<<<<
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block
 */
    __pyx_t_15 = (__pyx_v_self->current_block != Py_None);
    __pyx_t_7 = (__pyx_t_15 != 0);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L89_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1631
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->current_dir_info); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1631, __pyx_L1_error)
    if (__pyx_t_7) {
    } else {
      __pyx_t_2 = __pyx_t_7;
      goto __pyx_L92_bool_binop_done;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1632
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block
 *                 and self.current_dir_info[0][0] != self.current_block[0]):             # <<<<<<<<<<<<<<
 *                 # Work around pyrex broken heuristic - current_dirname has
 *                 # the same scope as current_dirname_c
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->current_block); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1631, __pyx_L1_error)
    if (__pyx_t_7) {
    } else {
      __pyx_t_2 = __pyx_t_7;
      goto __pyx_L92_bool_binop_done;
    }
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_17 = __Pyx_GetItemInt(__pyx_t_13, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_self->current_block, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = PyObject_RichCompare(__pyx_t_17, __pyx_t_13, Py_NE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1632, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1632, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_2 = __pyx_t_7;
    __pyx_L92_bool_binop_done:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1631
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
 */
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1635
 *                 # Work around pyrex broken heuristic - current_dirname has
 *                 # the same scope as current_dirname_c
 *                 current_dirname = self.current_dir_info[0][0]             # <<<<<<<<<<<<<<
 *                 current_dirname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_dirname)
 */
      __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1635, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF_SET(__pyx_v_current_dirname, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1636
 *                 # the same scope as current_dirname_c
 *                 current_dirname = self.current_dir_info[0][0]
 *                 current_dirname_c = PyBytes_AS_STRING_void(             # <<<<<<<<<<<<<<
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]
 */
      __pyx_v_current_dirname_c = PyBytes_AS_STRING(((void *)__pyx_v_current_dirname));

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1638
 *                 current_dirname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]             # <<<<<<<<<<<<<<
 *                 current_blockname_c = PyBytes_AS_STRING_void(
 *                     <void *>current_blockname)
 */
      __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_self->current_block, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1638, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_XDECREF_SET(__pyx_v_current_blockname, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1639
 *                     <void *>current_dirname)
 *                 current_blockname = self.current_block[0]
 *                 current_blockname_c = PyBytes_AS_STRING_void(             # <<<<<<<<<<<<<<
 *                     <void *>current_blockname)
 *                 # In the python generator we evaluate this if block once per
 */
      __pyx_v_current_blockname_c = PyBytes_AS_STRING(((void *)__pyx_v_current_blockname));

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1648
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
 *                     PyBytes_Size(current_blockname)) < 0:             # <<<<<<<<<<<<<<
 *                     # filesystem data refers to paths not covered by the
 *                     # dirblock.  this has two possibilities:
 */
      __pyx_t_2 = ((__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__cmp_by_dirs(__pyx_v_current_dirname_c, PyBytes_Size(__pyx_v_current_dirname), __pyx_v_current_blockname_c, PyBytes_Size(__pyx_v_current_blockname)) < 0) != 0);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1645
 *                 # evaluated once per path: we could cache the result before
 *                 # doing the while loop and probably save time.
 *                 if _cmp_by_dirs(current_dirname_c,             # <<<<<<<<<<<<<<
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
 */
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1659
 *                     # recurse into unknown directories.
 *                     # We are doing a loop
 *                     while self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         # dont descend into this unversioned path if it is
 */
        while (1) {
          __pyx_t_13 = __pyx_v_self->current_dir_list;
          __Pyx_INCREF(__pyx_t_13);
          __pyx_t_12 = PyObject_Length(__pyx_t_13); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1659, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_2 = ((__pyx_v_self->path_index < __pyx_t_12) != 0);
          if (!__pyx_t_2) break;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1660
 *                     # We are doing a loop
 *                     while self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 */
          __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1660, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_XDECREF_SET(__pyx_v_current_path_info, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1663
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory',             # <<<<<<<<<<<<<<
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
 */
          __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1663, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_13, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1663, __pyx_L1_error)
          if (!__pyx_t_7) {
          } else {
            __pyx_t_2 = __pyx_t_7;
            goto __pyx_L99_bool_binop_done;
          }
          __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_13, __pyx_kp_s_tree_reference, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1663, __pyx_L1_error)
          __pyx_t_2 = __pyx_t_7;
          __pyx_L99_bool_binop_done:;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_7 = (__pyx_t_2 != 0);
          if (__pyx_t_7) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1665
 *                         if current_path_info[2] in ('directory',
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
 */
            if (unlikely(__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1665, __pyx_L1_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1666
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
 */
            __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1663
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory',             # <<<<<<<<<<<<<<
 *                                                     'tree-reference'):
 *                             del self.current_dir_list[self.path_index]
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1667
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1             # <<<<<<<<<<<<<<
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 */
          __pyx_v_self->path_index = (__pyx_v_self->path_index + 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1668
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
 */
          __pyx_t_7 = (__pyx_v_self->want_unversioned != 0);
          if (__pyx_t_7) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1669
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0])[0]):
 */
            __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1669, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_13, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1669, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (__pyx_t_7) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1670
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0])[0]):
 *                                     current_path_info = current_path_info[:2] + \
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_directory_is_tree_reference); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1670, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1671
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0])[0]):             # <<<<<<<<<<<<<<
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]
 */
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1671, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_v_self->utf8_decode);
              __pyx_t_4 = __pyx_v_self->utf8_decode; __pyx_t_3 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
                __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
                if (likely(__pyx_t_3)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                  __Pyx_INCREF(__pyx_t_3);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_4, function);
                }
              }
              __pyx_t_17 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1671, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_17, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1671, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              __pyx_t_17 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
                __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_14);
                if (likely(__pyx_t_17)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                  __Pyx_INCREF(__pyx_t_17);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_14, function);
                }
              }
              __pyx_t_13 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_17, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_4);
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1670, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1670
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0])[0]):
 *                                     current_path_info = current_path_info[:2] + \
 */
              __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1670, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (__pyx_t_7) {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1672
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0])[0]):
 *                                     current_path_info = current_path_info[:2] + \             # <<<<<<<<<<<<<<
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(
 */
                __pyx_t_13 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 0, 2, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1672, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_13);

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1673
 *                                     self.utf8_decode(current_path_info[0])[0]):
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 */
                __pyx_t_14 = PyNumber_Add(__pyx_t_13, __pyx_tuple__12); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1672, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __pyx_t_13 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 3, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1673, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_13);
                __pyx_t_4 = PyNumber_Add(__pyx_t_14, __pyx_t_13); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1673, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_4);
                __pyx_t_4 = 0;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1670
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                                     self.utf8_decode(current_path_info[0])[0]):
 *                                     current_path_info = current_path_info[:2] + \
 */
              }

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1669
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:
 *                             if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                                 if self.tree._directory_is_tree_reference(
 *                                     self.utf8_decode(current_path_info[0])[0]):
 */
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1675
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             return TreeChange(
 */
            __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_stat); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_S_ISREG); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_14, function);
              }
            }
            __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_13, __pyx_t_17) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_17);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1675, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (__pyx_t_2) {
            } else {
              __pyx_t_7 = __pyx_t_2;
              goto __pyx_L104_bool_binop_done;
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1676
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                             return TreeChange(
 *                                 None,
 */
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_stat); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_S_IEXEC); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = PyNumber_And(__pyx_t_14, __pyx_t_17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1676, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_7 = __pyx_t_2;
            __pyx_L104_bool_binop_done:;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1674
 *                                     current_path_info = current_path_info[:2] + \
 *                                         ('tree-reference',) + current_path_info[3:]
 *                             new_executable = bool(             # <<<<<<<<<<<<<<
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 */
            __pyx_t_4 = __Pyx_PyBool_FromLong((!(!__pyx_t_7))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1674, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_v_new_executable = __pyx_t_4;
            __pyx_t_4 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1677
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             return TreeChange(             # <<<<<<<<<<<<<<
 *                                 None,
 *                                 (None, self.utf8_decode(current_path_info[0])[0]),
 */
            __Pyx_XDECREF(__pyx_r);
            __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1677, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1679
 *                             return TreeChange(
 *                                 None,
 *                                 (None, self.utf8_decode(current_path_info[0])[0]),             # <<<<<<<<<<<<<<
 *                                 True,
 *                                 (False, False),
 */
            __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1679, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_INCREF(__pyx_v_self->utf8_decode);
            __pyx_t_5 = __pyx_v_self->utf8_decode; __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_5, function);
              }
            }
            __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_3, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_13);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1679, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_14, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1679, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1679, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_14, 0, Py_None);
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_5);
            __pyx_t_5 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1683
 *                                 (False, False),
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),             # <<<<<<<<<<<<<<
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
 */
            __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1683, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_INCREF(__pyx_v_self->utf8_decode);
            __pyx_t_3 = __pyx_v_self->utf8_decode; __pyx_t_1 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_5 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_13);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1683, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1683, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1683, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_5, 0, Py_None);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1684
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),
 *                                 (None, current_path_info[2]),             # <<<<<<<<<<<<<<
 *                                 (None, new_executable))
 *                     # This dir info has been handled, go to the next
 */
            __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1684, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1684, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_13, 0, Py_None);
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_3);
            __pyx_t_3 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1685
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))             # <<<<<<<<<<<<<<
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0
 */
            __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1685, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
            __Pyx_INCREF(__pyx_v_new_executable);
            __Pyx_GIVEREF(__pyx_v_new_executable);
            PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_new_executable);
            __pyx_t_1 = NULL;
            __pyx_t_8 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
                __pyx_t_8 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[9] = {__pyx_t_1, Py_None, __pyx_t_14, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_5, __pyx_t_13, __pyx_t_3};
              __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1677, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[9] = {__pyx_t_1, Py_None, __pyx_t_14, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_5, __pyx_t_13, __pyx_t_3};
              __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_8, 8+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1677, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            } else
            #endif
            {
              __pyx_t_16 = PyTuple_New(8+__pyx_t_8); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1677, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_16);
              if (__pyx_t_1) {
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_1); __pyx_t_1 = NULL;
              }
              __Pyx_INCREF(Py_None);
              __Pyx_GIVEREF(Py_None);
              PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_8, Py_None);
              __Pyx_GIVEREF(__pyx_t_14);
              PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_8, __pyx_t_14);
              __Pyx_INCREF(Py_True);
              __Pyx_GIVEREF(Py_True);
              PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_8, Py_True);
              __Pyx_INCREF(__pyx_tuple__14);
              __Pyx_GIVEREF(__pyx_tuple__14);
              PyTuple_SET_ITEM(__pyx_t_16, 3+__pyx_t_8, __pyx_tuple__14);
              __Pyx_INCREF(__pyx_tuple__6);
              __Pyx_GIVEREF(__pyx_tuple__6);
              PyTuple_SET_ITEM(__pyx_t_16, 4+__pyx_t_8, __pyx_tuple__6);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_16, 5+__pyx_t_8, __pyx_t_5);
              __Pyx_GIVEREF(__pyx_t_13);
              PyTuple_SET_ITEM(__pyx_t_16, 6+__pyx_t_8, __pyx_t_13);
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_16, 7+__pyx_t_8, __pyx_t_3);
              __pyx_t_14 = 0;
              __pyx_t_5 = 0;
              __pyx_t_13 = 0;
              __pyx_t_3 = 0;
              __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_16, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1677, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_r = __pyx_t_4;
            __pyx_t_4 = 0;
            goto __pyx_L0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1668
 *                             self.path_index = self.path_index - 1
 *                         self.path_index = self.path_index + 1
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             if current_path_info[2] == 'directory':
 *                                 if self.tree._directory_is_tree_reference(
 */
          }
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1687
 *                                 (None, new_executable))
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = None
 *                     try:
 */
        __pyx_v_self->path_index = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1688
 *                     # This dir info has been handled, go to the next
 *                     self.path_index = 0
 *                     self.current_dir_list = None             # <<<<<<<<<<<<<<
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)
 */
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->current_dir_list);
        __Pyx_DECREF(__pyx_v_self->current_dir_list);
        __pyx_v_self->current_dir_list = Py_None;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1689
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_9);
          /*try:*/ {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1690
 *                     self.current_dir_list = None
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:
 */
            __pyx_t_4 = __pyx_v_self->dir_iterator;
            __Pyx_INCREF(__pyx_t_4);
            __pyx_t_17 = __Pyx_PyIter_Next(__pyx_t_4); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1690, __pyx_L106_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GIVEREF(__pyx_t_17);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = __pyx_t_17;
            __pyx_t_17 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1691
 *                     try:
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                     except StopIteration, _:
 *                         self.current_dir_info = None
 */
            __pyx_t_17 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1691, __pyx_L106_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GIVEREF(__pyx_t_17);
            __Pyx_GOTREF(__pyx_v_self->current_dir_list);
            __Pyx_DECREF(__pyx_v_self->current_dir_list);
            __pyx_v_self->current_dir_list = __pyx_t_17;
            __pyx_t_17 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1689
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
 */
          }
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          goto __pyx_L113_try_end;
          __pyx_L106_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1692
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = None
 *                 else: #(dircmp > 0)
 */
          __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
          if (__pyx_t_8) {
            __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_17, &__pyx_t_4, &__pyx_t_16) < 0) __PYX_ERR(0, 1692, __pyx_L108_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_4);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1693
 *                         self.current_dir_list = self.current_dir_info[1]
 *                     except StopIteration, _:
 *                         self.current_dir_info = None             # <<<<<<<<<<<<<<
 *                 else: #(dircmp > 0)
 *                     # We have a dirblock entry for this location, but there
 */
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            __Pyx_GOTREF(__pyx_v_self->current_dir_info);
            __Pyx_DECREF(__pyx_v_self->current_dir_info);
            __pyx_v_self->current_dir_info = Py_None;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            goto __pyx_L107_exception_handled;
          }
          goto __pyx_L108_except_error;
          __pyx_L108_except_error:;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1689
 *                     self.path_index = 0
 *                     self.current_dir_list = None
 *                     try:             # <<<<<<<<<<<<<<
 *                         self.current_dir_info = next(self.dir_iterator)
 *                         self.current_dir_list = self.current_dir_info[1]
 */
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
          goto __pyx_L1_error;
          __pyx_L107_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_9);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_9);
          __pyx_L113_try_end:;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1645
 *                 # evaluated once per path: we could cache the result before
 *                 # doing the while loop and probably save time.
 *                 if _cmp_by_dirs(current_dirname_c,             # <<<<<<<<<<<<<<
 *                     PyBytes_Size(current_dirname),
 *                     current_blockname_c,
 */
        goto __pyx_L95;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1702
 *                     # need to handle all of the files that are contained
 *                     # within.
 *                     while self.current_block_pos < len(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                         self.current_block_pos = self.current_block_pos + 1
 */
      /*else*/ {
        while (1) {
          __pyx_t_16 = __pyx_v_self->current_block_list;
          __Pyx_INCREF(__pyx_t_16);
          __pyx_t_12 = PyObject_Length(__pyx_t_16); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1702, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_7 = ((__pyx_v_self->current_block_pos < __pyx_t_12) != 0);
          if (!__pyx_t_7) break;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1703
 *                     # within.
 *                     while self.current_block_pos < len(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]             # <<<<<<<<<<<<<<
 *                         self.current_block_pos = self.current_block_pos + 1
 *                         # entry referring to file not present on disk.
 */
          __pyx_t_16 = __Pyx_GetItemInt(__pyx_v_self->current_block_list, __pyx_v_self->current_block_pos, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1703, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_XDECREF_SET(__pyx_v_current_entry, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1704
 *                     while self.current_block_pos < len(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                         self.current_block_pos = self.current_block_pos + 1             # <<<<<<<<<<<<<<
 *                         # entry referring to file not present on disk.
 *                         # advance the entry only, after processing.
 */
          __pyx_v_self->current_block_pos = (__pyx_v_self->current_block_pos + 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1707
 *                         # entry referring to file not present on disk.
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)             # <<<<<<<<<<<<<<
 *                         if changed is not None:
 *                             if changed:
 */
          __pyx_t_16 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, Py_None); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1707, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if ((likely(PyTuple_CheckExact(__pyx_t_16))) || (PyList_CheckExact(__pyx_t_16))) {
            PyObject* sequence = __pyx_t_16;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1707, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_17);
            #else
            __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1707, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_17 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1707, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_17);
            #endif
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_3 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1707, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext;
            index = 0; __pyx_t_4 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L118_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 1; __pyx_t_17 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_17)) goto __pyx_L118_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_17);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1707, __pyx_L1_error)
            __pyx_t_6 = NULL;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            goto __pyx_L119_unpacking_done;
            __pyx_L118_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_6 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1707, __pyx_L1_error)
            __pyx_L119_unpacking_done:;
          }
          __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_17);
          __pyx_t_17 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1708
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 */
          __pyx_t_7 = (__pyx_v_changed != Py_None);
          __pyx_t_2 = (__pyx_t_7 != 0);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1709
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:
 *                             if changed:             # <<<<<<<<<<<<<<
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
 */
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1709, __pyx_L1_error)
            if (__pyx_t_2) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1710
 *                         if changed is not None:
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                             if changed or self.include_unchanged:
 *                                 return result
 */
              __pyx_t_8 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1710, __pyx_L1_error)

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1709
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:
 *                             if changed:             # <<<<<<<<<<<<<<
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
 */
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1711
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 return result
 *                     self.block_index = self.block_index + 1
 */
            __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1711, __pyx_L1_error)
            if (!__pyx_t_7) {
            } else {
              __pyx_t_2 = __pyx_t_7;
              goto __pyx_L123_bool_binop_done;
            }
            __pyx_t_7 = (__pyx_v_self->include_unchanged != 0);
            __pyx_t_2 = __pyx_t_7;
            __pyx_L123_bool_binop_done:;
            if (__pyx_t_2) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1712
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:
 *                                 return result             # <<<<<<<<<<<<<<
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()
 */
              __Pyx_XDECREF(__pyx_r);
              __Pyx_INCREF(__pyx_v_result);
              __pyx_r = __pyx_v_result;
              goto __pyx_L0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1711
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 *                             if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 return result
 *                     self.block_index = self.block_index + 1
 */
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1708
 *                         # advance the entry only, after processing.
 *                         result, changed = self._process_entry(current_entry, None)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             if changed:
 *                                 self._gather_result_for_consistency(result)
 */
          }
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1713
 *                             if changed or self.include_unchanged:
 *                                 return result
 *                     self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir
 */
        __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1714
 *                                 return result
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()             # <<<<<<<<<<<<<<
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
 */
        __pyx_t_8 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 1714, __pyx_L1_error)
      }
      __pyx_L95:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1715
 *                     self.block_index = self.block_index + 1
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir             # <<<<<<<<<<<<<<
 *             result = self._loop_one_block()
 *             if result is not None:
 */
      goto __pyx_L87_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1631
 *             or self.current_block is not None):
 *             # Uncommon case - a missing directory or an unversioned directory:
 *             if (self.current_dir_info and self.current_block             # <<<<<<<<<<<<<<
 *                 and self.current_dir_info[0][0] != self.current_block[0]):
 *                 # Work around pyrex broken heuristic - current_dirname has
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1716
 *                     self._update_current_block()
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()             # <<<<<<<<<<<<<<
 *             if result is not None:
 *                 return result
 */
    __pyx_t_16 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_loop_one_block(__pyx_v_self); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1716, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_16);
    __pyx_t_16 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1717
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         if len(self.search_specific_files):
 */
    __pyx_t_2 = (__pyx_v_result != Py_None);
    __pyx_t_7 = (__pyx_t_2 != 0);
    if (__pyx_t_7) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1718
 *             result = self._loop_one_block()
 *             if result is not None:
 *                 return result             # <<<<<<<<<<<<<<
 *         if len(self.search_specific_files):
 *             # More supplied paths to process
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1717
 *                 continue # next loop-on-block/dir
 *             result = self._loop_one_block()
 *             if result is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         if len(self.search_specific_files):
 */
    }
    __pyx_L87_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1719
 *             if result is not None:
 *                 return result
 *         if len(self.search_specific_files):             # <<<<<<<<<<<<<<
 *             # More supplied paths to process
 *             self.current_root = None
 */
  __pyx_t_16 = __pyx_v_self->search_specific_files;
  __Pyx_INCREF(__pyx_t_16);
  __pyx_t_12 = PyObject_Length(__pyx_t_16); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_7 = (__pyx_t_12 != 0);
  if (__pyx_t_7) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1721
 *         if len(self.search_specific_files):
 *             # More supplied paths to process
 *             self.current_root = None             # <<<<<<<<<<<<<<
 *             return self._iter_next()
 *         # Start expanding more conservatively, adding paths the user may not
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_root);
    __Pyx_DECREF(__pyx_v_self->current_root);
    __pyx_v_self->current_root = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1722
 *             # More supplied paths to process
 *             self.current_root = None
 *             return self._iter_next()             # <<<<<<<<<<<<<<
 *         # Start expanding more conservatively, adding paths the user may not
 *         # have intended but required for consistent deltas.
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_16 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_iter_next(__pyx_v_self); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_r = __pyx_t_16;
    __pyx_t_16 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1719
 *             if result is not None:
 *                 return result
 *         if len(self.search_specific_files):             # <<<<<<<<<<<<<<
 *             # More supplied paths to process
 *             self.current_root = None
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1725
 *         # Start expanding more conservatively, adding paths the user may not
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1             # <<<<<<<<<<<<<<
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()
 */
  __pyx_v_self->doing_consistency_expansion = 1;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1726
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->_pending_consistent_entries); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1726, __pyx_L1_error)
  __pyx_t_2 = ((!__pyx_t_7) != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1727
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()             # <<<<<<<<<<<<<<
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
 */
    __pyx_t_16 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_next_consistent_entries(__pyx_v_self); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_16);
    __Pyx_GOTREF(__pyx_v_self->_pending_consistent_entries);
    __Pyx_DECREF(__pyx_v_self->_pending_consistent_entries);
    __pyx_v_self->_pending_consistent_entries = __pyx_t_16;
    __pyx_t_16 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1726
 *         # have intended but required for consistent deltas.
 *         self.doing_consistency_expansion = 1
 *         if not self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1728
 *         if not self._pending_consistent_entries:
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:             # <<<<<<<<<<<<<<
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->_pending_consistent_entries); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1728, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1729
 *             self._pending_consistent_entries = self._next_consistent_entries()
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()             # <<<<<<<<<<<<<<
 *             if changed is not None:
 *                 return result
 */
    __pyx_t_16 = __Pyx_PyObject_Pop(__pyx_v_self->_pending_consistent_entries); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1729, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if ((likely(PyTuple_CheckExact(__pyx_t_16))) || (PyList_CheckExact(__pyx_t_16))) {
      PyObject* sequence = __pyx_t_16;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1729, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_17 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_17 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_17 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1729, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1729, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1729, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_17 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_17)) goto __pyx_L130_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_17);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L130_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1729, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L131_unpacking_done;
      __pyx_L130_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1729, __pyx_L1_error)
      __pyx_L131_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_17);
    __pyx_t_17 = 0;
    __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1730
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         raise StopIteration()
 */
    __pyx_t_2 = (__pyx_v_changed != Py_None);
    __pyx_t_7 = (__pyx_t_2 != 0);
    if (__pyx_t_7) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1731
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:
 *                 return result             # <<<<<<<<<<<<<<
 *         raise StopIteration()
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1730
 *         while self._pending_consistent_entries:
 *             result, changed = self._pending_consistent_entries.pop()
 *             if changed is not None:             # <<<<<<<<<<<<<<
 *                 return result
 *         raise StopIteration()
 */
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1732
 *             if changed is not None:
 *                 return result
 *         raise StopIteration()             # <<<<<<<<<<<<<<
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 */
  __pyx_t_16 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_Raise(__pyx_t_16, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __PYX_ERR(0, 1732, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1442
 *         return self._iter_next()
 * 
 *     cdef _iter_next(self):             # <<<<<<<<<<<<<<
 *         """Iterate over the changes."""
 *         # This function single steps through an iterator. As such while loops
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._iter_next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_dirname);
  __Pyx_XDECREF(__pyx_v_current_blockname);
  __Pyx_XDECREF(__pyx_v_searched_specific_files);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_root_stat);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_new_executable);
  __Pyx_XDECREF(__pyx_v_e_winerror);
  __Pyx_XDECREF(__pyx_v_win_errors);
  __Pyx_XDECREF(__pyx_v_bzr_index);
  __Pyx_XDECREF(__pyx_v_initial_key);
  __Pyx_XDECREF(__pyx_v_current_path_info);
  __Pyx_XDECREF(__pyx_v_current_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1734
 *         raise StopIteration()
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):             # <<<<<<<<<<<<<<
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0])[0]):
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_current_path_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_maybe_tree_ref", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1735
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0])[0]):
 *             return current_path_info[:2] + \
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_directory_is_tree_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1736
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0])[0]):             # <<<<<<<<<<<<<<
 *             return current_path_info[:2] + \
 *                 ('tree-reference',) + current_path_info[3:]
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->utf8_decode);
  __pyx_t_5 = __pyx_v_self->utf8_decode; __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_3, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1735
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0])[0]):
 *             return current_path_info[:2] + \
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1737
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0])[0]):
 *             return current_path_info[:2] + \             # <<<<<<<<<<<<<<
 *                 ('tree-reference',) + current_path_info[3:]
 *         else:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 0, 2, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1738
 *             self.utf8_decode(current_path_info[0])[0]):
 *             return current_path_info[:2] + \
 *                 ('tree-reference',) + current_path_info[3:]             # <<<<<<<<<<<<<<
 *         else:
 *             return current_path_info
 */
    __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_current_path_info, 3, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1738, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1735
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):
 *         if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *             self.utf8_decode(current_path_info[0])[0]):
 *             return current_path_info[:2] + \
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1740
 *                 ('tree-reference',) + current_path_info[3:]
 *         else:
 *             return current_path_info             # <<<<<<<<<<<<<<
 * 
 *     cdef object _loop_one_block(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_current_path_info);
    __pyx_r = __pyx_v_current_path_info;
    goto __pyx_L0;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1734
 *         raise StopIteration()
 * 
 *     cdef object _maybe_tree_ref(self, current_path_info):             # <<<<<<<<<<<<<<
 *         if self.tree._directory_is_tree_reference(
 *             self.utf8_decode(current_path_info[0])[0]):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._maybe_tree_ref", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1742
 *             return current_path_info
 * 
 *     cdef object _loop_one_block(self):             # <<<<<<<<<<<<<<
 *             # current_dir_info and current_block refer to the same directory -
 *             # this is the common case code.
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_current_entry = 0;
  PyObject *__pyx_v_current_path_info = 0;
  int __pyx_v_path_handled;
  char __pyx_v_minikind;
  int __pyx_v_cmp_result;
  PyObject *__pyx_v_disk_kind = NULL;
  long __pyx_v_advance_entry;
  long __pyx_v_advance_path;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  PyObject *__pyx_v_new_executable = NULL;
  PyObject *__pyx_v_relpath_unicode = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  char __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  Py_ssize_t __pyx_t_19;
  __Pyx_RefNannySetupContext("_loop_one_block", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1755
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 */
  __pyx_t_2 = (__pyx_v_self->current_block != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1756
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):             # <<<<<<<<<<<<<<
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 *                     self.current_block_pos)
 */
  __pyx_t_4 = __pyx_v_self->current_block_list;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_3 = ((__pyx_v_self->current_block_pos < PyList_GET_SIZE(__pyx_t_4)) != 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1755
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 */
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1757
 *             if (self.current_block is not None and
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,             # <<<<<<<<<<<<<<
 *                     self.current_block_pos)
 *                 # accomodate pyrex
 */
    __pyx_t_4 = __pyx_v_self->current_block_list;
    __Pyx_INCREF(__pyx_t_4);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1758
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 *                     self.current_block_pos)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(current_entry)
 */
    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_v_self->current_block_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_current_entry = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1760
 *                     self.current_block_pos)
 *                 # accomodate pyrex
 *                 Py_INCREF(current_entry)             # <<<<<<<<<<<<<<
 *             else:
 *                 current_entry = None
 */
    Py_INCREF(__pyx_v_current_entry);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1755
 *             # PyBytes_AsStringAndSize(disk_kind, &temp_str, &temp_str_length)
 *             # if not strncmp(temp_str, "directory", temp_str_length):
 *             if (self.current_block is not None and             # <<<<<<<<<<<<<<
 *                 self.current_block_pos < PyList_GET_SIZE(self.current_block_list)):
 *                 current_entry = PyList_GET_ITEM(self.current_block_list,
 */
    goto __pyx_L3;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1762
 *                 Py_INCREF(current_entry)
 *             else:
 *                 current_entry = None             # <<<<<<<<<<<<<<
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_current_entry = Py_None;
  }
  __pyx_L3:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1763
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 */
  __pyx_t_3 = (__pyx_v_self->current_dir_info != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1764
 *                 current_entry = None
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):             # <<<<<<<<<<<<<<
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 *                     self.path_index)
 */
  __pyx_t_5 = __pyx_v_self->current_dir_list;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_2 = ((__pyx_v_self->path_index < PyList_GET_SIZE(__pyx_t_5)) != 0);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1763
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 */
  if (__pyx_t_1) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1765
 *             if (self.current_dir_info is not None and
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,             # <<<<<<<<<<<<<<
 *                     self.path_index)
 *                 # accomodate pyrex
 */
    __pyx_t_5 = __pyx_v_self->current_dir_list;
    __Pyx_INCREF(__pyx_t_5);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1766
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 *                     self.path_index)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_v_self->path_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_current_path_info = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1768
 *                     self.path_index)
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)             # <<<<<<<<<<<<<<
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
 *                 # accomodate pyrex
 */
    Py_INCREF(__pyx_v_current_path_info);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1769
 *                 # accomodate pyrex
 *                 Py_INCREF(current_path_info)
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)             # <<<<<<<<<<<<<<
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
 */
    __pyx_t_4 = PyTuple_GET_ITEM(__pyx_v_current_path_info, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_disk_kind = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1771
 *                 disk_kind = PyTuple_GET_ITEM(current_path_info, 2)
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)             # <<<<<<<<<<<<<<
 *                 if disk_kind == "directory":
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 */
    Py_INCREF(__pyx_v_disk_kind);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1772
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":             # <<<<<<<<<<<<<<
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
 */
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_disk_kind, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1772, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1773
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":
 *                     current_path_info = self._maybe_tree_ref(current_path_info)             # <<<<<<<<<<<<<<
 *             else:
 *                 current_path_info = None
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_maybe_tree_ref(__pyx_v_self, __pyx_v_current_path_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1772
 *                 # accomodate pyrex
 *                 Py_INCREF(disk_kind)
 *                 if disk_kind == "directory":             # <<<<<<<<<<<<<<
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1763
 *             else:
 *                 current_entry = None
 *             if (self.current_dir_info is not None and             # <<<<<<<<<<<<<<
 *                 self.path_index < PyList_GET_SIZE(self.current_dir_list)):
 *                 current_path_info = PyList_GET_ITEM(self.current_dir_list,
 */
    goto __pyx_L6;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1775
 *                     current_path_info = self._maybe_tree_ref(current_path_info)
 *             else:
 *                 current_path_info = None             # <<<<<<<<<<<<<<
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_current_path_info = Py_None;
  }
  __pyx_L6:;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1776
 *             else:
 *                 current_path_info = None
 *             while (current_entry is not None or current_path_info is not None):             # <<<<<<<<<<<<<<
 *                 advance_entry = -1
 *                 advance_path = -1
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_current_entry != Py_None);
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_current_path_info != Py_None);
    __pyx_t_2 = (__pyx_t_3 != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1777
 *                 current_path_info = None
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1             # <<<<<<<<<<<<<<
 *                 advance_path = -1
 *                 result = None
 */
    __pyx_v_advance_entry = -1L;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1778
 *             while (current_entry is not None or current_path_info is not None):
 *                 advance_entry = -1
 *                 advance_path = -1             # <<<<<<<<<<<<<<
 *                 result = None
 *                 changed = None
 */
    __pyx_v_advance_path = -1L;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1779
 *                 advance_entry = -1
 *                 advance_path = -1
 *                 result = None             # <<<<<<<<<<<<<<
 *                 changed = None
 *                 path_handled = 0
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_result, Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1780
 *                 advance_path = -1
 *                 result = None
 *                 changed = None             # <<<<<<<<<<<<<<
 *                 path_handled = 0
 *                 if current_entry is None:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_XDECREF_SET(__pyx_v_changed, Py_None);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1781
 *                 result = None
 *                 changed = None
 *                 path_handled = 0             # <<<<<<<<<<<<<<
 *                 if current_entry is None:
 *                     # unversioned -  the check for path_handled when the path
 */
    __pyx_v_path_handled = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1782
 *                 changed = None
 *                 path_handled = 0
 *                 if current_entry is None:             # <<<<<<<<<<<<<<
 *                     # unversioned -  the check for path_handled when the path
 *                     # is advanced will yield this path if needed.
 */
    __pyx_t_1 = (__pyx_v_current_entry == Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {
      goto __pyx_L14;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1786
 *                     # is advanced will yield this path if needed.
 *                     pass
 *                 elif current_path_info is None:             # <<<<<<<<<<<<<<
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,
 */
    __pyx_t_2 = (__pyx_v_current_path_info == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1788
 *                 elif current_path_info is None:
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                         current_path_info)
 *                 else:
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, __pyx_v_current_path_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1788, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
        PyObject* sequence = __pyx_t_4;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1788, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1788, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L15_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 1788, __pyx_L1_error)
        __pyx_t_8 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L16_unpacking_done;
        __pyx_L15_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1788, __pyx_L1_error)
        __pyx_L16_unpacking_done:;
      }
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_5);
      __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1786
 *                     # is advanced will yield this path if needed.
 *                     pass
 *                 elif current_path_info is None:             # <<<<<<<<<<<<<<
 *                     # no path is fine: the per entry code will handle it.
 *                     result, changed = self._process_entry(current_entry,
 */
      goto __pyx_L14;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1791
 *                         current_path_info)
 *                 else:
 *                     minikind = _minikind_from_string(             # <<<<<<<<<<<<<<
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 */
    /*else*/ {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1792
 *                 else:
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])             # <<<<<<<<<<<<<<
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1792, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1792, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1792, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1791
 *                         current_path_info)
 *                 else:
 *                     minikind = _minikind_from_string(             # <<<<<<<<<<<<<<
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 */
      __pyx_t_9 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx__minikind_from_string(__pyx_t_4); if (unlikely(__pyx_t_9 == ((char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1791, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_minikind = __pyx_t_9;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1793
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -             # <<<<<<<<<<<<<<
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_current_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1794
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))             # <<<<<<<<<<<<<<
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
 *                         # The current path on disk doesn't match the dirblock
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_5, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1794, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1793
 *                     minikind = _minikind_from_string(
 *                         current_entry[1][self.target_index][0])
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -             # <<<<<<<<<<<<<<
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):
 */
      __pyx_t_7 = PyNumber_Subtract(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1793, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_v_cmp_result = __pyx_t_10;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1795
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):             # <<<<<<<<<<<<<<
 *                         # The current path on disk doesn't match the dirblock
 *                         # record. Either the dirblock record is marked as
 */
      __pyx_t_2 = (__pyx_v_cmp_result != 0);
      if (!__pyx_t_2) {
      } else {
        __pyx_t_1 = __pyx_t_2;
        goto __pyx_L18_bool_binop_done;
      }
      switch (__pyx_v_minikind) {
        case 'a':
        case 'r':
        __pyx_t_2 = 1;
        break;
        default:
        __pyx_t_2 = 0;
        break;
      }
      __pyx_t_1 = __pyx_t_2;
      __pyx_L18_bool_binop_done:;
      if (__pyx_t_1) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1804
 *                         # Compare the basename for these files to determine
 *                         # which comes first
 *                         if cmp_result < 0:             # <<<<<<<<<<<<<<
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
 */
        __pyx_t_1 = ((__pyx_v_cmp_result < 0) != 0);
        if (__pyx_t_1) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1807
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
 *                             advance_entry = 0             # <<<<<<<<<<<<<<
 *                         else:
 *                             # entry referring to file not present on disk.
 */
          __pyx_v_advance_entry = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1804
 *                         # Compare the basename for these files to determine
 *                         # which comes first
 *                         if cmp_result < 0:             # <<<<<<<<<<<<<<
 *                             # extra file on disk: pass for now, but only
 *                             # increment the path, not the entry
 */
          goto __pyx_L20;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1811
 *                             # entry referring to file not present on disk.
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                                 None)
 *                             advance_path = 0
 */
        /*else*/ {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1812
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,
 *                                 None)             # <<<<<<<<<<<<<<
 *                             advance_path = 0
 *                     else:
 */
          __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, Py_None); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1811, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
            PyObject* sequence = __pyx_t_7;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1811, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_6);
            #else
            __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1811, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1811, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1811, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext;
            index = 0; __pyx_t_4 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L21_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L21_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_6);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1811, __pyx_L1_error)
            __pyx_t_8 = NULL;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            goto __pyx_L22_unpacking_done;
            __pyx_L21_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_8 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1811, __pyx_L1_error)
            __pyx_L22_unpacking_done:;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1811
 *                             # entry referring to file not present on disk.
 *                             # advance the entry only, after processing.
 *                             result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                                 None)
 *                             advance_path = 0
 */
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1813
 *                             result, changed = self._process_entry(current_entry,
 *                                 None)
 *                             advance_path = 0             # <<<<<<<<<<<<<<
 *                     else:
 *                         # paths are the same,and the dirstate entry is not
 */
          __pyx_v_advance_path = 0;
        }
        __pyx_L20:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1795
 *                     cmp_result = ((current_path_info[1] > current_entry[0][1]) -
 *                                   (current_path_info[1] < current_entry[0][1]))
 *                     if (cmp_result or minikind == b'a' or minikind == b'r'):             # <<<<<<<<<<<<<<
 *                         # The current path on disk doesn't match the dirblock
 *                         # record. Either the dirblock record is marked as
 */
        goto __pyx_L17;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1817
 *                         # paths are the same,and the dirstate entry is not
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                             current_path_info)
 *                         if changed is not None:
 */
      /*else*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1818
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)             # <<<<<<<<<<<<<<
 *                         if changed is not None:
 *                             path_handled = -1
 */
        __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_current_entry, __pyx_v_current_path_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1817, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1817, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          #else
          __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1817, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1817, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1817, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext;
          index = 0; __pyx_t_6 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_6)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 1; __pyx_t_4 = __pyx_t_8(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L23_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_5), 2) < 0) __PYX_ERR(0, 1817, __pyx_L1_error)
          __pyx_t_8 = NULL;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          goto __pyx_L24_unpacking_done;
          __pyx_L23_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_8 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1817, __pyx_L1_error)
          __pyx_L24_unpacking_done:;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1817
 *                         # paths are the same,and the dirstate entry is not
 *                         # absent or renamed.
 *                         result, changed = self._process_entry(current_entry,             # <<<<<<<<<<<<<<
 *                             current_path_info)
 *                         if changed is not None:
 */
        __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_DECREF_SET(__pyx_v_changed, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1819
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
 */
        __pyx_t_1 = (__pyx_v_changed != Py_None);
        __pyx_t_2 = (__pyx_t_1 != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1820
 *                             current_path_info)
 *                         if changed is not None:
 *                             path_handled = -1             # <<<<<<<<<<<<<<
 *                             if not changed and not self.include_unchanged:
 *                                 changed = None
 */
          __pyx_v_path_handled = -1;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1821
 *                         if changed is not None:
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 changed = None
 *                 # >- loop control starts here:
 */
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1821, __pyx_L1_error)
          __pyx_t_3 = ((!__pyx_t_1) != 0);
          if (__pyx_t_3) {
          } else {
            __pyx_t_2 = __pyx_t_3;
            goto __pyx_L27_bool_binop_done;
          }
          __pyx_t_3 = ((!(__pyx_v_self->include_unchanged != 0)) != 0);
          __pyx_t_2 = __pyx_t_3;
          __pyx_L27_bool_binop_done:;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1822
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
 *                                 changed = None             # <<<<<<<<<<<<<<
 *                 # >- loop control starts here:
 *                 # >- entry
 */
            __Pyx_INCREF(Py_None);
            __Pyx_DECREF_SET(__pyx_v_changed, Py_None);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1821
 *                         if changed is not None:
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:             # <<<<<<<<<<<<<<
 *                                 changed = None
 *                 # >- loop control starts here:
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1819
 *                         result, changed = self._process_entry(current_entry,
 *                             current_path_info)
 *                         if changed is not None:             # <<<<<<<<<<<<<<
 *                             path_handled = -1
 *                             if not changed and not self.include_unchanged:
 */
        }
      }
      __pyx_L17:;
    }
    __pyx_L14:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1825
 *                 # >- loop control starts here:
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:             # <<<<<<<<<<<<<<
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 */
    __pyx_t_3 = (__pyx_v_advance_entry != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L30_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_current_entry != Py_None);
    __pyx_t_1 = (__pyx_t_3 != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L30_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1826
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1             # <<<<<<<<<<<<<<
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]
 */
      __pyx_v_self->current_block_pos = (__pyx_v_self->current_block_pos + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1827
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
 */
      __pyx_t_7 = __pyx_v_self->current_block_list;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_2 = ((__pyx_v_self->current_block_pos < PyList_GET_SIZE(__pyx_t_7)) != 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1828
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 *                         current_entry = self.current_block_list[self.current_block_pos]             # <<<<<<<<<<<<<<
 *                     else:
 *                         current_entry = None
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->current_block_list, __pyx_v_self->current_block_pos, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1828, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_current_entry, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1827
 *                 if advance_entry and current_entry is not None:
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):             # <<<<<<<<<<<<<<
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
 */
        goto __pyx_L32;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1830
 *                         current_entry = self.current_block_list[self.current_block_pos]
 *                     else:
 *                         current_entry = None             # <<<<<<<<<<<<<<
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
 */
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_current_entry, Py_None);
      }
      __pyx_L32:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1825
 *                 # >- loop control starts here:
 *                 # >- entry
 *                 if advance_entry and current_entry is not None:             # <<<<<<<<<<<<<<
 *                     self.current_block_pos = self.current_block_pos + 1
 *                     if self.current_block_pos < PyList_GET_SIZE(self.current_block_list):
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1832
 *                         current_entry = None
 *                 # >- path
 *                 if advance_path and current_path_info is not None:             # <<<<<<<<<<<<<<
 *                     if not path_handled:
 *                         # unversioned in all regards
 */
    __pyx_t_1 = (__pyx_v_advance_path != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L34_bool_binop_done;
    }
    __pyx_t_1 = (__pyx_v_current_path_info != Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L34_bool_binop_done:;
    if (__pyx_t_2) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1833
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
 *                     if not path_handled:             # <<<<<<<<<<<<<<
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
 */
      __pyx_t_2 = ((!(__pyx_v_path_handled != 0)) != 0);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1835
 *                     if not path_handled:
 *                         # unversioned in all regards
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 */
        __pyx_t_2 = (__pyx_v_self->want_unversioned != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1837
 *                         if self.want_unversioned:
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             try:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_stat); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_S_ISREG); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1837, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_3) {
          } else {
            __pyx_t_2 = __pyx_t_3;
            goto __pyx_L38_bool_binop_done;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1838
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)             # <<<<<<<<<<<<<<
 *                             try:
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 */
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_stat); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_S_IEXEC); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_current_path_info, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyNumber_And(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1838, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_2 = __pyx_t_3;
          __pyx_L38_bool_binop_done:;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1836
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
 *                             new_executable = bool(             # <<<<<<<<<<<<<<
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 */
          __pyx_t_7 = __Pyx_PyBool_FromLong((!(!__pyx_t_2))); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1836, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_XDECREF_SET(__pyx_v_new_executable, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1839
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             try:             # <<<<<<<<<<<<<<
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 *                             except UnicodeDecodeError, _:
 */
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_11);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            /*try:*/ {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1840
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             try:
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]             # <<<<<<<<<<<<<<
 *                             except UnicodeDecodeError, _:
 *                                 raise errors.BadFilenameEncoding(
 */
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1840, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_v_self->utf8_decode);
              __pyx_t_6 = __pyx_v_self->utf8_decode; __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_6, function);
                }
              }
              __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1840, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
              __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1840, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_XDECREF_SET(__pyx_v_relpath_unicode, __pyx_t_6);
              __pyx_t_6 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1839
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             try:             # <<<<<<<<<<<<<<
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 *                             except UnicodeDecodeError, _:
 */
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L47_try_end;
            __pyx_L40_error:;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1841
 *                             try:
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 *                             except UnicodeDecodeError, _:             # <<<<<<<<<<<<<<
 *                                 raise errors.BadFilenameEncoding(
 *                                     current_path_info[0], osutils._fs_enc)
 */
            __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
            if (__pyx_t_10) {
              __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._loop_one_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_5) < 0) __PYX_ERR(0, 1841, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_6);
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_7);
              __pyx_v__ = __pyx_t_7;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1842
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 *                             except UnicodeDecodeError, _:
 *                                 raise errors.BadFilenameEncoding(             # <<<<<<<<<<<<<<
 *                                     current_path_info[0], osutils._fs_enc)
 *                             if changed is not None:
 */
              __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_errors); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_BadFilenameEncoding); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1843
 *                             except UnicodeDecodeError, _:
 *                                 raise errors.BadFilenameEncoding(
 *                                     current_path_info[0], osutils._fs_enc)             # <<<<<<<<<<<<<<
 *                             if changed is not None:
 *                                 raise AssertionError(
 */
              __pyx_t_14 = __Pyx_GetItemInt(__pyx_v_current_path_info, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1843, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_osutils); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1843, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_fs_enc); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1843, __pyx_L42_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __pyx_t_16 = NULL;
              __pyx_t_10 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_16)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_16);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                  __pyx_t_10 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_15)) {
                PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_14, __pyx_t_17};
                __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
                PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_14, __pyx_t_17};
                __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              } else
              #endif
              {
                __pyx_t_18 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_18);
                if (__pyx_t_16) {
                  __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_16); __pyx_t_16 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_14);
                PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_10, __pyx_t_14);
                __Pyx_GIVEREF(__pyx_t_17);
                PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_10, __pyx_t_17);
                __pyx_t_14 = 0;
                __pyx_t_17 = 0;
                __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_18, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1842, __pyx_L42_except_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              }
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_Raise(__pyx_t_4, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __PYX_ERR(0, 1842, __pyx_L42_except_error)
            }
            goto __pyx_L42_except_error;
            __pyx_L42_except_error:;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1839
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 *                                 and stat.S_IEXEC & current_path_info[3].st_mode)
 *                             try:             # <<<<<<<<<<<<<<
 *                                 relpath_unicode = self.utf8_decode(current_path_info[0])[0]
 *                             except UnicodeDecodeError, _:
 */
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
            goto __pyx_L1_error;
            __pyx_L47_try_end:;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1844
 *                                 raise errors.BadFilenameEncoding(
 *                                     current_path_info[0], osutils._fs_enc)
 *                             if changed is not None:             # <<<<<<<<<<<<<<
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
 */
          __pyx_t_2 = (__pyx_v_changed != Py_None);
          __pyx_t_3 = (__pyx_t_2 != 0);
          if (unlikely(__pyx_t_3)) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1846
 *                             if changed is not None:
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)             # <<<<<<<<<<<<<<
 *                             result = TreeChange(
 *                                 None,
 */
            __pyx_t_5 = __Pyx_PyString_FormatSafe(__pyx_kp_s_result_is_not_None_r, __pyx_v_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1846, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1845
 *                                     current_path_info[0], osutils._fs_enc)
 *                             if changed is not None:
 *                                 raise AssertionError(             # <<<<<<<<<<<<<<
 *                                     "result is not None: %r" % result)
 *                             result = TreeChange(
 */
            __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1845, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_Raise(__pyx_t_7, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __PYX_ERR(0, 1845, __pyx_L1_error)

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1844
 *                                 raise errors.BadFilenameEncoding(
 *                                     current_path_info[0], osutils._fs_enc)
 *                             if changed is not None:             # <<<<<<<<<<<<<<
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1847
 *                                 raise AssertionError(
 *                                     "result is not None: %r" % result)
 *                             result = TreeChange(             # <<<<<<<<<<<<<<
 *                                 None,
 *                                 (None, relpath_unicode),
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1847, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1849
 *                             result = TreeChange(
 *                                 None,
 *                                 (None, relpath_unicode),             # <<<<<<<<<<<<<<
 *                                 True,
 *                                 (False, False),
 */
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1849, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          PyTuple_SET_ITEM(__pyx_t_6, 0, Py_None);
          __Pyx_INCREF(__pyx_v_relpath_unicode);
          __Pyx_GIVEREF(__pyx_v_relpath_unicode);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_relpath_unicode);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1853
 *                                 (False, False),
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),             # <<<<<<<<<<<<<<
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
 */
          __pyx_t_15 = __Pyx_GetItemInt(__pyx_v_current_path_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_INCREF(__pyx_v_self->utf8_decode);
          __pyx_t_18 = __pyx_v_self->utf8_decode; __pyx_t_17 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_18, function);
            }
          }
          __pyx_t_4 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_17, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_15);
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          __pyx_t_18 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1853, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          PyTuple_SET_ITEM(__pyx_t_4, 0, Py_None);
          __Pyx_GIVEREF(__pyx_t_18);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_18);
          __pyx_t_18 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1854
 *                                 (None, None),
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),
 *                                 (None, current_path_info[2]),             # <<<<<<<<<<<<<<
 *                                 (None, new_executable))
 *                             changed = True
 */
          __pyx_t_18 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1854, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1854, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          PyTuple_SET_ITEM(__pyx_t_15, 0, Py_None);
          __Pyx_GIVEREF(__pyx_t_18);
          PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_18);
          __pyx_t_18 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1855
 *                                 (None, self.utf8_decode(current_path_info[1])[0]),
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))             # <<<<<<<<<<<<<<
 *                             changed = True
 *                         # dont descend into this unversioned path if it is
 */
          __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1855, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_18);
          __Pyx_INCREF(Py_None);
          __Pyx_GIVEREF(Py_None);
          PyTuple_SET_ITEM(__pyx_t_18, 0, Py_None);
          __Pyx_INCREF(__pyx_v_new_executable);
          __Pyx_GIVEREF(__pyx_v_new_executable);
          PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_v_new_executable);
          __pyx_t_17 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
              __pyx_t_10 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[9] = {__pyx_t_17, Py_None, __pyx_t_6, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_4, __pyx_t_15, __pyx_t_18};
            __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 8+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1847, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[9] = {__pyx_t_17, Py_None, __pyx_t_6, Py_True, __pyx_tuple__14, __pyx_tuple__6, __pyx_t_4, __pyx_t_15, __pyx_t_18};
            __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_10, 8+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1847, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          } else
          #endif
          {
            __pyx_t_14 = PyTuple_New(8+__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1847, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            if (__pyx_t_17) {
              __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_17); __pyx_t_17 = NULL;
            }
            __Pyx_INCREF(Py_None);
            __Pyx_GIVEREF(Py_None);
            PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_10, Py_None);
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_10, __pyx_t_6);
            __Pyx_INCREF(Py_True);
            __Pyx_GIVEREF(Py_True);
            PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_10, Py_True);
            __Pyx_INCREF(__pyx_tuple__14);
            __Pyx_GIVEREF(__pyx_tuple__14);
            PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_10, __pyx_tuple__14);
            __Pyx_INCREF(__pyx_tuple__6);
            __Pyx_GIVEREF(__pyx_tuple__6);
            PyTuple_SET_ITEM(__pyx_t_14, 4+__pyx_t_10, __pyx_tuple__6);
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_14, 5+__pyx_t_10, __pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_14, 6+__pyx_t_10, __pyx_t_15);
            __Pyx_GIVEREF(__pyx_t_18);
            PyTuple_SET_ITEM(__pyx_t_14, 7+__pyx_t_10, __pyx_t_18);
            __pyx_t_6 = 0;
            __pyx_t_4 = 0;
            __pyx_t_15 = 0;
            __pyx_t_18 = 0;
            __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_14, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1847, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1856
 *                                 (None, current_path_info[2]),
 *                                 (None, new_executable))
 *                             changed = True             # <<<<<<<<<<<<<<
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 */
          __Pyx_INCREF(Py_True);
          __Pyx_DECREF_SET(__pyx_v_changed, Py_True);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1835
 *                     if not path_handled:
 *                         # unversioned in all regards
 *                         if self.want_unversioned:             # <<<<<<<<<<<<<<
 *                             new_executable = bool(
 *                                 stat.S_ISREG(current_path_info[3].st_mode)
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1859
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):             # <<<<<<<<<<<<<<
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1859, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1859, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1860
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):
 *                             del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                             self.path_index = self.path_index - 1
 *                     # dont descend the disk iterator into any tree
 */
          if (unlikely(__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1860, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1861
 *                         if current_path_info[2] in ('directory'):
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
 */
          __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1859
 *                         # dont descend into this unversioned path if it is
 *                         # a dir
 *                         if current_path_info[2] in ('directory'):             # <<<<<<<<<<<<<<
 *                             del self.current_dir_list[self.path_index]
 *                             self.path_index = self.path_index - 1
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1833
 *                 # >- path
 *                 if advance_path and current_path_info is not None:
 *                     if not path_handled:             # <<<<<<<<<<<<<<
 *                         # unversioned in all regards
 *                         if self.want_unversioned:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1864
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
 */
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_kp_s_tree_reference, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1864, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1865
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':
 *                         del self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
 */
        if (unlikely(__Pyx_DelItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0)) __PYX_ERR(0, 1865, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1866
 *                     if current_path_info[2] == 'tree-reference':
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1             # <<<<<<<<<<<<<<
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):
 */
        __pyx_v_self->path_index = (__pyx_v_self->path_index - 1);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1864
 *                     # dont descend the disk iterator into any tree
 *                     # paths.
 *                     if current_path_info[2] == 'tree-reference':             # <<<<<<<<<<<<<<
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1867
 *                         del self.current_dir_list[self.path_index]
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1             # <<<<<<<<<<<<<<
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
 */
      __pyx_v_self->path_index = (__pyx_v_self->path_index + 1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1868
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
 */
      __pyx_t_7 = __pyx_v_self->current_dir_list;
      __Pyx_INCREF(__pyx_t_7);
      __pyx_t_19 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_19 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1868, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_2 = ((__pyx_v_self->path_index < __pyx_t_19) != 0);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1869
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]             # <<<<<<<<<<<<<<
 *                         if current_path_info[2] == 'directory':
 *                             current_path_info = self._maybe_tree_ref(
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_self->current_dir_list, __pyx_v_self->path_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1869, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1870
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                             current_path_info = self._maybe_tree_ref(
 *                                 current_path_info)
 */
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_current_path_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1870, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1871
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
 *                             current_path_info = self._maybe_tree_ref(             # <<<<<<<<<<<<<<
 *                                 current_path_info)
 *                     else:
 */
          __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_maybe_tree_ref(__pyx_v_self, __pyx_v_current_path_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1871, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF_SET(__pyx_v_current_path_info, __pyx_t_7);
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1870
 *                     if self.path_index < len(self.current_dir_list):
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *                             current_path_info = self._maybe_tree_ref(
 *                                 current_path_info)
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1868
 *                         self.path_index = self.path_index - 1
 *                     self.path_index = self.path_index + 1
 *                     if self.path_index < len(self.current_dir_list):             # <<<<<<<<<<<<<<
 *                         current_path_info = self.current_dir_list[self.path_index]
 *                         if current_path_info[2] == 'directory':
 */
        goto __pyx_L53;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1874
 *                                 current_path_info)
 *                     else:
 *                         current_path_info = None             # <<<<<<<<<<<<<<
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
 */
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_DECREF_SET(__pyx_v_current_path_info, Py_None);
      }
      __pyx_L53:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1832
 *                         current_entry = None
 *                 # >- path
 *                 if advance_path and current_path_info is not None:             # <<<<<<<<<<<<<<
 *                     if not path_handled:
 *                         # unversioned in all regards
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1875
 *                     else:
 *                         current_path_info = None
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     # Found a result on this pass, yield it
 *                     if changed:
 */
    __pyx_t_2 = (__pyx_v_changed != Py_None);
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1877
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 */
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1877, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1878
 *                     # Found a result on this pass, yield it
 *                     if changed:
 *                         self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if changed or self.include_unchanged:
 *                         return result
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1878, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1877
 *                 if changed is not None:
 *                     # Found a result on this pass, yield it
 *                     if changed:             # <<<<<<<<<<<<<<
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1879
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             if self.current_block is not None:
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1879, __pyx_L1_error)
      if (!__pyx_t_2) {
      } else {
        __pyx_t_3 = __pyx_t_2;
        goto __pyx_L58_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_3 = __pyx_t_2;
      __pyx_L58_bool_binop_done:;
      if (__pyx_t_3) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1880
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:
 *                         return result             # <<<<<<<<<<<<<<
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_result);
        __pyx_r = __pyx_v_result;
        goto __pyx_L0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1879
 *                     if changed:
 *                         self._gather_result_for_consistency(result)
 *                     if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                         return result
 *             if self.current_block is not None:
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1875
 *                     else:
 *                         current_path_info = None
 *                 if changed is not None:             # <<<<<<<<<<<<<<
 *                     # Found a result on this pass, yield it
 *                     if changed:
 */
    }
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1881
 *                     if changed or self.include_unchanged:
 *                         return result
 *             if self.current_block is not None:             # <<<<<<<<<<<<<<
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 */
  __pyx_t_3 = (__pyx_v_self->current_block != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1882
 *                         return result
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1             # <<<<<<<<<<<<<<
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:
 */
    __pyx_v_self->block_index = (__pyx_v_self->block_index + 1);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1883
 *             if self.current_block is not None:
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()             # <<<<<<<<<<<<<<
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0
 */
    __pyx_t_10 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_update_current_block(__pyx_v_self); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1883, __pyx_L1_error)

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1881
 *                     if changed or self.include_unchanged:
 *                         return result
 *             if self.current_block is not None:             # <<<<<<<<<<<<<<
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1884
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 */
  __pyx_t_2 = (__pyx_v_self->current_dir_info != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1885
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0             # <<<<<<<<<<<<<<
 *                 self.current_dir_list = None
 *                 try:
 */
    __pyx_v_self->path_index = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1886
 *             if self.current_dir_info is not None:
 *                 self.path_index = 0
 *                 self.current_dir_list = None             # <<<<<<<<<<<<<<
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->current_dir_list);
    __Pyx_DECREF(__pyx_v_self->current_dir_list);
    __pyx_v_self->current_dir_list = Py_None;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1887
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      /*try:*/ {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1888
 *                 self.current_dir_list = None
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)             # <<<<<<<<<<<<<<
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:
 */
        __pyx_t_7 = __pyx_v_self->dir_iterator;
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_5 = __Pyx_PyIter_Next(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1888, __pyx_L62_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_v_self->current_dir_info);
        __Pyx_DECREF(__pyx_v_self->current_dir_info);
        __pyx_v_self->current_dir_info = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1889
 *                 try:
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]             # <<<<<<<<<<<<<<
 *                 except StopIteration, _:
 *                     self.current_dir_info = None
 */
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_self->current_dir_info, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1889, __pyx_L62_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_v_self->current_dir_list);
        __Pyx_DECREF(__pyx_v_self->current_dir_list);
        __pyx_v_self->current_dir_list = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1887
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L67_try_end;
      __pyx_L62_error:;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1890
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = None
 * 
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._loop_one_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_14) < 0) __PYX_ERR(0, 1890, __pyx_L64_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v__ = __pyx_t_7;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1891
 *                     self.current_dir_list = self.current_dir_info[1]
 *                 except StopIteration, _:
 *                     self.current_dir_info = None             # <<<<<<<<<<<<<<
 * 
 *     cdef object _next_consistent_entries(self):
 */
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->current_dir_info);
        __Pyx_DECREF(__pyx_v_self->current_dir_info);
        __pyx_v_self->current_dir_info = Py_None;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L63_exception_handled;
      }
      goto __pyx_L64_except_error;
      __pyx_L64_except_error:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1887
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 *                 try:             # <<<<<<<<<<<<<<
 *                     self.current_dir_info = next(self.dir_iterator)
 *                     self.current_dir_list = self.current_dir_info[1]
 */
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
      goto __pyx_L1_error;
      __pyx_L63_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ExceptionReset(__pyx_t_13, __pyx_t_12, __pyx_t_11);
      __pyx_L67_try_end:;
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1884
 *                 self.block_index = self.block_index + 1
 *                 self._update_current_block()
 *             if self.current_dir_info is not None:             # <<<<<<<<<<<<<<
 *                 self.path_index = 0
 *                 self.current_dir_list = None
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1742
 *             return current_path_info
 * 
 *     cdef object _loop_one_block(self):             # <<<<<<<<<<<<<<
 *             # current_dir_info and current_block refer to the same directory -
 *             # this is the common case code.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._loop_one_block", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current_entry);
  __Pyx_XDECREF(__pyx_v_current_path_info);
  __Pyx_XDECREF(__pyx_v_disk_kind);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_new_executable);
  __Pyx_XDECREF(__pyx_v_relpath_unicode);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1893
 *                     self.current_dir_info = None
 * 
 *     cdef object _next_consistent_entries(self):             # <<<<<<<<<<<<<<
 *         """Grabs the next specific file parent case to consider.
 * 
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_results = NULL;
  PyObject *__pyx_v_path_utf8 = NULL;
  PyObject *__pyx_v_path_entries = NULL;
  PyObject *__pyx_v_selected_entries = NULL;
  int __pyx_v_found_item;
  PyObject *__pyx_v_candidate_entry = NULL;
  PyObject *__pyx_v_path_info = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_changed = NULL;
  PyObject *__pyx_v_entry_path_utf8 = NULL;
  PyObject *__pyx_v_initial_key = NULL;
  PyObject *__pyx_v_block_index = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_current_block = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("_next_consistent_entries", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1898
 *         :return: A list of the results, each of which is as for _process_entry.
 *         """
 *         results = []             # <<<<<<<<<<<<<<
 *         while self.search_specific_file_parents:
 *             # Process the parent directories for the paths we were iterating.
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1899
 *         """
 *         results = []
 *         while self.search_specific_file_parents:             # <<<<<<<<<<<<<<
 *             # Process the parent directories for the paths we were iterating.
 *             # Even in extremely large trees this should be modest, so currently
 */
  while (1) {
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_self->search_specific_file_parents); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1899, __pyx_L1_error)
    if (!__pyx_t_2) break;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1903
 *             # Even in extremely large trees this should be modest, so currently
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()             # <<<<<<<<<<<<<<
 *             if path_utf8 in self.searched_exact_paths:
 *                 # We've examined this path.
 */
    __pyx_t_1 = __Pyx_PyObject_Pop(__pyx_v_self->search_specific_file_parents); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_path_utf8, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1904
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()
 *             if path_utf8 in self.searched_exact_paths:             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
 */
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_path_utf8, __pyx_v_self->searched_exact_paths, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1904, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1906
 *             if path_utf8 in self.searched_exact_paths:
 *                 # We've examined this path.
 *                 continue             # <<<<<<<<<<<<<<
 *             if is_inside_any(self.searched_specific_files, path_utf8):
 *                 # We've examined this path.
 */
      goto __pyx_L3_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1904
 *             # no attempt is made to optimise.
 *             path_utf8 = self.search_specific_file_parents.pop()
 *             if path_utf8 in self.searched_exact_paths:             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1907
 *                 # We've examined this path.
 *                 continue
 *             if is_inside_any(self.searched_specific_files, path_utf8):             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_is_inside_any); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self->searched_specific_files, __pyx_v_path_utf8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1907, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_v_self->searched_specific_files, __pyx_v_path_utf8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1907, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_self->searched_specific_files);
      __Pyx_GIVEREF(__pyx_v_self->searched_specific_files);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_self->searched_specific_files);
      __Pyx_INCREF(__pyx_v_path_utf8);
      __Pyx_GIVEREF(__pyx_v_path_utf8);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_path_utf8);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1907, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1909
 *             if is_inside_any(self.searched_specific_files, path_utf8):
 *                 # We've examined this path.
 *                 continue             # <<<<<<<<<<<<<<
 *             path_entries = self.state._entries_for_path(path_utf8)
 *             # We need either one or two entries. If the path in
 */
      goto __pyx_L3_continue;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1907
 *                 # We've examined this path.
 *                 continue
 *             if is_inside_any(self.searched_specific_files, path_utf8):             # <<<<<<<<<<<<<<
 *                 # We've examined this path.
 *                 continue
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1910
 *                 # We've examined this path.
 *                 continue
 *             path_entries = self.state._entries_for_path(path_utf8)             # <<<<<<<<<<<<<<
 *             # We need either one or two entries. If the path in
 *             # self.target_index has moved (so the entry in source_index is in
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_entries_for_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_v_path_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_path_utf8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_path_entries, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1915
 *             # 'ar') then we need to also look for the entry for this path in
 *             # self.source_index, to output the appropriate delete-or-rename.
 *             selected_entries = []             # <<<<<<<<<<<<<<
 *             found_item = False
 *             for candidate_entry in path_entries:
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_selected_entries, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1916
 *             # self.source_index, to output the appropriate delete-or-rename.
 *             selected_entries = []
 *             found_item = False             # <<<<<<<<<<<<<<
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
 */
    __pyx_v_found_item = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1917
 *             selected_entries = []
 *             found_item = False
 *             for candidate_entry in path_entries:             # <<<<<<<<<<<<<<
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 */
    if (likely(PyList_CheckExact(__pyx_v_path_entries)) || PyTuple_CheckExact(__pyx_v_path_entries)) {
      __pyx_t_1 = __pyx_v_path_entries; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_path_entries); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1917, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1917, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1917, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1917, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1917, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1917, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1917, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_candidate_entry, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1919
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1919, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_n_b_a, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1919, __pyx_L1_error)
      if (__pyx_t_2) {
      } else {
        __pyx_t_3 = __pyx_t_2;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_n_b_r, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1919, __pyx_L1_error)
      __pyx_t_3 = __pyx_t_2;
      __pyx_L10_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1920
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 *                     found_item = True             # <<<<<<<<<<<<<<
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 */
        __pyx_v_found_item = 1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1921
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)             # <<<<<<<<<<<<<<
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_selected_entries, __pyx_v_candidate_entry); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1921, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1919
 *             for candidate_entry in path_entries:
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     selected_entries.append(candidate_entry)
 */
        goto __pyx_L9;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1923
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1923, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = (__pyx_t_4 != Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = (__pyx_t_3 != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_2 = __pyx_t_11;
        goto __pyx_L12_bool_binop_done;
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1924
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):             # <<<<<<<<<<<<<<
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1924, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_n_b_a, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1924, __pyx_L1_error)
      if (__pyx_t_3) {
      } else {
        __pyx_t_11 = __pyx_t_3;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_n_b_r, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1924, __pyx_L1_error)
      __pyx_t_11 = __pyx_t_3;
      __pyx_L14_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_3 = (__pyx_t_11 != 0);
      __pyx_t_2 = __pyx_t_3;
      __pyx_L12_bool_binop_done:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1923
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 */
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1925
 *                 elif (self.source_index is not None and
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True             # <<<<<<<<<<<<<<
 *                     if candidate_entry[1][self.target_index][0] == b'a':
 *                         # Deleted, emit it here.
 */
        __pyx_v_found_item = 1;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1926
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':             # <<<<<<<<<<<<<<
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)
 */
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_4, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1926, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_4, __pyx_n_b_a, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1926, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1928
 *                     if candidate_entry[1][self.target_index][0] == b'a':
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)             # <<<<<<<<<<<<<<
 *                     else:
 *                         # renamed, emit it when we process the directory it
 */
          __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_selected_entries, __pyx_v_candidate_entry); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1928, __pyx_L1_error)

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1926
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 *                     if candidate_entry[1][self.target_index][0] == b'a':             # <<<<<<<<<<<<<<
 *                         # Deleted, emit it here.
 *                         selected_entries.append(candidate_entry)
 */
          goto __pyx_L16;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1932
 *                         # renamed, emit it when we process the directory it
 *                         # ended up at.
 *                         self.search_specific_file_parents.add(             # <<<<<<<<<<<<<<
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:
 */
        /*else*/ {
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_file_parents, __pyx_n_s_add); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1932, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1933
 *                         # ended up at.
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])             # <<<<<<<<<<<<<<
 *             if not found_item:
 *                 raise AssertionError(
 */
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_candidate_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1933, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_self->target_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1933, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_12, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1933, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          __pyx_t_4 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_12, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1932, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __pyx_L16:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1923
 *                     selected_entries.append(candidate_entry)
 *                 # Find entries present in source at this path:
 *                 elif (self.source_index is not None and             # <<<<<<<<<<<<<<
 *                     candidate_entry[1][self.source_index][0] not in (b'a', b'r')):
 *                     found_item = True
 */
      }
      __pyx_L9:;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1917
 *             selected_entries = []
 *             found_item = False
 *             for candidate_entry in path_entries:             # <<<<<<<<<<<<<<
 *                 # Find entries present in target at this path:
 *                 if candidate_entry[1][self.target_index][0] not in (b'a', b'r'):
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1934
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
 */
    __pyx_t_2 = ((!(__pyx_v_found_item != 0)) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1937
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))             # <<<<<<<<<<<<<<
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
 */
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1937, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_path_utf8);
      __Pyx_GIVEREF(__pyx_v_path_utf8);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_path_utf8);
      __Pyx_INCREF(__pyx_v_path_entries);
      __Pyx_GIVEREF(__pyx_v_path_entries);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_path_entries);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1936
 *             if not found_item:
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (             # <<<<<<<<<<<<<<
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 */
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Missing_entry_for_specific_path, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1936, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1935
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:
 *                 raise AssertionError(             # <<<<<<<<<<<<<<
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 1935, __pyx_L1_error)

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1934
 *                         self.search_specific_file_parents.add(
 *                             candidate_entry[1][self.target_index][1])
 *             if not found_item:             # <<<<<<<<<<<<<<
 *                 raise AssertionError(
 *                     "Missing entry for specific path parent %r, %r" % (
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1938
 *                     "Missing entry for specific path parent %r, %r" % (
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))             # <<<<<<<<<<<<<<
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_path_utf8, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_n_s_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_utf8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_path_info(__pyx_v_self, __pyx_v_path_utf8, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_path_info, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1939
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:             # <<<<<<<<<<<<<<
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
 */
    __pyx_t_4 = __pyx_v_selected_entries; __Pyx_INCREF(__pyx_t_4); __pyx_t_8 = 0;
    for (;;) {
      if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_4)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_1); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 1939, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1939, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1940
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:             # <<<<<<<<<<<<<<
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_1, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1940, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_7, __pyx_v_self->seen_ids, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1940, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1941
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:
 *                     continue             # <<<<<<<<<<<<<<
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:
 */
        goto __pyx_L18_continue;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1940
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:
 *                 if entry[0][2] in self.seen_ids:             # <<<<<<<<<<<<<<
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1942
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)             # <<<<<<<<<<<<<<
 *                 if changed is None:
 *                     raise AssertionError(
 */
      __pyx_t_7 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_process_entry(__pyx_v_self, __pyx_v_entry, __pyx_v_path_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1942, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
        PyObject* sequence = __pyx_t_7;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1942, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1942, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1942, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_12 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1942, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_12)->tp_iternext;
        index = 0; __pyx_t_1 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_1)) goto __pyx_L21_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_1);
        index = 1; __pyx_t_5 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_5)) goto __pyx_L21_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_12), 2) < 0) __PYX_ERR(0, 1942, __pyx_L1_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L22_unpacking_done;
        __pyx_L21_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1942, __pyx_L1_error)
        __pyx_L22_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_result, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_XDECREF_SET(__pyx_v_changed, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1943
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:             # <<<<<<<<<<<<<<
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "
 */
      __pyx_t_3 = (__pyx_v_changed == Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (unlikely(__pyx_t_2)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1947
 *                         "Got entry<->path mismatch for specific path "
 *                         "%r entry %r path_info %r " % (
 *                         path_utf8, entry, path_info))             # <<<<<<<<<<<<<<
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
 */
        __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1947, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_path_utf8);
        __Pyx_GIVEREF(__pyx_v_path_utf8);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_path_utf8);
        __Pyx_INCREF(__pyx_v_entry);
        __Pyx_GIVEREF(__pyx_v_entry);
        PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_entry);
        __Pyx_INCREF(__pyx_v_path_info);
        __Pyx_GIVEREF(__pyx_v_path_info);
        PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_v_path_info);

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1946
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "
 *                         "%r entry %r path_info %r " % (             # <<<<<<<<<<<<<<
 *                         path_utf8, entry, path_info))
 *                 # Only include changes - we're outside the users requested
 */
        __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Got_entry_path_mismatch_for_spec, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1946, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1944
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:
 *                     raise AssertionError(             # <<<<<<<<<<<<<<
 *                         "Got entry<->path mismatch for specific path "
 *                         "%r entry %r path_info %r " % (
 */
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1944, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 1944, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1943
 *                     continue
 *                 result, changed = self._process_entry(entry, path_info)
 *                 if changed is None:             # <<<<<<<<<<<<<<
 *                     raise AssertionError(
 *                         "Got entry<->path mismatch for specific path "
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1950
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1950, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1951
 *                 # expansion.
 *                 if changed:
 *                     self._gather_result_for_consistency(result)             # <<<<<<<<<<<<<<
 *                     if (result.kind[0] == 'directory' and
 *                         result.kind[1] != 'directory'):
 */
        __pyx_t_6 = ((struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self->__pyx_vtab)->_gather_result_for_consistency(__pyx_v_self, __pyx_v_result); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1951, __pyx_L1_error)

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1952
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_kind); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1952, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_5, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1952, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (__pyx_t_3) {
        } else {
          __pyx_t_2 = __pyx_t_3;
          goto __pyx_L26_bool_binop_done;
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1953
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
 *                         result.kind[1] != 'directory'):             # <<<<<<<<<<<<<<
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_result, __pyx_n_s_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1953, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1953, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_t_7, __pyx_n_s_directory, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1953, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_2 = __pyx_t_3;
        __pyx_L26_bool_binop_done:;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1952
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
 */
        if (__pyx_t_2) {

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1956
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
 *                         if entry[1][self.source_index][0] == b'r':             # <<<<<<<<<<<<<<
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]
 */
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1956, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1956, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1956, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_2 = (__Pyx_PyBytes_Equals(__pyx_t_7, __pyx_n_b_r, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1956, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1958
 *                         if entry[1][self.source_index][0] == b'r':
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]             # <<<<<<<<<<<<<<
 *                         else:
 *                             entry_path_utf8 = path_utf8
 */
            __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1958, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1958, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1958, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF_SET(__pyx_v_entry_path_utf8, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1956
 *                         # This stopped being a directory, the old children have
 *                         # to be included.
 *                         if entry[1][self.source_index][0] == b'r':             # <<<<<<<<<<<<<<
 *                             # renamed, take the source path
 *                             entry_path_utf8 = entry[1][self.source_index][1]
 */
            goto __pyx_L28;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1960
 *                             entry_path_utf8 = entry[1][self.source_index][1]
 *                         else:
 *                             entry_path_utf8 = path_utf8             # <<<<<<<<<<<<<<
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(
 */
          /*else*/ {
            __Pyx_INCREF(__pyx_v_path_utf8);
            __Pyx_XDECREF_SET(__pyx_v_entry_path_utf8, __pyx_v_path_utf8);
          }
          __pyx_L28:;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1961
 *                         else:
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')             # <<<<<<<<<<<<<<
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
 */
          __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1961, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_v_entry_path_utf8);
          __Pyx_GIVEREF(__pyx_v_entry_path_utf8);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_entry_path_utf8);
          __Pyx_INCREF(__pyx_kp_b__4);
          __Pyx_GIVEREF(__pyx_kp_b__4);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_kp_b__4);
          __Pyx_INCREF(__pyx_kp_b__4);
          __Pyx_GIVEREF(__pyx_kp_b__4);
          PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_kp_b__4);
          __Pyx_XDECREF_SET(__pyx_v_initial_key, ((PyObject*)__pyx_t_7));
          __pyx_t_7 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1962
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(             # <<<<<<<<<<<<<<
 *                             initial_key)
 *                         if block_index == 0:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_find_block_index_from_key); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1962, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1963
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)             # <<<<<<<<<<<<<<
 *                         if block_index == 0:
 *                             # The children of the root are in block index 1.
 */
          __pyx_t_1 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_7 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_initial_key) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_initial_key);
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1962, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
            PyObject* sequence = __pyx_t_7;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1962, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_1);
            #else
            __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1962, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_12 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1962, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_13 = Py_TYPE(__pyx_t_12)->tp_iternext;
            index = 0; __pyx_t_5 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_5)) goto __pyx_L29_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_5);
            index = 1; __pyx_t_1 = __pyx_t_13(__pyx_t_12); if (unlikely(!__pyx_t_1)) goto __pyx_L29_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_1);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_12), 2) < 0) __PYX_ERR(0, 1962, __pyx_L1_error)
            __pyx_t_13 = NULL;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            goto __pyx_L30_unpacking_done;
            __pyx_L29_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_13 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1962, __pyx_L1_error)
            __pyx_L30_unpacking_done:;
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1962
 *                             entry_path_utf8 = path_utf8
 *                         initial_key = (entry_path_utf8, b'', b'')
 *                         block_index, _ = self.state._find_block_index_from_key(             # <<<<<<<<<<<<<<
 *                             initial_key)
 *                         if block_index == 0:
 */
          __Pyx_XDECREF_SET(__pyx_v_block_index, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1964
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
 *                         if block_index == 0:             # <<<<<<<<<<<<<<
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
 */
          __pyx_t_7 = __Pyx_PyInt_EqObjC(__pyx_v_block_index, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1964, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1964, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1966
 *                         if block_index == 0:
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1             # <<<<<<<<<<<<<<
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):
 */
            __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_v_block_index, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1966, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF_SET(__pyx_v_block_index, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1964
 *                         block_index, _ = self.state._find_block_index_from_key(
 *                             initial_key)
 *                         if block_index == 0:             # <<<<<<<<<<<<<<
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1967
 *                             # The children of the root are in block index 1.
 *                             block_index = block_index + 1
 *                         current_block = None             # <<<<<<<<<<<<<<
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 */
          __Pyx_INCREF(Py_None);
          __Pyx_XDECREF_SET(__pyx_v_current_block, Py_None);

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1968
 *                             block_index = block_index + 1
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):             # <<<<<<<<<<<<<<
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
 */
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1968, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_14 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1968, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = PyInt_FromSsize_t(__pyx_t_14); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1968, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_block_index, __pyx_t_7, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1968, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1968, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1969
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]             # <<<<<<<<<<<<<<
 *                             if not is_inside(
 *                                 entry_path_utf8, current_block[0]):
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->state, __pyx_n_s_dirblocks_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1969, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_block_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1969, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF_SET(__pyx_v_current_block, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1970
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
 */
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_is_inside); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1970, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1971
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
 *                                 entry_path_utf8, current_block[0]):             # <<<<<<<<<<<<<<
 *                                 # No entries for this directory at all.
 *                                 current_block = None
 */
            __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_current_block, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1971, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_12 = NULL;
            __pyx_t_6 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_12)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_12);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
                __pyx_t_6 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_entry_path_utf8, __pyx_t_5};
              __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1970, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
              PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_entry_path_utf8, __pyx_t_5};
              __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1970, __pyx_L1_error)
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            } else
            #endif
            {
              __pyx_t_15 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1970, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              if (__pyx_t_12) {
                __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_12); __pyx_t_12 = NULL;
              }
              __Pyx_INCREF(__pyx_v_entry_path_utf8);
              __Pyx_GIVEREF(__pyx_v_entry_path_utf8);
              PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_6, __pyx_v_entry_path_utf8);
              __Pyx_GIVEREF(__pyx_t_5);
              PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_6, __pyx_t_5);
              __pyx_t_5 = 0;
              __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_15, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1970, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1970
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
 */
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1970, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_3 = ((!__pyx_t_2) != 0);
            if (__pyx_t_3) {

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1973
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
 *                                 current_block = None             # <<<<<<<<<<<<<<
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
 */
              __Pyx_INCREF(Py_None);
              __Pyx_DECREF_SET(__pyx_v_current_block, Py_None);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1970
 *                         if block_index < len(self.state._dirblocks):
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(             # <<<<<<<<<<<<<<
 *                                 entry_path_utf8, current_block[0]):
 *                                 # No entries for this directory at all.
 */
            }

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1968
 *                             block_index = block_index + 1
 *                         current_block = None
 *                         if block_index < len(self.state._dirblocks):             # <<<<<<<<<<<<<<
 *                             current_block = self.state._dirblocks[block_index]
 *                             if not is_inside(
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1974
 *                                 # No entries for this directory at all.
 *                                 current_block = None
 *                         if current_block is not None:             # <<<<<<<<<<<<<<
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 */
          __pyx_t_3 = (__pyx_v_current_block != Py_None);
          __pyx_t_2 = (__pyx_t_3 != 0);
          if (__pyx_t_2) {

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1975
 *                                 current_block = None
 *                         if current_block is not None:
 *                             for entry in current_block[1]:             # <<<<<<<<<<<<<<
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 *                                     # Not in the source tree, so doesn't have to be
 */
            __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_current_block, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1975, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
              __pyx_t_1 = __pyx_t_7; __Pyx_INCREF(__pyx_t_1); __pyx_t_14 = 0;
              __pyx_t_9 = NULL;
            } else {
              __pyx_t_14 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1975, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1975, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            for (;;) {
              if (likely(!__pyx_t_9)) {
                if (likely(PyList_CheckExact(__pyx_t_1))) {
                  if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_1)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_7 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1975, __pyx_L1_error)
                  #else
                  __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1975, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  #endif
                } else {
                  if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_14); __Pyx_INCREF(__pyx_t_7); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1975, __pyx_L1_error)
                  #else
                  __pyx_t_7 = PySequence_ITEM(__pyx_t_1, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1975, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_7);
                  #endif
                }
              } else {
                __pyx_t_7 = __pyx_t_9(__pyx_t_1);
                if (unlikely(!__pyx_t_7)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 1975, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_7);
              }
              __Pyx_DECREF_SET(__pyx_v_entry, __pyx_t_7);
              __pyx_t_7 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1976
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
 */
              __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_entry, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1976, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_7, __pyx_v_self->source_index, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1976, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_15, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1976, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_t_7, __pyx_n_b_a, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1976, __pyx_L1_error)
              if (!__pyx_t_3) {
              } else {
                __pyx_t_2 = __pyx_t_3;
                goto __pyx_L38_bool_binop_done;
              }
              __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_t_7, __pyx_n_b_r, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1976, __pyx_L1_error)
              __pyx_t_2 = __pyx_t_3;
              __pyx_L38_bool_binop_done:;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_3 = (__pyx_t_2 != 0);
              if (__pyx_t_3) {

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1979
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
 *                                     continue             # <<<<<<<<<<<<<<
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(
 */
                goto __pyx_L35_continue;

                /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1976
 *                         if current_block is not None:
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):             # <<<<<<<<<<<<<<
 *                                     # Not in the source tree, so doesn't have to be
 *                                     # included.
 */
              }

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1981
 *                                     continue
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(             # <<<<<<<<<<<<<<
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:
 */
              __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->search_specific_file_parents, __pyx_n_s_add); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1981, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1982
 *                                 # Path of the entry itself.
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))             # <<<<<<<<<<<<<<
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))
 */
              __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_entry, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1982, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_12 = __Pyx_PyObject_GetSlice(__pyx_t_5, 0, 2, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1982, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PySequence_Tuple(__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1982, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_12 = __Pyx_PyObject_Call(__pyx_v_self->pathjoin, __pyx_t_5, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1982, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_15);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_15, function);
                }
              }
              __pyx_t_7 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_5, __pyx_t_12) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_t_12);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1981, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1975
 *                                 current_block = None
 *                         if current_block is not None:
 *                             for entry in current_block[1]:             # <<<<<<<<<<<<<<
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 *                                     # Not in the source tree, so doesn't have to be
 */
              __pyx_L35_continue:;
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1974
 *                                 # No entries for this directory at all.
 *                                 current_block = None
 *                         if current_block is not None:             # <<<<<<<<<<<<<<
 *                             for entry in current_block[1]:
 *                                 if entry[1][self.source_index][0] in (b'a', b'r'):
 */
          }

          /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1952
 *                 if changed:
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and             # <<<<<<<<<<<<<<
 *                         result.kind[1] != 'directory'):
 *                         # This stopped being a directory, the old children have
 */
        }

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1950
 *                 # Only include changes - we're outside the users requested
 *                 # expansion.
 *                 if changed:             # <<<<<<<<<<<<<<
 *                     self._gather_result_for_consistency(result)
 *                     if (result.kind[0] == 'directory' and
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1983
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_changed); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1983, __pyx_L1_error)
      if (!__pyx_t_2) {
      } else {
        __pyx_t_3 = __pyx_t_2;
        goto __pyx_L41_bool_binop_done;
      }
      __pyx_t_2 = (__pyx_v_self->include_unchanged != 0);
      __pyx_t_3 = __pyx_t_2;
      __pyx_L41_bool_binop_done:;
      if (__pyx_t_3) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1984
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))             # <<<<<<<<<<<<<<
 *             self.searched_exact_paths.add(path_utf8)
 *         return results
 */
        __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_result);
        __Pyx_GIVEREF(__pyx_v_result);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_result);
        __Pyx_INCREF(__pyx_v_changed);
        __Pyx_GIVEREF(__pyx_v_changed);
        PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_changed);
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_results, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1984, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1983
 *                                 self.search_specific_file_parents.add(
 *                                     self.pathjoin(*entry[0][:2]))
 *                 if changed or self.include_unchanged:             # <<<<<<<<<<<<<<
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1939
 *                     path_utf8, path_entries))
 *             path_info = self._path_info(path_utf8, path_utf8.decode('utf8'))
 *             for entry in selected_entries:             # <<<<<<<<<<<<<<
 *                 if entry[0][2] in self.seen_ids:
 *                     continue
 */
      __pyx_L18_continue:;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1985
 *                 if changed or self.include_unchanged:
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)             # <<<<<<<<<<<<<<
 *         return results
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->searched_exact_paths, __pyx_n_s_add); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_v_path_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path_utf8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L3_continue:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1986
 *                     results.append((result, changed))
 *             self.searched_exact_paths.add(path_utf8)
 *         return results             # <<<<<<<<<<<<<<
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_results);
  __pyx_r = __pyx_v_results;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1893
 *                     self.current_dir_info = None
 * 
 *     cdef object _next_consistent_entries(self):             # <<<<<<<<<<<<<<
 *         """Grabs the next specific file parent case to consider.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._next_consistent_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_results);
  __Pyx_XDECREF(__pyx_v_path_utf8);
  __Pyx_XDECREF(__pyx_v_path_entries);
  __Pyx_XDECREF(__pyx_v_selected_entries);
  __Pyx_XDECREF(__pyx_v_candidate_entry);
  __Pyx_XDECREF(__pyx_v_path_info);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_changed);
  __Pyx_XDECREF(__pyx_v_entry_path_utf8);
  __Pyx_XDECREF(__pyx_v_initial_key);
  __Pyx_XDECREF(__pyx_v_block_index);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_current_block);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1988
 *         return results
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):             # <<<<<<<<<<<<<<
 *         """Generate path_info for unicode_path.
 * 
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v_utf8_path, PyObject *__pyx_v_unicode_path) {
  PyObject *__pyx_v_abspath = NULL;
  PyObject *__pyx_v_stat = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_utf8_basename = NULL;
  PyObject *__pyx_v_dir_info = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("_path_info", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1993
 *         :return: None if unicode_path does not exist, or a path_info tuple.
 *         """
 *         abspath = self.tree.abspath(unicode_path)             # <<<<<<<<<<<<<<
 *         try:
 *             stat = os.lstat(abspath)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_abspath); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_unicode_path) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_unicode_path);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_abspath = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1994
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    /*try:*/ {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1995
 *         abspath = self.tree.abspath(unicode_path)
 *         try:
 *             stat = os.lstat(abspath)             # <<<<<<<<<<<<<<
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1995, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_lstat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1995, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_v_abspath) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_abspath);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1995, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_stat = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1994
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 */
    }
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1996
 *         try:
 *             stat = os.lstat(abspath)
 *         except OSError, e:             # <<<<<<<<<<<<<<
 *             if e.errno == errno.ENOENT:
 *                 # the path does not exist.
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OSError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._path_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 1996, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_e = __pyx_t_3;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1997
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                 # the path does not exist.
 *                 return None
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_e, __pyx_n_s_errno); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1997, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_errno); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1997, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_ENOENT); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1997, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_8, __pyx_t_10, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1997, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1997, __pyx_L5_except_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (likely(__pyx_t_11)) {

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1999
 *             if e.errno == errno.ENOENT:
 *                 # the path does not exist.
 *                 return None             # <<<<<<<<<<<<<<
 *             else:
 *                 raise
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L6_except_return;

        /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1997
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 *             if e.errno == errno.ENOENT:             # <<<<<<<<<<<<<<
 *                 # the path does not exist.
 *                 return None
 */
      }

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2001
 *                 return None
 *             else:
 *                 raise             # <<<<<<<<<<<<<<
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,
 */
      /*else*/ {
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_3);
        __Pyx_XGIVEREF(__pyx_t_2);
        __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_2);
        __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_2 = 0; 
        __PYX_ERR(0, 2001, __pyx_L5_except_error)
      }
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1994
 *         """
 *         abspath = self.tree.abspath(unicode_path)
 *         try:             # <<<<<<<<<<<<<<
 *             stat = os.lstat(abspath)
 *         except OSError, e:
 */
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2002
 *             else:
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]             # <<<<<<<<<<<<<<
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_utf8_path, __pyx_n_s_rsplit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_utf8_basename = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2004
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,             # <<<<<<<<<<<<<<
 *             abspath)
 *         if dir_info[2] == 'directory':
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_osutils); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_file_kind_from_stat_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_stat, __pyx_n_s_st_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2003
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]
 *         dir_info = (utf8_path, utf8_basename,             # <<<<<<<<<<<<<<
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
 */
  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_utf8_path);
  __Pyx_GIVEREF(__pyx_v_utf8_path);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_utf8_path);
  __Pyx_INCREF(__pyx_v_utf8_basename);
  __Pyx_GIVEREF(__pyx_v_utf8_basename);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_utf8_basename);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_stat);
  __Pyx_GIVEREF(__pyx_v_stat);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_v_stat);
  __Pyx_INCREF(__pyx_v_abspath);
  __Pyx_GIVEREF(__pyx_v_abspath);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_v_abspath);
  __pyx_t_2 = 0;
  __pyx_v_dir_info = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2006
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
 *         if dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 */
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_dir_info, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_directory, Py_EQ)); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2007
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->tree, __pyx_n_s_directory_is_tree_reference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2008
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):             # <<<<<<<<<<<<<<
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]
 */
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_unicode_path) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_unicode_path);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2007
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 */
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 2007, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_11) {

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2009
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info
 */
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 0, 2, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2010
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *         return dir_info
 */
      __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_tuple__12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_self->root_dir_info, 3, 0, NULL, NULL, &__pyx_slice__13, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2009
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info
 */
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_v_self->root_dir_info);
      __Pyx_DECREF(__pyx_v_self->root_dir_info);
      __pyx_v_self->root_dir_info = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2007
 *             abspath)
 *         if dir_info[2] == 'directory':
 *             if self.tree._directory_is_tree_reference(             # <<<<<<<<<<<<<<
 *                 unicode_path):
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 */
    }

    /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2006
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 *             abspath)
 *         if dir_info[2] == 'directory':             # <<<<<<<<<<<<<<
 *             if self.tree._directory_is_tree_reference(
 *                 unicode_path):
 */
  }

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2011
 *                 self.root_dir_info = self.root_dir_info[:2] + \
 *                     ('tree-reference',) + self.root_dir_info[3:]
 *         return dir_info             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dir_info);
  __pyx_r = __pyx_v_dir_info;
  goto __pyx_L0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1988
 *         return results
 * 
 *     cdef object _path_info(self, utf8_path, unicode_path):             # <<<<<<<<<<<<<<
 *         """Generate path_info for unicode_path.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC._path_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_abspath);
  __Pyx_XDECREF(__pyx_v_stat);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_utf8_basename);
  __Pyx_XDECREF(__pyx_v_dir_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1005
 *     cdef object use_filesystem_for_exec
 *     cdef object utf8_decode
 *     cdef readonly object searched_specific_files             # <<<<<<<<<<<<<<
 *     cdef readonly object searched_exact_paths
 *     cdef object search_specific_files
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->searched_specific_files);
  __pyx_r = __pyx_v_self->searched_specific_files;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "breezy/bzr/_dirstate_helpers_pyx.pyx":1006
 *     cdef object utf8_decode
 *     cdef readonly object searched_specific_files
 *     cdef readonly object searched_exact_paths             # <<<<<<<<<<<<<<
 *     cdef object search_specific_files
 *     # The parents up to the root of the paths we are searching.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths___get__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->searched_exact_paths);
  __pyx_r = __pyx_v_self->searched_exact_paths;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_8__reduce_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->block_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->current_block_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->doing_consistency_expansion); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->include_unchanged); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->partial); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_self->path_index); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->root_entries_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->root_entries_pos); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_int(__pyx_v_self->source_index); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_self->target_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->want_unversioned); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(39); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_INCREF(__pyx_v_self->_pending_consistent_entries);
  __Pyx_GIVEREF(__pyx_v_self->_pending_consistent_entries);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v_self->_pending_consistent_entries);
  __Pyx_INCREF(__pyx_v_self->bisect_left);
  __Pyx_GIVEREF(__pyx_v_self->bisect_left);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_v_self->bisect_left);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->current_block);
  __Pyx_GIVEREF(__pyx_v_self->current_block);
  PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_v_self->current_block);
  __Pyx_INCREF(__pyx_v_self->current_block_list);
  __Pyx_GIVEREF(__pyx_v_self->current_block_list);
  PyTuple_SET_ITEM(__pyx_t_12, 4, __pyx_v_self->current_block_list);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_12, 5, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->current_dir_info);
  __Pyx_GIVEREF(__pyx_v_self->current_dir_info);
  PyTuple_SET_ITEM(__pyx_t_12, 6, __pyx_v_self->current_dir_info);
  __Pyx_INCREF(__pyx_v_self->current_dir_list);
  __Pyx_GIVEREF(__pyx_v_self->current_dir_list);
  PyTuple_SET_ITEM(__pyx_t_12, 7, __pyx_v_self->current_dir_list);
  __Pyx_INCREF(__pyx_v_self->current_root);
  __Pyx_GIVEREF(__pyx_v_self->current_root);
  PyTuple_SET_ITEM(__pyx_t_12, 8, __pyx_v_self->current_root);
  __Pyx_INCREF(__pyx_v_self->current_root_unicode);
  __Pyx_GIVEREF(__pyx_v_self->current_root_unicode);
  PyTuple_SET_ITEM(__pyx_t_12, 9, __pyx_v_self->current_root_unicode);
  __Pyx_INCREF(__pyx_v_self->dir_iterator);
  __Pyx_GIVEREF(__pyx_v_self->dir_iterator);
  PyTuple_SET_ITEM(__pyx_t_12, 10, __pyx_v_self->dir_iterator);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_12, 11, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->fstat);
  __Pyx_GIVEREF(__pyx_v_self->fstat);
  PyTuple_SET_ITEM(__pyx_t_12, 12, __pyx_v_self->fstat);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_12, 13, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->last_source_parent);
  __Pyx_GIVEREF(__pyx_v_self->last_source_parent);
  PyTuple_SET_ITEM(__pyx_t_12, 14, __pyx_v_self->last_source_parent);
  __Pyx_INCREF(__pyx_v_self->last_target_parent);
  __Pyx_GIVEREF(__pyx_v_self->last_target_parent);
  PyTuple_SET_ITEM(__pyx_t_12, 15, __pyx_v_self->last_target_parent);
  __Pyx_INCREF(__pyx_v_self->new_dirname_to_file_id);
  __Pyx_GIVEREF(__pyx_v_self->new_dirname_to_file_id);
  PyTuple_SET_ITEM(__pyx_t_12, 16, __pyx_v_self->new_dirname_to_file_id);
  __Pyx_INCREF(__pyx_v_self->old_dirname_to_file_id);
  __Pyx_GIVEREF(__pyx_v_self->old_dirname_to_file_id);
  PyTuple_SET_ITEM(__pyx_t_12, 17, __pyx_v_self->old_dirname_to_file_id);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_12, 18, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_12, 19, __pyx_t_6);
  __Pyx_INCREF(__pyx_v_self->pathjoin);
  __Pyx_GIVEREF(__pyx_v_self->pathjoin);
  PyTuple_SET_ITEM(__pyx_t_12, 20, __pyx_v_self->pathjoin);
  __Pyx_INCREF(__pyx_v_self->root_abspath);
  __Pyx_GIVEREF(__pyx_v_self->root_abspath);
  PyTuple_SET_ITEM(__pyx_t_12, 21, __pyx_v_self->root_abspath);
  __Pyx_INCREF(__pyx_v_self->root_dir_info);
  __Pyx_GIVEREF(__pyx_v_self->root_dir_info);
  PyTuple_SET_ITEM(__pyx_t_12, 22, __pyx_v_self->root_dir_info);
  __Pyx_INCREF(__pyx_v_self->root_entries);
  __Pyx_GIVEREF(__pyx_v_self->root_entries);
  PyTuple_SET_ITEM(__pyx_t_12, 23, __pyx_v_self->root_entries);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_12, 24, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_12, 25, __pyx_t_8);
  __Pyx_INCREF(__pyx_v_self->search_specific_file_parents);
  __Pyx_GIVEREF(__pyx_v_self->search_specific_file_parents);
  PyTuple_SET_ITEM(__pyx_t_12, 26, __pyx_v_self->search_specific_file_parents);
  __Pyx_INCREF(__pyx_v_self->search_specific_files);
  __Pyx_GIVEREF(__pyx_v_self->search_specific_files);
  PyTuple_SET_ITEM(__pyx_t_12, 27, __pyx_v_self->search_specific_files);
  __Pyx_INCREF(__pyx_v_self->searched_exact_paths);
  __Pyx_GIVEREF(__pyx_v_self->searched_exact_paths);
  PyTuple_SET_ITEM(__pyx_t_12, 28, __pyx_v_self->searched_exact_paths);
  __Pyx_INCREF(__pyx_v_self->searched_specific_files);
  __Pyx_GIVEREF(__pyx_v_self->searched_specific_files);
  PyTuple_SET_ITEM(__pyx_t_12, 29, __pyx_v_self->searched_specific_files);
  __Pyx_INCREF(__pyx_v_self->seen_ids);
  __Pyx_GIVEREF(__pyx_v_self->seen_ids);
  PyTuple_SET_ITEM(__pyx_t_12, 30, __pyx_v_self->seen_ids);
  __Pyx_INCREF(__pyx_v_self->sha_file);
  __Pyx_GIVEREF(__pyx_v_self->sha_file);
  PyTuple_SET_ITEM(__pyx_t_12, 31, __pyx_v_self->sha_file);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_12, 32, __pyx_t_9);
  __Pyx_INCREF(__pyx_v_self->state);
  __Pyx_GIVEREF(__pyx_v_self->state);
  PyTuple_SET_ITEM(__pyx_t_12, 33, __pyx_v_self->state);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_12, 34, __pyx_t_10);
  __Pyx_INCREF(__pyx_v_self->tree);
  __Pyx_GIVEREF(__pyx_v_self->tree);
  PyTuple_SET_ITEM(__pyx_t_12, 35, __pyx_v_self->tree);
  __Pyx_INCREF(__pyx_v_self->use_filesystem_for_exec);
  __Pyx_GIVEREF(__pyx_v_self->use_filesystem_for_exec);
  PyTuple_SET_ITEM(__pyx_t_12, 36, __pyx_v_self->use_filesystem_for_exec);
  __Pyx_INCREF(__pyx_v_self->utf8_decode);
  __Pyx_GIVEREF(__pyx_v_self->utf8_decode);
  PyTuple_SET_ITEM(__pyx_t_12, 37, __pyx_v_self->utf8_decode);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_12, 38, __pyx_t_11);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_12);
  __pyx_t_12 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_12 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_v__dict = __pyx_t_12;
  __pyx_t_12 = 0;

  /* "(tree fragment)":7
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_13 = (__pyx_v__dict != Py_None);
  __pyx_t_14 = (__pyx_t_13 != 0);
  if (__pyx_t_14) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_12 = PyTuple_New(1); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_v__dict);
    __pyx_t_11 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_11));
    __pyx_t_11 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._pending_consistent_entries, self.bisect_left, self.block_index, self.current_block, self.current_block_list, self.current_block_pos, self.current_dir_info, self.current_dir_list, self.current_root, self.current_root_unicode, self.dir_iterator, self.doing_consistency_expansion, self.fstat, self.include_unchanged, self.last_source_parent, self.last_target_parent, self.new_dirname_to_file_id, self.old_dirname_to_file_id, self.partial, self.path_index, self.pathjoin, self.root_abspath, self.root_dir_info, self.root_entries, self.root_entries_len, self.root_entries_pos, self.search_specific_file_parents, self.search_specific_files, self.searched_exact_paths, self.searched_specific_files, self.seen_ids, self.sha_file, self.source_index, self.state, self.target_index, self.tree, self.use_filesystem_for_exec, self.utf8_decode, self.want_unversioned)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, None), state
 */
  /*else*/ {
    __pyx_t_13 = (__pyx_v_self->_pending_consistent_entries != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->bisect_left != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->current_block != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->current_block_list != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->current_dir_info != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->current_dir_list != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->current_root != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->current_root_unicode != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->dir_iterator != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->fstat != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->last_source_parent != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->last_target_parent != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->new_dirname_to_file_id != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->old_dirname_to_file_id != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->pathjoin != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->root_abspath != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->root_dir_info != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->root_entries != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->search_specific_file_parents != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->search_specific_files != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->searched_exact_paths != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->searched_specific_files != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->seen_ids != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->sha_file != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->state != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->tree != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    if (!__pyx_t_13) {
    } else {
      __pyx_t_14 = __pyx_t_13;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_13 = (__pyx_v_self->use_filesystem_for_exec != Py_None);
    __pyx_t_15 = (__pyx_t_13 != 0);
    if (!__pyx_t_15) {
    } else {
      __pyx_t_14 = __pyx_t_15;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_15 = (__pyx_v_self->utf8_decode != Py_None);
    __pyx_t_13 = (__pyx_t_15 != 0);
    __pyx_t_14 = __pyx_t_13;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_14;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, None), state
 *     else:
 */
  __pyx_t_14 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_14) {

    /* "(tree fragment)":13
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_pyx_unpickle_ProcessEntryC); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_35494349);
    __Pyx_GIVEREF(__pyx_int_35494349);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_35494349);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_12, 2, Py_None);
    __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_12);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_v_state);
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_10;
    __pyx_t_10 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._pending_consistent_entries is not None or self.bisect_left is not None or self.current_block is not None or self.current_block_list is not None or self.current_dir_info is not None or self.current_dir_list is not None or self.current_root is not None or self.current_root_unicode is not None or self.dir_iterator is not None or self.fstat is not None or self.last_source_parent is not None or self.last_target_parent is not None or self.new_dirname_to_file_id is not None or self.old_dirname_to_file_id is not None or self.pathjoin is not None or self.root_abspath is not None or self.root_dir_info is not None or self.root_entries is not None or self.search_specific_file_parents is not None or self.search_specific_files is not None or self.searched_exact_paths is not None or self.searched_specific_files is not None or self.seen_ids is not None or self.sha_file is not None or self.state is not None or self.tree is not None or self.use_filesystem_for_exec is not None or self.utf8_decode is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, None), state
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_pyx_unpickle_ProcessEntryC); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_12, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_35494349);
    __Pyx_GIVEREF(__pyx_int_35494349);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_int_35494349);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_v_state);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_12);
    __pyx_t_10 = 0;
    __pyx_t_12 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_10__setstate_cython__(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ProcessEntryC, (type(self), 0x21d99cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ProcessEntryC__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.ProcessEntryC.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader = {"__pyx_unpickle_Reader", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Reader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Reader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Reader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Reader", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_18__pyx_unpickle_Reader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x65fc0dd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x65fc0dd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x65fc0dd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 *     __pyx_result = Reader.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x65fc0dd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x65, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x65fc0dd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 *     __pyx_result = Reader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x65fc0dd = (cur_cstr, end_cstr, next, state, text, text_cstr, text_size))" % __pyx_checksum)
 *     __pyx_result = Reader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_Reader__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Reader__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->cur_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->end_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->next = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->state);
  __Pyx_DECREF(__pyx_v___pyx_result->state);
  __pyx_v___pyx_result->state = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->text);
  __Pyx_DECREF(__pyx_v___pyx_result->text);
  __pyx_v___pyx_result->text = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->text_cstr = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->text_size = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 7) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_Reader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ProcessEntryC(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC = {"__pyx_unpickle_ProcessEntryC", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ProcessEntryC", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ProcessEntryC", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ProcessEntryC") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ProcessEntryC", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6breezy_3bzr_21_dirstate_helpers_pyx_20__pyx_unpickle_ProcessEntryC(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x21d99cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x21d99cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x21d99cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x21d99cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x21, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x21d99cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x21d99cd = (_pending_consistent_entries, bisect_left, block_index, current_block, current_block_list, current_block_pos, current_dir_info, current_dir_list, current_root, current_root_unicode, dir_iterator, doing_consistency_expansion, fstat, include_unchanged, last_source_parent, last_target_parent, new_dirname_to_file_id, old_dirname_to_file_id, partial, path_index, pathjoin, root_abspath, root_dir_info, root_entries, root_entries_len, root_entries_pos, search_specific_file_parents, search_specific_files, searched_exact_paths, searched_specific_files, seen_ids, sha_file, source_index, state, target_index, tree, use_filesystem_for_exec, utf8_decode, want_unversioned))" % __pyx_checksum)
 *     __pyx_result = ProcessEntryC.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ProcessEntryC(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx___pyx_unpickle_ProcessEntryC__set_state(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ProcessEntryC__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[39])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_pending_consistent_entries);
  __Pyx_DECREF(__pyx_v___pyx_result->_pending_consistent_entries);
  __pyx_v___pyx_result->_pending_consistent_entries = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bisect_left);
  __Pyx_DECREF(__pyx_v___pyx_result->bisect_left);
  __pyx_v___pyx_result->bisect_left = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->block_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_block);
  __Pyx_DECREF(__pyx_v___pyx_result->current_block);
  __pyx_v___pyx_result->current_block = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_block_list);
  __Pyx_DECREF(__pyx_v___pyx_result->current_block_list);
  __pyx_v___pyx_result->current_block_list = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->current_block_pos = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_dir_info);
  __Pyx_DECREF(__pyx_v___pyx_result->current_dir_info);
  __pyx_v___pyx_result->current_dir_info = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_dir_list);
  __Pyx_DECREF(__pyx_v___pyx_result->current_dir_list);
  __pyx_v___pyx_result->current_dir_list = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 8, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_root);
  __Pyx_DECREF(__pyx_v___pyx_result->current_root);
  __pyx_v___pyx_result->current_root = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 9, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->current_root_unicode);
  __Pyx_DECREF(__pyx_v___pyx_result->current_root_unicode);
  __pyx_v___pyx_result->current_root_unicode = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 10, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->dir_iterator);
  __Pyx_DECREF(__pyx_v___pyx_result->dir_iterator);
  __pyx_v___pyx_result->dir_iterator = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 11, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->doing_consistency_expansion = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 12, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->fstat);
  __Pyx_DECREF(__pyx_v___pyx_result->fstat);
  __pyx_v___pyx_result->fstat = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 13, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->include_unchanged = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 14, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->last_source_parent);
  __Pyx_DECREF(__pyx_v___pyx_result->last_source_parent);
  __pyx_v___pyx_result->last_source_parent = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 15, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->last_target_parent);
  __Pyx_DECREF(__pyx_v___pyx_result->last_target_parent);
  __pyx_v___pyx_result->last_target_parent = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 16, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->new_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v___pyx_result->new_dirname_to_file_id);
  __pyx_v___pyx_result->new_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 17, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->old_dirname_to_file_id);
  __Pyx_DECREF(__pyx_v___pyx_result->old_dirname_to_file_id);
  __pyx_v___pyx_result->old_dirname_to_file_id = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 18, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->partial = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 19, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->path_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 20, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->pathjoin);
  __Pyx_DECREF(__pyx_v___pyx_result->pathjoin);
  __pyx_v___pyx_result->pathjoin = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 21, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_abspath);
  __Pyx_DECREF(__pyx_v___pyx_result->root_abspath);
  __pyx_v___pyx_result->root_abspath = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 22, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_dir_info);
  __Pyx_DECREF(__pyx_v___pyx_result->root_dir_info);
  __pyx_v___pyx_result->root_dir_info = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 23, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->root_entries);
  __Pyx_DECREF(__pyx_v___pyx_result->root_entries);
  __pyx_v___pyx_result->root_entries = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 24, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->root_entries_len = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 25, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->root_entries_pos = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 26, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->search_specific_file_parents);
  __Pyx_DECREF(__pyx_v___pyx_result->search_specific_file_parents);
  __pyx_v___pyx_result->search_specific_file_parents = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 27, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->search_specific_files);
  __Pyx_DECREF(__pyx_v___pyx_result->search_specific_files);
  __pyx_v___pyx_result->search_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 28, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->searched_exact_paths);
  __Pyx_DECREF(__pyx_v___pyx_result->searched_exact_paths);
  __pyx_v___pyx_result->searched_exact_paths = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 29, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->searched_specific_files);
  __Pyx_DECREF(__pyx_v___pyx_result->searched_specific_files);
  __pyx_v___pyx_result->searched_specific_files = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 30, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->seen_ids);
  __Pyx_DECREF(__pyx_v___pyx_result->seen_ids);
  __pyx_v___pyx_result->seen_ids = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 31, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->sha_file);
  __Pyx_DECREF(__pyx_v___pyx_result->sha_file);
  __pyx_v___pyx_result->sha_file = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 32, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->source_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 33, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->state);
  __Pyx_DECREF(__pyx_v___pyx_result->state);
  __pyx_v___pyx_result->state = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 34, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->target_index = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 35, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->tree);
  __Pyx_DECREF(__pyx_v___pyx_result->tree);
  __pyx_v___pyx_result->tree = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 36, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->use_filesystem_for_exec);
  __Pyx_DECREF(__pyx_v___pyx_result->use_filesystem_for_exec);
  __pyx_v___pyx_result->use_filesystem_for_exec = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 37, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->utf8_decode);
  __Pyx_DECREF(__pyx_v___pyx_result->utf8_decode);
  __pyx_v___pyx_result->utf8_decode = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 38, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->want_unversioned = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[39])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 39) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[39])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 39, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[39])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ProcessEntryC__set_state(<ProcessEntryC> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ProcessEntryC__set_state(ProcessEntryC __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._pending_consistent_entries = __pyx_state[0]; __pyx_result.bisect_left = __pyx_state[1]; __pyx_result.block_index = __pyx_state[2]; __pyx_result.current_block = __pyx_state[3]; __pyx_result.current_block_list = __pyx_state[4]; __pyx_result.current_block_pos = __pyx_state[5]; __pyx_result.current_dir_info = __pyx_state[6]; __pyx_result.current_dir_list = __pyx_state[7]; __pyx_result.current_root = __pyx_state[8]; __pyx_result.current_root_unicode = __pyx_state[9]; __pyx_result.dir_iterator = __pyx_state[10]; __pyx_result.doing_consistency_expansion = __pyx_state[11]; __pyx_result.fstat = __pyx_state[12]; __pyx_result.include_unchanged = __pyx_state[13]; __pyx_result.last_source_parent = __pyx_state[14]; __pyx_result.last_target_parent = __pyx_state[15]; __pyx_result.new_dirname_to_file_id = __pyx_state[16]; __pyx_result.old_dirname_to_file_id = __pyx_state[17]; __pyx_result.partial = __pyx_state[18]; __pyx_result.path_index = __pyx_state[19]; __pyx_result.pathjoin = __pyx_state[20]; __pyx_result.root_abspath = __pyx_state[21]; __pyx_result.root_dir_info = __pyx_state[22]; __pyx_result.root_entries = __pyx_state[23]; __pyx_result.root_entries_len = __pyx_state[24]; __pyx_result.root_entries_pos = __pyx_state[25]; __pyx_result.search_specific_file_parents = __pyx_state[26]; __pyx_result.search_specific_files = __pyx_state[27]; __pyx_result.searched_exact_paths = __pyx_state[28]; __pyx_result.searched_specific_files = __pyx_state[29]; __pyx_result.seen_ids = __pyx_state[30]; __pyx_result.sha_file = __pyx_state[31]; __pyx_result.source_index = __pyx_state[32]; __pyx_result.state = __pyx_state[33]; __pyx_result.target_index = __pyx_state[34]; __pyx_result.tree = __pyx_state[35]; __pyx_result.use_filesystem_for_exec = __pyx_state[36]; __pyx_result.utf8_decode = __pyx_state[37]; __pyx_result.want_unversioned = __pyx_state[38]
 *     if len(__pyx_state) > 39 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("breezy.bzr._dirstate_helpers_pyx.__pyx_unpickle_ProcessEntryC__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
 */

static CYTHON_INLINE void *__pyx_f_6breezy_3bzr_12_str_helpers__my_memrchr(void *__pyx_v_s, int __pyx_v_c, size_t __pyx_v_n) {
  char *__pyx_v_pos;
  char *__pyx_v_start;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_my_memrchr", 0);

  /* "_str_helpers.pxd":32
 *     cdef char *start
 * 
 *     start = <char*>s             # <<<<<<<<<<<<<<
 *     pos = start + n - 1
 *     while pos >= start:
 */
  __pyx_v_start = ((char *)__pyx_v_s);

  /* "_str_helpers.pxd":33
 * 
 *     start = <char*>s
 *     pos = start + n - 1             # <<<<<<<<<<<<<<
 *     while pos >= start:
 *         if pos[0] == c:
 */
  __pyx_v_pos = ((__pyx_v_start + __pyx_v_n) - 1);

  /* "_str_helpers.pxd":34
 *     start = <char*>s
 *     pos = start + n - 1
 *     while pos >= start:             # <<<<<<<<<<<<<<
 *         if pos[0] == c:
 *             return <void*>pos
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_pos >= __pyx_v_start) != 0);
    if (!__pyx_t_1) break;

    /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    __pyx_t_1 = (((__pyx_v_pos[0]) == __pyx_v_c) != 0);
    if (__pyx_t_1) {

      /* "_str_helpers.pxd":36
 *     while pos >= start:
 *         if pos[0] == c:
 *             return <void*>pos             # <<<<<<<<<<<<<<
 *         pos = pos - 1
 *     return NULL
 */
      __pyx_r = ((void *)__pyx_v_pos);
      goto __pyx_L0;

      /* "_str_helpers.pxd":35
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    }

    /* "_str_helpers.pxd":37
 *         if pos[0] == c:
 *             return <void*>pos
 *         pos = pos - 1             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
    __pyx_v_pos = (__pyx_v_pos - 1);
  }

  /* "_str_helpers.pxd":38
 *             return <void*>pos
 *         pos = pos - 1
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "_str_helpers.pxd":27
 * 
 * 
 * cdef inline void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     cdef char *pos
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("safe_string_from_size", 0);

  /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":44
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_FromStringAndSize(s, size)
 * 
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_str_helpers.pxd":43
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 43, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 43, __pyx_L1_error)

    /* "_str_helpers.pxd":42
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  }

  /* "_str_helpers.pxd":45
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_FromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_FromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":41
 * 
 * 
 * cdef inline object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

static CYTHON_INLINE PyObject *__pyx_f_6breezy_3bzr_12_str_helpers_safe_interned_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("safe_interned_string_from_size", 0);

  /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "_str_helpers.pxd":51
 *     if size < 0:
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)             # <<<<<<<<<<<<<<
 *     return PyBytes_InternFromStringAndSize(s, size)
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 51, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "_str_helpers.pxd":50
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 50, __pyx_L1_error)

    /* "_str_helpers.pxd":49
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 */
  }

  /* "_str_helpers.pxd":52
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d' % size)
 *     return PyBytes_InternFromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyBytes_InternFromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("breezy.bzr._str_helpers.safe_interned_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_Reader __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  p->state = Py_None; Py_INCREF(Py_None);
  p->text = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->state);
  Py_CLEAR(p->text);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  if (p->state) {
    e = (*v)(p->state, a); if (e) return e;
  }
  if (p->text) {
    e = (*v)(p->text, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_Reader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *)o;
  tmp = ((PyObject*)p->state);
  p->state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->text);
  p->text = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_Reader[] = {
  {"_parse_dirblocks", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_3_parse_dirblocks, METH_NOARGS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_2_parse_dirblocks},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._dirstate_helpers_pyx.Reader", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Maintain the current location, and return fields as you parse them.", /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_Reader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;

static PyObject *__pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o);
  p->__pyx_vtab = __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  p->old_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  p->new_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  p->last_source_parent = Py_None; Py_INCREF(Py_None);
  p->last_target_parent = Py_None; Py_INCREF(Py_None);
  p->use_filesystem_for_exec = Py_None; Py_INCREF(Py_None);
  p->utf8_decode = Py_None; Py_INCREF(Py_None);
  p->searched_specific_files = Py_None; Py_INCREF(Py_None);
  p->searched_exact_paths = Py_None; Py_INCREF(Py_None);
  p->search_specific_files = Py_None; Py_INCREF(Py_None);
  p->search_specific_file_parents = Py_None; Py_INCREF(Py_None);
  p->state = Py_None; Py_INCREF(Py_None);
  p->current_root = Py_None; Py_INCREF(Py_None);
  p->current_root_unicode = Py_None; Py_INCREF(Py_None);
  p->root_entries = Py_None; Py_INCREF(Py_None);
  p->root_abspath = Py_None; Py_INCREF(Py_None);
  p->tree = Py_None; Py_INCREF(Py_None);
  p->dir_iterator = Py_None; Py_INCREF(Py_None);
  p->current_block = Py_None; Py_INCREF(Py_None);
  p->current_block_list = Py_None; Py_INCREF(Py_None);
  p->current_dir_info = Py_None; Py_INCREF(Py_None);
  p->current_dir_list = Py_None; Py_INCREF(Py_None);
  p->_pending_consistent_entries = Py_None; Py_INCREF(Py_None);
  p->root_dir_info = Py_None; Py_INCREF(Py_None);
  p->bisect_left = Py_None; Py_INCREF(Py_None);
  p->pathjoin = Py_None; Py_INCREF(Py_None);
  p->fstat = Py_None; Py_INCREF(Py_None);
  p->seen_ids = Py_None; Py_INCREF(Py_None);
  p->sha_file = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o) {
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->old_dirname_to_file_id);
  Py_CLEAR(p->new_dirname_to_file_id);
  Py_CLEAR(p->last_source_parent);
  Py_CLEAR(p->last_target_parent);
  Py_CLEAR(p->use_filesystem_for_exec);
  Py_CLEAR(p->utf8_decode);
  Py_CLEAR(p->searched_specific_files);
  Py_CLEAR(p->searched_exact_paths);
  Py_CLEAR(p->search_specific_files);
  Py_CLEAR(p->search_specific_file_parents);
  Py_CLEAR(p->state);
  Py_CLEAR(p->current_root);
  Py_CLEAR(p->current_root_unicode);
  Py_CLEAR(p->root_entries);
  Py_CLEAR(p->root_abspath);
  Py_CLEAR(p->tree);
  Py_CLEAR(p->dir_iterator);
  Py_CLEAR(p->current_block);
  Py_CLEAR(p->current_block_list);
  Py_CLEAR(p->current_dir_info);
  Py_CLEAR(p->current_dir_list);
  Py_CLEAR(p->_pending_consistent_entries);
  Py_CLEAR(p->root_dir_info);
  Py_CLEAR(p->bisect_left);
  Py_CLEAR(p->pathjoin);
  Py_CLEAR(p->fstat);
  Py_CLEAR(p->seen_ids);
  Py_CLEAR(p->sha_file);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  if (p->old_dirname_to_file_id) {
    e = (*v)(p->old_dirname_to_file_id, a); if (e) return e;
  }
  if (p->new_dirname_to_file_id) {
    e = (*v)(p->new_dirname_to_file_id, a); if (e) return e;
  }
  if (p->last_source_parent) {
    e = (*v)(p->last_source_parent, a); if (e) return e;
  }
  if (p->last_target_parent) {
    e = (*v)(p->last_target_parent, a); if (e) return e;
  }
  if (p->use_filesystem_for_exec) {
    e = (*v)(p->use_filesystem_for_exec, a); if (e) return e;
  }
  if (p->utf8_decode) {
    e = (*v)(p->utf8_decode, a); if (e) return e;
  }
  if (p->searched_specific_files) {
    e = (*v)(p->searched_specific_files, a); if (e) return e;
  }
  if (p->searched_exact_paths) {
    e = (*v)(p->searched_exact_paths, a); if (e) return e;
  }
  if (p->search_specific_files) {
    e = (*v)(p->search_specific_files, a); if (e) return e;
  }
  if (p->search_specific_file_parents) {
    e = (*v)(p->search_specific_file_parents, a); if (e) return e;
  }
  if (p->state) {
    e = (*v)(p->state, a); if (e) return e;
  }
  if (p->current_root) {
    e = (*v)(p->current_root, a); if (e) return e;
  }
  if (p->current_root_unicode) {
    e = (*v)(p->current_root_unicode, a); if (e) return e;
  }
  if (p->root_entries) {
    e = (*v)(p->root_entries, a); if (e) return e;
  }
  if (p->root_abspath) {
    e = (*v)(p->root_abspath, a); if (e) return e;
  }
  if (p->tree) {
    e = (*v)(p->tree, a); if (e) return e;
  }
  if (p->dir_iterator) {
    e = (*v)(p->dir_iterator, a); if (e) return e;
  }
  if (p->current_block) {
    e = (*v)(p->current_block, a); if (e) return e;
  }
  if (p->current_block_list) {
    e = (*v)(p->current_block_list, a); if (e) return e;
  }
  if (p->current_dir_info) {
    e = (*v)(p->current_dir_info, a); if (e) return e;
  }
  if (p->current_dir_list) {
    e = (*v)(p->current_dir_list, a); if (e) return e;
  }
  if (p->_pending_consistent_entries) {
    e = (*v)(p->_pending_consistent_entries, a); if (e) return e;
  }
  if (p->root_dir_info) {
    e = (*v)(p->root_dir_info, a); if (e) return e;
  }
  if (p->bisect_left) {
    e = (*v)(p->bisect_left, a); if (e) return e;
  }
  if (p->pathjoin) {
    e = (*v)(p->pathjoin, a); if (e) return e;
  }
  if (p->fstat) {
    e = (*v)(p->fstat, a); if (e) return e;
  }
  if (p->seen_ids) {
    e = (*v)(p->seen_ids, a); if (e) return e;
  }
  if (p->sha_file) {
    e = (*v)(p->sha_file, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *p = (struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *)o;
  tmp = ((PyObject*)p->old_dirname_to_file_id);
  p->old_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->new_dirname_to_file_id);
  p->new_dirname_to_file_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->last_source_parent);
  p->last_source_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->last_target_parent);
  p->last_target_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->use_filesystem_for_exec);
  p->use_filesystem_for_exec = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->utf8_decode);
  p->utf8_decode = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->searched_specific_files);
  p->searched_specific_files = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->searched_exact_paths);
  p->searched_exact_paths = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->search_specific_files);
  p->search_specific_files = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->search_specific_file_parents);
  p->search_specific_file_parents = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->state);
  p->state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_root);
  p->current_root = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_root_unicode);
  p->current_root_unicode = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_entries);
  p->root_entries = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_abspath);
  p->root_abspath = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->tree);
  p->tree = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->dir_iterator);
  p->dir_iterator = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_block);
  p->current_block = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_block_list);
  p->current_block_list = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_dir_info);
  p->current_dir_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->current_dir_list);
  p->current_dir_list = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pending_consistent_entries);
  p->_pending_consistent_entries = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->root_dir_info);
  p->root_dir_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->bisect_left);
  p->bisect_left = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pathjoin);
  p->pathjoin = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fstat);
  p->fstat = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->seen_ids);
  p->seen_ids = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->sha_file);
  p->sha_file = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_specific_files(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_23searched_specific_files_1__get__(o);
}

static PyObject *__pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_exact_paths(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_20searched_exact_paths_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__(self);}

static PyMethodDef __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC[] = {
  {"iter_changes", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_5iter_changes, METH_NOARGS, 0},
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC[] = {
  {(char *)"searched_specific_files", __pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_specific_files, 0, (char *)0, 0},
  {(char *)"searched_exact_paths", __pyx_getprop_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_searched_exact_paths, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = {
  PyVarObject_HEAD_INIT(0, 0)
  "breezy.bzr._dirstate_helpers_pyx.ProcessEntryC", /*tp_name*/
  sizeof(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_traverse*/
  __pyx_tp_clear_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_3__iter__, /*tp_iter*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_7__next__, /*tp_iternext*/
  __pyx_methods_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"update_entry", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6breezy_3bzr_21_dirstate_helpers_pyx_17update_entry, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6breezy_3bzr_21_dirstate_helpers_pyx_16update_entry},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__dirstate_helpers_pyx(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__dirstate_helpers_pyx},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_dirstate_helpers_pyx",
    __pyx_k_Helper_functions_for_DirState_Th, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_BadFileKindError, __pyx_k_BadFileKindError, sizeof(__pyx_k_BadFileKindError), 0, 0, 1, 1},
  {&__pyx_n_s_BadFilenameEncoding, __pyx_k_BadFilenameEncoding, sizeof(__pyx_k_BadFilenameEncoding), 0, 0, 1, 1},
  {&__pyx_kp_s_Bad_parse_we_expected_to_end_on, __pyx_k_Bad_parse_we_expected_to_end_on, sizeof(__pyx_k_Bad_parse_we_expected_to_end_on), 0, 0, 1, 0},
  {&__pyx_n_s_BzrError, __pyx_k_BzrError, sizeof(__pyx_k_BzrError), 0, 0, 1, 1},
  {&__pyx_kp_s_Could_not_find_target_parent_in, __pyx_k_Could_not_find_target_parent_in, sizeof(__pyx_k_Could_not_find_target_parent_in), 0, 0, 1, 0},
  {&__pyx_n_s_DirState, __pyx_k_DirState, sizeof(__pyx_k_DirState), 0, 0, 1, 1},
  {&__pyx_n_s_DirstateCorrupt, __pyx_k_DirstateCorrupt, sizeof(__pyx_k_DirstateCorrupt), 0, 0, 1, 1},
  {&__pyx_n_s_EINVAL, __pyx_k_EINVAL, sizeof(__pyx_k_EINVAL), 0, 0, 1, 1},
  {&__pyx_n_s_ENOENT, __pyx_k_ENOENT, sizeof(__pyx_k_ENOENT), 0, 0, 1, 1},
  {&__pyx_n_s_ENOTDIR, __pyx_k_ENOTDIR, sizeof(__pyx_k_ENOTDIR), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_single_character_string, __pyx_k_Expected_single_character_string, sizeof(__pyx_k_Expected_single_character_string), 0, 0, 1, 0},
  {&__pyx_kp_s_First_character_should_be_null_n, __pyx_k_First_character_should_be_null_n, sizeof(__pyx_k_First_character_should_be_null_n), 0, 0, 1, 0},
  {&__pyx_kp_s_Got_entry_path_mismatch_for_spec, __pyx_k_Got_entry_path_mismatch_for_spec, sizeof(__pyx_k_Got_entry_path_mismatch_for_spec), 0, 0, 1, 0},
  {&__pyx_n_s_IN_MEMORY_UNMODIFIED, __pyx_k_IN_MEMORY_UNMODIFIED, sizeof(__pyx_k_IN_MEMORY_UNMODIFIED), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x21, __pyx_k_Incompatible_checksums_s_vs_0x21, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x21), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x65, __pyx_k_Incompatible_checksums_s_vs_0x65, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x65), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_Missing_entry_for_specific_path, __pyx_k_Missing_entry_for_specific_path, sizeof(__pyx_k_Missing_entry_for_specific_path), 0, 0, 1, 0},
  {&__pyx_n_s_NULLSTAT, __pyx_k_NULLSTAT, sizeof(__pyx_k_NULLSTAT), 0, 0, 1, 1},
  {&__pyx_n_s_NULL_PARENT_DETAILS, __pyx_k_NULL_PARENT_DETAILS, sizeof(__pyx_k_NULL_PARENT_DETAILS), 0, 0, 1, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_ProcessEntryC, __pyx_k_ProcessEntryC, sizeof(__pyx_k_ProcessEntryC), 0, 0, 1, 1},
  {&__pyx_n_s_Reader, __pyx_k_Reader, sizeof(__pyx_k_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_S_IEXEC, __pyx_k_S_IEXEC, sizeof(__pyx_k_S_IEXEC), 0, 0, 1, 1},
  {&__pyx_n_s_S_ISREG, __pyx_k_S_ISREG, sizeof(__pyx_k_S_ISREG), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TreeChange, __pyx_k_TreeChange, sizeof(__pyx_k_TreeChange), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UnicodeDecodeError, __pyx_k_UnicodeDecodeError, sizeof(__pyx_k_UnicodeDecodeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unsupported_target_index_d, __pyx_k_Unsupported_target_index_d, sizeof(__pyx_k_Unsupported_target_index_d), 0, 0, 1, 0},
  {&__pyx_kp_s_We_could_not_find_the_parent_ent, __pyx_k_We_could_not_find_the_parent_ent, sizeof(__pyx_k_We_could_not_find_the_parent_ent), 0, 0, 1, 0},
  {&__pyx_kp_s_We_read_the_wrong_number_of_entr, __pyx_k_We_read_the_wrong_number_of_entr, sizeof(__pyx_k_We_read_the_wrong_number_of_entr), 0, 0, 1, 0},
  {&__pyx_kp_b__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 0, 0, 0},
  {&__pyx_kp_b__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 0, 0},
  {&__pyx_kp_b__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 0, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_n_b_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 0, 1},
  {&__pyx_n_s_absent, __pyx_k_absent, sizeof(__pyx_k_absent), 0, 0, 1, 1},
  {&__pyx_n_s_abspath, __pyx_k_abspath, sizeof(__pyx_k_abspath), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_b2a_base64, __pyx_k_b2a_base64, sizeof(__pyx_k_b2a_base64), 0, 0, 1, 1},
  {&__pyx_n_s_binascii, __pyx_k_binascii, sizeof(__pyx_k_binascii), 0, 0, 1, 1},
  {&__pyx_n_s_bisect, __pyx_k_bisect, sizeof(__pyx_k_bisect), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_dirblock, __pyx_k_bisect_dirblock, sizeof(__pyx_k_bisect_dirblock), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_left, __pyx_k_bisect_left, sizeof(__pyx_k_bisect_left), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_path_left, __pyx_k_bisect_path_left, sizeof(__pyx_k_bisect_path_left), 0, 0, 1, 1},
  {&__pyx_n_s_bisect_path_right, __pyx_k_bisect_path_right, sizeof(__pyx_k_bisect_path_right), 0, 0, 1, 1},
  {&__pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_k_breezy_bzr__dirstate_helpers_pyx, sizeof(__pyx_k_breezy_bzr__dirstate_helpers_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2, __pyx_k_breezy_bzr__dirstate_helpers_pyx_2, sizeof(__pyx_k_breezy_bzr__dirstate_helpers_pyx_2), 0, 0, 1, 1},
  {&__pyx_kp_b_bzr, __pyx_k_bzr, sizeof(__pyx_k_bzr), 0, 0, 0, 0},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_c_2, __pyx_k_c_2, sizeof(__pyx_k_c_2), 0, 0, 1, 1},
  {&__pyx_n_s_cache, __pyx_k_cache, sizeof(__pyx_k_cache), 0, 0, 1, 1},
  {&__pyx_n_s_cache_utf8, __pyx_k_cache_utf8, sizeof(__pyx_k_cache_utf8), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cur, __pyx_k_cur, sizeof(__pyx_k_cur), 0, 0, 1, 1},
  {&__pyx_n_s_cur_cstr, __pyx_k_cur_cstr, sizeof(__pyx_k_cur_cstr), 0, 0, 1, 1},
  {&__pyx_n_s_cur_size, __pyx_k_cur_size, sizeof(__pyx_k_cur_size), 0, 0, 1, 1},
  {&__pyx_n_s_cutoff_time, __pyx_k_cutoff_time, sizeof(__pyx_k_cutoff_time), 0, 0, 1, 1},
  {&__pyx_n_b_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 0, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dirblock_state, __pyx_k_dirblock_state, sizeof(__pyx_k_dirblock_state), 0, 0, 1, 1},
  {&__pyx_n_s_dirblocks, __pyx_k_dirblocks, sizeof(__pyx_k_dirblocks), 0, 0, 1, 1},
  {&__pyx_n_s_dirblocks_2, __pyx_k_dirblocks_2, sizeof(__pyx_k_dirblocks_2), 0, 0, 1, 1},
  {&__pyx_n_s_directory, __pyx_k_directory, sizeof(__pyx_k_directory), 0, 0, 1, 1},
  {&__pyx_n_s_directory_is_tree_reference, __pyx_k_directory_is_tree_reference, sizeof(__pyx_k_directory_is_tree_reference), 0, 0, 1, 1},
  {&__pyx_n_s_dirname, __pyx_k_dirname, sizeof(__pyx_k_dirname), 0, 0, 1, 1},
  {&__pyx_n_s_dirname_cstr, __pyx_k_dirname_cstr, sizeof(__pyx_k_dirname_cstr), 0, 0, 1, 1},
  {&__pyx_n_s_dirname_size, __pyx_k_dirname_size, sizeof(__pyx_k_dirname_size), 0, 0, 1, 1},
  {&__pyx_n_s_dirstate, __pyx_k_dirstate, sizeof(__pyx_k_dirstate), 0, 0, 1, 1},
  {&__pyx_kp_s_don_t_know_how_to_compare_source, __pyx_k_don_t_know_how_to_compare_source, sizeof(__pyx_k_don_t_know_how_to_compare_source), 0, 0, 1, 0},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encode_2, __pyx_k_encode_2, sizeof(__pyx_k_encode_2), 0, 0, 1, 1},
  {&__pyx_n_s_end_of_header, __pyx_k_end_of_header, sizeof(__pyx_k_end_of_header), 0, 0, 1, 1},
  {&__pyx_n_s_ensure_block, __pyx_k_ensure_block, sizeof(__pyx_k_ensure_block), 0, 0, 1, 1},
  {&__pyx_n_s_entries_for_path, __pyx_k_entries_for_path, sizeof(__pyx_k_entries_for_path), 0, 0, 1, 1},
  {&__pyx_n_s_entry, __pyx_k_entry, sizeof(__pyx_k_entry), 0, 0, 1, 1},
  {&__pyx_kp_s_entry_s_s_is_considered_renamed, __pyx_k_entry_s_s_is_considered_renamed, sizeof(__pyx_k_entry_s_s_is_considered_renamed), 0, 0, 1, 0},
  {&__pyx_n_s_errno, __pyx_k_errno, sizeof(__pyx_k_errno), 0, 0, 1, 1},
  {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
  {&__pyx_n_b_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 0, 1},
  {&__pyx_kp_s_failed_to_find_trailing_NULL_0_T, __pyx_k_failed_to_find_trailing_NULL_0_T, sizeof(__pyx_k_failed_to_find_trailing_NULL_0_T), 0, 0, 1, 0},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_file_id, __pyx_k_file_id, sizeof(__pyx_k_file_id), 0, 0, 1, 1},
  {&__pyx_n_s_file_kind_from_stat_mode, __pyx_k_file_kind_from_stat_mode, sizeof(__pyx_k_file_kind_from_stat_mode), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_find_block_index_from_key, __pyx_k_find_block_index_from_key, sizeof(__pyx_k_find_block_index_from_key), 0, 0, 1, 1},
  {&__pyx_n_s_found, __pyx_k_found, sizeof(__pyx_k_found), 0, 0, 1, 1},
  {&__pyx_n_s_fs_enc, __pyx_k_fs_enc, sizeof(__pyx_k_fs_enc), 0, 0, 1, 1},
  {&__pyx_n_s_fstat, __pyx_k_fstat, sizeof(__pyx_k_fstat), 0, 0, 1, 1},
  {&__pyx_n_s_get_block_entry_index, __pyx_k_get_block_entry_index, sizeof(__pyx_k_get_block_entry_index), 0, 0, 1, 1},
  {&__pyx_n_s_get_entry, __pyx_k_get_entry, sizeof(__pyx_k_get_entry), 0, 0, 1, 1},
  {&__pyx_kp_s_get_next_called_when_cur_str_is, __pyx_k_get_next_called_when_cur_str_is, sizeof(__pyx_k_get_next_called_when_cur_str_is), 0, 0, 1, 0},
  {&__pyx_kp_s_get_next_called_when_there_are_n, __pyx_k_get_next_called_when_there_are_n, sizeof(__pyx_k_get_next_called_when_there_are_n), 0, 0, 1, 0},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hi, __pyx_k_hi, sizeof(__pyx_k_hi), 0, 0, 1, 1},
  {&__pyx_n_s_hi_2, __pyx_k_hi_2, sizeof(__pyx_k_hi_2), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_include_unchanged, __pyx_k_include_unchanged, sizeof(__pyx_k_include_unchanged), 0, 0, 1, 1},
  {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
  {&__pyx_n_s_is_executable, __pyx_k_is_executable, sizeof(__pyx_k_is_executable), 0, 0, 1, 1},
  {&__pyx_n_s_is_inside, __pyx_k_is_inside, sizeof(__pyx_k_is_inside), 0, 0, 1, 1},
  {&__pyx_n_s_is_inside_any, __pyx_k_is_inside_any, sizeof(__pyx_k_is_inside_any), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_b_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 0, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_lo, __pyx_k_lo, sizeof(__pyx_k_lo), 0, 0, 1, 1},
  {&__pyx_n_s_lo_2, __pyx_k_lo_2, sizeof(__pyx_k_lo_2), 0, 0, 1, 1},
  {&__pyx_n_s_lstat, __pyx_k_lstat, sizeof(__pyx_k_lstat), 0, 0, 1, 1},
  {&__pyx_n_s_lt_by_dirs, __pyx_k_lt_by_dirs, sizeof(__pyx_k_lt_by_dirs), 0, 0, 1, 1},
  {&__pyx_n_s_lt_path_by_dirblock, __pyx_k_lt_path_by_dirblock, sizeof(__pyx_k_lt_path_by_dirblock), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mark_modified, __pyx_k_mark_modified, sizeof(__pyx_k_mark_modified), 0, 0, 1, 1},
  {&__pyx_n_s_mid, __pyx_k_mid, sizeof(__pyx_k_mid), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_ne, __pyx_k_ne, sizeof(__pyx_k_ne), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_num_entries, __pyx_k_num_entries, sizeof(__pyx_k_num_entries), 0, 0, 1, 1},
  {&__pyx_n_s_num_present_parents, __pyx_k_num_present_parents, sizeof(__pyx_k_num_present_parents), 0, 0, 1, 1},
  {&__pyx_n_s_observed_sha1, __pyx_k_observed_sha1, sizeof(__pyx_k_observed_sha1), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_osutils, __pyx_k_osutils, sizeof(__pyx_k_osutils), 0, 0, 1, 1},
  {&__pyx_n_s_pack_stat, __pyx_k_pack_stat, sizeof(__pyx_k_pack_stat), 0, 0, 1, 1},
  {&__pyx_n_s_parent_directories, __pyx_k_parent_directories, sizeof(__pyx_k_parent_directories), 0, 0, 1, 1},
  {&__pyx_n_s_parse_dirblocks, __pyx_k_parse_dirblocks, sizeof(__pyx_k_parse_dirblocks), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_path1, __pyx_k_path1, sizeof(__pyx_k_path1), 0, 0, 1, 1},
  {&__pyx_kp_s_path1_must_be_a_bytes_string_no, __pyx_k_path1_must_be_a_bytes_string_no, sizeof(__pyx_k_path1_must_be_a_bytes_string_no), 0, 0, 1, 0},
  {&__pyx_kp_s_path1_must_be_a_plain_string_no, __pyx_k_path1_must_be_a_plain_string_no, sizeof(__pyx_k_path1_must_be_a_plain_string_no), 0, 0, 1, 0},
  {&__pyx_n_s_path2, __pyx_k_path2, sizeof(__pyx_k_path2), 0, 0, 1, 1},
  {&__pyx_kp_s_path2_must_be_a_bytes_string_no, __pyx_k_path2_must_be_a_bytes_string_no, sizeof(__pyx_k_path2_must_be_a_bytes_string_no), 0, 0, 1, 0},
  {&__pyx_kp_s_path2_must_be_a_plain_string_no, __pyx_k_path2_must_be_a_plain_string_no, sizeof(__pyx_k_path2_must_be_a_plain_string_no), 0, 0, 1, 0},
  {&__pyx_n_s_path_cstr, __pyx_k_path_cstr, sizeof(__pyx_k_path_cstr), 0, 0, 1, 1},
  {&__pyx_n_s_path_size, __pyx_k_path_size, sizeof(__pyx_k_path_size), 0, 0, 1, 1},
  {&__pyx_n_s_path_utf8, __pyx_k_path_utf8, sizeof(__pyx_k_path_utf8), 0, 0, 1, 1},
  {&__pyx_n_s_pathjoin, __pyx_k_pathjoin, sizeof(__pyx_k_pathjoin), 0, 0, 1, 1},
  {&__pyx_n_s_paths, __pyx_k_paths, sizeof(__pyx_k_paths), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_platform, __pyx_k_platform, sizeof(__pyx_k_platform), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_prefix, __pyx_k_prefix, sizeof(__pyx_k_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_py_memrchr, __pyx_k_py_memrchr, sizeof(__pyx_k_py_memrchr), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ProcessEntryC, __pyx_k_pyx_unpickle_ProcessEntryC, sizeof(__pyx_k_pyx_unpickle_ProcessEntryC), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Reader, __pyx_k_pyx_unpickle_Reader, sizeof(__pyx_k_pyx_unpickle_Reader), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_b_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 0, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_read_dirblocks, __pyx_k_read_dirblocks, sizeof(__pyx_k_read_dirblocks), 0, 0, 1, 1},
  {&__pyx_n_s_read_link, __pyx_k_read_link, sizeof(__pyx_k_read_link), 0, 0, 1, 1},
  {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_relocated, __pyx_k_relocated, sizeof(__pyx_k_relocated), 0, 0, 1, 1},
  {&__pyx_kp_s_result_is_not_None_r, __pyx_k_result_is_not_None_r, sizeof(__pyx_k_result_is_not_None_r), 0, 0, 1, 0},
  {&__pyx_n_s_rsplit, __pyx_k_rsplit, sizeof(__pyx_k_rsplit), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_s_2, __pyx_k_s_2, sizeof(__pyx_k_s_2), 0, 0, 1, 1},
  {&__pyx_n_s_search_specific_files, __pyx_k_search_specific_files, sizeof(__pyx_k_search_specific_files), 0, 0, 1, 1},
  {&__pyx_n_s_seek, __pyx_k_seek, sizeof(__pyx_k_seek), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sha1_file, __pyx_k_sha1_file, sizeof(__pyx_k_sha1_file), 0, 0, 1, 1},
  {&__pyx_n_s_sha1_provider, __pyx_k_sha1_provider, sizeof(__pyx_k_sha1_provider), 0, 0, 1, 1},
  {&__pyx_n_s_sha_cutoff_time, __pyx_k_sha_cutoff_time, sizeof(__pyx_k_sha_cutoff_time), 0, 0, 1, 1},
  {&__pyx_n_s_sha_file, __pyx_k_sha_file, sizeof(__pyx_k_sha_file), 0, 0, 1, 1},
  {&__pyx_n_s_source_index, __pyx_k_source_index, sizeof(__pyx_k_source_index), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_split_root_dirblock_into_conten, __pyx_k_split_root_dirblock_into_conten, sizeof(__pyx_k_split_root_dirblock_into_conten), 0, 0, 1, 1},
  {&__pyx_n_s_splitpath, __pyx_k_splitpath, sizeof(__pyx_k_splitpath), 0, 0, 1, 1},
  {&__pyx_n_s_st_ctime, __pyx_k_st_ctime, sizeof(__pyx_k_st_ctime), 0, 0, 1, 1},
  {&__pyx_n_s_st_dev, __pyx_k_st_dev, sizeof(__pyx_k_st_dev), 0, 0, 1, 1},
  {&__pyx_n_s_st_ino, __pyx_k_st_ino, sizeof(__pyx_k_st_ino), 0, 0, 1, 1},
  {&__pyx_n_s_st_mode, __pyx_k_st_mode, sizeof(__pyx_k_st_mode), 0, 0, 1, 1},
  {&__pyx_n_s_st_mtime, __pyx_k_st_mtime, sizeof(__pyx_k_st_mtime), 0, 0, 1, 1},
  {&__pyx_n_s_st_size, __pyx_k_st_size, sizeof(__pyx_k_st_size), 0, 0, 1, 1},
  {&__pyx_n_s_stat, __pyx_k_stat, sizeof(__pyx_k_stat), 0, 0, 1, 1},
  {&__pyx_n_s_stat_and_sha1, __pyx_k_stat_and_sha1, sizeof(__pyx_k_stat_and_sha1), 0, 0, 1, 1},
  {&__pyx_n_s_stat_value, __pyx_k_stat_value, sizeof(__pyx_k_stat_value), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_state_file, __pyx_k_state_file, sizeof(__pyx_k_state_file), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_symlink, __pyx_k_symlink, sizeof(__pyx_k_symlink), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_target_index, __pyx_k_target_index, sizeof(__pyx_k_target_index), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_tree, __pyx_k_tree, sizeof(__pyx_k_tree), 0, 0, 1, 1},
  {&__pyx_kp_s_tree_reference, __pyx_k_tree_reference, sizeof(__pyx_k_tree_reference), 0, 0, 1, 0},
  {&__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_k_tried_to_create_a_string_with_an, sizeof(__pyx_k_tried_to_create_a_string_with_an), 0, 0, 1, 0},
  {&__pyx_kp_s_unsupported_target_index, __pyx_k_unsupported_target_index, sizeof(__pyx_k_unsupported_target_index), 0, 0, 1, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_use_filesystem_for_exec, __pyx_k_use_filesystem_for_exec, sizeof(__pyx_k_use_filesystem_for_exec), 0, 0, 1, 1},
  {&__pyx_n_s_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 0, 1, 1},
  {&__pyx_n_s_utf8_decode, __pyx_k_utf8_decode, sizeof(__pyx_k_utf8_decode), 0, 0, 1, 1},
  {&__pyx_n_s_walkdirs_utf8, __pyx_k_walkdirs_utf8, sizeof(__pyx_k_walkdirs_utf8), 0, 0, 1, 1},
  {&__pyx_n_s_want_unversioned, __pyx_k_want_unversioned, sizeof(__pyx_k_want_unversioned), 0, 0, 1, 1},
  {&__pyx_n_s_win32, __pyx_k_win32, sizeof(__pyx_k_win32), 0, 0, 1, 1},
  {&__pyx_n_s_winerror, __pyx_k_winerror, sizeof(__pyx_k_winerror), 0, 0, 1, 1},
  {&__pyx_kp_s_you_must_pass_a_python_list_for, __pyx_k_you_must_pass_a_python_list_for, sizeof(__pyx_k_you_must_pass_a_python_list_for), 0, 0, 1, 0},
  {&__pyx_kp_s_you_must_pass_a_python_list_for_2, __pyx_k_you_must_pass_a_python_list_for_2, sizeof(__pyx_k_you_must_pass_a_python_list_for_2), 0, 0, 1, 0},
  {&__pyx_kp_s_you_must_pass_a_string_for_dirna, __pyx_k_you_must_pass_a_string_for_dirna, sizeof(__pyx_k_you_must_pass_a_string_for_dirna), 0, 0, 1, 0},
  {&__pyx_kp_s_you_must_pass_a_string_for_path, __pyx_k_you_must_pass_a_string_for_path, sizeof(__pyx_k_you_must_pass_a_string_for_path), 0, 0, 1, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 250, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 543, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 983, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 1513, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 1523, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 1602, __pyx_L1_error)
  __pyx_builtin_UnicodeDecodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeDecodeError); if (!__pyx_builtin_UnicodeDecodeError) __PYX_ERR(0, 1841, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":543
 * 
 *         if self.cur_cstr == NULL:
 *             raise AssertionError('get_next() called when cur_str is NULL')             # <<<<<<<<<<<<<<
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_get_next_called_when_cur_str_is); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":545
 *             raise AssertionError('get_next() called when cur_str is NULL')
 *         elif self.cur_cstr >= self.end_cstr:
 *             raise AssertionError('get_next() called when there are no chars'             # <<<<<<<<<<<<<<
 *                                  ' left')
 *         next = self.cur_cstr
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_get_next_called_when_there_are_n); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":823
 *     aliased[5] = htonl(PyInt_AsUnsignedLongMask(stat_value.st_mode))
 *     packed = PyBytes_FromStringAndSize(result, 6*4)
 *     return _encode(packed)[:-1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_slice__5 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(0, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1174
 *                 # update the source details variable to be the real
 *                 # location.
 *                 if old_entry == (None, None):             # <<<<<<<<<<<<<<
 *                     raise DirstateCorrupt(self.state._filename,
 *                         "entry '%s/%s' is considered renamed from %r"
 */
  __pyx_tuple__6 = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1292
 *                 )
 *             if not changed and not self.include_unchanged:
 *                 return None, False             # <<<<<<<<<<<<<<
 *             else:
 *                 if old_path is None:
 */
  __pyx_tuple__7 = PyTuple_Pack(2, Py_None, Py_False); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 1292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1309
 *                        (old_path_u, path_u),
 *                        content_change,
 *                        (True, True),             # <<<<<<<<<<<<<<
 *                        (source_parent_id, target_parent_id),
 *                        (self.utf8_decode(old_basename)[0], self.utf8_decode(entry[0][1])[0]),
 */
  __pyx_tuple__8 = PyTuple_Pack(2, Py_True, Py_True); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 1309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1342
 *                        (None, self.utf8_decode(path)[0]),
 *                        True,
 *                        (False, True),             # <<<<<<<<<<<<<<
 *                        (None, parent_id),
 *                        (None, self.utf8_decode(entry[0][1])[0]),
 */
  __pyx_tuple__9 = PyTuple_Pack(2, Py_False, Py_True); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 1342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1371
 *                    (self.utf8_decode(old_path)[0], None),
 *                    True,
 *                    (True, False),             # <<<<<<<<<<<<<<
 *                    (parent_id, None),
 *                    (self.utf8_decode(entry[0][1])[0], None),
 */
  __pyx_tuple__10 = PyTuple_Pack(2, Py_True, Py_False); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1537
 *                     if self.tree._directory_is_tree_reference(
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \             # <<<<<<<<<<<<<<
 *                             ('tree-reference',) + self.root_dir_info[3:]
 *             if not self.root_entries and not self.root_dir_info:
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1538
 *                         self.current_root_unicode):
 *                         self.root_dir_info = self.root_dir_info[:2] + \
 *                             ('tree-reference',) + self.root_dir_info[3:]             # <<<<<<<<<<<<<<
 *             if not self.root_entries and not self.root_dir_info:
 *                 # this specified path is not present at all, skip it.
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_tree_reference); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_slice__13 = PySlice_New(__pyx_int_3, Py_None, Py_None); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1569
 *                        (None, self.current_root_unicode),
 *                        True,
 *                        (False, False),             # <<<<<<<<<<<<<<
 *                        (None, None),
 *                        (None, splitpath(self.current_root_unicode)[-1]),
 */
  __pyx_tuple__14 = PyTuple_Pack(2, Py_False, Py_False); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 1569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1615
 *                     if self.current_dir_info[0][0] == b'':
 *                         # remove .bzr from iteration
 *                         bzr_index = self.bisect_left(self.current_dir_list, (b'.bzr',))             # <<<<<<<<<<<<<<
 *                         if self.current_dir_list[bzr_index][0] != b'.bzr':
 *                             raise AssertionError()
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_b_bzr); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 1615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":2002
 *             else:
 *                 raise
 *         utf8_basename = utf8_path.rsplit(b'/', 1)[-1]             # <<<<<<<<<<<<<<
 *         dir_info = (utf8_path, utf8_basename,
 *             osutils.file_kind_from_stat_mode(stat.st_mode), stat,
 */
  __pyx_tuple__17 = PyTuple_Pack(2, __pyx_kp_b__16, __pyx_int_1); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 2002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":139
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(6, __pyx_n_s_s, __pyx_n_s_c, __pyx_n_s_s_2, __pyx_n_s_found, __pyx_n_s_length, __pyx_n_s_c_2); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_py_memrchr, 139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 139, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":234
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_s_path1, __pyx_n_s_path2); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_lt_by_dirs, 234, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 234, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":261
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_path1, __pyx_n_s_path2); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_lt_path_by_dirblock, 261, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 261, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":362
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(10, __pyx_n_s_paths, __pyx_n_s_path, __pyx_n_s_lo_2, __pyx_n_s_hi_2, __pyx_n_s_mid, __pyx_n_s_path_cstr, __pyx_n_s_path_size, __pyx_n_s_cur_cstr, __pyx_n_s_cur_size, __pyx_n_s_cur); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_bisect_path_left, 362, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 362, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":415
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(10, __pyx_n_s_paths, __pyx_n_s_path, __pyx_n_s_lo_2, __pyx_n_s_hi_2, __pyx_n_s_mid, __pyx_n_s_path_cstr, __pyx_n_s_path_size, __pyx_n_s_cur_cstr, __pyx_n_s_cur_size, __pyx_n_s_cur); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_bisect_path_right, 415, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 415, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":468
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(13, __pyx_n_s_dirblocks, __pyx_n_s_dirname, __pyx_n_s_lo, __pyx_n_s_hi, __pyx_n_s_cache, __pyx_n_s_lo_2, __pyx_n_s_hi_2, __pyx_n_s_mid, __pyx_n_s_dirname_cstr, __pyx_n_s_dirname_size, __pyx_n_s_cur_cstr, __pyx_n_s_cur_size, __pyx_n_s_cur); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(5, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_bisect_dirblock, 468, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 468, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":764
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(3, __pyx_n_s_state, __pyx_n_s_text, __pyx_n_s_reader); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_read_dirblocks, 764, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 764, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":826
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_n_s_stat_value); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_breezy_bzr__dirstate_helpers_pyx, __pyx_n_s_pack_stat, 826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 826, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__34 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Reader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__36 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ProcessEntryC, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_35494349 = PyInt_FromLong(35494349L); if (unlikely(!__pyx_int_35494349)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_106938589 = PyInt_FromLong(106938589L); if (unlikely(!__pyx_int_106938589)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated = Py_None; Py_INCREF(Py_None);
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = &__pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.get_next = (char *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.get_next_str = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader_get_next_str;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader._init = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__init;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_Reader._get_entry = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_Reader *, int, void **, int *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_6Reader__get_entry;
  if (PyType_Ready(&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.tp_dictoffset && __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader.tp_dict, __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Reader, (PyObject *)&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader) < 0) __PYX_ERR(0, 517, __pyx_L1_error)
  __pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_Reader = &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_Reader;
  __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = &__pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._process_entry = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__process_entry;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._gather_result_for_consistency = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__gather_result_for_consistency;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._update_current_block = (int (*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__update_current_block;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._iter_next = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__iter_next;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._maybe_tree_ref = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__maybe_tree_ref;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._loop_one_block = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__loop_one_block;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._next_consistent_entries = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__next_consistent_entries;
  __pyx_vtable_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC._path_info = (PyObject *(*)(struct __pyx_obj_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC *, PyObject *, PyObject *))__pyx_f_6breezy_3bzr_21_dirstate_helpers_pyx_13ProcessEntryC__path_info;
  if (PyType_Ready(&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < 0) __PYX_ERR(0, 994, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC.tp_dictoffset && __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC.tp_dict, __pyx_vtabptr_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < 0) __PYX_ERR(0, 994, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ProcessEntryC, (PyObject *)&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < 0) __PYX_ERR(0, 994, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC) < 0) __PYX_ERR(0, 994, __pyx_L1_error)
  __pyx_ptype_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC = &__pyx_type_6breezy_3bzr_21_dirstate_helpers_pyx_ProcessEntryC;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("breezy._static_tuple_c"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_6breezy_15_static_tuple_c_StaticTuple = __Pyx_ImportType(__pyx_t_1, "breezy._static_tuple_c", "StaticTuple", sizeof(StaticTuple), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_6breezy_15_static_tuple_c_StaticTuple) __PYX_ERR(3, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_dirstate_helpers_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_dirstate_helpers_pyx(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__dirstate_helpers_pyx(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__dirstate_helpers_pyx(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__dirstate_helpers_pyx(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_dirstate_helpers_pyx' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__dirstate_helpers_pyx(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_dirstate_helpers_pyx", __pyx_methods, __pyx_k_Helper_functions_for_DirState_Th, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_breezy__bzr___dirstate_helpers_pyx) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "breezy.bzr._dirstate_helpers_pyx")) {
      if (unlikely(PyDict_SetItemString(modules, "breezy.bzr._dirstate_helpers_pyx", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":24
 * from __future__ import absolute_import
 * 
 * import binascii             # <<<<<<<<<<<<<<
 * import bisect
 * import errno
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_binascii, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_binascii, __pyx_t_1) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":25
 * 
 * import binascii
 * import bisect             # <<<<<<<<<<<<<<
 * import errno
 * import os
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_bisect, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":26
 * import binascii
 * import bisect
 * import errno             # <<<<<<<<<<<<<<
 * import os
 * import stat
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_errno, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_errno, __pyx_t_1) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":27
 * import bisect
 * import errno
 * import os             # <<<<<<<<<<<<<<
 * import stat
 * import sys
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":28
 * import errno
 * import os
 * import stat             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_stat, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stat, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":29
 * import os
 * import stat
 * import sys             # <<<<<<<<<<<<<<
 * 
 * from .. import cache_utf8, errors, osutils
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":31
 * import sys
 * 
 * from .. import cache_utf8, errors, osutils             # <<<<<<<<<<<<<<
 * from .dirstate import DirState, DirstateCorrupt
 * from ..osutils import parent_directories, pathjoin, splitpath, is_inside_any, is_inside
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_cache_utf8);
  __Pyx_GIVEREF(__pyx_n_s_cache_utf8);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_cache_utf8);
  __Pyx_INCREF(__pyx_n_s_errors);
  __Pyx_GIVEREF(__pyx_n_s_errors);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_errors);
  __Pyx_INCREF(__pyx_n_s_osutils);
  __Pyx_GIVEREF(__pyx_n_s_osutils);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_osutils);
  __pyx_t_2 = __Pyx_Import(__pyx_kp_s__4, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_cache_utf8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cache_utf8, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_errors); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_errors, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_osutils); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_osutils, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":32
 * 
 * from .. import cache_utf8, errors, osutils
 * from .dirstate import DirState, DirstateCorrupt             # <<<<<<<<<<<<<<
 * from ..osutils import parent_directories, pathjoin, splitpath, is_inside_any, is_inside
 * from ..tree import TreeChange
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_DirState);
  __Pyx_GIVEREF(__pyx_n_s_DirState);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_DirState);
  __Pyx_INCREF(__pyx_n_s_DirstateCorrupt);
  __Pyx_GIVEREF(__pyx_n_s_DirstateCorrupt);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_DirstateCorrupt);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_dirstate, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_DirState); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DirState, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_DirstateCorrupt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DirstateCorrupt, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":33
 * from .. import cache_utf8, errors, osutils
 * from .dirstate import DirState, DirstateCorrupt
 * from ..osutils import parent_directories, pathjoin, splitpath, is_inside_any, is_inside             # <<<<<<<<<<<<<<
 * from ..tree import TreeChange
 * 
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_parent_directories);
  __Pyx_GIVEREF(__pyx_n_s_parent_directories);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_parent_directories);
  __Pyx_INCREF(__pyx_n_s_pathjoin);
  __Pyx_GIVEREF(__pyx_n_s_pathjoin);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_pathjoin);
  __Pyx_INCREF(__pyx_n_s_splitpath);
  __Pyx_GIVEREF(__pyx_n_s_splitpath);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_splitpath);
  __Pyx_INCREF(__pyx_n_s_is_inside_any);
  __Pyx_GIVEREF(__pyx_n_s_is_inside_any);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_is_inside_any);
  __Pyx_INCREF(__pyx_n_s_is_inside);
  __Pyx_GIVEREF(__pyx_n_s_is_inside);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_is_inside);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_osutils, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_parent_directories); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parent_directories, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_pathjoin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pathjoin, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_splitpath); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_splitpath, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_is_inside_any); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_inside_any, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_is_inside); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_is_inside, __pyx_t_1) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":34
 * from .dirstate import DirState, DirstateCorrupt
 * from ..osutils import parent_directories, pathjoin, splitpath, is_inside_any, is_inside
 * from ..tree import TreeChange             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_TreeChange);
  __Pyx_GIVEREF(__pyx_n_s_TreeChange);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_TreeChange);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_tree, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_TreeChange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TreeChange, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":42
 * # XXX: Perhaps we could get it from a windows header ?
 * cdef int ERROR_PATH_NOT_FOUND
 * ERROR_PATH_NOT_FOUND = 3             # <<<<<<<<<<<<<<
 * cdef int ERROR_DIRECTORY
 * ERROR_DIRECTORY = 267
 */
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_PATH_NOT_FOUND = 3;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":44
 * ERROR_PATH_NOT_FOUND = 3
 * cdef int ERROR_DIRECTORY
 * ERROR_DIRECTORY = 267             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "python-compat.h":
 */
  __pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx_ERROR_DIRECTORY = 0x10B;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":136
 *     )
 * 
 * import_static_tuple_c()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = import_static_tuple_c(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 136, __pyx_L1_error)

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":139
 * 
 * 
 * def _py_memrchr(s, c):             # <<<<<<<<<<<<<<
 *     """Just to expose _my_memrchr for testing.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_1_py_memrchr, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_memrchr, __pyx_t_1) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":234
 * 
 * 
 * def lt_by_dirs(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths directory by directory.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_3lt_by_dirs, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lt_by_dirs, __pyx_t_1) < 0) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":261
 * 
 * 
 * def _lt_path_by_dirblock(path1, path2):             # <<<<<<<<<<<<<<
 *     """Compare two paths based on what directory they are in.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_5_lt_path_by_dirblock, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lt_path_by_dirblock, __pyx_t_1) < 0) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":362
 * 
 * 
 * def _bisect_path_left(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_7_bisect_path_left, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect_path_left, __pyx_t_1) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":415
 * 
 * 
 * def _bisect_path_right(paths, path):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert path into paths.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_9_bisect_path_right, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect_path_right, __pyx_t_1) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":468
 * 
 * 
 * def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache=None):             # <<<<<<<<<<<<<<
 *     """Return the index where to insert dirname into the dirblocks.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_11bisect_dirblock, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bisect_dirblock, __pyx_t_1) < 0) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":764
 * 
 * 
 * def _read_dirblocks(state):             # <<<<<<<<<<<<<<
 *     """Read in the dirblocks for the given DirState object.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_13_read_dirblocks, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_read_dirblocks, __pyx_t_1) < 0) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":797
 * 
 * 
 * _encode = binascii.b2a_base64             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_binascii); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_b2a_base64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_encode, __pyx_t_2) < 0) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":826
 * 
 * 
 * def pack_stat(stat_value):             # <<<<<<<<<<<<<<
 *     """Convert stat value into a packed representation quickly with pyrex"""
 *     return _pack_stat(stat_value)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_15pack_stat, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pack_stat, __pyx_t_2) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":959
 * cdef object _kind_relocated
 * cdef object _kind_tree_reference
 * _kind_absent = "absent"             # <<<<<<<<<<<<<<
 * _kind_file = "file"
 * _kind_directory = "directory"
 */
  __Pyx_INCREF(__pyx_n_s_absent);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_absent, __pyx_n_s_absent);
  __Pyx_GIVEREF(__pyx_n_s_absent);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":960
 * cdef object _kind_tree_reference
 * _kind_absent = "absent"
 * _kind_file = "file"             # <<<<<<<<<<<<<<
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"
 */
  __Pyx_INCREF(__pyx_n_s_file);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_file, __pyx_n_s_file);
  __Pyx_GIVEREF(__pyx_n_s_file);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":961
 * _kind_absent = "absent"
 * _kind_file = "file"
 * _kind_directory = "directory"             # <<<<<<<<<<<<<<
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"
 */
  __Pyx_INCREF(__pyx_n_s_directory);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_directory, __pyx_n_s_directory);
  __Pyx_GIVEREF(__pyx_n_s_directory);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":962
 * _kind_file = "file"
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"             # <<<<<<<<<<<<<<
 * _kind_relocated = "relocated"
 * _kind_tree_reference = "tree-reference"
 */
  __Pyx_INCREF(__pyx_n_s_symlink);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_symlink, __pyx_n_s_symlink);
  __Pyx_GIVEREF(__pyx_n_s_symlink);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":963
 * _kind_directory = "directory"
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"             # <<<<<<<<<<<<<<
 * _kind_tree_reference = "tree-reference"
 * 
 */
  __Pyx_INCREF(__pyx_n_s_relocated);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_relocated, __pyx_n_s_relocated);
  __Pyx_GIVEREF(__pyx_n_s_relocated);

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":964
 * _kind_symlink = "symlink"
 * _kind_relocated = "relocated"
 * _kind_tree_reference = "tree-reference"             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_kp_s_tree_reference);
  __Pyx_XGOTREF(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference);
  __Pyx_DECREF_SET(__pyx_v_6breezy_3bzr_21_dirstate_helpers_pyx__kind_tree_reference, __pyx_kp_s_tree_reference);
  __Pyx_GIVEREF(__pyx_kp_s_tree_reference);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Reader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_19__pyx_unpickle_Reader, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Reader, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Reader__set_state(<Reader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Reader__set_state(Reader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.cur_cstr = __pyx_state[0]; __pyx_result.end_cstr = __pyx_state[1]; __pyx_result.next = __pyx_state[2]; __pyx_result.state = __pyx_state[3]; __pyx_result.text = __pyx_state[4]; __pyx_result.text_cstr = __pyx_state[5]; __pyx_result.text_size = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6breezy_3bzr_21_dirstate_helpers_pyx_21__pyx_unpickle_ProcessEntryC, NULL, __pyx_n_s_breezy_bzr__dirstate_helpers_pyx_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ProcessEntryC, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "breezy/bzr/_dirstate_helpers_pyx.pyx":1
 * # Copyright (C) 2007-2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "_str_helpers.pxd":48
 * 
 * 
 * cdef inline object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init breezy.bzr._dirstate_helpers_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init breezy.bzr._dirstate_helpers_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* bytes_tailmatch */
static int __Pyx_PyBytes_SingleTailmatch(PyObject* self, PyObject* arg,
                                         Py_ssize_t start, Py_ssize_t end, int direction) {
    const char* self_ptr = PyBytes_AS_STRING(self);
    Py_ssize_t self_len = PyBytes_GET_SIZE(self);
    const char* sub_ptr;
    Py_ssize_t sub_len;
    int retval;
    Py_buffer view;
    view.obj = NULL;
    if ( PyBytes_Check(arg) ) {
        sub_ptr = PyBytes_AS_STRING(arg);
        sub_len = PyBytes_GET_SIZE(arg);
    }
#if PY_MAJOR_VERSION < 3
    else if ( PyUnicode_Check(arg) ) {
        return (int) PyUnicode_Tailmatch(self, arg, start, end, direction);
    }
#endif
    else {
        if (unlikely(PyObject_GetBuffer(self, &view, PyBUF_SIMPLE) == -1))
            return -1;
        sub_ptr = (const char*) view.buf;
        sub_len = view.len;
    }
    if (end > self_len)
        end = self_len;
    else if (end < 0)
        end += self_len;
    if (end < 0)
        end = 0;
    if (start < 0)
        start += self_len;
    if (start < 0)
        start = 0;
    if (direction > 0) {
        if (end-sub_len > start)
            start = end - sub_len;
    }
    if (start + sub_len <= end)
        retval = !memcmp(self_ptr+start, sub_ptr, (size_t)sub_len);
    else
        retval = 0;
    if (view.obj)
        PyBuffer_Release(&view);
    return retval;
}
static int __Pyx_PyBytes_TailmatchTuple(PyObject* self, PyObject* substrings,
                                        Py_ssize_t start, Py_ssize_t end, int direction) {
    Py_ssize_t i, count = PyTuple_GET_SIZE(substrings);
    for (i = 0; i < count; i++) {
        int result;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        result = __Pyx_PyBytes_SingleTailmatch(self, PyTuple_GET_ITEM(substrings, i),
                                               start, end, direction);
#else
        PyObject* sub = PySequence_ITEM(substrings, i);
        if (unlikely(!sub)) return -1;
        result = __Pyx_PyBytes_SingleTailmatch(self, sub, start, end, direction);
        Py_DECREF(sub);
#endif
        if (result) {
            return result;
        }
    }
    return 0;
}
static int __Pyx_PyBytes_Tailmatch(PyObject* self, PyObject* substr,
                                   Py_ssize_t start, Py_ssize_t end, int direction) {
    if (unlikely(PyTuple_Check(substr))) {
        return __Pyx_PyBytes_TailmatchTuple(self, substr, start, end, direction);
    }
    return __Pyx_PyBytes_SingleTailmatch(self, substr, start, end, direction);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        Py_SIZE(L) -= 1;
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_Occurred();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    PyErr_Format(PyExc_TypeError,
        "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS
        next = iternext(iterator);
        if (likely(next))
            return next;
        #if PY_VERSION_HEX >= 0x02070000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
        #endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* DelItemInt */
static int __Pyx_DelItem_Generic(PyObject *o, PyObject *j) {
    int r;
    if (!j) return -1;
    r = PyObject_DelItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_DelItemInt_Fast(PyObject *o, Py_ssize_t i,
                                               CYTHON_UNUSED int is_list, CYTHON_NCP_UNUSED int wraparound) {
#if !CYTHON_USE_TYPE_SLOTS
    if (is_list || PySequence_Check(o)) {
        return PySequence_DelItem(o, i);
    }
#else
    PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
    if (likely(m && m->sq_ass_item)) {
        if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
            Py_ssize_t l = m->sq_length(o);
            if (likely(l >= 0)) {
                i += l;
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    return -1;
                PyErr_Clear();
            }
        }
        return m->sq_ass_item(o, i, (PyObject *)NULL);
    }
#endif
    return __Pyx_DelItem_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_ptrdiff_t(ptrdiff_t value) {
    const ptrdiff_t neg_one = (ptrdiff_t) ((ptrdiff_t) 0 - (ptrdiff_t) 1), const_zero = (ptrdiff_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(ptrdiff_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(ptrdiff_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(ptrdiff_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(ptrdiff_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(ptrdiff_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(ptrdiff_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
    const unsigned long neg_one = (unsigned long) ((unsigned long) 0 - (unsigned long) 1), const_zero = (unsigned long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
